// SPDX-License-Identifier: GPL-2.0-only
/*
 *  Driver for Goodix Touchscreens
 *
 *  Copyright (c) 2014 Red Hat Inc.
 *  Copyright (c) 2015 K. Merker <merker@debian.org>
 *
 *  This code is based on gt9xx.c authored by andrew@goodix.com:
 *
 *  2010 - 2012 Goodix Technology.
 */


#include <drm/drmP.h>
#include <linux/kernel.h>
#include <linux/dmi.h>
#include <linux/firmware.h>
#include <linux/gpio/consumer.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/input/mt.h>
#include <linux/input/touchscreen.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/regulator/consumer.h>
#include <linux/slab.h>
#include <linux/acpi.h>
#include <linux/of.h>
#include <linux/pm_runtime.h>
#include <asm/unaligned.h>

struct goodix_ts_data;

struct goodix_chip_data {
	u16 config_addr;
	int config_len;
	int (*check_config)(struct goodix_ts_data *, const struct firmware *);
};

struct goodix_ts_data {
	struct i2c_client *client;
	struct input_dev *input_dev;
	const struct goodix_chip_data *chip;
	struct touchscreen_properties prop;
	struct device_node *disp_node;
	struct notifier_block drmnb;
	unsigned int max_touch_num;
	unsigned int int_trigger_type;
	int substitute_i2c_address;
	struct regulator *avdd28;
	struct regulator *vddio;
	struct gpio_desc *gpiod_int;
	struct gpio_desc *gpiod_rst;
	u16 id;
	u16 version;
	const char *cfg_name;
	struct completion firmware_loading_complete;
	unsigned long irq_flags;
	unsigned int contact_size;
	atomic_t esd_timeout;
	struct delayed_work esd_work;
	bool suspended;
	atomic_t open_count;
	/* Protects power management calls and access to suspended flag */
	struct mutex mutex;
	struct mutex irq_enable_mutex;
	unsigned char irq_active;
	unsigned char drm_disabled_irq;
	unsigned char irq_requested;
	unsigned char wake_irq_requested;
};

#define GOODIX_GPIO_INT_NAME		"irq"
#define GOODIX_GPIO_RST_NAME		"reset"
#define GOODIX_DEVICE_ESD_TIMEOUT_PROPERTY     "esd-recovery-timeout-ms"

#define GOODIX_MAX_HEIGHT		4096
#define GOODIX_MAX_WIDTH		4096
#define GOODIX_INT_TRIGGER		1
#define GOODIX_CONTACT_SIZE		8
#define GOODIX_MAX_CONTACT_SIZE		9
#define GOODIX_MAX_CONTACTS		10

#define GOODIX_CONFIG_MAX_LENGTH	240
#define GOODIX_CONFIG_911_LENGTH	186
#define GOODIX_CONFIG_967_LENGTH	228

/* Register defines */
#define GOODIX_REG_COMMAND		0x8040
#define GOODIX_CMD_SCREEN_OFF		0x05
#define GOODIX_CMD_ESD_ENABLED		0xAA
#define GOODIX_REG_ESD_CHECK		0x8041

#define GOODIX_READ_COOR_ADDR		0x814E
#define GOODIX_GT1X_REG_CONFIG_DATA	0x8050
#define GOODIX_GT9X_REG_CONFIG_DATA	0x8047
#define GOODIX_REG_ID			0x8140

#define GOODIX_BUFFER_STATUS_READY	BIT(7)
#define GOODIX_BUFFER_STATUS_TIMEOUT	20

#define RESOLUTION_LOC		1
#define MAX_CONTACTS_LOC	5
#define TRIGGER_LOC		6
#define GOODIX_AUTOSUSPEND_DELAY_MS	2000

static int goodix_check_cfg_8(struct goodix_ts_data *ts,
			const struct firmware *cfg);
static int goodix_check_cfg_16(struct goodix_ts_data *ts,
			const struct firmware *cfg);

static const struct goodix_chip_data gt1x_chip_data = {
	.config_addr		= GOODIX_GT1X_REG_CONFIG_DATA,
	.config_len		= GOODIX_CONFIG_MAX_LENGTH,
	.check_config		= goodix_check_cfg_16,
};

static const struct goodix_chip_data gt911_chip_data = {
	.config_addr		= GOODIX_GT9X_REG_CONFIG_DATA,
	.config_len		= GOODIX_CONFIG_911_LENGTH,
	.check_config		= goodix_check_cfg_8,
};

static const struct goodix_chip_data gt967_chip_data = {
	.config_addr		= GOODIX_GT9X_REG_CONFIG_DATA,
	.config_len		= GOODIX_CONFIG_967_LENGTH,
	.check_config		= goodix_check_cfg_8,
};

static const struct goodix_chip_data gt9x_chip_data = {
	.config_addr		= GOODIX_GT9X_REG_CONFIG_DATA,
	.config_len		= GOODIX_CONFIG_MAX_LENGTH,
	.check_config		= goodix_check_cfg_8,
};

static const unsigned long goodix_irq_flags[] = {
	IRQ_TYPE_EDGE_RISING,
	IRQ_TYPE_EDGE_FALLING,
	IRQ_TYPE_LEVEL_LOW,
	IRQ_TYPE_LEVEL_HIGH,
};

/*
 * Those tablets have their coordinates origin at the bottom right
 * of the tablet, as if rotated 180 degrees
 */
static const struct dmi_system_id rotated_screen[] = {
#if defined(CONFIG_DMI) && defined(CONFIG_X86)
	{
		.ident = "Teclast X89",
		.matches = {
			/* tPAD is too generic, also match on bios date */
			DMI_MATCH(DMI_BOARD_VENDOR, "TECLAST"),
			DMI_MATCH(DMI_BOARD_NAME, "tPAD"),
			DMI_MATCH(DMI_BIOS_DATE, "12/19/2014"),
		},
	},
	{
		.ident = "Teclast X98 Pro",
		.matches = {
			/*
			 * Only match BIOS date, because the manufacturers
			 * BIOS does not report the board name at all
			 * (sometimes)...
			 */
			DMI_MATCH(DMI_BOARD_VENDOR, "TECLAST"),
			DMI_MATCH(DMI_BIOS_DATE, "10/28/2015"),
		},
	},
	{
		.ident = "WinBook TW100",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "WinBook"),
			DMI_MATCH(DMI_PRODUCT_NAME, "TW100")
		}
	},
	{
		.ident = "WinBook TW700",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "WinBook"),
			DMI_MATCH(DMI_PRODUCT_NAME, "TW700")
		},
	},
#endif
	{}
};

static const struct dmi_system_id nine_bytes_report[] = {
#if defined(CONFIG_DMI) && defined(CONFIG_X86)
	{
		.ident = "Lenovo YogaBook",
		/* YB1-X91L/F and YB1-X90L/F */
		.matches = {
			DMI_MATCH(DMI_PRODUCT_NAME, "Lenovo YB1-X9")
		}
	},
#endif
	{}
};

/*
 * Those tablets have their x coordinate inverted
 */
static const struct dmi_system_id inverted_x_screen[] = {
#if defined(CONFIG_DMI) && defined(CONFIG_X86)
	{
		.ident = "Cube I15-TC",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "Cube"),
			DMI_MATCH(DMI_PRODUCT_NAME, "I15-TC")
		},
	},
#endif
	{}
};

/**
 * goodix_i2c_read - read data from a register of the i2c slave device.
 *
 * @client: i2c device.
 * @reg: the register to read from.
 * @buf: raw write data buffer.
 * @len: length of the buffer to write
 */
static int goodix_i2c_read(struct i2c_client *client,
			   u16 reg, u8 *buf, int len)
{
	struct i2c_msg msgs[2];
	__be16 wbuf = cpu_to_be16(reg);
	int ret;

	msgs[0].flags = 0;
	msgs[0].addr  = client->addr;
	msgs[0].len   = 2;
	msgs[0].buf   = (u8 *)&wbuf;

	msgs[1].flags = I2C_M_RD;
	msgs[1].addr  = client->addr;
	msgs[1].len   = len;
	msgs[1].buf   = buf;

	ret = i2c_transfer(client->adapter, msgs, 2);
	return ret < 0 ? ret : (ret != ARRAY_SIZE(msgs) ? -EIO : 0);
}

/**
 * goodix_i2c_write - write data to a register of the i2c slave device.
 *
 * @client: i2c device.
 * @reg: the register to write to.
 * @buf: raw data buffer to write.
 * @len: length of the buffer to write
 */
static int goodix_i2c_write(struct i2c_client *client, u16 reg, const u8 *buf,
			    unsigned len)
{
	u8 *addr_buf;
	struct i2c_msg msg;
	int ret;

	addr_buf = kmalloc(len + 2, GFP_KERNEL);
	if (!addr_buf)
		return -ENOMEM;

	addr_buf[0] = reg >> 8;
	addr_buf[1] = reg & 0xFF;
	memcpy(&addr_buf[2], buf, len);

	msg.flags = 0;
	msg.addr = client->addr;
	msg.buf = addr_buf;
	msg.len = len + 2;

	ret = i2c_transfer(client->adapter, &msg, 1);
	kfree(addr_buf);
	return ret < 0 ? ret : (ret != 1 ? -EIO : 0);
}

static int goodix_i2c_write_u8(struct i2c_client *client, u16 reg, u8 value)
{
	return goodix_i2c_write(client, reg, &value, sizeof(value));
}

static const struct goodix_chip_data *goodix_get_chip_data(u16 id)
{
	switch (id) {
	case 1151:
	case 5663:
	case 5688:
		return &gt1x_chip_data;

	case 911:
	case 9271:
	case 9110:
	case 9157:
	case 927:
	case 928:
		return &gt911_chip_data;

	case 912:
	case 967:
		return &gt967_chip_data;

	default:
		return &gt9x_chip_data;
	}
}

static int goodix_set_power_state(struct goodix_ts_data *ts, bool on)
{
	int error;

	if (on) {
		error = pm_runtime_get_sync(&ts->client->dev);
	} else {
		pm_runtime_mark_last_busy(&ts->client->dev);
		error = pm_runtime_put_autosuspend(&ts->client->dev);
	}

	if (error < 0) {
		dev_err(&ts->client->dev,
			"failed to change power state to %d\n", on);
		if (on)
			pm_runtime_put_noidle(&ts->client->dev);

		return error;
	}

	return 0;
}

static int goodix_ts_read_input_report(struct goodix_ts_data *ts, u8 *data)
{
	unsigned long max_timeout;
	int touch_num;
	int error;

	/*
	 * The 'buffer status' bit, which indicates that the data is valid, is
	 * not set as soon as the interrupt is raised, but slightly after.
	 * This takes around 10 ms to happen, so we poll for 20 ms.
	 */
	max_timeout = jiffies + msecs_to_jiffies(GOODIX_BUFFER_STATUS_TIMEOUT);
	do {
		error = goodix_i2c_read(ts->client, GOODIX_READ_COOR_ADDR,
					data, ts->contact_size + 1);
		if (error) {
			dev_err(&ts->client->dev, "I2C transfer error: %d\n",
					error);
			return error;
		}

		if (data[0] & GOODIX_BUFFER_STATUS_READY) {
			touch_num = data[0] & 0x0f;
			if (touch_num > ts->max_touch_num)
				return -EPROTO;

			if (touch_num > 1) {
				data += 1 + ts->contact_size;
				error = goodix_i2c_read(ts->client,
						GOODIX_READ_COOR_ADDR +
							1 + ts->contact_size,
						data,
						ts->contact_size *
							(touch_num - 1));
				if (error)
					return error;
			}

			return touch_num;
		}
		if (time_after(jiffies, max_timeout))
			break;
		usleep_range(1000, 2000); /* Poll every 1 - 2 ms */
	} while (1);

	/*
	 * The Goodix panel will send spurious interrupts after a
	 * 'finger up' event, which will always cause a timeout.
	 */
	return -EAGAIN;
}

static void goodix_ts_report_touch_8b(struct goodix_ts_data *ts, u8 *coor_data)
{
	int id = coor_data[0] & 0x0F;
	int input_x = get_unaligned_le16(&coor_data[1]);
	int input_y = get_unaligned_le16(&coor_data[3]);
	int input_w = get_unaligned_le16(&coor_data[5]);

	input_mt_slot(ts->input_dev, id);
	input_mt_report_slot_state(ts->input_dev, MT_TOOL_FINGER, true);
	touchscreen_report_pos(ts->input_dev, &ts->prop,
			       input_x, input_y, true);
	input_report_abs(ts->input_dev, ABS_MT_TOUCH_MAJOR, input_w);
	input_report_abs(ts->input_dev, ABS_MT_WIDTH_MAJOR, input_w);
}

static void goodix_ts_report_touch_9b(struct goodix_ts_data *ts, u8 *coor_data)
{
	int id = coor_data[1] & 0x0F;
	int input_x = get_unaligned_le16(&coor_data[3]);
	int input_y = get_unaligned_le16(&coor_data[5]);
	int input_w = get_unaligned_le16(&coor_data[7]);

	input_mt_slot(ts->input_dev, id);
	input_mt_report_slot_state(ts->input_dev, MT_TOOL_FINGER, true);
	touchscreen_report_pos(ts->input_dev, &ts->prop,
			       input_x, input_y, true);
	input_report_abs(ts->input_dev, ABS_MT_TOUCH_MAJOR, input_w);
	input_report_abs(ts->input_dev, ABS_MT_WIDTH_MAJOR, input_w);
}

/**
 * goodix_process_events - Process incoming events
 *
 * @ts: our goodix_ts_data pointer
 *
 * Called when the IRQ is triggered. Read the current device state, and push
 * the input events to the user space.
 */
static void goodix_process_events(struct goodix_ts_data *ts)
{
	u8  point_data[1 + GOODIX_MAX_CONTACT_SIZE * GOODIX_MAX_CONTACTS];
	int touch_num;
	int i;

	touch_num = goodix_ts_read_input_report(ts, point_data);
	if (touch_num < 0)
		return;

	/*
	 * Bit 4 of the first byte reports the status of the capacitive
	 * Windows/Home button.
	 */
	input_report_key(ts->input_dev, KEY_LEFTMETA, point_data[0] & BIT(4));

	for (i = 0; i < touch_num; i++)
		if (ts->contact_size == 9)
			goodix_ts_report_touch_9b(ts,
				&point_data[1 + ts->contact_size * i]);
		else
			goodix_ts_report_touch_8b(ts,
				&point_data[1 + ts->contact_size * i]);

	input_mt_sync_frame(ts->input_dev);
	input_sync(ts->input_dev);
}

/**
 * goodix_ts_irq_handler - The IRQ handler
 *
 * @irq: interrupt number.
 * @dev_id: private data pointer.
 */
static irqreturn_t goodix_ts_irq_handler(int irq, void *dev_id)
{
	struct goodix_ts_data *ts = dev_id;

	goodix_process_events(ts);

	if (goodix_i2c_write_u8(ts->client, GOODIX_READ_COOR_ADDR, 0) < 0)
		dev_err(&ts->client->dev, "I2C write end_cmd error\n");

	return IRQ_HANDLED;
}

static void goodix_disable_irq(struct goodix_ts_data *ts)
{
	mutex_lock(&ts->irq_enable_mutex);
	if (ts->irq_active) {
		ts->irq_active = 0;
		disable_irq(ts->client->irq);
	}
	mutex_unlock(&ts->irq_enable_mutex);
}

static int goodix_enable_irq(struct goodix_ts_data *ts)
{
	mutex_lock(&ts->irq_enable_mutex);
	if (!ts->irq_active) {
		enable_irq(ts->client->irq);
		ts->irq_active = 1;
	}
	mutex_unlock(&ts->irq_enable_mutex);
	return 0;
}

static int goodix_check_cfg_8(struct goodix_ts_data *ts,
			const struct firmware *cfg)
{
	int i, raw_cfg_len = cfg->size - 2;
	u8 check_sum = 0;

	for (i = 0; i < raw_cfg_len; i++)
		check_sum += cfg->data[i];
	check_sum = (~check_sum) + 1;
	if (check_sum != cfg->data[raw_cfg_len]) {
		dev_err(&ts->client->dev,
			"The checksum of the config fw is not correct");
		return -EINVAL;
	}

	if (cfg->data[raw_cfg_len + 1] != 1) {
		dev_err(&ts->client->dev,
			"Config fw must have Config_Fresh register set");
		return -EINVAL;
	}

	return 0;
}

static int goodix_check_cfg_16(struct goodix_ts_data *ts,
			const struct firmware *cfg)
{
	int i, raw_cfg_len = cfg->size - 3;
	u16 check_sum = 0;

	for (i = 0; i < raw_cfg_len; i += 2)
		check_sum += get_unaligned_be16(&cfg->data[i]);
	check_sum = (~check_sum) + 1;
	if (check_sum != get_unaligned_be16(&cfg->data[raw_cfg_len])) {
		dev_err(&ts->client->dev,
			"The checksum of the config fw is not correct");
		return -EINVAL;
	}

	if (cfg->data[raw_cfg_len + 2] != 1) {
		dev_err(&ts->client->dev,
			"Config fw must have Config_Fresh register set");
		return -EINVAL;
	}

	return 0;
}

/**
 * goodix_check_cfg - Checks if config fw is valid
 *
 * @ts: goodix_ts_data pointer
 * @cfg: firmware config data
 */
static int goodix_check_cfg(struct goodix_ts_data *ts,
			    const struct firmware *cfg)
{
	if (cfg->size > GOODIX_CONFIG_MAX_LENGTH) {
		dev_err(&ts->client->dev,
			"The length of the config fw is not correct");
		return -EINVAL;
	}

	return ts->chip->check_config(ts, cfg);
}

/**
 * goodix_send_cfg - Write fw config to device
 *
 * @ts: goodix_ts_data pointer
 * @cfg: config firmware to write to device
 */
static int goodix_send_cfg(struct goodix_ts_data *ts,
			   const struct firmware *cfg)
{
	int error;

	error = goodix_check_cfg(ts, cfg);
	if (error)
		return error;

	error = goodix_i2c_write(ts->client, ts->chip->config_addr, cfg->data,
				 cfg->size);
	if (error) {
		dev_err(&ts->client->dev, "Failed to write config data: %d",
			error);
		return error;
	}
	dev_dbg(&ts->client->dev, "Config sent successfully.");

	/* Let the firmware reconfigure itself, so sleep for 10ms */
	usleep_range(10000, 11000);

	return 0;
}

static int set_reset_output_val(struct goodix_ts_data *ts, int val)
{
	int ret;

	ret = gpiod_direction_output(ts->gpiod_rst, val);
#if 0
	if (ts->substitute_i2c_address) {
		struct i2c_msg msg;
		unsigned char buf[4];

		/* reg = <0x1f>, 9 - output high, 1 - output low*/
		buf[0] = 0x1f;
		buf[1] = val ? 9 : 1;

		msg.flags = 0;
		msg.addr = ts->substitute_i2c_address;
		msg.buf = buf;
		msg.len = 2;

		ret = i2c_transfer(ts->client->adapter, &msg, 1);
		return ret < 0 ? ret : (ret != 1 ? -EIO : 0);
	}
#endif
	return ret;
}

/* reg = <0x1d>, 3 - input, 9 - output high, 1 - output low*/

static int set_int_output_val(struct goodix_ts_data *ts, int val)
{
	int ret;

	if (ts->substitute_i2c_address) {
		struct i2c_msg msg;
		unsigned char buf[4];

		buf[0] = 0x1d;
		buf[1] = val ? 9 : 1;

		msg.flags = 0;
		msg.addr = ts->substitute_i2c_address;
		msg.buf = buf;
		msg.len = 2;

		ret = i2c_transfer(ts->client->adapter, &msg, 1);
		return ret < 0 ? ret : (ret != 1 ? -EIO : 0);
	}
	ret = gpiod_direction_output(ts->gpiod_int, val);
	return ret;
}

static int set_int_input(struct goodix_ts_data *ts)
{
	int ret;

	ret = gpiod_direction_input(ts->gpiod_int);
	if (ts->substitute_i2c_address) {
		struct i2c_msg msg;
		unsigned char buf[4];

		buf[0] = 0x1d;
		buf[1] = 3;

		msg.flags = 0;
		msg.addr = ts->substitute_i2c_address;
		msg.buf = buf;
		msg.len = 2;

		ret = i2c_transfer(ts->client->adapter, &msg, 1);
		return ret < 0 ? ret : (ret != 1 ? -EIO : 0);
	}
	return ret;
}

static int goodix_int_sync(struct goodix_ts_data *ts)
{
	int error;

	error = set_int_output_val(ts, 0);
	if (error)
		return error;

	msleep(50);				/* T5: 50ms */

	return set_int_input(ts);
}

static int goodix_request_irq(struct goodix_ts_data *ts)
{
	int ret;

	ret = devm_request_threaded_irq(&ts->client->dev, ts->client->irq,
			NULL, goodix_ts_irq_handler,
			ts->irq_flags, ts->client->name, ts);
	if (ret < 0)
		dev_err(&ts->client->dev, "request_threaded_irq failed(%d)\n", ret);
	else
		ts->irq_requested = 1;
	return ret;
}

static int goodix_release_irq(struct goodix_ts_data *ts)
{
	unsigned char irq_requested = ts->irq_requested;

	if (irq_requested) {
		ts->irq_requested = 0;
		goodix_disable_irq(ts);
		devm_free_irq(&ts->client->dev, ts->client->irq, ts);
	}
	return irq_requested;
}

/**
 * goodix_reset - Reset device during power on
 *
 * @ts: goodix_ts_data pointer
 */
static int goodix_reset(struct goodix_ts_data *ts)
{
	unsigned char irq_was_requested = goodix_release_irq(ts);
	int error;

	/* begin select I2C slave addr */
	error = set_reset_output_val(ts, 0);
	if (error)
		return error;

	msleep(20);				/* T2: > 10ms */

	/* HIGH: 0x28/0x29, LOW: 0xBA/0xBB */
	error = set_int_output_val(ts, ts->client->addr == 0x14);
	if (error)
		return error;

	usleep_range(100, 2000);		/* T3: > 100us */

	error = set_reset_output_val(ts, 1);
	if (error)
		return error;

	usleep_range(6000, 10000);		/* T4: > 5ms */

	error = goodix_int_sync(ts);
	if (error)
		return error;

	if (irq_was_requested)
		goodix_request_irq(ts);
	return 0;
}

static ssize_t goodix_dump_config_show(struct device *dev,
				       struct device_attribute *attr, char *buf)
{
	struct goodix_ts_data *ts = dev_get_drvdata(dev);
	u8 config[GOODIX_CONFIG_MAX_LENGTH];
	int error, count = 0, i;

	wait_for_completion(&ts->firmware_loading_complete);

	error = goodix_set_power_state(ts, true);
	if (error)
		return error;

	error = goodix_i2c_read(ts->client, GOODIX_GT9X_REG_CONFIG_DATA,
				config, ts->chip->config_len);
	if (error) {
		dev_warn(&ts->client->dev,
			 "Error reading config (%d)\n",  error);
		return error;
	}

	goodix_set_power_state(ts, false);

	for (i = 0; i < ts->chip->config_len; i++)
		count += scnprintf(buf + count, PAGE_SIZE - count, "%02x ",
				   config[i]);
	return count;
}

static void goodix_disable_esd(struct goodix_ts_data *ts)
{
	if (!atomic_read(&ts->esd_timeout))
		return;
	cancel_delayed_work_sync(&ts->esd_work);
}

static int goodix_enable_esd(struct goodix_ts_data *ts)
{
	int error, esd_timeout;

	esd_timeout = atomic_read(&ts->esd_timeout);
	if (!esd_timeout)
		return 0;

	error = goodix_i2c_write_u8(ts->client, GOODIX_REG_ESD_CHECK,
				    GOODIX_CMD_ESD_ENABLED);
	if (error) {
		dev_err(&ts->client->dev, "Failed to enable ESD: %d\n", error);
		return error;
	}

	schedule_delayed_work(&ts->esd_work, round_jiffies_relative(
			      msecs_to_jiffies(esd_timeout)));
	return 0;
}

static void goodix_esd_work(struct work_struct *work)
{
	struct goodix_ts_data *ts = container_of(work, struct goodix_ts_data,
						 esd_work.work);
	int retries = 3, error;
	u8 esd_data[2];
	const struct firmware *cfg = NULL;

	wait_for_completion(&ts->firmware_loading_complete);

	while (--retries) {
		error = goodix_i2c_read(ts->client, GOODIX_REG_COMMAND,
					esd_data, sizeof(esd_data));
		if (error)
			continue;
		if (esd_data[0] != GOODIX_CMD_ESD_ENABLED &&
		    esd_data[1] == GOODIX_CMD_ESD_ENABLED) {
			/* feed the watchdog */
			goodix_i2c_write_u8(ts->client,
					    GOODIX_REG_COMMAND,
					    GOODIX_CMD_ESD_ENABLED);
			break;
		}
	}

	if (!retries) {
		dev_dbg(&ts->client->dev, "Performing ESD recovery.\n");
		goodix_disable_irq(ts);
		goodix_reset(ts);
		error = request_firmware(&cfg, ts->cfg_name, &ts->client->dev);
		if (!error) {
			goodix_send_cfg(ts, cfg);
			release_firmware(cfg);
		}
		goodix_enable_irq(ts);
		goodix_enable_esd(ts);
		return;
	}

	schedule_delayed_work(&ts->esd_work, round_jiffies_relative(
			      msecs_to_jiffies(atomic_read(&ts->esd_timeout))));
}

static ssize_t goodix_esd_timeout_show(struct device *dev,
				       struct device_attribute *attr, char *buf)
{
	struct goodix_ts_data *ts = dev_get_drvdata(dev);

	return scnprintf(buf, PAGE_SIZE, "%d\n", atomic_read(&ts->esd_timeout));
}

static ssize_t goodix_esd_timeout_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct goodix_ts_data *ts = dev_get_drvdata(dev);
	int error, esd_timeout, new_esd_timeout;

	error = kstrtouint(buf, 10, &new_esd_timeout);
	if (error)
		return error;

	esd_timeout = atomic_read(&ts->esd_timeout);
	if (esd_timeout && !new_esd_timeout &&
	    pm_runtime_active(&ts->client->dev))
		goodix_disable_esd(ts);

	atomic_set(&ts->esd_timeout, new_esd_timeout);
	if (!esd_timeout && new_esd_timeout &&
	    pm_runtime_active(&ts->client->dev))
		goodix_enable_esd(ts);

	return count;
}

static DEVICE_ATTR(dump_config, S_IRUGO, goodix_dump_config_show, NULL);
/* ESD timeout in ms. Default disabled (0). Recommended 2000 ms. */
static DEVICE_ATTR(esd_timeout, S_IRUGO | S_IWUSR, goodix_esd_timeout_show,
		   goodix_esd_timeout_store);

static struct attribute *goodix_attrs[] = {
	&dev_attr_dump_config.attr,
	&dev_attr_esd_timeout.attr,
	NULL
};

static const struct attribute_group goodix_attr_group = {
	.attrs = goodix_attrs,
};

static int goodix_open(struct input_dev *input_dev)
{
	struct goodix_ts_data *ts = input_get_drvdata(input_dev);
	int error;

	if (!ts->gpiod_int || !ts->gpiod_rst)
		return 0;

	wait_for_completion(&ts->firmware_loading_complete);

	error = goodix_set_power_state(ts, true);
	if (error)
		return error;
	atomic_inc(&ts->open_count);
	return 0;
}

static void goodix_close(struct input_dev *input_dev)
{
	struct goodix_ts_data *ts = input_get_drvdata(input_dev);

	if (!ts->gpiod_int || !ts->gpiod_rst)
		return;

	goodix_set_power_state(ts, false);
	atomic_dec(&ts->open_count);
}

/**
 * goodix_get_gpio_config - Get GPIO config from ACPI/DT
 *
 * @ts: goodix_ts_data pointer
 */
static int goodix_get_gpio_config(struct goodix_ts_data *ts)
{
	int error;
	struct device *dev;
	struct gpio_desc *gpiod;

	if (!ts->client)
		return -EINVAL;
	dev = &ts->client->dev;

	ts->avdd28 = devm_regulator_get(dev, "AVDD28");
	if (IS_ERR(ts->avdd28)) {
		error = PTR_ERR(ts->avdd28);
		if (error != -EPROBE_DEFER)
			dev_err(dev,
				"Failed to get AVDD28 regulator: %d\n", error);
		return error;
	}

	ts->vddio = devm_regulator_get(dev, "VDDIO");
	if (IS_ERR(ts->vddio)) {
		error = PTR_ERR(ts->vddio);
		if (error != -EPROBE_DEFER)
			dev_err(dev,
				"Failed to get VDDIO regulator: %d\n", error);
		return error;
	}

	/* Get the interrupt GPIO pin number */
	gpiod = devm_gpiod_get_optional(dev, GOODIX_GPIO_INT_NAME, GPIOD_IN);
	if (IS_ERR(gpiod)) {
		error = PTR_ERR(gpiod);
		if (error != -EPROBE_DEFER)
			dev_dbg(dev, "Failed to get %s GPIO: %d\n",
				GOODIX_GPIO_INT_NAME, error);
		return error;
	}

	ts->gpiod_int = gpiod;

	/* Get the reset line GPIO pin number */
	gpiod = devm_gpiod_get_optional(dev, GOODIX_GPIO_RST_NAME, GPIOD_IN);
	if (IS_ERR(gpiod)) {
		error = PTR_ERR(gpiod);
		if (error != -EPROBE_DEFER)
			dev_dbg(dev, "Failed to get %s GPIO: %d\n",
				GOODIX_GPIO_RST_NAME, error);
		return error;
	}

	ts->gpiod_rst = gpiod;

	error = of_property_read_u32_index(dev->of_node,
			"substitute-i2c-address",
				0, &ts->substitute_i2c_address);
	if (ts->substitute_i2c_address) {
		if (set_int_input(ts)) {
			ts->substitute_i2c_address = 0;
			dev_info(dev, "disabling substitute_i2c_address\n");
		} else {
			dev_info(dev, "substitute_i2c_address=0x%x\n",
					ts->substitute_i2c_address);

		}
	}
	return 0;
}

/**
 * goodix_read_config - Read the embedded configuration of the panel
 *
 * @ts: our goodix_ts_data pointer
 *
 * Must be called during probe
 */
static void goodix_read_config(struct goodix_ts_data *ts)
{
	u8 config[GOODIX_CONFIG_MAX_LENGTH];
	int x_max, y_max;
	int error;

	error = goodix_i2c_read(ts->client, ts->chip->config_addr,
				config, ts->chip->config_len);
	if (error) {
		dev_warn(&ts->client->dev, "Error reading config: %d\n",
			 error);
		ts->int_trigger_type = GOODIX_INT_TRIGGER;
		ts->max_touch_num = GOODIX_MAX_CONTACTS;
		return;
	}

	ts->int_trigger_type = config[TRIGGER_LOC] & 0x03;
	ts->max_touch_num = config[MAX_CONTACTS_LOC] & 0x0f;

	x_max = get_unaligned_le16(&config[RESOLUTION_LOC]);
	y_max = get_unaligned_le16(&config[RESOLUTION_LOC + 2]);
	if (x_max && y_max) {
		input_abs_set_max(ts->input_dev, ABS_MT_POSITION_X, x_max - 1);
		input_abs_set_max(ts->input_dev, ABS_MT_POSITION_Y, y_max - 1);
	}
}

/**
 * goodix_read_version - Read goodix touchscreen version
 *
 * @ts: our goodix_ts_data pointer
 */
static int goodix_read_version(struct goodix_ts_data *ts)
{
	int error;
	u8 buf[6];
	char id_str[5];

	error = goodix_i2c_read(ts->client, GOODIX_REG_ID, buf, sizeof(buf));
	if (error) {
		dev_err(&ts->client->dev, "read version failed: %d\n", error);
		return error;
	}

	memcpy(id_str, buf, 4);
	id_str[4] = 0;
	if (kstrtou16(id_str, 10, &ts->id))
		ts->id = 0x1001;

	ts->version = get_unaligned_le16(&buf[4]);

	dev_info(&ts->client->dev, "ID %d, version: %04x\n", ts->id,
		 ts->version);

	return 0;
}

/**
 * goodix_i2c_test - I2C test function to check if the device answers.
 *
 * @client: the i2c client
 */
static int goodix_i2c_test(struct i2c_client *client)
{
	int retry = 0;
	int error;
	u8 test;

	while (retry++ < 2) {
		error = goodix_i2c_read(client, GOODIX_REG_ID,
					&test, 1);
		if (!error)
			return 0;

		dev_err(&client->dev, "i2c test failed attempt %d: %d\n",
			retry, error);
		msleep(20);
	}

	return error;
}

/**
 * goodix_configure_dev - Finish device initialization
 *
 * @ts: our goodix_ts_data pointer
 *
 * Must be called from probe to finish initialization of the device.
 * Contains the common initialization code for both devices that
 * declare gpio pins and devices that do not. It is either called
 * directly from probe or from request_firmware_wait callback.
 */
static int goodix_configure_dev(struct goodix_ts_data *ts)
{
	ts->int_trigger_type = GOODIX_INT_TRIGGER;
	ts->max_touch_num = GOODIX_MAX_CONTACTS;

	ts->input_dev = devm_input_allocate_device(&ts->client->dev);
	if (!ts->input_dev) {
		dev_err(&ts->client->dev, "Failed to allocate input device.");
		return -ENOMEM;
	}

	ts->input_dev->name = "Goodix Capacitive TouchScreen";
	ts->input_dev->phys = "input/ts";
	ts->input_dev->id.bustype = BUS_I2C;
	ts->input_dev->id.vendor = 0x0416;
	ts->input_dev->id.product = ts->id;
	ts->input_dev->id.version = ts->version;
	ts->input_dev->open = goodix_open;
	ts->input_dev->close = goodix_close;
	input_set_drvdata(ts->input_dev, ts);

	/* Capacitive Windows/Home button on some devices */
	input_set_capability(ts->input_dev, EV_KEY, KEY_LEFTMETA);

	input_set_capability(ts->input_dev, EV_ABS, ABS_MT_POSITION_X);
	input_set_capability(ts->input_dev, EV_ABS, ABS_MT_POSITION_Y);
	input_set_abs_params(ts->input_dev, ABS_MT_WIDTH_MAJOR, 0, 255, 0, 0);
	input_set_abs_params(ts->input_dev, ABS_MT_TOUCH_MAJOR, 0, 255, 0, 0);

	/* Read configuration and apply touchscreen parameters */
	goodix_read_config(ts);
	return 0;
}

static int ts_drm_event(struct notifier_block *nb, unsigned long event, void *data)
{
	struct drm_device *drm_dev = data;
	struct device_node *node = drm_dev->dev->of_node;
	struct goodix_ts_data *ts = container_of(nb, struct goodix_ts_data, drmnb);
	struct device *dev = &ts->client->dev;

	dev_dbg(dev, "%s: event %lx\n", __func__, event);

	if (node != ts->disp_node) {
		dev_info(dev, "%s: event %lx, (%s)%p (%s)%p\n", __func__, event, node->name, node, ts->disp_node->name, ts->disp_node);
		return 0;
	}

	switch (event) {
	case DRM_MODE_DPMS_ON:
		dev_dbg(dev, "%s: ON %lx\n", __func__, event);
		if (ts->drm_disabled_irq) {
			ts->drm_disabled_irq = 0;
			goodix_enable_irq(ts);
		}
		break;
	case DRM_MODE_DPMS_STANDBY:
		break;
	case DRM_MODE_DPMS_SUSPEND:
	case DRM_MODE_DPMS_OFF:
		dev_dbg(dev, "%s: OFF %lx\n", __func__, event);
		if (!ts->drm_disabled_irq) {
			ts->drm_disabled_irq = 1;
			goodix_disable_irq(ts);
		}
		break;
	default:
		dev_info(dev, "%s: unknown event %lx\n", __func__, event);
	}

	return 0;
}

static int goodix_finish_setup(struct goodix_ts_data *ts)
{
	int error;
        struct device_node *np = ts->client->dev.of_node;

	/* Try overriding touchscreen parameters via device properties */
	touchscreen_parse_properties(ts->input_dev, true, &ts->prop);

	if (!ts->prop.max_x || !ts->prop.max_y || !ts->max_touch_num) {
		dev_err(&ts->client->dev,
			"Invalid config (%d, %d, %d), using defaults\n",
			ts->prop.max_x, ts->prop.max_y, ts->max_touch_num);
		ts->prop.max_x = GOODIX_MAX_WIDTH - 1;
		ts->prop.max_y = GOODIX_MAX_HEIGHT - 1;
		ts->max_touch_num = GOODIX_MAX_CONTACTS;
		input_abs_set_max(ts->input_dev,
				  ABS_MT_POSITION_X, ts->prop.max_x);
		input_abs_set_max(ts->input_dev,
				  ABS_MT_POSITION_Y, ts->prop.max_y);
	}

	if (dmi_check_system(rotated_screen)) {
		ts->prop.invert_x = true;
		ts->prop.invert_y = true;
		dev_dbg(&ts->client->dev,
			"Applying '180 degrees rotated screen' quirk\n");
	}

	if (dmi_check_system(nine_bytes_report)) {
		ts->contact_size = 9;

		dev_dbg(&ts->client->dev,
			"Non-standard 9-bytes report format quirk\n");
	}

	if (dmi_check_system(inverted_x_screen)) {
		ts->prop.invert_x = true;
		dev_dbg(&ts->client->dev,
			"Applying 'inverted x screen' quirk\n");
	}

	error = input_mt_init_slots(ts->input_dev, ts->max_touch_num,
				    INPUT_MT_DIRECT | INPUT_MT_DROP_UNUSED);
	if (error) {
		dev_err(&ts->client->dev,
			"Failed to initialize MT slots: %d", error);
		return error;
	}

	error = input_register_device(ts->input_dev);
	if (error) {
		dev_err(&ts->client->dev,
			"Failed to register input device: %d", error);
		return error;
	}

	if (device_property_read_bool(ts->input_dev->dev.parent, "edge-failling-trigger"))
		ts->int_trigger_type = GOODIX_INT_TRIGGER;

	ts->irq_flags = goodix_irq_flags[ts->int_trigger_type] | IRQF_ONESHOT;
	irq_set_status_flags(ts->client->irq, IRQ_NOAUTOEN);
	error = goodix_request_irq(ts);
	if (error < 0)
		return error;
	ts->disp_node = of_parse_phandle(np, "display", 0);
	if (ts->disp_node) {
		ts->drmnb.notifier_call = ts_drm_event;
		error = drm_register_client(&ts->drmnb);
		if (error < 0)
			dev_err(&ts->client->dev, "drm_register_client failed(%d)\n", error);
	}

	error = goodix_enable_irq(ts);
	if (error) {
		dev_err(&ts->client->dev, "enable IRQ failed: %d\n", error);
		return error;
	}

	return 0;
}

/**
 * goodix_config_cb - Callback to finish device init
 *
 * @ts: our goodix_ts_data pointer
 *
 * request_firmware_wait callback that finishes
 * initialization of the device. This will only be called
 * when ts->gpiod_int and ts->gpiod_rst are properly initialized.
 */
static void goodix_config_cb(const struct firmware *cfg, void *ctx)
{
	struct goodix_ts_data *ts = ctx;
	int error;

	if (cfg) {
		/* send device configuration to the firmware */
		error = goodix_send_cfg(ts, cfg);
		if (error)
			goto err_release_cfg;
	}

	error = goodix_configure_dev(ts);
	if (error)
		goto err_release_cfg;

	goodix_enable_esd(ts);

	pm_runtime_set_autosuspend_delay(&ts->client->dev,
					 GOODIX_AUTOSUSPEND_DELAY_MS);
	pm_runtime_use_autosuspend(&ts->client->dev);
	error = pm_runtime_set_active(&ts->client->dev);
	if (error) {
		dev_err(&ts->client->dev, "failed to set active: %d\n", error);
		goto err_release_cfg;
	}
	pm_runtime_enable(&ts->client->dev);
	/* Must not suspend immediately after device initialization */
	pm_runtime_mark_last_busy(&ts->client->dev);
	pm_request_autosuspend(&ts->client->dev);

	release_firmware(cfg);
	complete_all(&ts->firmware_loading_complete);
	goodix_finish_setup(ts);
	return;

err_release_cfg:
	release_firmware(cfg);
	complete_all(&ts->firmware_loading_complete);
}

static void goodix_disable_regulators(void *arg)
{
	struct goodix_ts_data *ts = arg;

	regulator_disable(ts->vddio);
	regulator_disable(ts->avdd28);
}

static int goodix_ts_probe(struct i2c_client *client,
			   const struct i2c_device_id *id)
{
	struct goodix_ts_data *ts;
	int error, esd_timeout;

	dev_dbg(&client->dev, "I2C Address: 0x%02x\n", client->addr);

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		dev_err(&client->dev, "I2C check functionality failed.\n");
		return -ENXIO;
	}

	ts = devm_kzalloc(&client->dev, sizeof(*ts), GFP_KERNEL);
	if (!ts)
		return -ENOMEM;

	ts->client = client;
	i2c_set_clientdata(client, ts);
	init_completion(&ts->firmware_loading_complete);
	ts->contact_size = GOODIX_CONTACT_SIZE;
	INIT_DELAYED_WORK(&ts->esd_work, goodix_esd_work);
	mutex_init(&ts->mutex);
	mutex_init(&ts->irq_enable_mutex);

	error = goodix_get_gpio_config(ts);
	if (error)
		return error;

	/* power up the controller */
	error = regulator_enable(ts->avdd28);
	if (error) {
		dev_err(&client->dev,
			"Failed to enable AVDD28 regulator: %d\n",
			error);
		return error;
	}

	error = regulator_enable(ts->vddio);
	if (error) {
		dev_err(&client->dev,
			"Failed to enable VDDIO regulator: %d\n",
			error);
		regulator_disable(ts->avdd28);
		return error;
	}

	error = devm_add_action_or_reset(&client->dev,
					 goodix_disable_regulators, ts);
	if (error)
		return error;

	if (ts->gpiod_int && ts->gpiod_rst) {
		/* reset the controller */
		error = goodix_reset(ts);
		if (error) {
			dev_err(&client->dev, "Controller reset failed.\n");
			return error;
		}
	}

	error = goodix_i2c_test(client);
	if (error) {
		dev_err(&client->dev, "I2C communication failure: %d\n", error);
		return error;
	}

	error = goodix_read_version(ts);
	if (error) {
		dev_err(&client->dev, "Read version failed.\n");
		return error;
	}

	ts->chip = goodix_get_chip_data(ts->id);

	if (ts->gpiod_int && ts->gpiod_rst) {
		error = device_property_read_u32(&ts->client->dev,
					GOODIX_DEVICE_ESD_TIMEOUT_PROPERTY,
					&esd_timeout);
		if (!error)
			atomic_set(&ts->esd_timeout, esd_timeout);

		error = sysfs_create_group(&client->dev.kobj,
					   &goodix_attr_group);
		if (error) {
			dev_err(&client->dev,
				"Failed to create sysfs group: %d\n",
				error);
			return error;
		}

		/* update device config */
		ts->cfg_name = devm_kasprintf(&client->dev, GFP_KERNEL,
					      "goodix_%d_cfg.bin", ts->id);
		if (!ts->cfg_name) {
			error = -ENOMEM;
			goto err_sysfs_remove_group;
		}

		error = request_firmware_nowait(THIS_MODULE, true, ts->cfg_name,
						&client->dev, GFP_KERNEL, ts,
						goodix_config_cb);
		if (error) {
			dev_err(&client->dev,
				"Failed to invoke firmware loader: %d\n",
				error);
			goto err_sysfs_remove_group;
		}
	} else {
		error = goodix_configure_dev(ts);
		if (error)
			return error;
		error = goodix_finish_setup(ts);
		if (error)
			return error;
	}
	return 0;

err_sysfs_remove_group:
	if (ts->gpiod_int && ts->gpiod_rst)
		sysfs_remove_group(&client->dev.kobj, &goodix_attr_group);
	return error;
}

static int goodix_ts_remove(struct i2c_client *client)
{
	struct goodix_ts_data *ts = i2c_get_clientdata(client);

	goodix_disable_irq(ts);
	if (ts->client->irq)
		irq_set_irq_type(ts->client->irq, IRQ_TYPE_NONE);
	if (!ts->gpiod_int || !ts->gpiod_rst)
		return 0;

	drm_unregister_client(&ts->drmnb);
	wait_for_completion(&ts->firmware_loading_complete);

	pm_runtime_disable(&client->dev);
	pm_runtime_set_suspended(&client->dev);
	pm_runtime_put_noidle(&client->dev);

	sysfs_remove_group(&client->dev.kobj, &goodix_attr_group);
	goodix_disable_esd(ts);

	return 0;
}

static int __maybe_unused goodix_sleep(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct goodix_ts_data *ts = i2c_get_clientdata(client);
	unsigned char irq_was_requested;
	int error = 0;

	/* We need gpio pins to suspend/resume */
	if (!ts->gpiod_int || !ts->gpiod_rst) {
		disable_irq(client->irq);
		return 0;
	}

	wait_for_completion(&ts->firmware_loading_complete);

	mutex_lock(&ts->mutex);

	if (ts->suspended)
		goto out_error;

	goodix_disable_esd(ts);

	/* disable IRQ as IRQ pin is used as output in the suspend sequence */
	irq_was_requested = goodix_release_irq(ts);
	ts->wake_irq_requested = irq_was_requested;

	/* Output LOW on the INT pin for 5 ms */
	error = set_int_output_val(ts, 0);
	if (error) {
		goodix_enable_irq(ts);
		goto out_error;
	}

	usleep_range(5000, 6000);

	error = goodix_i2c_write_u8(ts->client, GOODIX_REG_COMMAND,
				    GOODIX_CMD_SCREEN_OFF);
	if (error) {
		dev_err(&ts->client->dev, "Screen off command failed\n");
		set_int_input(ts);
		if (irq_was_requested) {
			goodix_request_irq(ts);
			goodix_enable_irq(ts);
		}
		error = -EAGAIN;
		goto out_error;
	}

	/*
	 * The datasheet specifies that the interval between sending screen-off
	 * command and wake-up should be longer than 58 ms. To avoid waking up
	 * sooner, delay 58ms here.
	 */
	msleep(58);
	ts->suspended = true;
	mutex_unlock(&ts->mutex);

	return 0;

out_error:
	mutex_unlock(&ts->mutex);
	return error;
}

static int __maybe_unused goodix_wakeup(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct goodix_ts_data *ts = i2c_get_clientdata(client);
	int error = 0;

	if (!ts->gpiod_int || !ts->gpiod_rst) {
		enable_irq(client->irq);
		return 0;
	}

	mutex_lock(&ts->mutex);

	if (!ts->suspended)
		goto out_error;

	/*
	 * Exit sleep mode by outputting HIGH level to INT pin
	 * for 2ms~5ms.
	 */
	error = set_int_output_val(ts, 1);
	if (error)
		goto out_error;

	usleep_range(2000, 5000);

	error = goodix_int_sync(ts);
	if (error)
		goto out_error;

	if (ts->wake_irq_requested) {
		ts->wake_irq_requested = 0;
		error = goodix_request_irq(ts);
		if (error)
			goto out_error;
		error = goodix_enable_irq(ts);
		if (error)
			goto out_error;
	}

	error = goodix_enable_esd(ts);
	if (error)
		goto out_error;

	ts->suspended = false;
	mutex_unlock(&ts->mutex);

	return 0;

out_error:
	mutex_unlock(&ts->mutex);
	return error;
}

static int __maybe_unused goodix_resume(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct goodix_ts_data *ts = i2c_get_clientdata(client);

	if (!atomic_read(&ts->open_count))
		return 0;

	return goodix_wakeup(dev);
}

static const struct dev_pm_ops goodix_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(goodix_sleep, goodix_resume)
	SET_RUNTIME_PM_OPS(goodix_sleep, goodix_wakeup, NULL)
};

static const struct i2c_device_id goodix_ts_id[] = {
	{ "GDIX1001:00", 0 },
	{ "gt9271", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, goodix_ts_id);

#ifdef CONFIG_ACPI
static const struct acpi_device_id goodix_acpi_match[] = {
	{ "GDIX1001", 0 },
	{ "GDIX1002", 0 },
	{ }
};
MODULE_DEVICE_TABLE(acpi, goodix_acpi_match);
#endif

#ifdef CONFIG_OF
static const struct of_device_id goodix_of_match[] = {
	{ .compatible = "goodix,gt1151" },
	{ .compatible = "goodix,gt5663" },
	{ .compatible = "goodix,gt5688" },
	{ .compatible = "goodix,gt911" },
	{ .compatible = "goodix,gt9110" },
	{ .compatible = "goodix,gt912" },
	{ .compatible = "goodix,gt9157" },
	{ .compatible = "goodix,gt927" },
	{ .compatible = "goodix,gt9271" },
	{ .compatible = "goodix,gt928" },
	{ .compatible = "goodix,gt967" },
	{ }
};
MODULE_DEVICE_TABLE(of, goodix_of_match);
#endif

static struct i2c_driver goodix_ts_driver = {
	.probe = goodix_ts_probe,
	.remove = goodix_ts_remove,
	.id_table = goodix_ts_id,
	.driver = {
		.name = "Goodix-TS",
		.acpi_match_table = ACPI_PTR(goodix_acpi_match),
		.of_match_table = of_match_ptr(goodix_of_match),
		.pm = &goodix_pm_ops,
	},
};
module_i2c_driver(goodix_ts_driver);

MODULE_AUTHOR("Benjamin Tissoires <benjamin.tissoires@gmail.com>");
MODULE_AUTHOR("Bastien Nocera <hadess@hadess.net>");
MODULE_DESCRIPTION("Goodix touchscreen driver");
MODULE_LICENSE("GPL v2");
