/*
 * Copyright (C) 2012-2014 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/of_device.h>
#include <linux/i2c.h>
#include <linux/of_gpio.h>
#include <linux/pinctrl/consumer.h>
#include <linux/regmap.h>
#include <linux/regulator/consumer.h>
#include <linux/fsl_devices.h>
#include <media/v4l2-chip-ident.h>
#include "v4l2-int-device.h"
#include "mxc_v4l2_capture.h"

#define OV5642_VOLTAGE_ANALOG               2800000
#define OV5642_VOLTAGE_DIGITAL_CORE         1500000
#define OV5642_VOLTAGE_DIGITAL_IO           1800000

#define MIN_FPS 15
#define MAX_FPS 30
#define DEFAULT_FPS 30

#define OV5642_XCLK_MIN 6000000
#define OV5642_XCLK_MAX 24000000

/* OV5640 focus area values */
#define OV5642_AF_ZONE_ARRAY_WIDTH	80
#define OV5642_AF_ZONE_ARRAY_HEIGHT	60

/* OV5642 Camera Auto Focus Registers */
#define REG_CMD_MAIN			0x3024
#define REG_STA_FOCUS			0x3027
#define REG_STA_ZONE			0x3026
#define REG_CMD_TAG			0x3025
#define REG_CMD_PARA3 			0x5082
#define REG_CMD_PARA2 			0x5083
#define REG_CMD_PARA1			0x5084
#define REG_CMD_PARA0 			0x5085

/* OV5642 Auto Focus Commands and Responses */
#define	S_STARTUP			0xFA
#define	S_FIRWARE			0xFF
#define	S_STARTUP			0xFA
#define	S_ERROR				0xFE
#define	S_DRVICERR			0xEE
#define	S_IDLE				0x00
#define	S_FOCUSING			0x01
#define	S_FOCUSED			0x02
#define	S_CAPTURE			0x12
#define	S_STEP				0x20

/* OV5642 AWB Registers */
#define REG_AWB_MANUAL			0x3406
#define REG_AWB_R_GAIN_HIGH		0x3400
#define REG_AWB_R_GAIN_LOW		0x3401
#define REG_AWB_G_GAIN_HIGH		0x3402
#define REG_AWB_G_GAIN_LOW		0x3403
#define REG_AWB_B_GAIN_HIGH		0x3404
#define REG_AWB_B_GAIN_LOW		0x3405

/* OV5642 CMD Registers */
#define	CMD_ENABLE_OVERLAY		0x01
#define	CMD_DISABLE_OVERLAY		0x02
#define	CMD_SINGLE_FOCUS_MODE		0x03
#define	CMD_CONST_FOCUS_MODE		0x04
#define	CMD_STEP_FOCUS_MODE		0x05
#define	CMD_PAUSE			0x06
#define	CMD_IDLE_MODE			0x08
#define	CMD_SET_ZONE_MODE		0x10
#define	CMD_UPDATE_ZONE_MODE		0x12
#define	CMD_MOTOR_MODE			0x20
#define	CMD_SCAN_MODE			0x30

#define OV5642_CHIP_ID_HIGH_BYTE	0x300A
#define OV5642_CHIP_ID_LOW_BYTE		0x300B

enum ov5642_mode {
	ov5642_mode_MIN = 0,
	ov5642_mode_VGA_640_480 = 0,
	ov5642_mode_QVGA_320_240 = 1,
	ov5642_mode_NTSC_720_480 = 2,
	ov5642_mode_PAL_720_576 = 3,
	ov5642_mode_720P_1280_720 = 4,
	ov5642_mode_1080P_1920_1080 = 5,
	ov5642_mode_QSXGA_2592_1944 = 6,
	ov5642_mode_QCIF_176_144 = 7,
	ov5642_mode_XGA_1024_768 = 8,
	ov5642_mode_MAX = 8
};

enum ov5642_frame_rate {
	ov5642_15_fps,
	ov5642_30_fps
};

static int ov5642_framerates[] = {
	[ov5642_15_fps] = 15,
	[ov5642_30_fps] = 30,
};

struct reg_value {
	u16 u16RegAddr;
	u8 u8Val;
	u8 u8Mask;
	u32 u32Delay_ms;
};

struct ov5642_mode_info {
	enum ov5642_mode mode;
	u32 width;
	u32 height;
	struct reg_value *init_data_ptr;
	u32 init_data_size;
};

/*!
 * Maintains the information on the current state of the sesor.
 */
static struct sensor_data ov5642_data;
struct gpio_desc *gpiod_pwdn;
struct gpio_desc *gpiod_rst;
struct gpio_desc *gpiod_rst2;
static int focus_mode = V4L2_CID_AUTO_FOCUS_STOP;
static int focus_range = V4L2_AUTO_FOCUS_RANGE_NORMAL;
static uint16_t roi_x = 0;
static uint16_t roi_y = 0;

static struct reg_value ov5642_rot_none_VGA[] = {
	{0x3818, 0xc1, 0x00, 0x00}, {0x3621, 0x87, 0x00, 0x00},
};

static struct reg_value ov5642_rot_vert_flip_VGA[] = {
	{0x3818, 0x20, 0x60, 0x00}, {0x3621, 0x20, 0x20, 0x00},
};

static struct reg_value ov5642_rot_horiz_flip_VGA[] = {
	{0x3818, 0x81, 0x00, 0x01}, {0x3621, 0xa7, 0x00, 0x00},
};

static struct reg_value ov5642_rot_180_VGA[] = {
	{0x3818, 0x60, 0x60, 0x00}, {0x3621, 0x00, 0x20, 0x00},
};


static struct reg_value ov5642_rot_none_FULL[] = {
	{0x3818, 0xc0, 0x00, 0x00}, {0x3621, 0x09, 0x00, 0x00},
};

static struct reg_value ov5642_rot_vert_flip_FULL[] = {
	{0x3818, 0x20, 0x60, 0x01}, {0x3621, 0x20, 0x20, 0x00},
};

static struct reg_value ov5642_rot_horiz_flip_FULL[] = {
	{0x3818, 0x80, 0x00, 0x01}, {0x3621, 0x29, 0x00, 0x00},
};

static struct reg_value ov5642_rot_180_FULL[] = {
	{0x3818, 0x60, 0x60, 0x00}, {0x3621, 0x00, 0x20, 0x00},
};


static struct reg_value ov5642_initial_setting[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c00, 0x04, 0, 0}, {0x3c01, 0x80, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0},
	{0x5182, 0x00, 0, 0}, {0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5001, 0xff, 0, 0}, {0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0},
	{0x5505, 0x7f, 0, 0}, {0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0},
	{0x4610, 0x00, 0, 0}, {0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0},
	{0x3808, 0x02, 0, 0}, {0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0},
	{0x380b, 0xe0, 0, 0}, {0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0},
	{0x501f, 0x00, 0, 0}, {0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0},
	{0x350b, 0x00, 0, 0}, {0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3030, 0x0b, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 300},
	{0x302c, 0x40, 0x60, 0},
};

static struct reg_value ov5642_af_firmware[] = {
	{0x8000,0x02, 0, 0}, {0x8001,0x00, 0, 0}, {0x8002,0x06, 0, 0},
	{0x8003,0x02, 0, 0}, {0x8004,0x0c, 0, 0}, {0x8005,0xa4, 0, 0},
	{0x8006,0x78, 0, 0}, {0x8007,0x7f, 0, 0}, {0x8008,0xe4, 0, 0},
	{0x8009,0xf6, 0, 0}, {0x800a,0xd8, 0, 0}, {0x800b,0xfd, 0, 0},
	{0x800c,0x75, 0, 0}, {0x800d,0x81, 0, 0}, {0x800e,0x7e, 0, 0},
	{0x800f,0x02, 0, 0}, {0x8010,0x14, 0, 0}, {0x8011,0xc7, 0, 0},
	{0x8012,0x00, 0, 0}, {0x8013,0x02, 0, 0}, {0x8014,0x15, 0, 0},
	{0x8015,0x13, 0, 0}, {0x8016,0xe0, 0, 0}, {0x8017,0xf5, 0, 0},
	{0x8018,0x72, 0, 0}, {0x8019,0xa3, 0, 0}, {0x801a,0xe0, 0, 0},
	{0x801b,0xf5, 0, 0}, {0x801c,0x73, 0, 0}, {0x801d,0xae, 0, 0},
	{0x801e,0x6a, 0, 0}, {0x801f,0xe4, 0, 0}, {0x8020,0x85, 0, 0},
	{0x8021,0x6b, 0, 0}, {0x8022,0x55, 0, 0}, {0x8023,0x8e, 0, 0},
	{0x8024,0x54, 0, 0}, {0x8025,0xf5, 0, 0}, {0x8026,0x53, 0, 0},
	{0x8027,0xf5, 0, 0}, {0x8028,0x52, 0, 0}, {0x8029,0xab, 0, 0},
	{0x802a,0x55, 0, 0}, {0x802b,0xaa, 0, 0}, {0x802c,0x54, 0, 0},
	{0x802d,0xa9, 0, 0}, {0x802e,0x53, 0, 0}, {0x802f,0xa8, 0, 0},
	{0x8030,0x52, 0, 0}, {0x8031,0xaf, 0, 0}, {0x8032,0x2c, 0, 0},
	{0x8033,0xfc, 0, 0}, {0x8034,0xfd, 0, 0}, {0x8035,0xfe, 0, 0},
	{0x8036,0x12, 0, 0}, {0x8037,0x08, 0, 0}, {0x8038,0x98, 0, 0},
	{0x8039,0x8f, 0, 0}, {0x803a,0x55, 0, 0}, {0x803b,0x8e, 0, 0},
	{0x803c,0x54, 0, 0}, {0x803d,0x8d, 0, 0}, {0x803e,0x53, 0, 0},
	{0x803f,0x8c, 0, 0}, {0x8040,0x52, 0, 0}, {0x8041,0xaf, 0, 0},
	{0x8042,0x55, 0, 0}, {0x8043,0xae, 0, 0}, {0x8044,0x54, 0, 0},
	{0x8045,0xad, 0, 0}, {0x8046,0x53, 0, 0}, {0x8047,0xac, 0, 0},
	{0x8048,0x52, 0, 0}, {0x8049,0x8f, 0, 0}, {0x804a,0x2b, 0, 0},
	{0x804b,0x8e, 0, 0}, {0x804c,0x2a, 0, 0}, {0x804d,0x8d, 0, 0},
	{0x804e,0x29, 0, 0}, {0x804f,0x8c, 0, 0}, {0x8050,0x28, 0, 0},
	{0x8051,0xae, 0, 0}, {0x8052,0x6c, 0, 0}, {0x8053,0xe4, 0, 0},
	{0x8054,0x85, 0, 0}, {0x8055,0x6d, 0, 0}, {0x8056,0x55, 0, 0},
	{0x8057,0x8e, 0, 0}, {0x8058,0x54, 0, 0}, {0x8059,0xf5, 0, 0},
	{0x805a,0x53, 0, 0}, {0x805b,0xf5, 0, 0}, {0x805c,0x52, 0, 0},
	{0x805d,0xab, 0, 0}, {0x805e,0x55, 0, 0}, {0x805f,0xaa, 0, 0},
	{0x8060,0x54, 0, 0}, {0x8061,0xa9, 0, 0}, {0x8062,0x53, 0, 0},
	{0x8063,0xa8, 0, 0}, {0x8064,0x52, 0, 0}, {0x8065,0xaf, 0, 0},
	{0x8066,0x2d, 0, 0}, {0x8067,0xfc, 0, 0}, {0x8068,0xfd, 0, 0},
	{0x8069,0xfe, 0, 0}, {0x806a,0x12, 0, 0}, {0x806b,0x08, 0, 0},
	{0x806c,0x98, 0, 0}, {0x806d,0x8f, 0, 0}, {0x806e,0x55, 0, 0},
	{0x806f,0x8e, 0, 0}, {0x8070,0x54, 0, 0}, {0x8071,0x8d, 0, 0},
	{0x8072,0x53, 0, 0}, {0x8073,0x8c, 0, 0}, {0x8074,0x52, 0, 0},
	{0x8075,0xe5, 0, 0}, {0x8076,0x2b, 0, 0}, {0x8077,0x25, 0, 0},
	{0x8078,0x55, 0, 0}, {0x8079,0xf5, 0, 0}, {0x807a,0x2b, 0, 0},
	{0x807b,0xe5, 0, 0}, {0x807c,0x2a, 0, 0}, {0x807d,0x35, 0, 0},
	{0x807e,0x54, 0, 0}, {0x807f,0xf5, 0, 0}, {0x8080,0x2a, 0, 0},
	{0x8081,0xe5, 0, 0}, {0x8082,0x29, 0, 0}, {0x8083,0x35, 0, 0},
	{0x8084,0x53, 0, 0}, {0x8085,0xf5, 0, 0}, {0x8086,0x29, 0, 0},
	{0x8087,0xe5, 0, 0}, {0x8088,0x28, 0, 0}, {0x8089,0x35, 0, 0},
	{0x808a,0x52, 0, 0}, {0x808b,0xf5, 0, 0}, {0x808c,0x28, 0, 0},
	{0x808d,0xae, 0, 0}, {0x808e,0x6e, 0, 0}, {0x808f,0xe4, 0, 0},
	{0x8090,0x85, 0, 0}, {0x8091,0x6f, 0, 0}, {0x8092,0x55, 0, 0},
	{0x8093,0x8e, 0, 0}, {0x8094,0x54, 0, 0}, {0x8095,0xf5, 0, 0},
	{0x8096,0x53, 0, 0}, {0x8097,0xf5, 0, 0}, {0x8098,0x52, 0, 0},
	{0x8099,0xab, 0, 0}, {0x809a,0x55, 0, 0}, {0x809b,0xaa, 0, 0},
	{0x809c,0x54, 0, 0}, {0x809d,0xa9, 0, 0}, {0x809e,0x53, 0, 0},
	{0x809f,0xa8, 0, 0}, {0x80a0,0x52, 0, 0}, {0x80a1,0xaf, 0, 0},
	{0x80a2,0x2e, 0, 0}, {0x80a3,0xfc, 0, 0}, {0x80a4,0xfd, 0, 0},
	{0x80a5,0xfe, 0, 0}, {0x80a6,0x12, 0, 0}, {0x80a7,0x08, 0, 0},
	{0x80a8,0x98, 0, 0}, {0x80a9,0x8f, 0, 0}, {0x80aa,0x55, 0, 0},
	{0x80ab,0x8e, 0, 0}, {0x80ac,0x54, 0, 0}, {0x80ad,0x8d, 0, 0},
	{0x80ae,0x53, 0, 0}, {0x80af,0x8c, 0, 0}, {0x80b0,0x52, 0, 0},
	{0x80b1,0xe5, 0, 0}, {0x80b2,0x2b, 0, 0}, {0x80b3,0x25, 0, 0},
	{0x80b4,0x55, 0, 0}, {0x80b5,0xf5, 0, 0}, {0x80b6,0x2b, 0, 0},
	{0x80b7,0xe5, 0, 0}, {0x80b8,0x2a, 0, 0}, {0x80b9,0x35, 0, 0},
	{0x80ba,0x54, 0, 0}, {0x80bb,0xf5, 0, 0}, {0x80bc,0x2a, 0, 0},
	{0x80bd,0xe5, 0, 0}, {0x80be,0x29, 0, 0}, {0x80bf,0x35, 0, 0},
	{0x80c0,0x53, 0, 0}, {0x80c1,0xf5, 0, 0}, {0x80c2,0x29, 0, 0},
	{0x80c3,0xe5, 0, 0}, {0x80c4,0x28, 0, 0}, {0x80c5,0x35, 0, 0},
	{0x80c6,0x52, 0, 0}, {0x80c7,0xf5, 0, 0}, {0x80c8,0x28, 0, 0},
	{0x80c9,0xae, 0, 0}, {0x80ca,0x70, 0, 0}, {0x80cb,0xe4, 0, 0},
	{0x80cc,0x85, 0, 0}, {0x80cd,0x71, 0, 0}, {0x80ce,0x55, 0, 0},
	{0x80cf,0x8e, 0, 0}, {0x80d0,0x54, 0, 0}, {0x80d1,0xf5, 0, 0},
	{0x80d2,0x53, 0, 0}, {0x80d3,0xf5, 0, 0}, {0x80d4,0x52, 0, 0},
	{0x80d5,0xab, 0, 0}, {0x80d6,0x55, 0, 0}, {0x80d7,0xaa, 0, 0},
	{0x80d8,0x54, 0, 0}, {0x80d9,0xa9, 0, 0}, {0x80da,0x53, 0, 0},
	{0x80db,0xa8, 0, 0}, {0x80dc,0x52, 0, 0}, {0x80dd,0xaf, 0, 0},
	{0x80de,0x2f, 0, 0}, {0x80df,0xfc, 0, 0}, {0x80e0,0xfd, 0, 0},
	{0x80e1,0xfe, 0, 0}, {0x80e2,0x12, 0, 0}, {0x80e3,0x08, 0, 0},
	{0x80e4,0x98, 0, 0}, {0x80e5,0x8f, 0, 0}, {0x80e6,0x55, 0, 0},
	{0x80e7,0x8e, 0, 0}, {0x80e8,0x54, 0, 0}, {0x80e9,0x8d, 0, 0},
	{0x80ea,0x53, 0, 0}, {0x80eb,0x8c, 0, 0}, {0x80ec,0x52, 0, 0},
	{0x80ed,0xe5, 0, 0}, {0x80ee,0x2b, 0, 0}, {0x80ef,0x25, 0, 0},
	{0x80f0,0x55, 0, 0}, {0x80f1,0xf5, 0, 0}, {0x80f2,0x2b, 0, 0},
	{0x80f3,0xe5, 0, 0}, {0x80f4,0x2a, 0, 0}, {0x80f5,0x35, 0, 0},
	{0x80f6,0x54, 0, 0}, {0x80f7,0xf5, 0, 0}, {0x80f8,0x2a, 0, 0},
	{0x80f9,0xe5, 0, 0}, {0x80fa,0x29, 0, 0}, {0x80fb,0x35, 0, 0},
	{0x80fc,0x53, 0, 0}, {0x80fd,0xf5, 0, 0}, {0x80fe,0x29, 0, 0},
	{0x80ff,0xe5, 0, 0}, {0x8100,0x28, 0, 0}, {0x8101,0x35, 0, 0},
	{0x8102,0x52, 0, 0}, {0x8103,0xf5, 0, 0}, {0x8104,0x28, 0, 0},
	{0x8105,0xae, 0, 0}, {0x8106,0x72, 0, 0}, {0x8107,0xe4, 0, 0},
	{0x8108,0x85, 0, 0}, {0x8109,0x73, 0, 0}, {0x810a,0x55, 0, 0},
	{0x810b,0x8e, 0, 0}, {0x810c,0x54, 0, 0}, {0x810d,0xf5, 0, 0},
	{0x810e,0x53, 0, 0}, {0x810f,0xf5, 0, 0}, {0x8110,0x52, 0, 0},
	{0x8111,0xab, 0, 0}, {0x8112,0x55, 0, 0}, {0x8113,0xaa, 0, 0},
	{0x8114,0x54, 0, 0}, {0x8115,0xa9, 0, 0}, {0x8116,0x53, 0, 0},
	{0x8117,0xa8, 0, 0}, {0x8118,0x52, 0, 0}, {0x8119,0xaf, 0, 0},
	{0x811a,0x30, 0, 0}, {0x811b,0xfc, 0, 0}, {0x811c,0xfd, 0, 0},
	{0x811d,0xfe, 0, 0}, {0x811e,0x12, 0, 0}, {0x811f,0x08, 0, 0},
	{0x8120,0x98, 0, 0}, {0x8121,0x8f, 0, 0}, {0x8122,0x55, 0, 0},
	{0x8123,0x8e, 0, 0}, {0x8124,0x54, 0, 0}, {0x8125,0x8d, 0, 0},
	{0x8126,0x53, 0, 0}, {0x8127,0x8c, 0, 0}, {0x8128,0x52, 0, 0},
	{0x8129,0xe5, 0, 0}, {0x812a,0x2b, 0, 0}, {0x812b,0x25, 0, 0},
	{0x812c,0x55, 0, 0}, {0x812d,0xf5, 0, 0}, {0x812e,0x2b, 0, 0},
	{0x812f,0xe5, 0, 0}, {0x8130,0x2a, 0, 0}, {0x8131,0x35, 0, 0},
	{0x8132,0x54, 0, 0}, {0x8133,0xf5, 0, 0}, {0x8134,0x2a, 0, 0},
	{0x8135,0xe5, 0, 0}, {0x8136,0x29, 0, 0}, {0x8137,0x35, 0, 0},
	{0x8138,0x53, 0, 0}, {0x8139,0xf5, 0, 0}, {0x813a,0x29, 0, 0},
	{0x813b,0xe5, 0, 0}, {0x813c,0x28, 0, 0}, {0x813d,0x35, 0, 0},
	{0x813e,0x52, 0, 0}, {0x813f,0xf5, 0, 0}, {0x8140,0x28, 0, 0},
	{0x8141,0x22, 0, 0}, {0x8142,0xab, 0, 0}, {0x8143,0x0d, 0, 0},
	{0x8144,0xaa, 0, 0}, {0x8145,0x0c, 0, 0}, {0x8146,0xa9, 0, 0},
	{0x8147,0x0b, 0, 0}, {0x8148,0xa8, 0, 0}, {0x8149,0x0a, 0, 0},
	{0x814a,0xfc, 0, 0}, {0x814b,0xfd, 0, 0}, {0x814c,0xfe, 0, 0},
	{0x814d,0x12, 0, 0}, {0x814e,0x08, 0, 0}, {0x814f,0x98, 0, 0},
	{0x8150,0x8f, 0, 0}, {0x8151,0x0d, 0, 0}, {0x8152,0x8e, 0, 0},
	{0x8153,0x0c, 0, 0}, {0x8154,0x8d, 0, 0}, {0x8155,0x0b, 0, 0},
	{0x8156,0x8c, 0, 0}, {0x8157,0x0a, 0, 0}, {0x8158,0x7b, 0, 0},
	{0x8159,0x40, 0, 0}, {0x815a,0xe4, 0, 0}, {0x815b,0xfa, 0, 0},
	{0x815c,0xf9, 0, 0}, {0x815d,0xf8, 0, 0}, {0x815e,0x12, 0, 0},
	{0x815f,0x09, 0, 0}, {0x8160,0x23, 0, 0}, {0x8161,0x8f, 0, 0},
	{0x8162,0x0d, 0, 0}, {0x8163,0x8e, 0, 0}, {0x8164,0x0c, 0, 0},
	{0x8165,0x8d, 0, 0}, {0x8166,0x0b, 0, 0}, {0x8167,0x8c, 0, 0},
	{0x8168,0x0a, 0, 0}, {0x8169,0x22, 0, 0}, {0x816a,0xd2, 0, 0},
	{0x816b,0x29, 0, 0}, {0x816c,0x90, 0, 0}, {0x816d,0x30, 0, 0},
	{0x816e,0x1b, 0, 0}, {0x816f,0xe5, 0, 0}, {0x8170,0x25, 0, 0},
	{0x8171,0xf0, 0, 0}, {0x8172,0x22, 0, 0}, {0x8173,0x85, 0, 0},
	{0x8174,0x49, 0, 0}, {0x8175,0x82, 0, 0}, {0x8176,0x85, 0, 0},
	{0x8177,0x48, 0, 0}, {0x8178,0x83, 0, 0}, {0x8179,0xe5, 0, 0},
	{0x817a,0x7c, 0, 0}, {0x817b,0x75, 0, 0}, {0x817c,0xf0, 0, 0},
	{0x817d,0x02, 0, 0}, {0x817e,0x12, 0, 0}, {0x817f,0x09, 0, 0},
	{0x8180,0xeb, 0, 0}, {0x8181,0xe4, 0, 0}, {0x8182,0x93, 0, 0},
	{0x8183,0xfe, 0, 0}, {0x8184,0x74, 0, 0}, {0x8185,0x01, 0, 0},
	{0x8186,0x93, 0, 0}, {0x8187,0xff, 0, 0}, {0x8188,0x85, 0, 0},
	{0x8189,0x49, 0, 0}, {0x818a,0x82, 0, 0}, {0x818b,0x85, 0, 0},
	{0x818c,0x48, 0, 0}, {0x818d,0x83, 0, 0}, {0x818e,0xe4, 0, 0},
	{0x818f,0x93, 0, 0}, {0x8190,0xfc, 0, 0}, {0x8191,0x74, 0, 0},
	{0x8192,0x01, 0, 0}, {0x8193,0x93, 0, 0}, {0x8194,0xfd, 0, 0},
	{0x8195,0xc3, 0, 0}, {0x8196,0xef, 0, 0}, {0x8197,0x9d, 0, 0},
	{0x8198,0xff, 0, 0}, {0x8199,0xee, 0, 0}, {0x819a,0x9c, 0, 0},
	{0x819b,0x22, 0, 0}, {0x819c,0xfe, 0, 0}, {0x819d,0xe4, 0, 0},
	{0x819e,0xfc, 0, 0}, {0x819f,0xfd, 0, 0}, {0x81a0,0xe5, 0, 0},
	{0x81a1,0x3f, 0, 0}, {0x81a2,0x2f, 0, 0}, {0x81a3,0xf5, 0, 0},
	{0x81a4,0x3f, 0, 0}, {0x81a5,0xe5, 0, 0}, {0x81a6,0x3e, 0, 0},
	{0x81a7,0x3e, 0, 0}, {0x81a8,0xf5, 0, 0}, {0x81a9,0x3e, 0, 0},
	{0x81aa,0xed, 0, 0}, {0x81ab,0x35, 0, 0}, {0x81ac,0x3d, 0, 0},
	{0x81ad,0xf5, 0, 0}, {0x81ae,0x3d, 0, 0}, {0x81af,0xec, 0, 0},
	{0x81b0,0x35, 0, 0}, {0x81b1,0x3c, 0, 0}, {0x81b2,0xf5, 0, 0},
	{0x81b3,0x3c, 0, 0}, {0x81b4,0xaf, 0, 0}, {0x81b5,0x3f, 0, 0},
	{0x81b6,0xae, 0, 0}, {0x81b7,0x3e, 0, 0}, {0x81b8,0xfc, 0, 0},
	{0x81b9,0xad, 0, 0}, {0x81ba,0x3d, 0, 0}, {0x81bb,0x78, 0, 0},
	{0x81bc,0x08, 0, 0}, {0x81bd,0x12, 0, 0}, {0x81be,0x09, 0, 0},
	{0x81bf,0xc8, 0, 0}, {0x81c0,0x8f, 0, 0}, {0x81c1,0x3f, 0, 0},
	{0x81c2,0x8e, 0, 0}, {0x81c3,0x3e, 0, 0}, {0x81c4,0x8d, 0, 0},
	{0x81c5,0x3d, 0, 0}, {0x81c6,0x8c, 0, 0}, {0x81c7,0x3c, 0, 0},
	{0x81c8,0x22, 0, 0}, {0x81c9,0xaf, 0, 0}, {0x81ca,0x2b, 0, 0},
	{0x81cb,0xae, 0, 0}, {0x81cc,0x2a, 0, 0}, {0x81cd,0xad, 0, 0},
	{0x81ce,0x29, 0, 0}, {0x81cf,0xac, 0, 0}, {0x81d0,0x28, 0, 0},
	{0x81d1,0x78, 0, 0}, {0x81d2,0x06, 0, 0}, {0x81d3,0x12, 0, 0},
	{0x81d4,0x09, 0, 0}, {0x81d5,0xb5, 0, 0}, {0x81d6,0x8f, 0, 0},
	{0x81d7,0x2b, 0, 0}, {0x81d8,0x8e, 0, 0}, {0x81d9,0x2a, 0, 0},
	{0x81da,0x8d, 0, 0}, {0x81db,0x29, 0, 0}, {0x81dc,0x8c, 0, 0},
	{0x81dd,0x28, 0, 0}, {0x81de,0xd3, 0, 0}, {0x81df,0xe5, 0, 0},
	{0x81e0,0x29, 0, 0}, {0x81e1,0x94, 0, 0}, {0x81e2,0x00, 0, 0},
	{0x81e3,0xe5, 0, 0}, {0x81e4,0x28, 0, 0}, {0x81e5,0x94, 0, 0},
	{0x81e6,0x00, 0, 0}, {0x81e7,0x22, 0, 0}, {0x81e8,0xe5, 0, 0},
	{0x81e9,0x0b, 0, 0}, {0x81ea,0x24, 0, 0}, {0x81eb,0x01, 0, 0},
	{0x81ec,0xff, 0, 0}, {0x81ed,0xe4, 0, 0}, {0x81ee,0x33, 0, 0},
	{0x81ef,0xfe, 0, 0}, {0x81f0,0x22, 0, 0}, {0x81f1,0x12, 0, 0},
	{0x81f2,0x08, 0, 0}, {0x81f3,0x98, 0, 0}, {0x81f4,0x8f, 0, 0},
	{0x81f5,0x69, 0, 0}, {0x81f6,0x8e, 0, 0}, {0x81f7,0x68, 0, 0},
	{0x81f8,0x8d, 0, 0}, {0x81f9,0x67, 0, 0}, {0x81fa,0x8c, 0, 0},
	{0x81fb,0x66, 0, 0}, {0x81fc,0xaf, 0, 0}, {0x81fd,0x69, 0, 0},
	{0x81fe,0xae, 0, 0}, {0x81ff,0x68, 0, 0}, {0x8200,0xad, 0, 0},
	{0x8201,0x67, 0, 0}, {0x8202,0xac, 0, 0}, {0x8203,0x66, 0, 0},
	{0x8204,0x22, 0, 0}, {0x8205,0xe0, 0, 0}, {0x8206,0x44, 0, 0},
	{0x8207,0x01, 0, 0}, {0x8208,0xf0, 0, 0}, {0x8209,0xe0, 0, 0},
	{0x820a,0x44, 0, 0}, {0x820b,0x02, 0, 0}, {0x820c,0xf0, 0, 0},
	{0x820d,0xe0, 0, 0}, {0x820e,0x44, 0, 0}, {0x820f,0x04, 0, 0},
	{0x8210,0xf0, 0, 0}, {0x8211,0x22, 0, 0}, {0x8212,0xd2, 0, 0},
	{0x8213,0x09, 0, 0}, {0x8214,0x90, 0, 0}, {0x8215,0x30, 0, 0},
	{0x8216,0x18, 0, 0}, {0x8217,0xe5, 0, 0}, {0x8218,0x21, 0, 0},
	{0x8219,0xf0, 0, 0}, {0x821a,0x22, 0, 0}, {0x821b,0xe4, 0, 0},
	{0x821c,0x85, 0, 0}, {0x821d,0x11, 0, 0}, {0x821e,0x0d, 0, 0},
	{0x821f,0x85, 0, 0}, {0x8220,0x10, 0, 0}, {0x8221,0x0c, 0, 0},
	{0x8222,0xf5, 0, 0}, {0x8223,0x0b, 0, 0}, {0x8224,0xf5, 0, 0},
	{0x8225,0x0a, 0, 0}, {0x8226,0xab, 0, 0}, {0x8227,0x0d, 0, 0},
	{0x8228,0xaa, 0, 0}, {0x8229,0x0c, 0, 0}, {0x822a,0xa9, 0, 0},
	{0x822b,0x0b, 0, 0}, {0x822c,0xa8, 0, 0}, {0x822d,0x0a, 0, 0},
	{0x822e,0x22, 0, 0}, {0x822f,0x90, 0, 0}, {0x8230,0x30, 0, 0},
	{0x8231,0x42, 0, 0}, {0x8232,0xe0, 0, 0}, {0x8233,0xf5, 0, 0},
	{0x8234,0x22, 0, 0}, {0x8235,0x75, 0, 0}, {0x8236,0x51, 0, 0},
	{0x8237,0x0a, 0, 0}, {0x8238,0x22, 0, 0}, {0x8239,0x85, 0, 0},
	{0x823a,0x49, 0, 0}, {0x823b,0x82, 0, 0}, {0x823c,0x85, 0, 0},
	{0x823d,0x48, 0, 0}, {0x823e,0x83, 0, 0}, {0x823f,0x22, 0, 0},
	{0x8240,0xf5, 0, 0}, {0x8241,0x82, 0, 0}, {0x8242,0xe4, 0, 0},
	{0x8243,0x3a, 0, 0}, {0x8244,0xf5, 0, 0}, {0x8245,0x83, 0, 0},
	{0x8246,0x02, 0, 0}, {0x8247,0x09, 0, 0}, {0x8248,0xdb, 0, 0},
	{0x8249,0x8f, 0, 0}, {0x824a,0x0a, 0, 0}, {0x824b,0x74, 0, 0},
	{0x824c,0x4a, 0, 0}, {0x824d,0x2f, 0, 0}, {0x824e,0xf8, 0, 0},
	{0x824f,0xe6, 0, 0}, {0x8250,0x22, 0, 0}, {0x8251,0xc2, 0, 0},
	{0x8252,0x07, 0, 0}, {0x8253,0xc2, 0, 0}, {0x8254,0x06, 0, 0},
	{0x8255,0xc2, 0, 0}, {0x8256,0x02, 0, 0}, {0x8257,0xc2, 0, 0},
	{0x8258,0x01, 0, 0}, {0x8259,0xc2, 0, 0}, {0x825a,0x00, 0, 0},
	{0x825b,0xc2, 0, 0}, {0x825c,0x03, 0, 0}, {0x825d,0xd2, 0, 0},
	{0x825e,0x04, 0, 0}, {0x825f,0x22, 0, 0}, {0x8260,0xe5, 0, 0},
	{0x8261,0x16, 0, 0}, {0x8262,0x25, 0, 0}, {0x8263,0xe0, 0, 0},
	{0x8264,0x25, 0, 0}, {0x8265,0xe0, 0, 0}, {0x8266,0x22, 0, 0},
	{0x8267,0x12, 0, 0}, {0x8268,0x09, 0, 0}, {0x8269,0x23, 0, 0},
	{0x826a,0x8f, 0, 0}, {0x826b,0x69, 0, 0}, {0x826c,0x8e, 0, 0},
	{0x826d,0x68, 0, 0}, {0x826e,0x8d, 0, 0}, {0x826f,0x67, 0, 0},
	{0x8270,0x8c, 0, 0}, {0x8271,0x66, 0, 0}, {0x8272,0x22, 0, 0},
	{0x8273,0xe4, 0, 0}, {0x8274,0x85, 0, 0}, {0x8275,0x0f, 0, 0},
	{0x8276,0x0d, 0, 0}, {0x8277,0x85, 0, 0}, {0x8278,0x0e, 0, 0},
	{0x8279,0x0c, 0, 0}, {0x827a,0xf5, 0, 0}, {0x827b,0x0b, 0, 0},
	{0x827c,0xf5, 0, 0}, {0x827d,0x0a, 0, 0}, {0x827e,0x22, 0, 0},
	{0x827f,0x90, 0, 0}, {0x8280,0x11, 0, 0}, {0x8281,0x67, 0, 0},
	{0x8282,0xe4, 0, 0}, {0x8283,0x93, 0, 0}, {0x8284,0xff, 0, 0},
	{0x8285,0x90, 0, 0}, {0x8286,0x30, 0, 0}, {0x8287,0x0a, 0, 0},
	{0x8288,0xe0, 0, 0}, {0x8289,0x22, 0, 0}, {0x828a,0xc2, 0, 0},
	{0x828b,0x02, 0, 0}, {0x828c,0xc2, 0, 0}, {0x828d,0x01, 0, 0},
	{0x828e,0xd2, 0, 0}, {0x828f,0x00, 0, 0}, {0x8290,0xc2, 0, 0},
	{0x8291,0x03, 0, 0}, {0x8292,0xc2, 0, 0}, {0x8293,0x04, 0, 0},
	{0x8294,0x22, 0, 0}, {0x8295,0x75, 0, 0}, {0x8296,0xf0, 0, 0},
	{0x8297,0x02, 0, 0}, {0x8298,0x02, 0, 0}, {0x8299,0x09, 0, 0},
	{0x829a,0xeb, 0, 0}, {0x829b,0xd2, 0, 0}, {0x829c,0x02, 0, 0},
	{0x829d,0xd2, 0, 0}, {0x829e,0x01, 0, 0}, {0x829f,0xc2, 0, 0},
	{0x82a0,0x00, 0, 0}, {0x82a1,0x22, 0, 0}, {0x82a2,0x74, 0, 0},
	{0x82a3,0x4a, 0, 0}, {0x82a4,0x25, 0, 0}, {0x82a5,0x0a, 0, 0},
	{0x82a6,0xf8, 0, 0}, {0x82a7,0xe6, 0, 0}, {0x82a8,0x22, 0, 0},
	{0x82a9,0xd3, 0, 0}, {0x82aa,0xe5, 0, 0}, {0x82ab,0x0b, 0, 0},
	{0x82ac,0x94, 0, 0}, {0x82ad,0xff, 0, 0}, {0x82ae,0xe5, 0, 0},
	{0x82af,0x0a, 0, 0}, {0x82b0,0x94, 0, 0}, {0x82b1,0x00, 0, 0},
	{0x82b2,0x22, 0, 0}, {0x82b3,0xd3, 0, 0}, {0x82b4,0xe5, 0, 0},
	{0x82b5,0x69, 0, 0}, {0x82b6,0x94, 0, 0}, {0x82b7,0xff, 0, 0},
	{0x82b8,0xe5, 0, 0}, {0x82b9,0x68, 0, 0}, {0x82ba,0x94, 0, 0},
	{0x82bb,0x03, 0, 0}, {0x82bc,0x22, 0, 0}, {0x82bd,0x30, 0, 0},
	{0x82be,0x18, 0, 0}, {0x82bf,0x4d, 0, 0}, {0x82c0,0x20, 0, 0},
	{0x82c1,0x19, 0, 0}, {0x82c2,0x4a, 0, 0}, {0x82c3,0x75, 0, 0},
	{0x82c4,0x0a, 0, 0}, {0x82c5,0x02, 0, 0}, {0x82c6,0x12, 0, 0},
	{0x82c7,0x02, 0, 0}, {0x82c8,0xa2, 0, 0}, {0x82c9,0xff, 0, 0},
	{0x82ca,0xe5, 0, 0}, {0x82cb,0x4a, 0, 0}, {0x82cc,0xd3, 0, 0},
	{0x82cd,0x9f, 0, 0}, {0x82ce,0x40, 0, 0}, {0x82cf,0x04, 0, 0},
	{0x82d0,0x7f, 0, 0}, {0x82d1,0x00, 0, 0}, {0x82d2,0x80, 0, 0},
	{0x82d3,0x02, 0, 0}, {0x82d4,0xaf, 0, 0}, {0x82d5,0x0a, 0, 0},
	{0x82d6,0x12, 0, 0}, {0x82d7,0x02, 0, 0}, {0x82d8,0x49, 0, 0},
	{0x82d9,0xff, 0, 0}, {0x82da,0xe5, 0, 0}, {0x82db,0x4b, 0, 0},
	{0x82dc,0xd3, 0, 0}, {0x82dd,0x9f, 0, 0}, {0x82de,0x40, 0, 0},
	{0x82df,0x04, 0, 0}, {0x82e0,0x7f, 0, 0}, {0x82e1,0x01, 0, 0},
	{0x82e2,0x80, 0, 0}, {0x82e3,0x02, 0, 0}, {0x82e4,0xaf, 0, 0},
	{0x82e5,0x0a, 0, 0}, {0x82e6,0x12, 0, 0}, {0x82e7,0x02, 0, 0},
	{0x82e8,0x49, 0, 0}, {0x82e9,0xff, 0, 0}, {0x82ea,0xe5, 0, 0},
	{0x82eb,0x4d, 0, 0}, {0x82ec,0xd3, 0, 0}, {0x82ed,0x9f, 0, 0},
	{0x82ee,0x40, 0, 0}, {0x82ef,0x04, 0, 0}, {0x82f0,0x7f, 0, 0},
	{0x82f1,0x03, 0, 0}, {0x82f2,0x80, 0, 0}, {0x82f3,0x02, 0, 0},
	{0x82f4,0xaf, 0, 0}, {0x82f5,0x0a, 0, 0}, {0x82f6,0x12, 0, 0},
	{0x82f7,0x02, 0, 0}, {0x82f8,0x49, 0, 0}, {0x82f9,0xff, 0, 0},
	{0x82fa,0xe5, 0, 0}, {0x82fb,0x4e, 0, 0}, {0x82fc,0xd3, 0, 0},
	{0x82fd,0x9f, 0, 0}, {0x82fe,0x40, 0, 0}, {0x82ff,0x04, 0, 0},
	{0x8300,0x7f, 0, 0}, {0x8301,0x04, 0, 0}, {0x8302,0x80, 0, 0},
	{0x8303,0x02, 0, 0}, {0x8304,0xaf, 0, 0}, {0x8305,0x0a, 0, 0},
	{0x8306,0x12, 0, 0}, {0x8307,0x02, 0, 0}, {0x8308,0x49, 0, 0},
	{0x8309,0xf5, 0, 0}, {0x830a,0x0b, 0, 0}, {0x830b,0x80, 0, 0},
	{0x830c,0x06, 0, 0}, {0x830d,0x85, 0, 0}, {0x830e,0x79, 0, 0},
	{0x830f,0x0a, 0, 0}, {0x8310,0x85, 0, 0}, {0x8311,0x4f, 0, 0},
	{0x8312,0x0b, 0, 0}, {0x8313,0x7f, 0, 0}, {0x8314,0x01, 0, 0},
	{0x8315,0xe4, 0, 0}, {0x8316,0xfe, 0, 0}, {0x8317,0x12, 0, 0},
	{0x8318,0x02, 0, 0}, {0x8319,0xa2, 0, 0}, {0x831a,0xfd, 0, 0},
	{0x831b,0xe5, 0, 0}, {0x831c,0x0b, 0, 0}, {0x831d,0xc3, 0, 0},
	{0x831e,0x9d, 0, 0}, {0x831f,0x50, 0, 0}, {0x8320,0x04, 0, 0},
	{0x8321,0x7d, 0, 0}, {0x8322,0x01, 0, 0}, {0x8323,0x80, 0, 0},
	{0x8324,0x02, 0, 0}, {0x8325,0x7d, 0, 0}, {0x8326,0xff, 0, 0},
	{0x8327,0xac, 0, 0}, {0x8328,0x0b, 0, 0}, {0x8329,0xe5, 0, 0},
	{0x832a,0x4e, 0, 0}, {0x832b,0xb5, 0, 0}, {0x832c,0x0b, 0, 0},
	{0x832d,0x03, 0, 0}, {0x832e,0xd3, 0, 0}, {0x832f,0x80, 0, 0},
	{0x8330,0x01, 0, 0}, {0x8331,0xc3, 0, 0}, {0x8332,0x92, 0, 0},
	{0x8333,0x1f, 0, 0}, {0x8334,0xe5, 0, 0}, {0x8335,0x4d, 0, 0},
	{0x8336,0xb5, 0, 0}, {0x8337,0x0b, 0, 0}, {0x8338,0x03, 0, 0},
	{0x8339,0xd3, 0, 0}, {0x833a,0x80, 0, 0}, {0x833b,0x01, 0, 0},
	{0x833c,0xc3, 0, 0}, {0x833d,0x92, 0, 0}, {0x833e,0x1e, 0, 0},
	{0x833f,0xe5, 0, 0}, {0x8340,0x4c, 0, 0}, {0x8341,0xb5, 0, 0},
	{0x8342,0x0b, 0, 0}, {0x8343,0x03, 0, 0}, {0x8344,0xd3, 0, 0},
	{0x8345,0x80, 0, 0}, {0x8346,0x01, 0, 0}, {0x8347,0xc3, 0, 0},
	{0x8348,0x92, 0, 0}, {0x8349,0x1d, 0, 0}, {0x834a,0xe5, 0, 0},
	{0x834b,0x4b, 0, 0}, {0x834c,0xb5, 0, 0}, {0x834d,0x0b, 0, 0},
	{0x834e,0x03, 0, 0}, {0x834f,0xd3, 0, 0}, {0x8350,0x80, 0, 0},
	{0x8351,0x01, 0, 0}, {0x8352,0xc3, 0, 0}, {0x8353,0x92, 0, 0},
	{0x8354,0x1c, 0, 0}, {0x8355,0xe5, 0, 0}, {0x8356,0x4a, 0, 0},
	{0x8357,0xb5, 0, 0}, {0x8358,0x0b, 0, 0}, {0x8359,0x03, 0, 0},
	{0x835a,0xd3, 0, 0}, {0x835b,0x80, 0, 0}, {0x835c,0x01, 0, 0},
	{0x835d,0xc3, 0, 0}, {0x835e,0x92, 0, 0}, {0x835f,0x1b, 0, 0},
	{0x8360,0xe5, 0, 0}, {0x8361,0x30, 0, 0}, {0x8362,0xd3, 0, 0},
	{0x8363,0x94, 0, 0}, {0x8364,0x00, 0, 0}, {0x8365,0x40, 0, 0},
	{0x8366,0x04, 0, 0}, {0x8367,0xa2, 0, 0}, {0x8368,0x1f, 0, 0},
	{0x8369,0x80, 0, 0}, {0x836a,0x01, 0, 0}, {0x836b,0xc3, 0, 0},
	{0x836c,0x92, 0, 0}, {0x836d,0x1f, 0, 0}, {0x836e,0xe5, 0, 0},
	{0x836f,0x2f, 0, 0}, {0x8370,0xd3, 0, 0}, {0x8371,0x94, 0, 0},
	{0x8372,0x00, 0, 0}, {0x8373,0x40, 0, 0}, {0x8374,0x04, 0, 0},
	{0x8375,0xa2, 0, 0}, {0x8376,0x1e, 0, 0}, {0x8377,0x80, 0, 0},
	{0x8378,0x01, 0, 0}, {0x8379,0xc3, 0, 0}, {0x837a,0x92, 0, 0},
	{0x837b,0x1e, 0, 0}, {0x837c,0xe5, 0, 0}, {0x837d,0x2e, 0, 0},
	{0x837e,0xd3, 0, 0}, {0x837f,0x94, 0, 0}, {0x8380,0x00, 0, 0},
	{0x8381,0x40, 0, 0}, {0x8382,0x04, 0, 0}, {0x8383,0xa2, 0, 0},
	{0x8384,0x1d, 0, 0}, {0x8385,0x80, 0, 0}, {0x8386,0x01, 0, 0},
	{0x8387,0xc3, 0, 0}, {0x8388,0x92, 0, 0}, {0x8389,0x1d, 0, 0},
	{0x838a,0xe5, 0, 0}, {0x838b,0x2d, 0, 0}, {0x838c,0xd3, 0, 0},
	{0x838d,0x94, 0, 0}, {0x838e,0x00, 0, 0}, {0x838f,0x40, 0, 0},
	{0x8390,0x04, 0, 0}, {0x8391,0xa2, 0, 0}, {0x8392,0x1c, 0, 0},
	{0x8393,0x80, 0, 0}, {0x8394,0x01, 0, 0}, {0x8395,0xc3, 0, 0},
	{0x8396,0x92, 0, 0}, {0x8397,0x1c, 0, 0}, {0x8398,0xe5, 0, 0},
	{0x8399,0x2c, 0, 0}, {0x839a,0xd3, 0, 0}, {0x839b,0x94, 0, 0},
	{0x839c,0x00, 0, 0}, {0x839d,0x40, 0, 0}, {0x839e,0x04, 0, 0},
	{0x839f,0xa2, 0, 0}, {0x83a0,0x1b, 0, 0}, {0x83a1,0x80, 0, 0},
	{0x83a2,0x01, 0, 0}, {0x83a3,0xc3, 0, 0}, {0x83a4,0x92, 0, 0},
	{0x83a5,0x1b, 0, 0}, {0x83a6,0xe5, 0, 0}, {0x83a7,0x23, 0, 0},
	{0x83a8,0x54, 0, 0}, {0x83a9,0xf8, 0, 0}, {0x83aa,0x70, 0, 0},
	{0x83ab,0x5b, 0, 0}, {0x83ac,0xbf, 0, 0}, {0x83ad,0x01, 0, 0},
	{0x83ae,0x08, 0, 0}, {0x83af,0xed, 0, 0}, {0x83b0,0xf4, 0, 0},
	{0x83b1,0x04, 0, 0}, {0x83b2,0xfd, 0, 0}, {0x83b3,0x7f, 0, 0},
	{0x83b4,0x02, 0, 0}, {0x83b5,0x80, 0, 0}, {0x83b6,0x06, 0, 0},
	{0x83b7,0xbf, 0, 0}, {0x83b8,0x02, 0, 0}, {0x83b9,0x02, 0, 0},
	{0x83ba,0x7f, 0, 0}, {0x83bb,0x01, 0, 0}, {0x83bc,0x0e, 0, 0},
	{0x83bd,0xd3, 0, 0}, {0x83be,0xed, 0, 0}, {0x83bf,0x64, 0, 0},
	{0x83c0,0x80, 0, 0}, {0x83c1,0x94, 0, 0}, {0x83c2,0x80, 0, 0},
	{0x83c3,0x40, 0, 0}, {0x83c4,0x18, 0, 0}, {0x83c5,0xec, 0, 0},
	{0x83c6,0x2e, 0, 0}, {0x83c7,0xf5, 0, 0}, {0x83c8,0x0b, 0, 0},
	{0x83c9,0xd3, 0, 0}, {0x83ca,0x95, 0, 0}, {0x83cb,0x7c, 0, 0},
	{0x83cc,0x40, 0, 0}, {0x83cd,0x2c, 0, 0}, {0x83ce,0x7d, 0, 0},
	{0x83cf,0xff, 0, 0}, {0x83d0,0xef, 0, 0}, {0x83d1,0x60, 0, 0},
	{0x83d2,0x04, 0, 0}, {0x83d3,0x7b, 0, 0}, {0x83d4,0x00, 0, 0},
	{0x83d5,0x80, 0, 0}, {0x83d6,0x02, 0, 0}, {0x83d7,0x7b, 0, 0},
	{0x83d8,0x03, 0, 0}, {0x83d9,0xaf, 0, 0}, {0x83da,0x03, 0, 0},
	{0x83db,0x80, 0, 0}, {0x83dc,0x18, 0, 0}, {0x83dd,0xec, 0, 0},
	{0x83de,0xc3, 0, 0}, {0x83df,0x9e, 0, 0}, {0x83e0,0x50, 0, 0},
	{0x83e1,0x13, 0, 0}, {0x83e2,0x7d, 0, 0}, {0x83e3,0x01, 0, 0},
	{0x83e4,0xef, 0, 0}, {0x83e5,0x60, 0, 0}, {0x83e6,0x04, 0, 0},
	{0x83e7,0x7b, 0, 0}, {0x83e8,0x00, 0, 0}, {0x83e9,0x80, 0, 0},
	{0x83ea,0x02, 0, 0}, {0x83eb,0x7b, 0, 0}, {0x83ec,0x03, 0, 0},
	{0x83ed,0xaf, 0, 0}, {0x83ee,0x03, 0, 0}, {0x83ef,0xec, 0, 0},
	{0x83f0,0x2e, 0, 0}, {0x83f1,0xf5, 0, 0}, {0x83f2,0x0b, 0, 0},
	{0x83f3,0x80, 0, 0}, {0x83f4,0x05, 0, 0}, {0x83f5,0xc3, 0, 0},
	{0x83f6,0xec, 0, 0}, {0x83f7,0x9e, 0, 0}, {0x83f8,0xf5, 0, 0},
	{0x83f9,0x0b, 0, 0}, {0x83fa,0xef, 0, 0}, {0x83fb,0x64, 0, 0},
	{0x83fc,0x03, 0, 0}, {0x83fd,0x60, 0, 0}, {0x83fe,0x03, 0, 0},
	{0x83ff,0x02, 0, 0}, {0x8400,0x03, 0, 0}, {0x8401,0x29, 0, 0},
	{0x8402,0x12, 0, 0}, {0x8403,0x02, 0, 0}, {0x8404,0xa2, 0, 0},
	{0x8405,0xf5, 0, 0}, {0x8406,0x0b, 0, 0}, {0x8407,0x12, 0, 0},
	{0x8408,0x01, 0, 0}, {0x8409,0xe8, 0, 0}, {0x840a,0xe5, 0, 0},
	{0x840b,0x4e, 0, 0}, {0x840c,0xb5, 0, 0}, {0x840d,0x07, 0, 0},
	{0x840e,0x07, 0, 0}, {0x840f,0xe4, 0, 0}, {0x8410,0xb5, 0, 0},
	{0x8411,0x06, 0, 0}, {0x8412,0x03, 0, 0}, {0x8413,0xd3, 0, 0},
	{0x8414,0x80, 0, 0}, {0x8415,0x02, 0, 0}, {0x8416,0xa2, 0, 0},
	{0x8417,0x1f, 0, 0}, {0x8418,0x92, 0, 0}, {0x8419,0x1f, 0, 0},
	{0x841a,0xe5, 0, 0}, {0x841b,0x4d, 0, 0}, {0x841c,0xb5, 0, 0},
	{0x841d,0x07, 0, 0}, {0x841e,0x07, 0, 0}, {0x841f,0xe4, 0, 0},
	{0x8420,0xb5, 0, 0}, {0x8421,0x06, 0, 0}, {0x8422,0x03, 0, 0},
	{0x8423,0xd3, 0, 0}, {0x8424,0x80, 0, 0}, {0x8425,0x02, 0, 0},
	{0x8426,0xa2, 0, 0}, {0x8427,0x1e, 0, 0}, {0x8428,0x92, 0, 0},
	{0x8429,0x1e, 0, 0}, {0x842a,0x12, 0, 0}, {0x842b,0x01, 0, 0},
	{0x842c,0xe8, 0, 0}, {0x842d,0xe5, 0, 0}, {0x842e,0x4c, 0, 0},
	{0x842f,0xb5, 0, 0}, {0x8430,0x07, 0, 0}, {0x8431,0x07, 0, 0},
	{0x8432,0xe4, 0, 0}, {0x8433,0xb5, 0, 0}, {0x8434,0x06, 0, 0},
	{0x8435,0x03, 0, 0}, {0x8436,0xd3, 0, 0}, {0x8437,0x80, 0, 0},
	{0x8438,0x02, 0, 0}, {0x8439,0xa2, 0, 0}, {0x843a,0x1d, 0, 0},
	{0x843b,0x92, 0, 0}, {0x843c,0x1d, 0, 0}, {0x843d,0xe5, 0, 0},
	{0x843e,0x4b, 0, 0}, {0x843f,0xb5, 0, 0}, {0x8440,0x07, 0, 0},
	{0x8441,0x07, 0, 0}, {0x8442,0xe4, 0, 0}, {0x8443,0xb5, 0, 0},
	{0x8444,0x06, 0, 0}, {0x8445,0x03, 0, 0}, {0x8446,0xd3, 0, 0},
	{0x8447,0x80, 0, 0}, {0x8448,0x02, 0, 0}, {0x8449,0xa2, 0, 0},
	{0x844a,0x1c, 0, 0}, {0x844b,0x92, 0, 0}, {0x844c,0x1c, 0, 0},
	{0x844d,0x12, 0, 0}, {0x844e,0x01, 0, 0}, {0x844f,0xe8, 0, 0},
	{0x8450,0xe5, 0, 0}, {0x8451,0x4a, 0, 0}, {0x8452,0xb5, 0, 0},
	{0x8453,0x07, 0, 0}, {0x8454,0x07, 0, 0}, {0x8455,0xe4, 0, 0},
	{0x8456,0xb5, 0, 0}, {0x8457,0x06, 0, 0}, {0x8458,0x03, 0, 0},
	{0x8459,0xd3, 0, 0}, {0x845a,0x80, 0, 0}, {0x845b,0x02, 0, 0},
	{0x845c,0xa2, 0, 0}, {0x845d,0x1b, 0, 0}, {0x845e,0x92, 0, 0},
	{0x845f,0x1b, 0, 0}, {0x8460,0xe5, 0, 0}, {0x8461,0x4e, 0, 0},
	{0x8462,0x12, 0, 0}, {0x8463,0x01, 0, 0}, {0x8464,0xea, 0, 0},
	{0x8465,0xad, 0, 0}, {0x8466,0x0b, 0, 0}, {0x8467,0x7c, 0, 0},
	{0x8468,0x00, 0, 0}, {0x8469,0xef, 0, 0}, {0x846a,0xb5, 0, 0},
	{0x846b,0x05, 0, 0}, {0x846c,0x07, 0, 0}, {0x846d,0xec, 0, 0},
	{0x846e,0xb5, 0, 0}, {0x846f,0x06, 0, 0}, {0x8470,0x03, 0, 0},
	{0x8471,0xd3, 0, 0}, {0x8472,0x80, 0, 0}, {0x8473,0x02, 0, 0},
	{0x8474,0xa2, 0, 0}, {0x8475,0x1f, 0, 0}, {0x8476,0x92, 0, 0},
	{0x8477,0x1f, 0, 0}, {0x8478,0xe5, 0, 0}, {0x8479,0x4d, 0, 0},
	{0x847a,0x12, 0, 0}, {0x847b,0x01, 0, 0}, {0x847c,0xea, 0, 0},
	{0x847d,0xef, 0, 0}, {0x847e,0xb5, 0, 0}, {0x847f,0x05, 0, 0},
	{0x8480,0x07, 0, 0}, {0x8481,0xee, 0, 0}, {0x8482,0xb5, 0, 0},
	{0x8483,0x04, 0, 0}, {0x8484,0x03, 0, 0}, {0x8485,0xd3, 0, 0},
	{0x8486,0x80, 0, 0}, {0x8487,0x02, 0, 0}, {0x8488,0xa2, 0, 0},
	{0x8489,0x1e, 0, 0}, {0x848a,0x92, 0, 0}, {0x848b,0x1e, 0, 0},
	{0x848c,0xe5, 0, 0}, {0x848d,0x4c, 0, 0}, {0x848e,0x12, 0, 0},
	{0x848f,0x01, 0, 0}, {0x8490,0xea, 0, 0}, {0x8491,0xad, 0, 0},
	{0x8492,0x0b, 0, 0}, {0x8493,0x7c, 0, 0}, {0x8494,0x00, 0, 0},
	{0x8495,0xef, 0, 0}, {0x8496,0xb5, 0, 0}, {0x8497,0x05, 0, 0},
	{0x8498,0x07, 0, 0}, {0x8499,0xec, 0, 0}, {0x849a,0xb5, 0, 0},
	{0x849b,0x06, 0, 0}, {0x849c,0x03, 0, 0}, {0x849d,0xd3, 0, 0},
	{0x849e,0x80, 0, 0}, {0x849f,0x02, 0, 0}, {0x84a0,0xa2, 0, 0},
	{0x84a1,0x1d, 0, 0}, {0x84a2,0x92, 0, 0}, {0x84a3,0x1d, 0, 0},
	{0x84a4,0xe5, 0, 0}, {0x84a5,0x4b, 0, 0}, {0x84a6,0x12, 0, 0},
	{0x84a7,0x01, 0, 0}, {0x84a8,0xea, 0, 0}, {0x84a9,0xef, 0, 0},
	{0x84aa,0xb5, 0, 0}, {0x84ab,0x05, 0, 0}, {0x84ac,0x07, 0, 0},
	{0x84ad,0xee, 0, 0}, {0x84ae,0xb5, 0, 0}, {0x84af,0x04, 0, 0},
	{0x84b0,0x03, 0, 0}, {0x84b1,0xd3, 0, 0}, {0x84b2,0x80, 0, 0},
	{0x84b3,0x02, 0, 0}, {0x84b4,0xa2, 0, 0}, {0x84b5,0x1c, 0, 0},
	{0x84b6,0x92, 0, 0}, {0x84b7,0x1c, 0, 0}, {0x84b8,0xe5, 0, 0},
	{0x84b9,0x4a, 0, 0}, {0x84ba,0x12, 0, 0}, {0x84bb,0x01, 0, 0},
	{0x84bc,0xea, 0, 0}, {0x84bd,0x7c, 0, 0}, {0x84be,0x00, 0, 0},
	{0x84bf,0xef, 0, 0}, {0x84c0,0xb5, 0, 0}, {0x84c1,0x0b, 0, 0},
	{0x84c2,0x07, 0, 0}, {0x84c3,0xec, 0, 0}, {0x84c4,0xb5, 0, 0},
	{0x84c5,0x06, 0, 0}, {0x84c6,0x03, 0, 0}, {0x84c7,0xd3, 0, 0},
	{0x84c8,0x80, 0, 0}, {0x84c9,0x02, 0, 0}, {0x84ca,0xa2, 0, 0},
	{0x84cb,0x1b, 0, 0}, {0x84cc,0x92, 0, 0}, {0x84cd,0x1b, 0, 0},
	{0x84ce,0xe5, 0, 0}, {0x84cf,0x30, 0, 0}, {0x84d0,0xd3, 0, 0},
	{0x84d1,0x94, 0, 0}, {0x84d2,0x00, 0, 0}, {0x84d3,0x40, 0, 0},
	{0x84d4,0x04, 0, 0}, {0x84d5,0xa2, 0, 0}, {0x84d6,0x1f, 0, 0},
	{0x84d7,0x80, 0, 0}, {0x84d8,0x01, 0, 0}, {0x84d9,0xc3, 0, 0},
	{0x84da,0x92, 0, 0}, {0x84db,0x1f, 0, 0}, {0x84dc,0xe5, 0, 0},
	{0x84dd,0x2f, 0, 0}, {0x84de,0xd3, 0, 0}, {0x84df,0x94, 0, 0},
	{0x84e0,0x00, 0, 0}, {0x84e1,0x40, 0, 0}, {0x84e2,0x04, 0, 0},
	{0x84e3,0xa2, 0, 0}, {0x84e4,0x1e, 0, 0}, {0x84e5,0x80, 0, 0},
	{0x84e6,0x01, 0, 0}, {0x84e7,0xc3, 0, 0}, {0x84e8,0x92, 0, 0},
	{0x84e9,0x1e, 0, 0}, {0x84ea,0xe5, 0, 0}, {0x84eb,0x2e, 0, 0},
	{0x84ec,0xd3, 0, 0}, {0x84ed,0x94, 0, 0}, {0x84ee,0x00, 0, 0},
	{0x84ef,0x40, 0, 0}, {0x84f0,0x04, 0, 0}, {0x84f1,0xa2, 0, 0},
	{0x84f2,0x1d, 0, 0}, {0x84f3,0x80, 0, 0}, {0x84f4,0x01, 0, 0},
	{0x84f5,0xc3, 0, 0}, {0x84f6,0x92, 0, 0}, {0x84f7,0x1d, 0, 0},
	{0x84f8,0xe5, 0, 0}, {0x84f9,0x2d, 0, 0}, {0x84fa,0xd3, 0, 0},
	{0x84fb,0x94, 0, 0}, {0x84fc,0x00, 0, 0}, {0x84fd,0x40, 0, 0},
	{0x84fe,0x04, 0, 0}, {0x84ff,0xa2, 0, 0}, {0x8500,0x1c, 0, 0},
	{0x8501,0x80, 0, 0}, {0x8502,0x01, 0, 0}, {0x8503,0xc3, 0, 0},
	{0x8504,0x92, 0, 0}, {0x8505,0x1c, 0, 0}, {0x8506,0xe5, 0, 0},
	{0x8507,0x2c, 0, 0}, {0x8508,0xd3, 0, 0}, {0x8509,0x94, 0, 0},
	{0x850a,0x00, 0, 0}, {0x850b,0x40, 0, 0}, {0x850c,0x04, 0, 0},
	{0x850d,0xa2, 0, 0}, {0x850e,0x1b, 0, 0}, {0x850f,0x80, 0, 0},
	{0x8510,0x01, 0, 0}, {0x8511,0xc3, 0, 0}, {0x8512,0x92, 0, 0},
	{0x8513,0x1b, 0, 0}, {0x8514,0x85, 0, 0}, {0x8515,0x0a, 0, 0},
	{0x8516,0x79, 0, 0}, {0x8517,0xe5, 0, 0}, {0x8518,0x7d, 0, 0},
	{0x8519,0xb5, 0, 0}, {0x851a,0x0b, 0, 0}, {0x851b,0x03, 0, 0},
	{0x851c,0x02, 0, 0}, {0x851d,0x16, 0, 0}, {0x851e,0x5f, 0, 0},
	{0x851f,0x85, 0, 0}, {0x8520,0x0b, 0, 0}, {0x8521,0x0c, 0, 0},
	{0x8522,0x12, 0, 0}, {0x8523,0x13, 0, 0}, {0x8524,0x64, 0, 0},
	{0x8525,0xd2, 0, 0}, {0x8526,0x02, 0, 0}, {0x8527,0xc2, 0, 0},
	{0x8528,0x01, 0, 0}, {0x8529,0xd2, 0, 0}, {0x852a,0x00, 0, 0},
	{0x852b,0x75, 0, 0}, {0x852c,0x31, 0, 0}, {0x852d,0x03, 0, 0},
	{0x852e,0x22, 0, 0}, {0x852f,0xe5, 0, 0}, {0x8530,0x7e, 0, 0},
	{0x8531,0x24, 0, 0}, {0x8532,0xfe, 0, 0}, {0x8533,0x60, 0, 0},
	{0x8534,0x27, 0, 0}, {0x8535,0x14, 0, 0}, {0x8536,0x60, 0, 0},
	{0x8537,0x31, 0, 0}, {0x8538,0x24, 0, 0}, {0x8539,0xf8, 0, 0},
	{0x853a,0x60, 0, 0}, {0x853b,0x3a, 0, 0}, {0x853c,0x14, 0, 0},
	{0x853d,0x60, 0, 0}, {0x853e,0x4b, 0, 0}, {0x853f,0x14, 0, 0},
	{0x8540,0x60, 0, 0}, {0x8541,0x59, 0, 0}, {0x8542,0x14, 0, 0},
	{0x8543,0x60, 0, 0}, {0x8544,0x6a, 0, 0}, {0x8545,0x24, 0, 0},
	{0x8546,0xfd, 0, 0}, {0x8547,0x70, 0, 0}, {0x8548,0x03, 0, 0},
	{0x8549,0x02, 0, 0}, {0x854a,0x05, 0, 0}, {0x854b,0xe4, 0, 0},
	{0x854c,0x24, 0, 0}, {0x854d,0x10, 0, 0}, {0x854e,0x60, 0, 0},
	{0x854f,0x03, 0, 0}, {0x8550,0x02, 0, 0}, {0x8551,0x06, 0, 0},
	{0x8552,0xdf, 0, 0}, {0x8553,0xe4, 0, 0}, {0x8554,0xf5, 0, 0},
	{0x8555,0x0a, 0, 0}, {0x8556,0x12, 0, 0}, {0x8557,0x12, 0, 0},
	{0x8558,0x9c, 0, 0}, {0x8559,0x02, 0, 0}, {0x855a,0x06, 0, 0},
	{0x855b,0xc8, 0, 0}, {0x855c,0x75, 0, 0}, {0x855d,0x0a, 0, 0},
	{0x855e,0x01, 0, 0}, {0x855f,0x12, 0, 0}, {0x8560,0x06, 0, 0},
	{0x8561,0xe0, 0, 0}, {0x8562,0xc2, 0, 0}, {0x8563,0x3c, 0, 0},
	{0x8564,0xd2, 0, 0}, {0x8565,0x3b, 0, 0}, {0x8566,0x02, 0, 0},
	{0x8567,0x06, 0, 0}, {0x8568,0xd9, 0, 0}, {0x8569,0x75, 0, 0},
	{0x856a,0x0a, 0, 0}, {0x856b,0x02, 0, 0}, {0x856c,0x12, 0, 0},
	{0x856d,0x06, 0, 0}, {0x856e,0xe0, 0, 0}, {0x856f,0xd2, 0, 0},
	{0x8570,0x3c, 0, 0}, {0x8571,0xc2, 0, 0}, {0x8572,0x3b, 0, 0},
	{0x8573,0x02, 0, 0}, {0x8574,0x06, 0, 0}, {0x8575,0xd9, 0, 0},
	{0x8576,0x30, 0, 0}, {0x8577,0x38, 0, 0}, {0x8578,0x0c, 0, 0},
	{0x8579,0x20, 0, 0}, {0x857a,0x3c, 0, 0}, {0x857b,0x03, 0, 0},
	{0x857c,0x30, 0, 0}, {0x857d,0x3b, 0, 0}, {0x857e,0x06, 0, 0},
	{0x857f,0xe4, 0, 0}, {0x8580,0xf5, 0, 0}, {0x8581,0x0a, 0, 0},
	{0x8582,0x12, 0, 0}, {0x8583,0x14, 0, 0}, {0x8584,0x78, 0, 0},
	{0x8585,0xd2, 0, 0}, {0x8586,0x18, 0, 0}, {0x8587,0xc2, 0, 0},
	{0x8588,0x19, 0, 0}, {0x8589,0x22, 0, 0}, {0x858a,0x30, 0, 0},
	{0x858b,0x38, 0, 0}, {0x858c,0x09, 0, 0}, {0x858d,0x20, 0, 0},
	{0x858e,0x3c, 0, 0}, {0x858f,0x03, 0, 0}, {0x8590,0x30, 0, 0},
	{0x8591,0x3b, 0, 0}, {0x8592,0x03, 0, 0}, {0x8593,0x12, 0, 0},
	{0x8594,0x06, 0, 0}, {0x8595,0xf1, 0, 0}, {0x8596,0xd2, 0, 0},
	{0x8597,0x18, 0, 0}, {0x8598,0xd2, 0, 0}, {0x8599,0x19, 0, 0},
	{0x859a,0x22, 0, 0}, {0x859b,0x30, 0, 0}, {0x859c,0x38, 0, 0},
	{0x859d,0x0c, 0, 0}, {0x859e,0x20, 0, 0}, {0x859f,0x3c, 0, 0},
	{0x85a0,0x03, 0, 0}, {0x85a1,0x30, 0, 0}, {0x85a2,0x3b, 0, 0},
	{0x85a3,0x06, 0, 0}, {0x85a4,0x75, 0, 0}, {0x85a5,0x0a, 0, 0},
	{0x85a6,0x02, 0, 0}, {0x85a7,0x12, 0, 0}, {0x85a8,0x14, 0, 0},
	{0x85a9,0x78, 0, 0}, {0x85aa,0xc2, 0, 0}, {0x85ab,0x18, 0, 0},
	{0x85ac,0xd2, 0, 0}, {0x85ad,0x19, 0, 0}, {0x85ae,0x22, 0, 0},
	{0x85af,0x20, 0, 0}, {0x85b0,0x3c, 0, 0}, {0x85b1,0x06, 0, 0},
	{0x85b2,0x20, 0, 0}, {0x85b3,0x3b, 0, 0}, {0x85b4,0x03, 0, 0},
	{0x85b5,0x02, 0, 0}, {0x85b6,0x06, 0, 0}, {0x85b7,0xdf, 0, 0},
	{0x85b8,0xe5, 0, 0}, {0x85b9,0x79, 0, 0}, {0x85ba,0xd3, 0, 0},
	{0x85bb,0x94, 0, 0}, {0x85bc,0x03, 0, 0}, {0x85bd,0x40, 0, 0},
	{0x85be,0x04, 0, 0}, {0x85bf,0x7f, 0, 0}, {0x85c0,0x00, 0, 0},
	{0x85c1,0x80, 0, 0}, {0x85c2,0x04, 0, 0}, {0x85c3,0xe5, 0, 0},
	{0x85c4,0x79, 0, 0}, {0x85c5,0x04, 0, 0}, {0x85c6,0xff, 0, 0},
	{0x85c7,0x8f, 0, 0}, {0x85c8,0x79, 0, 0}, {0x85c9,0x30, 0, 0},
	{0x85ca,0x18, 0, 0}, {0x85cb,0x06, 0, 0}, {0x85cc,0x30, 0, 0},
	{0x85cd,0x19, 0, 0}, {0x85ce,0x03, 0, 0}, {0x85cf,0x12, 0, 0},
	{0x85d0,0x06, 0, 0}, {0x85d1,0xf1, 0, 0}, {0x85d2,0x30, 0, 0},
	{0x85d3,0x18, 0, 0}, {0x85d4,0x03, 0, 0}, {0x85d5,0x02, 0, 0},
	{0x85d6,0x06, 0, 0}, {0x85d7,0xdf, 0, 0}, {0x85d8,0x20, 0, 0},
	{0x85d9,0x19, 0, 0}, {0x85da,0x03, 0, 0}, {0x85db,0x02, 0, 0},
	{0x85dc,0x06, 0, 0}, {0x85dd,0xdf, 0, 0}, {0x85de,0x75, 0, 0},
	{0x85df,0x0a, 0, 0}, {0x85e0,0x02, 0, 0}, {0x85e1,0x02, 0, 0},
	{0x85e2,0x14, 0, 0}, {0x85e3,0x78, 0, 0}, {0x85e4,0xe5, 0, 0},
	{0x85e5,0x68, 0, 0}, {0x85e6,0xd3, 0, 0}, {0x85e7,0x94, 0, 0},
	{0x85e8,0x38, 0, 0}, {0x85e9,0x40, 0, 0}, {0x85ea,0x04, 0, 0},
	{0x85eb,0x7f, 0, 0}, {0x85ec,0x34, 0, 0}, {0x85ed,0x80, 0, 0},
	{0x85ee,0x02, 0, 0}, {0x85ef,0xaf, 0, 0}, {0x85f0,0x68, 0, 0},
	{0x85f1,0x8f, 0, 0}, {0x85f2,0x68, 0, 0}, {0x85f3,0xe5, 0, 0},
	{0x85f4,0x68, 0, 0}, {0x85f5,0xc3, 0, 0}, {0x85f6,0x94, 0, 0},
	{0x85f7,0x08, 0, 0}, {0x85f8,0x50, 0, 0}, {0x85f9,0x04, 0, 0},
	{0x85fa,0x7f, 0, 0}, {0x85fb,0x08, 0, 0}, {0x85fc,0x80, 0, 0},
	{0x85fd,0x02, 0, 0}, {0x85fe,0xaf, 0, 0}, {0x85ff,0x68, 0, 0},
	{0x8600,0x8f, 0, 0}, {0x8601,0x68, 0, 0}, {0x8602,0xe5, 0, 0},
	{0x8603,0x69, 0, 0}, {0x8604,0xd3, 0, 0}, {0x8605,0x94, 0, 0},
	{0x8606,0x2a, 0, 0}, {0x8607,0x40, 0, 0}, {0x8608,0x04, 0, 0},
	{0x8609,0x7f, 0, 0}, {0x860a,0x2a, 0, 0}, {0x860b,0x80, 0, 0},
	{0x860c,0x02, 0, 0}, {0x860d,0xaf, 0, 0}, {0x860e,0x69, 0, 0},
	{0x860f,0x8f, 0, 0}, {0x8610,0x69, 0, 0}, {0x8611,0xe5, 0, 0},
	{0x8612,0x69, 0, 0}, {0x8613,0xc3, 0, 0}, {0x8614,0x94, 0, 0},
	{0x8615,0x06, 0, 0}, {0x8616,0x50, 0, 0}, {0x8617,0x04, 0, 0},
	{0x8618,0x7f, 0, 0}, {0x8619,0x06, 0, 0}, {0x861a,0x80, 0, 0},
	{0x861b,0x02, 0, 0}, {0x861c,0xaf, 0, 0}, {0x861d,0x69, 0, 0},
	{0x861e,0x8f, 0, 0}, {0x861f,0x69, 0, 0}, {0x8620,0xaf, 0, 0},
	{0x8621,0x68, 0, 0}, {0x8622,0xef, 0, 0}, {0x8623,0x24, 0, 0},
	{0x8624,0xf8, 0, 0}, {0x8625,0xff, 0, 0}, {0x8626,0xe4, 0, 0},
	{0x8627,0x34, 0, 0}, {0x8628,0xff, 0, 0}, {0x8629,0xfe, 0, 0},
	{0x862a,0xe4, 0, 0}, {0x862b,0x8f, 0, 0}, {0x862c,0x3f, 0, 0},
	{0x862d,0x8e, 0, 0}, {0x862e,0x3e, 0, 0}, {0x862f,0xf5, 0, 0},
	{0x8630,0x3d, 0, 0}, {0x8631,0xf5, 0, 0}, {0x8632,0x3c, 0, 0},
	{0x8633,0xac, 0, 0}, {0x8634,0x3c, 0, 0}, {0x8635,0x12, 0, 0},
	{0x8636,0x01, 0, 0}, {0x8637,0xb9, 0, 0}, {0x8638,0xaf, 0, 0},
	{0x8639,0x69, 0, 0}, {0x863a,0xef, 0, 0}, {0x863b,0x24, 0, 0},
	{0x863c,0xfa, 0, 0}, {0x863d,0xff, 0, 0}, {0x863e,0xe4, 0, 0},
	{0x863f,0x34, 0, 0}, {0x8640,0xff, 0, 0}, {0x8641,0x12, 0, 0},
	{0x8642,0x01, 0, 0}, {0x8643,0x9c, 0, 0}, {0x8644,0xaf, 0, 0},
	{0x8645,0x68, 0, 0}, {0x8646,0xef, 0, 0}, {0x8647,0x24, 0, 0},
	{0x8648,0x08, 0, 0}, {0x8649,0xff, 0, 0}, {0x864a,0xe4, 0, 0},
	{0x864b,0x33, 0, 0}, {0x864c,0x12, 0, 0}, {0x864d,0x01, 0, 0},
	{0x864e,0x9c, 0, 0}, {0x864f,0xaf, 0, 0}, {0x8650,0x69, 0, 0},
	{0x8651,0xef, 0, 0}, {0x8652,0x24, 0, 0}, {0x8653,0x06, 0, 0},
	{0x8654,0xff, 0, 0}, {0x8655,0xe4, 0, 0}, {0x8656,0x33, 0, 0},
	{0x8657,0xfe, 0, 0}, {0x8658,0xe4, 0, 0}, {0x8659,0xfc, 0, 0},
	{0x865a,0xfd, 0, 0}, {0x865b,0xe5, 0, 0}, {0x865c,0x3f, 0, 0},
	{0x865d,0x2f, 0, 0}, {0x865e,0xf5, 0, 0}, {0x865f,0x3f, 0, 0},
	{0x8660,0xe5, 0, 0}, {0x8661,0x3e, 0, 0}, {0x8662,0x3e, 0, 0},
	{0x8663,0xf5, 0, 0}, {0x8664,0x3e, 0, 0}, {0x8665,0xed, 0, 0},
	{0x8666,0x35, 0, 0}, {0x8667,0x3d, 0, 0}, {0x8668,0xf5, 0, 0},
	{0x8669,0x3d, 0, 0}, {0x866a,0xec, 0, 0}, {0x866b,0x35, 0, 0},
	{0x866c,0x3c, 0, 0}, {0x866d,0xf5, 0, 0}, {0x866e,0x3c, 0, 0},
	{0x866f,0xe4, 0, 0}, {0x8670,0x25, 0, 0}, {0x8671,0x3f, 0, 0},
	{0x8672,0xf5, 0, 0}, {0x8673,0x37, 0, 0}, {0x8674,0xe4, 0, 0},
	{0x8675,0x35, 0, 0}, {0x8676,0x3e, 0, 0}, {0x8677,0xf5, 0, 0},
	{0x8678,0x36, 0, 0}, {0x8679,0xe4, 0, 0}, {0x867a,0x35, 0, 0},
	{0x867b,0x3d, 0, 0}, {0x867c,0xf5, 0, 0}, {0x867d,0x35, 0, 0},
	{0x867e,0xe5, 0, 0}, {0x867f,0x3c, 0, 0}, {0x8680,0x34, 0, 0},
	{0x8681,0x08, 0, 0}, {0x8682,0xf5, 0, 0}, {0x8683,0x34, 0, 0},
	{0x8684,0xe4, 0, 0}, {0x8685,0x25, 0, 0}, {0x8686,0x3f, 0, 0},
	{0x8687,0xf5, 0, 0}, {0x8688,0x3b, 0, 0}, {0x8689,0xe4, 0, 0},
	{0x868a,0x35, 0, 0}, {0x868b,0x3e, 0, 0}, {0x868c,0xf5, 0, 0},
	{0x868d,0x3a, 0, 0}, {0x868e,0xe5, 0, 0}, {0x868f,0x3d, 0, 0},
	{0x8690,0x34, 0, 0}, {0x8691,0x06, 0, 0}, {0x8692,0xf5, 0, 0},
	{0x8693,0x39, 0, 0}, {0x8694,0xe4, 0, 0}, {0x8695,0x35, 0, 0},
	{0x8696,0x3c, 0, 0}, {0x8697,0xf5, 0, 0}, {0x8698,0x38, 0, 0},
	{0x8699,0xe5, 0, 0}, {0x869a,0x3f, 0, 0}, {0x869b,0x24, 0, 0},
	{0x869c,0xfa, 0, 0}, {0x869d,0xf5, 0, 0}, {0x869e,0x43, 0, 0},
	{0x869f,0xe5, 0, 0}, {0x86a0,0x3e, 0, 0}, {0x86a1,0x34, 0, 0},
	{0x86a2,0xff, 0, 0}, {0x86a3,0xf5, 0, 0}, {0x86a4,0x42, 0, 0},
	{0x86a5,0xe5, 0, 0}, {0x86a6,0x3d, 0, 0}, {0x86a7,0x34, 0, 0},
	{0x86a8,0xff, 0, 0}, {0x86a9,0xf5, 0, 0}, {0x86aa,0x41, 0, 0},
	{0x86ab,0xe5, 0, 0}, {0x86ac,0x3c, 0, 0}, {0x86ad,0x34, 0, 0},
	{0x86ae,0xff, 0, 0}, {0x86af,0xf5, 0, 0}, {0x86b0,0x40, 0, 0},
	{0x86b1,0xe4, 0, 0}, {0x86b2,0x25, 0, 0}, {0x86b3,0x3f, 0, 0},
	{0x86b4,0xf5, 0, 0}, {0x86b5,0x47, 0, 0}, {0x86b6,0xe5, 0, 0},
	{0x86b7,0x3e, 0, 0}, {0x86b8,0x34, 0, 0}, {0x86b9,0xf8, 0, 0},
	{0x86ba,0xf5, 0, 0}, {0x86bb,0x46, 0, 0}, {0x86bc,0xe5, 0, 0},
	{0x86bd,0x3d, 0, 0}, {0x86be,0x34, 0, 0}, {0x86bf,0xff, 0, 0},
	{0x86c0,0xf5, 0, 0}, {0x86c1,0x45, 0, 0}, {0x86c2,0xe5, 0, 0},
	{0x86c3,0x3c, 0, 0}, {0x86c4,0x34, 0, 0}, {0x86c5,0xff, 0, 0},
	{0x86c6,0xf5, 0, 0}, {0x86c7,0x44, 0, 0}, {0x86c8,0x75, 0, 0},
	{0x86c9,0x79, 0, 0}, {0x86ca,0x02, 0, 0}, {0x86cb,0x75, 0, 0},
	{0x86cc,0x0a, 0, 0}, {0x86cd,0x01, 0, 0}, {0x86ce,0x12, 0, 0},
	{0x86cf,0x14, 0, 0}, {0x86d0,0x78, 0, 0}, {0x86d1,0xd2, 0, 0},
	{0x86d2,0x18, 0, 0}, {0x86d3,0xd2, 0, 0}, {0x86d4,0x19, 0, 0},
	{0x86d5,0xc2, 0, 0}, {0x86d6,0x3c, 0, 0}, {0x86d7,0xc2, 0, 0},
	{0x86d8,0x3b, 0, 0}, {0x86d9,0xd2, 0, 0}, {0x86da,0x1a, 0, 0},
	{0x86db,0xd2, 0, 0}, {0x86dc,0x38, 0, 0}, {0x86dd,0xd2, 0, 0},
	{0x86de,0x30, 0, 0}, {0x86df,0x22, 0, 0}, {0x86e0,0x12, 0, 0},
	{0x86e1,0x12, 0, 0}, {0x86e2,0x9c, 0, 0}, {0x86e3,0x75, 0, 0},
	{0x86e4,0x79, 0, 0}, {0x86e5,0x02, 0, 0}, {0x86e6,0xe4, 0, 0},
	{0x86e7,0xf5, 0, 0}, {0x86e8,0x0a, 0, 0}, {0x86e9,0x12, 0, 0},
	{0x86ea,0x14, 0, 0}, {0x86eb,0x78, 0, 0}, {0x86ec,0xd2, 0, 0},
	{0x86ed,0x18, 0, 0}, {0x86ee,0xc2, 0, 0}, {0x86ef,0x19, 0, 0},
	{0x86f0,0x22, 0, 0}, {0x86f1,0x75, 0, 0}, {0x86f2,0x0a, 0, 0},
	{0x86f3,0x01, 0, 0}, {0x86f4,0x12, 0, 0}, {0x86f5,0x14, 0, 0},
	{0x86f6,0x78, 0, 0}, {0x86f7,0x22, 0, 0}, {0x86f8,0x90, 0, 0},
	{0x86f9,0x38, 0, 0}, {0x86fa,0x04, 0, 0}, {0x86fb,0xe0, 0, 0},
	{0x86fc,0xfe, 0, 0}, {0x86fd,0xa3, 0, 0}, {0x86fe,0xe0, 0, 0},
	{0x86ff,0xfd, 0, 0}, {0x8700,0xed, 0, 0}, {0x8701,0xff, 0, 0},
	{0x8702,0xee, 0, 0}, {0x8703,0x54, 0, 0}, {0x8704,0x0f, 0, 0},
	{0x8705,0xf5, 0, 0}, {0x8706,0x0e, 0, 0}, {0x8707,0x8f, 0, 0},
	{0x8708,0x0f, 0, 0}, {0x8709,0xa3, 0, 0}, {0x870a,0xe0, 0, 0},
	{0x870b,0xfe, 0, 0}, {0x870c,0xa3, 0, 0}, {0x870d,0xe0, 0, 0},
	{0x870e,0xfd, 0, 0}, {0x870f,0xed, 0, 0}, {0x8710,0xff, 0, 0},
	{0x8711,0xee, 0, 0}, {0x8712,0x54, 0, 0}, {0x8713,0x07, 0, 0},
	{0x8714,0xf5, 0, 0}, {0x8715,0x10, 0, 0}, {0x8716,0x8f, 0, 0},
	{0x8717,0x11, 0, 0}, {0x8718,0xe5, 0, 0}, {0x8719,0x0e, 0, 0},
	{0x871a,0xc4, 0, 0}, {0x871b,0xf8, 0, 0}, {0x871c,0x54, 0, 0},
	{0x871d,0xf0, 0, 0}, {0x871e,0xc8, 0, 0}, {0x871f,0x68, 0, 0},
	{0x8720,0xf5, 0, 0}, {0x8721,0x0e, 0, 0}, {0x8722,0xe5, 0, 0},
	{0x8723,0x0f, 0, 0}, {0x8724,0xc4, 0, 0}, {0x8725,0x54, 0, 0},
	{0x8726,0x0f, 0, 0}, {0x8727,0x48, 0, 0}, {0x8728,0xf5, 0, 0},
	{0x8729,0x0f, 0, 0}, {0x872a,0xe5, 0, 0}, {0x872b,0x10, 0, 0},
	{0x872c,0xc4, 0, 0}, {0x872d,0xf8, 0, 0}, {0x872e,0x54, 0, 0},
	{0x872f,0xf0, 0, 0}, {0x8730,0xc8, 0, 0}, {0x8731,0x68, 0, 0},
	{0x8732,0xf5, 0, 0}, {0x8733,0x10, 0, 0}, {0x8734,0xe5, 0, 0},
	{0x8735,0x11, 0, 0}, {0x8736,0xc4, 0, 0}, {0x8737,0x54, 0, 0},
	{0x8738,0x0f, 0, 0}, {0x8739,0x48, 0, 0}, {0x873a,0xf5, 0, 0},
	{0x873b,0x11, 0, 0}, {0x873c,0xe4, 0, 0}, {0x873d,0xf5, 0, 0},
	{0x873e,0x17, 0, 0}, {0x873f,0x75, 0, 0}, {0x8740,0x16, 0, 0},
	{0x8741,0x04, 0, 0}, {0x8742,0x12, 0, 0}, {0x8743,0x02, 0, 0},
	{0x8744,0x60, 0, 0}, {0x8745,0x24, 0, 0}, {0x8746,0x34, 0, 0},
	{0x8747,0xf8, 0, 0}, {0x8748,0xe6, 0, 0}, {0x8749,0xf5, 0, 0},
	{0x874a,0x12, 0, 0}, {0x874b,0x12, 0, 0}, {0x874c,0x02, 0, 0},
	{0x874d,0x60, 0, 0}, {0x874e,0x24, 0, 0}, {0x874f,0x35, 0, 0},
	{0x8750,0xf8, 0, 0}, {0x8751,0xe6, 0, 0}, {0x8752,0xf5, 0, 0},
	{0x8753,0x14, 0, 0}, {0x8754,0x12, 0, 0}, {0x8755,0x02, 0, 0},
	{0x8756,0x60, 0, 0}, {0x8757,0x24, 0, 0}, {0x8758,0x36, 0, 0},
	{0x8759,0xf8, 0, 0}, {0x875a,0xe6, 0, 0}, {0x875b,0xf5, 0, 0},
	{0x875c,0x13, 0, 0}, {0x875d,0x12, 0, 0}, {0x875e,0x02, 0, 0},
	{0x875f,0x60, 0, 0}, {0x8760,0x24, 0, 0}, {0x8761,0x37, 0, 0},
	{0x8762,0xf8, 0, 0}, {0x8763,0xe6, 0, 0}, {0x8764,0xf5, 0, 0},
	{0x8765,0x15, 0, 0}, {0x8766,0x12, 0, 0}, {0x8767,0x02, 0, 0},
	{0x8768,0x73, 0, 0}, {0x8769,0xaf, 0, 0}, {0x876a,0x12, 0, 0},
	{0x876b,0x12, 0, 0}, {0x876c,0x01, 0, 0}, {0x876d,0x42, 0, 0},
	{0x876e,0x8f, 0, 0}, {0x876f,0x12, 0, 0}, {0x8770,0x12, 0, 0},
	{0x8771,0x02, 0, 0}, {0x8772,0x73, 0, 0}, {0x8773,0xaf, 0, 0},
	{0x8774,0x13, 0, 0}, {0x8775,0x12, 0, 0}, {0x8776,0x01, 0, 0},
	{0x8777,0x42, 0, 0}, {0x8778,0x8f, 0, 0}, {0x8779,0x13, 0, 0},
	{0x877a,0x12, 0, 0}, {0x877b,0x02, 0, 0}, {0x877c,0x1b, 0, 0},
	{0x877d,0xaf, 0, 0}, {0x877e,0x14, 0, 0}, {0x877f,0xfc, 0, 0},
	{0x8780,0xfd, 0, 0}, {0x8781,0xfe, 0, 0}, {0x8782,0x12, 0, 0},
	{0x8783,0x08, 0, 0}, {0x8784,0x98, 0, 0}, {0x8785,0x12, 0, 0},
	{0x8786,0x01, 0, 0}, {0x8787,0x61, 0, 0}, {0x8788,0x7b, 0, 0},
	{0x8789,0x30, 0, 0}, {0x878a,0x12, 0, 0}, {0x878b,0x01, 0, 0},
	{0x878c,0x5a, 0, 0}, {0x878d,0x8f, 0, 0}, {0x878e,0x14, 0, 0},
	{0x878f,0x12, 0, 0}, {0x8790,0x02, 0, 0}, {0x8791,0x1b, 0, 0},
	{0x8792,0xaf, 0, 0}, {0x8793,0x15, 0, 0}, {0x8794,0xfc, 0, 0},
	{0x8795,0xfd, 0, 0}, {0x8796,0xfe, 0, 0}, {0x8797,0x12, 0, 0},
	{0x8798,0x08, 0, 0}, {0x8799,0x98, 0, 0}, {0x879a,0x12, 0, 0},
	{0x879b,0x01, 0, 0}, {0x879c,0x61, 0, 0}, {0x879d,0xe4, 0, 0},
	{0x879e,0x7b, 0, 0}, {0x879f,0x30, 0, 0}, {0x87a0,0x12, 0, 0},
	{0x87a1,0x01, 0, 0}, {0x87a2,0x5b, 0, 0}, {0x87a3,0x8f, 0, 0},
	{0x87a4,0x15, 0, 0}, {0x87a5,0xc3, 0, 0}, {0x87a6,0xe5, 0, 0},
	{0x87a7,0x13, 0, 0}, {0x87a8,0x95, 0, 0}, {0x87a9,0x12, 0, 0},
	{0x87aa,0xff, 0, 0}, {0x87ab,0x0f, 0, 0}, {0x87ac,0xef, 0, 0},
	{0x87ad,0xc3, 0, 0}, {0x87ae,0x13, 0, 0}, {0x87af,0xff, 0, 0},
	{0x87b0,0xc3, 0, 0}, {0x87b1,0x94, 0, 0}, {0x87b2,0x04, 0, 0},
	{0x87b3,0x50, 0, 0}, {0x87b4,0x27, 0, 0}, {0x87b5,0xe5, 0, 0},
	{0x87b6,0x12, 0, 0}, {0x87b7,0x9f, 0, 0}, {0x87b8,0x40, 0, 0},
	{0x87b9,0x06, 0, 0}, {0x87ba,0xe5, 0, 0}, {0x87bb,0x12, 0, 0},
	{0x87bc,0x9f, 0, 0}, {0x87bd,0xfe, 0, 0}, {0x87be,0x80, 0, 0},
	{0x87bf,0x02, 0, 0}, {0x87c0,0x7e, 0, 0}, {0x87c1,0x00, 0, 0},
	{0x87c2,0x8e, 0, 0}, {0x87c3,0x12, 0, 0}, {0x87c4,0xef, 0, 0},
	{0x87c5,0xfd, 0, 0}, {0x87c6,0xe5, 0, 0}, {0x87c7,0x13, 0, 0},
	{0x87c8,0x2d, 0, 0}, {0x87c9,0xfd, 0, 0}, {0x87ca,0xe4, 0, 0},
	{0x87cb,0x33, 0, 0}, {0x87cc,0xfc, 0, 0}, {0x87cd,0xc3, 0, 0},
	{0x87ce,0xed, 0, 0}, {0x87cf,0x95, 0, 0}, {0x87d0,0x0f, 0, 0},
	{0x87d1,0xec, 0, 0}, {0x87d2,0x95, 0, 0}, {0x87d3,0x0e, 0, 0},
	{0x87d4,0x50, 0, 0}, {0x87d5,0x02, 0, 0}, {0x87d6,0x80, 0, 0},
	{0x87d7,0x02, 0, 0}, {0x87d8,0xad, 0, 0}, {0x87d9,0x0f, 0, 0},
	{0x87da,0x8d, 0, 0}, {0x87db,0x13, 0, 0}, {0x87dc,0xc3, 0, 0},
	{0x87dd,0xe5, 0, 0}, {0x87de,0x15, 0, 0}, {0x87df,0x95, 0, 0},
	{0x87e0,0x14, 0, 0}, {0x87e1,0xff, 0, 0}, {0x87e2,0xc3, 0, 0},
	{0x87e3,0x94, 0, 0}, {0x87e4,0x04, 0, 0}, {0x87e5,0x50, 0, 0},
	{0x87e6,0x29, 0, 0}, {0x87e7,0xe5, 0, 0}, {0x87e8,0x14, 0, 0},
	{0x87e9,0x9f, 0, 0}, {0x87ea,0x40, 0, 0}, {0x87eb,0x06, 0, 0},
	{0x87ec,0xe5, 0, 0}, {0x87ed,0x14, 0, 0}, {0x87ee,0x9f, 0, 0},
	{0x87ef,0xfe, 0, 0}, {0x87f0,0x80, 0, 0}, {0x87f1,0x02, 0, 0},
	{0x87f2,0x7e, 0, 0}, {0x87f3,0x00, 0, 0}, {0x87f4,0x8e, 0, 0},
	{0x87f5,0x14, 0, 0}, {0x87f6,0xef, 0, 0}, {0x87f7,0xfd, 0, 0},
	{0x87f8,0xe5, 0, 0}, {0x87f9,0x15, 0, 0}, {0x87fa,0x2d, 0, 0},
	{0x87fb,0xfd, 0, 0}, {0x87fc,0xe4, 0, 0}, {0x87fd,0x33, 0, 0},
	{0x87fe,0xfc, 0, 0}, {0x87ff,0xc3, 0, 0}, {0x8800,0xed, 0, 0},
	{0x8801,0x95, 0, 0}, {0x8802,0x11, 0, 0}, {0x8803,0xec, 0, 0},
	{0x8804,0x95, 0, 0}, {0x8805,0x10, 0, 0}, {0x8806,0x50, 0, 0},
	{0x8807,0x04, 0, 0}, {0x8808,0xaf, 0, 0}, {0x8809,0x05, 0, 0},
	{0x880a,0x80, 0, 0}, {0x880b,0x02, 0, 0}, {0x880c,0xaf, 0, 0},
	{0x880d,0x11, 0, 0}, {0x880e,0x8f, 0, 0}, {0x880f,0x15, 0, 0},
	{0x8810,0xe5, 0, 0}, {0x8811,0x15, 0, 0}, {0x8812,0xd3, 0, 0},
	{0x8813,0x95, 0, 0}, {0x8814,0x17, 0, 0}, {0x8815,0x40, 0, 0},
	{0x8816,0x04, 0, 0}, {0x8817,0xaf, 0, 0}, {0x8818,0x15, 0, 0},
	{0x8819,0x80, 0, 0}, {0x881a,0x02, 0, 0}, {0x881b,0xaf, 0, 0},
	{0x881c,0x17, 0, 0}, {0x881d,0x8f, 0, 0}, {0x881e,0x17, 0, 0},
	{0x881f,0xd3, 0, 0}, {0x8820,0xe5, 0, 0}, {0x8821,0x16, 0, 0},
	{0x8822,0x64, 0, 0}, {0x8823,0x80, 0, 0}, {0x8824,0x94, 0, 0},
	{0x8825,0x80, 0, 0}, {0x8826,0x40, 0, 0}, {0x8827,0x04, 0, 0},
	{0x8828,0xaf, 0, 0}, {0x8829,0x15, 0, 0}, {0x882a,0x80, 0, 0},
	{0x882b,0x02, 0, 0}, {0x882c,0xaf, 0, 0}, {0x882d,0x17, 0, 0},
	{0x882e,0x8f, 0, 0}, {0x882f,0x15, 0, 0}, {0x8830,0xe5, 0, 0},
	{0x8831,0x16, 0, 0}, {0x8832,0xfd, 0, 0}, {0x8833,0x33, 0, 0},
	{0x8834,0x95, 0, 0}, {0x8835,0xe0, 0, 0}, {0x8836,0xfc, 0, 0},
	{0x8837,0xed, 0, 0}, {0x8838,0xae, 0, 0}, {0x8839,0x04, 0, 0},
	{0x883a,0x78, 0, 0}, {0x883b,0x02, 0, 0}, {0x883c,0xc3, 0, 0},
	{0x883d,0x33, 0, 0}, {0x883e,0xce, 0, 0}, {0x883f,0x33, 0, 0},
	{0x8840,0xce, 0, 0}, {0x8841,0xd8, 0, 0}, {0x8842,0xf9, 0, 0},
	{0x8843,0xff, 0, 0}, {0x8844,0x24, 0, 0}, {0x8845,0x01, 0, 0},
	{0x8846,0xfb, 0, 0}, {0x8847,0xee, 0, 0}, {0x8848,0x34, 0, 0},
	{0x8849,0x60, 0, 0}, {0x884a,0x8b, 0, 0}, {0x884b,0x82, 0, 0},
	{0x884c,0xf5, 0, 0}, {0x884d,0x83, 0, 0}, {0x884e,0xe5, 0, 0},
	{0x884f,0x12, 0, 0}, {0x8850,0xf0, 0, 0}, {0x8851,0xef, 0, 0},
	{0x8852,0x24, 0, 0}, {0x8853,0x02, 0, 0}, {0x8854,0xff, 0, 0},
	{0x8855,0xee, 0, 0}, {0x8856,0x34, 0, 0}, {0x8857,0x60, 0, 0},
	{0x8858,0x8f, 0, 0}, {0x8859,0x82, 0, 0}, {0x885a,0xf5, 0, 0},
	{0x885b,0x83, 0, 0}, {0x885c,0xe5, 0, 0}, {0x885d,0x14, 0, 0},
	{0x885e,0xf0, 0, 0}, {0x885f,0xed, 0, 0}, {0x8860,0xae, 0, 0},
	{0x8861,0x04, 0, 0}, {0x8862,0x78, 0, 0}, {0x8863,0x02, 0, 0},
	{0x8864,0xc3, 0, 0}, {0x8865,0x33, 0, 0}, {0x8866,0xce, 0, 0},
	{0x8867,0x33, 0, 0}, {0x8868,0xce, 0, 0}, {0x8869,0xd8, 0, 0},
	{0x886a,0xf9, 0, 0}, {0x886b,0xff, 0, 0}, {0x886c,0x24, 0, 0},
	{0x886d,0x03, 0, 0}, {0x886e,0xfd, 0, 0}, {0x886f,0xee, 0, 0},
	{0x8870,0x34, 0, 0}, {0x8871,0x60, 0, 0}, {0x8872,0x8d, 0, 0},
	{0x8873,0x82, 0, 0}, {0x8874,0xf5, 0, 0}, {0x8875,0x83, 0, 0},
	{0x8876,0xe5, 0, 0}, {0x8877,0x13, 0, 0}, {0x8878,0xf0, 0, 0},
	{0x8879,0xef, 0, 0}, {0x887a,0x24, 0, 0}, {0x887b,0x04, 0, 0},
	{0x887c,0xff, 0, 0}, {0x887d,0xee, 0, 0}, {0x887e,0x34, 0, 0},
	{0x887f,0x60, 0, 0}, {0x8880,0x8f, 0, 0}, {0x8881,0x82, 0, 0},
	{0x8882,0xf5, 0, 0}, {0x8883,0x83, 0, 0}, {0x8884,0xe5, 0, 0},
	{0x8885,0x15, 0, 0}, {0x8886,0xf0, 0, 0}, {0x8887,0x15, 0, 0},
	{0x8888,0x16, 0, 0}, {0x8889,0xc3, 0, 0}, {0x888a,0xe5, 0, 0},
	{0x888b,0x16, 0, 0}, {0x888c,0x64, 0, 0}, {0x888d,0x80, 0, 0},
	{0x888e,0x94, 0, 0}, {0x888f,0x80, 0, 0}, {0x8890,0x40, 0, 0},
	{0x8891,0x03, 0, 0}, {0x8892,0x02, 0, 0}, {0x8893,0x07, 0, 0},
	{0x8894,0x42, 0, 0}, {0x8895,0xc2, 0, 0}, {0x8896,0x30, 0, 0},
	{0x8897,0x22, 0, 0}, {0x8898,0xe8, 0, 0}, {0x8899,0x8f, 0, 0},
	{0x889a,0xf0, 0, 0}, {0x889b,0xa4, 0, 0}, {0x889c,0xcc, 0, 0},
	{0x889d,0x8b, 0, 0}, {0x889e,0xf0, 0, 0}, {0x889f,0xa4, 0, 0},
	{0x88a0,0x2c, 0, 0}, {0x88a1,0xfc, 0, 0}, {0x88a2,0xe9, 0, 0},
	{0x88a3,0x8e, 0, 0}, {0x88a4,0xf0, 0, 0}, {0x88a5,0xa4, 0, 0},
	{0x88a6,0x2c, 0, 0}, {0x88a7,0xfc, 0, 0}, {0x88a8,0x8a, 0, 0},
	{0x88a9,0xf0, 0, 0}, {0x88aa,0xed, 0, 0}, {0x88ab,0xa4, 0, 0},
	{0x88ac,0x2c, 0, 0}, {0x88ad,0xfc, 0, 0}, {0x88ae,0xea, 0, 0},
	{0x88af,0x8e, 0, 0}, {0x88b0,0xf0, 0, 0}, {0x88b1,0xa4, 0, 0},
	{0x88b2,0xcd, 0, 0}, {0x88b3,0xa8, 0, 0}, {0x88b4,0xf0, 0, 0},
	{0x88b5,0x8b, 0, 0}, {0x88b6,0xf0, 0, 0}, {0x88b7,0xa4, 0, 0},
	{0x88b8,0x2d, 0, 0}, {0x88b9,0xcc, 0, 0}, {0x88ba,0x38, 0, 0},
	{0x88bb,0x25, 0, 0}, {0x88bc,0xf0, 0, 0}, {0x88bd,0xfd, 0, 0},
	{0x88be,0xe9, 0, 0}, {0x88bf,0x8f, 0, 0}, {0x88c0,0xf0, 0, 0},
	{0x88c1,0xa4, 0, 0}, {0x88c2,0x2c, 0, 0}, {0x88c3,0xcd, 0, 0},
	{0x88c4,0x35, 0, 0}, {0x88c5,0xf0, 0, 0}, {0x88c6,0xfc, 0, 0},
	{0x88c7,0xeb, 0, 0}, {0x88c8,0x8e, 0, 0}, {0x88c9,0xf0, 0, 0},
	{0x88ca,0xa4, 0, 0}, {0x88cb,0xfe, 0, 0}, {0x88cc,0xa9, 0, 0},
	{0x88cd,0xf0, 0, 0}, {0x88ce,0xeb, 0, 0}, {0x88cf,0x8f, 0, 0},
	{0x88d0,0xf0, 0, 0}, {0x88d1,0xa4, 0, 0}, {0x88d2,0xcf, 0, 0},
	{0x88d3,0xc5, 0, 0}, {0x88d4,0xf0, 0, 0}, {0x88d5,0x2e, 0, 0},
	{0x88d6,0xcd, 0, 0}, {0x88d7,0x39, 0, 0}, {0x88d8,0xfe, 0, 0},
	{0x88d9,0xe4, 0, 0}, {0x88da,0x3c, 0, 0}, {0x88db,0xfc, 0, 0},
	{0x88dc,0xea, 0, 0}, {0x88dd,0xa4, 0, 0}, {0x88de,0x2d, 0, 0},
	{0x88df,0xce, 0, 0}, {0x88e0,0x35, 0, 0}, {0x88e1,0xf0, 0, 0},
	{0x88e2,0xfd, 0, 0}, {0x88e3,0xe4, 0, 0}, {0x88e4,0x3c, 0, 0},
	{0x88e5,0xfc, 0, 0}, {0x88e6,0x22, 0, 0}, {0x88e7,0x75, 0, 0},
	{0x88e8,0xf0, 0, 0}, {0x88e9,0x08, 0, 0}, {0x88ea,0x75, 0, 0},
	{0x88eb,0x82, 0, 0}, {0x88ec,0x00, 0, 0}, {0x88ed,0xef, 0, 0},
	{0x88ee,0x2f, 0, 0}, {0x88ef,0xff, 0, 0}, {0x88f0,0xee, 0, 0},
	{0x88f1,0x33, 0, 0}, {0x88f2,0xfe, 0, 0}, {0x88f3,0xcd, 0, 0},
	{0x88f4,0x33, 0, 0}, {0x88f5,0xcd, 0, 0}, {0x88f6,0xcc, 0, 0},
	{0x88f7,0x33, 0, 0}, {0x88f8,0xcc, 0, 0}, {0x88f9,0xc5, 0, 0},
	{0x88fa,0x82, 0, 0}, {0x88fb,0x33, 0, 0}, {0x88fc,0xc5, 0, 0},
	{0x88fd,0x82, 0, 0}, {0x88fe,0x9b, 0, 0}, {0x88ff,0xed, 0, 0},
	{0x8900,0x9a, 0, 0}, {0x8901,0xec, 0, 0}, {0x8902,0x99, 0, 0},
	{0x8903,0xe5, 0, 0}, {0x8904,0x82, 0, 0}, {0x8905,0x98, 0, 0},
	{0x8906,0x40, 0, 0}, {0x8907,0x0c, 0, 0}, {0x8908,0xf5, 0, 0},
	{0x8909,0x82, 0, 0}, {0x890a,0xee, 0, 0}, {0x890b,0x9b, 0, 0},
	{0x890c,0xfe, 0, 0}, {0x890d,0xed, 0, 0}, {0x890e,0x9a, 0, 0},
	{0x890f,0xfd, 0, 0}, {0x8910,0xec, 0, 0}, {0x8911,0x99, 0, 0},
	{0x8912,0xfc, 0, 0}, {0x8913,0x0f, 0, 0}, {0x8914,0xd5, 0, 0},
	{0x8915,0xf0, 0, 0}, {0x8916,0xd6, 0, 0}, {0x8917,0xe4, 0, 0},
	{0x8918,0xce, 0, 0}, {0x8919,0xfb, 0, 0}, {0x891a,0xe4, 0, 0},
	{0x891b,0xcd, 0, 0}, {0x891c,0xfa, 0, 0}, {0x891d,0xe4, 0, 0},
	{0x891e,0xcc, 0, 0}, {0x891f,0xf9, 0, 0}, {0x8920,0xa8, 0, 0},
	{0x8921,0x82, 0, 0}, {0x8922,0x22, 0, 0}, {0x8923,0xb8, 0, 0},
	{0x8924,0x00, 0, 0}, {0x8925,0xc1, 0, 0}, {0x8926,0xb9, 0, 0},
	{0x8927,0x00, 0, 0}, {0x8928,0x59, 0, 0}, {0x8929,0xba, 0, 0},
	{0x892a,0x00, 0, 0}, {0x892b,0x2d, 0, 0}, {0x892c,0xec, 0, 0},
	{0x892d,0x8b, 0, 0}, {0x892e,0xf0, 0, 0}, {0x892f,0x84, 0, 0},
	{0x8930,0xcf, 0, 0}, {0x8931,0xce, 0, 0}, {0x8932,0xcd, 0, 0},
	{0x8933,0xfc, 0, 0}, {0x8934,0xe5, 0, 0}, {0x8935,0xf0, 0, 0},
	{0x8936,0xcb, 0, 0}, {0x8937,0xf9, 0, 0}, {0x8938,0x78, 0, 0},
	{0x8939,0x18, 0, 0}, {0x893a,0xef, 0, 0}, {0x893b,0x2f, 0, 0},
	{0x893c,0xff, 0, 0}, {0x893d,0xee, 0, 0}, {0x893e,0x33, 0, 0},
	{0x893f,0xfe, 0, 0}, {0x8940,0xed, 0, 0}, {0x8941,0x33, 0, 0},
	{0x8942,0xfd, 0, 0}, {0x8943,0xec, 0, 0}, {0x8944,0x33, 0, 0},
	{0x8945,0xfc, 0, 0}, {0x8946,0xeb, 0, 0}, {0x8947,0x33, 0, 0},
	{0x8948,0xfb, 0, 0}, {0x8949,0x10, 0, 0}, {0x894a,0xd7, 0, 0},
	{0x894b,0x03, 0, 0}, {0x894c,0x99, 0, 0}, {0x894d,0x40, 0, 0},
	{0x894e,0x04, 0, 0}, {0x894f,0xeb, 0, 0}, {0x8950,0x99, 0, 0},
	{0x8951,0xfb, 0, 0}, {0x8952,0x0f, 0, 0}, {0x8953,0xd8, 0, 0},
	{0x8954,0xe5, 0, 0}, {0x8955,0xe4, 0, 0}, {0x8956,0xf9, 0, 0},
	{0x8957,0xfa, 0, 0}, {0x8958,0x22, 0, 0}, {0x8959,0x78, 0, 0},
	{0x895a,0x18, 0, 0}, {0x895b,0xef, 0, 0}, {0x895c,0x2f, 0, 0},
	{0x895d,0xff, 0, 0}, {0x895e,0xee, 0, 0}, {0x895f,0x33, 0, 0},
	{0x8960,0xfe, 0, 0}, {0x8961,0xed, 0, 0}, {0x8962,0x33, 0, 0},
	{0x8963,0xfd, 0, 0}, {0x8964,0xec, 0, 0}, {0x8965,0x33, 0, 0},
	{0x8966,0xfc, 0, 0}, {0x8967,0xc9, 0, 0}, {0x8968,0x33, 0, 0},
	{0x8969,0xc9, 0, 0}, {0x896a,0x10, 0, 0}, {0x896b,0xd7, 0, 0},
	{0x896c,0x05, 0, 0}, {0x896d,0x9b, 0, 0}, {0x896e,0xe9, 0, 0},
	{0x896f,0x9a, 0, 0}, {0x8970,0x40, 0, 0}, {0x8971,0x07, 0, 0},
	{0x8972,0xec, 0, 0}, {0x8973,0x9b, 0, 0}, {0x8974,0xfc, 0, 0},
	{0x8975,0xe9, 0, 0}, {0x8976,0x9a, 0, 0}, {0x8977,0xf9, 0, 0},
	{0x8978,0x0f, 0, 0}, {0x8979,0xd8, 0, 0}, {0x897a,0xe0, 0, 0},
	{0x897b,0xe4, 0, 0}, {0x897c,0xc9, 0, 0}, {0x897d,0xfa, 0, 0},
	{0x897e,0xe4, 0, 0}, {0x897f,0xcc, 0, 0}, {0x8980,0xfb, 0, 0},
	{0x8981,0x22, 0, 0}, {0x8982,0x75, 0, 0}, {0x8983,0xf0, 0, 0},
	{0x8984,0x10, 0, 0}, {0x8985,0xef, 0, 0}, {0x8986,0x2f, 0, 0},
	{0x8987,0xff, 0, 0}, {0x8988,0xee, 0, 0}, {0x8989,0x33, 0, 0},
	{0x898a,0xfe, 0, 0}, {0x898b,0xed, 0, 0}, {0x898c,0x33, 0, 0},
	{0x898d,0xfd, 0, 0}, {0x898e,0xcc, 0, 0}, {0x898f,0x33, 0, 0},
	{0x8990,0xcc, 0, 0}, {0x8991,0xc8, 0, 0}, {0x8992,0x33, 0, 0},
	{0x8993,0xc8, 0, 0}, {0x8994,0x10, 0, 0}, {0x8995,0xd7, 0, 0},
	{0x8996,0x07, 0, 0}, {0x8997,0x9b, 0, 0}, {0x8998,0xec, 0, 0},
	{0x8999,0x9a, 0, 0}, {0x899a,0xe8, 0, 0}, {0x899b,0x99, 0, 0},
	{0x899c,0x40, 0, 0}, {0x899d,0x0a, 0, 0}, {0x899e,0xed, 0, 0},
	{0x899f,0x9b, 0, 0}, {0x89a0,0xfd, 0, 0}, {0x89a1,0xec, 0, 0},
	{0x89a2,0x9a, 0, 0}, {0x89a3,0xfc, 0, 0}, {0x89a4,0xe8, 0, 0},
	{0x89a5,0x99, 0, 0}, {0x89a6,0xf8, 0, 0}, {0x89a7,0x0f, 0, 0},
	{0x89a8,0xd5, 0, 0}, {0x89a9,0xf0, 0, 0}, {0x89aa,0xda, 0, 0},
	{0x89ab,0xe4, 0, 0}, {0x89ac,0xcd, 0, 0}, {0x89ad,0xfb, 0, 0},
	{0x89ae,0xe4, 0, 0}, {0x89af,0xcc, 0, 0}, {0x89b0,0xfa, 0, 0},
	{0x89b1,0xe4, 0, 0}, {0x89b2,0xc8, 0, 0}, {0x89b3,0xf9, 0, 0},
	{0x89b4,0x22, 0, 0}, {0x89b5,0xe8, 0, 0}, {0x89b6,0x60, 0, 0},
	{0x89b7,0x0f, 0, 0}, {0x89b8,0xec, 0, 0}, {0x89b9,0xc3, 0, 0},
	{0x89ba,0x13, 0, 0}, {0x89bb,0xfc, 0, 0}, {0x89bc,0xed, 0, 0},
	{0x89bd,0x13, 0, 0}, {0x89be,0xfd, 0, 0}, {0x89bf,0xee, 0, 0},
	{0x89c0,0x13, 0, 0}, {0x89c1,0xfe, 0, 0}, {0x89c2,0xef, 0, 0},
	{0x89c3,0x13, 0, 0}, {0x89c4,0xff, 0, 0}, {0x89c5,0xd8, 0, 0},
	{0x89c6,0xf1, 0, 0}, {0x89c7,0x22, 0, 0}, {0x89c8,0xe8, 0, 0},
	{0x89c9,0x60, 0, 0}, {0x89ca,0x0f, 0, 0}, {0x89cb,0xef, 0, 0},
	{0x89cc,0xc3, 0, 0}, {0x89cd,0x33, 0, 0}, {0x89ce,0xff, 0, 0},
	{0x89cf,0xee, 0, 0}, {0x89d0,0x33, 0, 0}, {0x89d1,0xfe, 0, 0},
	{0x89d2,0xed, 0, 0}, {0x89d3,0x33, 0, 0}, {0x89d4,0xfd, 0, 0},
	{0x89d5,0xec, 0, 0}, {0x89d6,0x33, 0, 0}, {0x89d7,0xfc, 0, 0},
	{0x89d8,0xd8, 0, 0}, {0x89d9,0xf1, 0, 0}, {0x89da,0x22, 0, 0},
	{0x89db,0xe4, 0, 0}, {0x89dc,0x93, 0, 0}, {0x89dd,0xfc, 0, 0},
	{0x89de,0x74, 0, 0}, {0x89df,0x01, 0, 0}, {0x89e0,0x93, 0, 0},
	{0x89e1,0xfd, 0, 0}, {0x89e2,0x74, 0, 0}, {0x89e3,0x02, 0, 0},
	{0x89e4,0x93, 0, 0}, {0x89e5,0xfe, 0, 0}, {0x89e6,0x74, 0, 0},
	{0x89e7,0x03, 0, 0}, {0x89e8,0x93, 0, 0}, {0x89e9,0xff, 0, 0},
	{0x89ea,0x22, 0, 0}, {0x89eb,0xa4, 0, 0}, {0x89ec,0x25, 0, 0},
	{0x89ed,0x82, 0, 0}, {0x89ee,0xf5, 0, 0}, {0x89ef,0x82, 0, 0},
	{0x89f0,0xe5, 0, 0}, {0x89f1,0xf0, 0, 0}, {0x89f2,0x35, 0, 0},
	{0x89f3,0x83, 0, 0}, {0x89f4,0xf5, 0, 0}, {0x89f5,0x83, 0, 0},
	{0x89f6,0x22, 0, 0}, {0x89f7,0xd0, 0, 0}, {0x89f8,0x83, 0, 0},
	{0x89f9,0xd0, 0, 0}, {0x89fa,0x82, 0, 0}, {0x89fb,0xf8, 0, 0},
	{0x89fc,0xe4, 0, 0}, {0x89fd,0x93, 0, 0}, {0x89fe,0x70, 0, 0},
	{0x89ff,0x12, 0, 0}, {0x8a00,0x74, 0, 0}, {0x8a01,0x01, 0, 0},
	{0x8a02,0x93, 0, 0}, {0x8a03,0x70, 0, 0}, {0x8a04,0x0d, 0, 0},
	{0x8a05,0xa3, 0, 0}, {0x8a06,0xa3, 0, 0}, {0x8a07,0x93, 0, 0},
	{0x8a08,0xf8, 0, 0}, {0x8a09,0x74, 0, 0}, {0x8a0a,0x01, 0, 0},
	{0x8a0b,0x93, 0, 0}, {0x8a0c,0xf5, 0, 0}, {0x8a0d,0x82, 0, 0},
	{0x8a0e,0x88, 0, 0}, {0x8a0f,0x83, 0, 0}, {0x8a10,0xe4, 0, 0},
	{0x8a11,0x73, 0, 0}, {0x8a12,0x74, 0, 0}, {0x8a13,0x02, 0, 0},
	{0x8a14,0x93, 0, 0}, {0x8a15,0x68, 0, 0}, {0x8a16,0x60, 0, 0},
	{0x8a17,0xef, 0, 0}, {0x8a18,0xa3, 0, 0}, {0x8a19,0xa3, 0, 0},
	{0x8a1a,0xa3, 0, 0}, {0x8a1b,0x80, 0, 0}, {0x8a1c,0xdf, 0, 0},
	{0x8a1d,0x75, 0, 0}, {0x8a1e,0x0f, 0, 0}, {0x8a1f,0x0a, 0, 0},
	{0x8a20,0xa2, 0, 0}, {0x8a21,0xaf, 0, 0}, {0x8a22,0x92, 0, 0},
	{0x8a23,0x32, 0, 0}, {0x8a24,0xc2, 0, 0}, {0x8a25,0xaf, 0, 0},
	{0x8a26,0xc2, 0, 0}, {0x8a27,0x33, 0, 0}, {0x8a28,0x12, 0, 0},
	{0x8a29,0x01, 0, 0}, {0x8a2a,0x6a, 0, 0}, {0x8a2b,0x12, 0, 0},
	{0x8a2c,0x02, 0, 0}, {0x8a2d,0x12, 0, 0}, {0x8a2e,0x12, 0, 0},
	{0x8a2f,0x01, 0, 0}, {0x8a30,0x6a, 0, 0}, {0x8a31,0x75, 0, 0},
	{0x8a32,0x51, 0, 0}, {0x8a33,0x05, 0, 0}, {0x8a34,0xaf, 0, 0},
	{0x8a35,0x51, 0, 0}, {0x8a36,0x15, 0, 0}, {0x8a37,0x51, 0, 0},
	{0x8a38,0xef, 0, 0}, {0x8a39,0x70, 0, 0}, {0x8a3a,0xf9, 0, 0},
	{0x8a3b,0xd2, 0, 0}, {0x8a3c,0x28, 0, 0}, {0x8a3d,0x12, 0, 0},
	{0x8a3e,0x01, 0, 0}, {0x8a3f,0x6c, 0, 0}, {0x8a40,0x75, 0, 0},
	{0x8a41,0x51, 0, 0}, {0x8a42,0x0a, 0, 0}, {0x8a43,0xaf, 0, 0},
	{0x8a44,0x51, 0, 0}, {0x8a45,0x15, 0, 0}, {0x8a46,0x51, 0, 0},
	{0x8a47,0xef, 0, 0}, {0x8a48,0x70, 0, 0}, {0x8a49,0xf9, 0, 0},
	{0x8a4a,0xc2, 0, 0}, {0x8a4b,0x29, 0, 0}, {0x8a4c,0x12, 0, 0},
	{0x8a4d,0x01, 0, 0}, {0x8a4e,0x6c, 0, 0}, {0x8a4f,0x75, 0, 0},
	{0x8a50,0x51, 0, 0}, {0x8a51,0x05, 0, 0}, {0x8a52,0xaf, 0, 0},
	{0x8a53,0x51, 0, 0}, {0x8a54,0x15, 0, 0}, {0x8a55,0x51, 0, 0},
	{0x8a56,0xef, 0, 0}, {0x8a57,0x70, 0, 0}, {0x8a58,0xf9, 0, 0},
	{0x8a59,0xc2, 0, 0}, {0x8a5a,0x28, 0, 0}, {0x8a5b,0x12, 0, 0},
	{0x8a5c,0x01, 0, 0}, {0x8a5d,0x6c, 0, 0}, {0x8a5e,0x75, 0, 0},
	{0x8a5f,0x51, 0, 0}, {0x8a60,0x05, 0, 0}, {0x8a61,0xaf, 0, 0},
	{0x8a62,0x51, 0, 0}, {0x8a63,0x15, 0, 0}, {0x8a64,0x51, 0, 0},
	{0x8a65,0xef, 0, 0}, {0x8a66,0x70, 0, 0}, {0x8a67,0xf9, 0, 0},
	{0x8a68,0x75, 0, 0}, {0x8a69,0x10, 0, 0}, {0x8a6a,0x18, 0, 0},
	{0x8a6b,0x12, 0, 0}, {0x8a6c,0x0b, 0, 0}, {0x8a6d,0x5d, 0, 0},
	{0x8a6e,0x75, 0, 0}, {0x8a6f,0x51, 0, 0}, {0x8a70,0x0a, 0, 0},
	{0x8a71,0xaf, 0, 0}, {0x8a72,0x51, 0, 0}, {0x8a73,0x15, 0, 0},
	{0x8a74,0x51, 0, 0}, {0x8a75,0xef, 0, 0}, {0x8a76,0x70, 0, 0},
	{0x8a77,0xf9, 0, 0}, {0x8a78,0xd2, 0, 0}, {0x8a79,0x28, 0, 0},
	{0x8a7a,0x12, 0, 0}, {0x8a7b,0x01, 0, 0}, {0x8a7c,0x6c, 0, 0},
	{0x8a7d,0x12, 0, 0}, {0x8a7e,0x02, 0, 0}, {0x8a7f,0x2f, 0, 0},
	{0x8a80,0xaf, 0, 0}, {0x8a81,0x51, 0, 0}, {0x8a82,0x15, 0, 0},
	{0x8a83,0x51, 0, 0}, {0x8a84,0xef, 0, 0}, {0x8a85,0x70, 0, 0},
	{0x8a86,0xf9, 0, 0}, {0x8a87,0xc2, 0, 0}, {0x8a88,0x28, 0, 0},
	{0x8a89,0x12, 0, 0}, {0x8a8a,0x01, 0, 0}, {0x8a8b,0x6c, 0, 0},
	{0x8a8c,0x75, 0, 0}, {0x8a8d,0x51, 0, 0}, {0x8a8e,0x0a, 0, 0},
	{0x8a8f,0xaf, 0, 0}, {0x8a90,0x51, 0, 0}, {0x8a91,0x15, 0, 0},
	{0x8a92,0x51, 0, 0}, {0x8a93,0xef, 0, 0}, {0x8a94,0x70, 0, 0},
	{0x8a95,0xf9, 0, 0}, {0x8a96,0x30, 0, 0}, {0x8a97,0x11, 0, 0},
	{0x8a98,0x03, 0, 0}, {0x8a99,0x02, 0, 0}, {0x8a9a,0x0b, 0, 0},
	{0x8a9b,0x14, 0, 0}, {0x8a9c,0x12, 0, 0}, {0x8a9d,0x01, 0, 0},
	{0x8a9e,0x6a, 0, 0}, {0x8a9f,0x12, 0, 0}, {0x8aa0,0x02, 0, 0},
	{0x8aa1,0x12, 0, 0}, {0x8aa2,0xe5, 0, 0}, {0x8aa3,0x0d, 0, 0},
	{0x8aa4,0xf5, 0, 0}, {0x8aa5,0x10, 0, 0}, {0x8aa6,0x12, 0, 0},
	{0x8aa7,0x0b, 0, 0}, {0x8aa8,0x5d, 0, 0}, {0x8aa9,0x75, 0, 0},
	{0x8aaa,0x51, 0, 0}, {0x8aab,0x0a, 0, 0}, {0x8aac,0xaf, 0, 0},
	{0x8aad,0x51, 0, 0}, {0x8aae,0x15, 0, 0}, {0x8aaf,0x51, 0, 0},
	{0x8ab0,0xef, 0, 0}, {0x8ab1,0x70, 0, 0}, {0x8ab2,0xf9, 0, 0},
	{0x8ab3,0xd2, 0, 0}, {0x8ab4,0x28, 0, 0}, {0x8ab5,0x12, 0, 0},
	{0x8ab6,0x01, 0, 0}, {0x8ab7,0x6c, 0, 0}, {0x8ab8,0x12, 0, 0},
	{0x8ab9,0x02, 0, 0}, {0x8aba,0x2f, 0, 0}, {0x8abb,0xaf, 0, 0},
	{0x8abc,0x51, 0, 0}, {0x8abd,0x15, 0, 0}, {0x8abe,0x51, 0, 0},
	{0x8abf,0xef, 0, 0}, {0x8ac0,0x70, 0, 0}, {0x8ac1,0xf9, 0, 0},
	{0x8ac2,0xc2, 0, 0}, {0x8ac3,0x28, 0, 0}, {0x8ac4,0x12, 0, 0},
	{0x8ac5,0x01, 0, 0}, {0x8ac6,0x6c, 0, 0}, {0x8ac7,0x75, 0, 0},
	{0x8ac8,0x51, 0, 0}, {0x8ac9,0x0a, 0, 0}, {0x8aca,0xaf, 0, 0},
	{0x8acb,0x51, 0, 0}, {0x8acc,0x15, 0, 0}, {0x8acd,0x51, 0, 0},
	{0x8ace,0xef, 0, 0}, {0x8acf,0x70, 0, 0}, {0x8ad0,0xf9, 0, 0},
	{0x8ad1,0x30, 0, 0}, {0x8ad2,0x11, 0, 0}, {0x8ad3,0x04, 0, 0},
	{0x8ad4,0x15, 0, 0}, {0x8ad5,0x0f, 0, 0}, {0x8ad6,0x80, 0, 0},
	{0x8ad7,0x45, 0, 0}, {0x8ad8,0x12, 0, 0}, {0x8ad9,0x01, 0, 0},
	{0x8ada,0x6a, 0, 0}, {0x8adb,0x12, 0, 0}, {0x8adc,0x02, 0, 0},
	{0x8add,0x12, 0, 0}, {0x8ade,0x85, 0, 0}, {0x8adf,0x0e, 0, 0},
	{0x8ae0,0x10, 0, 0}, {0x8ae1,0x12, 0, 0}, {0x8ae2,0x15, 0, 0},
	{0x8ae3,0xa4, 0, 0}, {0x8ae4,0xc2, 0, 0}, {0x8ae5,0x09, 0, 0},
	{0x8ae6,0x12, 0, 0}, {0x8ae7,0x02, 0, 0}, {0x8ae8,0x14, 0, 0},
	{0x8ae9,0x75, 0, 0}, {0x8aea,0x51, 0, 0}, {0x8aeb,0x0a, 0, 0},
	{0x8aec,0xaf, 0, 0}, {0x8aed,0x51, 0, 0}, {0x8aee,0x15, 0, 0},
	{0x8aef,0x51, 0, 0}, {0x8af0,0xef, 0, 0}, {0x8af1,0x70, 0, 0},
	{0x8af2,0xf9, 0, 0}, {0x8af3,0xd2, 0, 0}, {0x8af4,0x28, 0, 0},
	{0x8af5,0x12, 0, 0}, {0x8af6,0x01, 0, 0}, {0x8af7,0x6c, 0, 0},
	{0x8af8,0x12, 0, 0}, {0x8af9,0x02, 0, 0}, {0x8afa,0x2f, 0, 0},
	{0x8afb,0xaf, 0, 0}, {0x8afc,0x51, 0, 0}, {0x8afd,0x15, 0, 0},
	{0x8afe,0x51, 0, 0}, {0x8aff,0xef, 0, 0}, {0x8b00,0x70, 0, 0},
	{0x8b01,0xf9, 0, 0}, {0x8b02,0xc2, 0, 0}, {0x8b03,0x28, 0, 0},
	{0x8b04,0x12, 0, 0}, {0x8b05,0x01, 0, 0}, {0x8b06,0x6c, 0, 0},
	{0x8b07,0x75, 0, 0}, {0x8b08,0x51, 0, 0}, {0x8b09,0x0a, 0, 0},
	{0x8b0a,0xaf, 0, 0}, {0x8b0b,0x51, 0, 0}, {0x8b0c,0x15, 0, 0},
	{0x8b0d,0x51, 0, 0}, {0x8b0e,0xef, 0, 0}, {0x8b0f,0x70, 0, 0},
	{0x8b10,0xf9, 0, 0}, {0x8b11,0x30, 0, 0}, {0x8b12,0x11, 0, 0},
	{0x8b13,0x06, 0, 0}, {0x8b14,0x15, 0, 0}, {0x8b15,0x0f, 0, 0},
	{0x8b16,0xd2, 0, 0}, {0x8b17,0x33, 0, 0}, {0x8b18,0x80, 0, 0},
	{0x8b19,0x03, 0, 0}, {0x8b1a,0xe4, 0, 0}, {0x8b1b,0xf5, 0, 0},
	{0x8b1c,0x0f, 0, 0}, {0x8b1d,0x12, 0, 0}, {0x8b1e,0x01, 0, 0},
	{0x8b1f,0x6a, 0, 0}, {0x8b20,0x12, 0, 0}, {0x8b21,0x02, 0, 0},
	{0x8b22,0x12, 0, 0}, {0x8b23,0xc2, 0, 0}, {0x8b24,0x29, 0, 0},
	{0x8b25,0x12, 0, 0}, {0x8b26,0x01, 0, 0}, {0x8b27,0x6c, 0, 0},
	{0x8b28,0x75, 0, 0}, {0x8b29,0x51, 0, 0}, {0x8b2a,0x05, 0, 0},
	{0x8b2b,0xaf, 0, 0}, {0x8b2c,0x51, 0, 0}, {0x8b2d,0x15, 0, 0},
	{0x8b2e,0x51, 0, 0}, {0x8b2f,0xef, 0, 0}, {0x8b30,0x70, 0, 0},
	{0x8b31,0xf9, 0, 0}, {0x8b32,0xd2, 0, 0}, {0x8b33,0x28, 0, 0},
	{0x8b34,0x12, 0, 0}, {0x8b35,0x01, 0, 0}, {0x8b36,0x6c, 0, 0},
	{0x8b37,0x75, 0, 0}, {0x8b38,0x51, 0, 0}, {0x8b39,0x05, 0, 0},
	{0x8b3a,0xaf, 0, 0}, {0x8b3b,0x51, 0, 0}, {0x8b3c,0x15, 0, 0},
	{0x8b3d,0x51, 0, 0}, {0x8b3e,0xef, 0, 0}, {0x8b3f,0x70, 0, 0},
	{0x8b40,0xf9, 0, 0}, {0x8b41,0x12, 0, 0}, {0x8b42,0x01, 0, 0},
	{0x8b43,0x6a, 0, 0}, {0x8b44,0x75, 0, 0}, {0x8b45,0x51, 0, 0},
	{0x8b46,0x05, 0, 0}, {0x8b47,0xaf, 0, 0}, {0x8b48,0x51, 0, 0},
	{0x8b49,0x15, 0, 0}, {0x8b4a,0x51, 0, 0}, {0x8b4b,0xef, 0, 0},
	{0x8b4c,0x70, 0, 0}, {0x8b4d,0xf9, 0, 0}, {0x8b4e,0xa2, 0, 0},
	{0x8b4f,0x32, 0, 0}, {0x8b50,0x92, 0, 0}, {0x8b51,0xaf, 0, 0},
	{0x8b52,0xe5, 0, 0}, {0x8b53,0x0f, 0, 0}, {0x8b54,0xd3, 0, 0},
	{0x8b55,0x94, 0, 0}, {0x8b56,0x00, 0, 0}, {0x8b57,0x40, 0, 0},
	{0x8b58,0x03, 0, 0}, {0x8b59,0x02, 0, 0}, {0x8b5a,0x0a, 0, 0},
	{0x8b5b,0x24, 0, 0}, {0x8b5c,0x22, 0, 0}, {0x8b5d,0x12, 0, 0},
	{0x8b5e,0x15, 0, 0}, {0x8b5f,0xa4, 0, 0}, {0x8b60,0xc2, 0, 0},
	{0x8b61,0x09, 0, 0}, {0x8b62,0x90, 0, 0}, {0x8b63,0x30, 0, 0},
	{0x8b64,0x18, 0, 0}, {0x8b65,0xe5, 0, 0}, {0x8b66,0x21, 0, 0},
	{0x8b67,0xf0, 0, 0}, {0x8b68,0x22, 0, 0}, {0x8b69,0xe5, 0, 0},
	{0x8b6a,0x33, 0, 0}, {0x8b6b,0x70, 0, 0}, {0x8b6c,0x03, 0, 0},
	{0x8b6d,0x02, 0, 0}, {0x8b6e,0x0c, 0, 0}, {0x8b6f,0xa3, 0, 0},
	{0x8b70,0xc2, 0, 0}, {0x8b71,0xaf, 0, 0}, {0x8b72,0xaf, 0, 0},
	{0x8b73,0x33, 0, 0}, {0x8b74,0xe4, 0, 0}, {0x8b75,0xf5, 0, 0},
	{0x8b76,0x33, 0, 0}, {0x8b77,0xd2, 0, 0}, {0x8b78,0xaf, 0, 0},
	{0x8b79,0x90, 0, 0}, {0x8b7a,0x30, 0, 0}, {0x8b7b,0x25, 0, 0},
	{0x8b7c,0xe0, 0, 0}, {0x8b7d,0xf5, 0, 0}, {0x8b7e,0x7e, 0, 0},
	{0x8b7f,0x90, 0, 0}, {0x8b80,0x50, 0, 0}, {0x8b81,0x82, 0, 0},
	{0x8b82,0xe0, 0, 0}, {0x8b83,0xf5, 0, 0}, {0x8b84,0x66, 0, 0},
	{0x8b85,0xa3, 0, 0}, {0x8b86,0xe0, 0, 0}, {0x8b87,0xf5, 0, 0},
	{0x8b88,0x67, 0, 0}, {0x8b89,0xa3, 0, 0}, {0x8b8a,0xe0, 0, 0},
	{0x8b8b,0xf5, 0, 0}, {0x8b8c,0x68, 0, 0}, {0x8b8d,0xa3, 0, 0},
	{0x8b8e,0xe0, 0, 0}, {0x8b8f,0xf5, 0, 0}, {0x8b90,0x69, 0, 0},
	{0x8b91,0xef, 0, 0}, {0x8b92,0x12, 0, 0}, {0x8b93,0x09, 0, 0},
	{0x8b94,0xf7, 0, 0}, {0x8b95,0x0b, 0, 0}, {0x8b96,0xbd, 0, 0},
	{0x8b97,0x03, 0, 0}, {0x8b98,0x0b, 0, 0}, {0x8b99,0xdb, 0, 0},
	{0x8b9a,0x04, 0, 0}, {0x8b9b,0x0b, 0, 0}, {0x8b9c,0xfc, 0, 0},
	{0x8b9d,0x05, 0, 0}, {0x8b9e,0x0c, 0, 0}, {0x8b9f,0x26, 0, 0},
	{0x8ba0,0x06, 0, 0}, {0x8ba1,0x0c, 0, 0}, {0x8ba2,0x14, 0, 0},
	{0x8ba3,0x08, 0, 0}, {0x8ba4,0x0c, 0, 0}, {0x8ba5,0x33, 0, 0},
	{0x8ba6,0x10, 0, 0}, {0x8ba7,0x0c, 0, 0}, {0x8ba8,0x47, 0, 0},
	{0x8ba9,0x12, 0, 0}, {0x8baa,0x0c, 0, 0}, {0x8bab,0x4c, 0, 0},
	{0x8bac,0x20, 0, 0}, {0x8bad,0x0c, 0, 0}, {0x8bae,0x5a, 0, 0},
	{0x8baf,0x21, 0, 0}, {0x8bb0,0x0c, 0, 0}, {0x8bb1,0x5f, 0, 0},
	{0x8bb2,0x30, 0, 0}, {0x8bb3,0x0c, 0, 0}, {0x8bb4,0x88, 0, 0},
	{0x8bb5,0x50, 0, 0}, {0x8bb6,0x0c, 0, 0}, {0x8bb7,0x6a, 0, 0},
	{0x8bb8,0xd8, 0, 0}, {0x8bb9,0x00, 0, 0}, {0x8bba,0x00, 0, 0},
	{0x8bbb,0x0c, 0, 0}, {0x8bbc,0x95, 0, 0}, {0x8bbd,0x20, 0, 0},
	{0x8bbe,0x05, 0, 0}, {0x8bbf,0x03, 0, 0}, {0x8bc0,0x02, 0, 0},
	{0x8bc1,0x0c, 0, 0}, {0x8bc2,0x95, 0, 0}, {0x8bc3,0x30, 0, 0},
	{0x8bc4,0x00, 0, 0}, {0x8bc5,0x03, 0, 0}, {0x8bc6,0x02, 0, 0},
	{0x8bc7,0x0c, 0, 0}, {0x8bc8,0x95, 0, 0}, {0x8bc9,0x20, 0, 0},
	{0x8bca,0x07, 0, 0}, {0x8bcb,0x06, 0, 0}, {0x8bcc,0x20, 0, 0},
	{0x8bcd,0x06, 0, 0}, {0x8bce,0x03, 0, 0}, {0x8bcf,0xc3, 0, 0},
	{0x8bd0,0x80, 0, 0}, {0x8bd1,0x01, 0, 0}, {0x8bd2,0xd3, 0, 0},
	{0x8bd3,0x92, 0, 0}, {0x8bd4,0x36, 0, 0}, {0x8bd5,0xd2, 0, 0},
	{0x8bd6,0x07, 0, 0}, {0x8bd7,0xc2, 0, 0}, {0x8bd8,0x06, 0, 0},
	{0x8bd9,0x80, 0, 0}, {0x8bda,0x1c, 0, 0}, {0x8bdb,0x20, 0, 0},
	{0x8bdc,0x05, 0, 0}, {0x8bdd,0x03, 0, 0}, {0x8bde,0x02, 0, 0},
	{0x8bdf,0x0c, 0, 0}, {0x8be0,0x95, 0, 0}, {0x8be1,0x30, 0, 0},
	{0x8be2,0x00, 0, 0}, {0x8be3,0x03, 0, 0}, {0x8be4,0x02, 0, 0},
	{0x8be5,0x0c, 0, 0}, {0x8be6,0x95, 0, 0}, {0x8be7,0x20, 0, 0},
	{0x8be8,0x07, 0, 0}, {0x8be9,0x06, 0, 0}, {0x8bea,0x20, 0, 0},
	{0x8beb,0x06, 0, 0}, {0x8bec,0x03, 0, 0}, {0x8bed,0xc3, 0, 0},
	{0x8bee,0x80, 0, 0}, {0x8bef,0x01, 0, 0}, {0x8bf0,0xd3, 0, 0},
	{0x8bf1,0x92, 0, 0}, {0x8bf2,0x36, 0, 0}, {0x8bf3,0xd2, 0, 0},
	{0x8bf4,0x07, 0, 0}, {0x8bf5,0xd2, 0, 0}, {0x8bf6,0x06, 0, 0},
	{0x8bf7,0x12, 0, 0}, {0x8bf8,0x02, 0, 0}, {0x8bf9,0x8a, 0, 0},
	{0x8bfa,0x80, 0, 0}, {0x8bfb,0x24, 0, 0}, {0x8bfc,0x20, 0, 0},
	{0x8bfd,0x05, 0, 0}, {0x8bfe,0x03, 0, 0}, {0x8bff,0x02, 0, 0},
	{0x8c00,0x0c, 0, 0}, {0x8c01,0x95, 0, 0}, {0x8c02,0x30, 0, 0},
	{0x8c03,0x00, 0, 0}, {0x8c04,0x03, 0, 0}, {0x8c05,0x02, 0, 0},
	{0x8c06,0x0c, 0, 0}, {0x8c07,0x95, 0, 0}, {0x8c08,0xc2, 0, 0},
	{0x8c09,0x07, 0, 0}, {0x8c0a,0xd2, 0, 0}, {0x8c0b,0x06, 0, 0},
	{0x8c0c,0x12, 0, 0}, {0x8c0d,0x02, 0, 0}, {0x8c0e,0x9b, 0, 0},
	{0x8c0f,0xc2, 0, 0}, {0x8c10,0x04, 0, 0}, {0x8c11,0x02, 0, 0},
	{0x8c12,0x0c, 0, 0}, {0x8c13,0x95, 0, 0}, {0x8c14,0x12, 0, 0},
	{0x8c15,0x02, 0, 0}, {0x8c16,0x51, 0, 0}, {0x8c17,0x30, 0, 0},
	{0x8c18,0x05, 0, 0}, {0x8c19,0x06, 0, 0}, {0x8c1a,0xe4, 0, 0},
	{0x8c1b,0xf5, 0, 0}, {0x8c1c,0x0c, 0, 0}, {0x8c1d,0x12, 0, 0},
	{0x8c1e,0x13, 0, 0}, {0x8c1f,0x64, 0, 0}, {0x8c20,0xc2, 0, 0},
	{0x8c21,0x31, 0, 0}, {0x8c22,0xd2, 0, 0}, {0x8c23,0x34, 0, 0},
	{0x8c24,0x80, 0, 0}, {0x8c25,0x6f, 0, 0}, {0x8c26,0x30, 0, 0},
	{0x8c27,0x07, 0, 0}, {0x8c28,0x6c, 0, 0}, {0x8c29,0x30, 0, 0},
	{0x8c2a,0x06, 0, 0}, {0x8c2b,0x69, 0, 0}, {0x8c2c,0x12, 0, 0},
	{0x8c2d,0x02, 0, 0}, {0x8c2e,0x8a, 0, 0}, {0x8c2f,0xd2, 0, 0},
	{0x8c30,0x31, 0, 0}, {0x8c31,0x80, 0, 0}, {0x8c32,0x62, 0, 0},
	{0x8c33,0x20, 0, 0}, {0x8c34,0x07, 0, 0}, {0x8c35,0x03, 0, 0},
	{0x8c36,0x30, 0, 0}, {0x8c37,0x06, 0, 0}, {0x8c38,0x09, 0, 0},
	{0x8c39,0xe5, 0, 0}, {0x8c3a,0x7e, 0, 0}, {0x8c3b,0x64, 0, 0},
	{0x8c3c,0x0e, 0, 0}, {0x8c3d,0x70, 0, 0}, {0x8c3e,0x56, 0, 0},
	{0x8c3f,0x20, 0, 0}, {0x8c40,0x00, 0, 0}, {0x8c41,0x53, 0, 0},
	{0x8c42,0x12, 0, 0}, {0x8c43,0x05, 0, 0}, {0x8c44,0x2f, 0, 0},
	{0x8c45,0x80, 0, 0}, {0x8c46,0x4e, 0, 0}, {0x8c47,0x12, 0, 0},
	{0x8c48,0x06, 0, 0}, {0x8c49,0xf8, 0, 0}, {0x8c4a,0x80, 0, 0},
	{0x8c4b,0x49, 0, 0}, {0x8c4c,0x30, 0, 0}, {0x8c4d,0x05, 0, 0},
	{0x8c4e,0x46, 0, 0}, {0x8c4f,0x20, 0, 0}, {0x8c50,0x07, 0, 0},
	{0x8c51,0x43, 0, 0}, {0x8c52,0x20, 0, 0}, {0x8c53,0x06, 0, 0},
	{0x8c54,0x40, 0, 0}, {0x8c55,0x12, 0, 0}, {0x8c56,0x16, 0, 0},
	{0x8c57,0xc4, 0, 0}, {0x8c58,0x80, 0, 0}, {0x8c59,0x3b, 0, 0},
	{0x8c5a,0x12, 0, 0}, {0x8c5b,0x10, 0, 0}, {0x8c5c,0xfe, 0, 0},
	{0x8c5d,0x80, 0, 0}, {0x8c5e,0x36, 0, 0}, {0x8c5f,0x20, 0, 0},
	{0x8c60,0x07, 0, 0}, {0x8c61,0x33, 0, 0}, {0x8c62,0x20, 0, 0},
	{0x8c63,0x06, 0, 0}, {0x8c64,0x30, 0, 0}, {0x8c65,0x12, 0, 0},
	{0x8c66,0x16, 0, 0}, {0x8c67,0xd3, 0, 0}, {0x8c68,0x80, 0, 0},
	{0x8c69,0x2b, 0, 0}, {0x8c6a,0xe5, 0, 0}, {0x8c6b,0x7e, 0, 0},
	{0x8c6c,0x64, 0, 0}, {0x8c6d,0x01, 0, 0}, {0x8c6e,0x70, 0, 0},
	{0x8c6f,0x25, 0, 0}, {0x8c70,0xd2, 0, 0}, {0x8c71,0x35, 0, 0},
	{0x8c72,0x90, 0, 0}, {0x8c73,0x50, 0, 0}, {0x8c74,0x82, 0, 0},
	{0x8c75,0xe5, 0, 0}, {0x8c76,0x74, 0, 0}, {0x8c77,0xf0, 0, 0},
	{0x8c78,0xa3, 0, 0}, {0x8c79,0xe5, 0, 0}, {0x8c7a,0x75, 0, 0},
	{0x8c7b,0xf0, 0, 0}, {0x8c7c,0xa3, 0, 0}, {0x8c7d,0xe5, 0, 0},
	{0x8c7e,0x76, 0, 0}, {0x8c7f,0xf0, 0, 0}, {0x8c80,0xa3, 0, 0},
	{0x8c81,0xe5, 0, 0}, {0x8c82,0x77, 0, 0}, {0x8c83,0xf0, 0, 0},
	{0x8c84,0xc2, 0, 0}, {0x8c85,0x35, 0, 0}, {0x8c86,0x80, 0, 0},
	{0x8c87,0x0d, 0, 0}, {0x8c88,0x90, 0, 0}, {0x8c89,0x50, 0, 0},
	{0x8c8a,0x82, 0, 0}, {0x8c8b,0x30, 0, 0}, {0x8c8c,0x33, 0, 0},
	{0x8c8d,0x05, 0, 0}, {0x8c8e,0x74, 0, 0}, {0x8c8f,0x55, 0, 0},
	{0x8c90,0xf0, 0, 0}, {0x8c91,0x80, 0, 0}, {0x8c92,0x02, 0, 0},
	{0x8c93,0xe4, 0, 0}, {0x8c94,0xf0, 0, 0}, {0x8c95,0x20, 0, 0},
	{0x8c96,0x07, 0, 0}, {0x8c97,0x06, 0, 0}, {0x8c98,0x30, 0, 0},
	{0x8c99,0x06, 0, 0}, {0x8c9a,0x03, 0, 0}, {0x8c9b,0x30, 0, 0},
	{0x8c9c,0x04, 0, 0}, {0x8c9d,0x05, 0, 0}, {0x8c9e,0x90, 0, 0},
	{0x8c9f,0x30, 0, 0}, {0x8ca0,0x25, 0, 0}, {0x8ca1,0xe4, 0, 0},
	{0x8ca2,0xf0, 0, 0}, {0x8ca3,0x22, 0, 0}, {0x8ca4,0xc0, 0, 0},
	{0x8ca5,0xe0, 0, 0}, {0x8ca6,0xc0, 0, 0}, {0x8ca7,0xf0, 0, 0},
	{0x8ca8,0xc0, 0, 0}, {0x8ca9,0x83, 0, 0}, {0x8caa,0xc0, 0, 0},
	{0x8cab,0x82, 0, 0}, {0x8cac,0xc0, 0, 0}, {0x8cad,0xd0, 0, 0},
	{0x8cae,0x75, 0, 0}, {0x8caf,0xd0, 0, 0}, {0x8cb0,0x00, 0, 0},
	{0x8cb1,0xc0, 0, 0}, {0x8cb2,0x00, 0, 0}, {0x8cb3,0xc0, 0, 0},
	{0x8cb4,0x01, 0, 0}, {0x8cb5,0xc0, 0, 0}, {0x8cb6,0x02, 0, 0},
	{0x8cb7,0xc0, 0, 0}, {0x8cb8,0x03, 0, 0}, {0x8cb9,0xc0, 0, 0},
	{0x8cba,0x04, 0, 0}, {0x8cbb,0xc0, 0, 0}, {0x8cbc,0x05, 0, 0},
	{0x8cbd,0xc0, 0, 0}, {0x8cbe,0x06, 0, 0}, {0x8cbf,0xc0, 0, 0},
	{0x8cc0,0x07, 0, 0}, {0x8cc1,0x90, 0, 0}, {0x8cc2,0x3f, 0, 0},
	{0x8cc3,0x0c, 0, 0}, {0x8cc4,0xe0, 0, 0}, {0x8cc5,0xf5, 0, 0},
	{0x8cc6,0x08, 0, 0}, {0x8cc7,0xe5, 0, 0}, {0x8cc8,0x08, 0, 0},
	{0x8cc9,0x20, 0, 0}, {0x8cca,0xe3, 0, 0}, {0x8ccb,0x03, 0, 0},
	{0x8ccc,0x02, 0, 0}, {0x8ccd,0x0d, 0, 0}, {0x8cce,0x55, 0, 0},
	{0x8ccf,0x30, 0, 0}, {0x8cd0,0x35, 0, 0}, {0x8cd1,0x03, 0, 0},
	{0x8cd2,0x02, 0, 0}, {0x8cd3,0x0d, 0, 0}, {0x8cd4,0x55, 0, 0},
	{0x8cd5,0x90, 0, 0}, {0x8cd6,0x60, 0, 0}, {0x8cd7,0x16, 0, 0},
	{0x8cd8,0xe0, 0, 0}, {0x8cd9,0xf5, 0, 0}, {0x8cda,0x6a, 0, 0},
	{0x8cdb,0xa3, 0, 0}, {0x8cdc,0xe0, 0, 0}, {0x8cdd,0xf5, 0, 0},
	{0x8cde,0x6b, 0, 0}, {0x8cdf,0x90, 0, 0}, {0x8ce0,0x60, 0, 0},
	{0x8ce1,0x1e, 0, 0}, {0x8ce2,0xe0, 0, 0}, {0x8ce3,0xf5, 0, 0},
	{0x8ce4,0x6c, 0, 0}, {0x8ce5,0xa3, 0, 0}, {0x8ce6,0xe0, 0, 0},
	{0x8ce7,0xf5, 0, 0}, {0x8ce8,0x6d, 0, 0}, {0x8ce9,0x90, 0, 0},
	{0x8cea,0x60, 0, 0}, {0x8ceb,0x26, 0, 0}, {0x8cec,0xe0, 0, 0},
	{0x8ced,0xf5, 0, 0}, {0x8cee,0x6e, 0, 0}, {0x8cef,0xa3, 0, 0},
	{0x8cf0,0xe0, 0, 0}, {0x8cf1,0xf5, 0, 0}, {0x8cf2,0x6f, 0, 0},
	{0x8cf3,0x90, 0, 0}, {0x8cf4,0x60, 0, 0}, {0x8cf5,0x2e, 0, 0},
	{0x8cf6,0xe0, 0, 0}, {0x8cf7,0xf5, 0, 0}, {0x8cf8,0x70, 0, 0},
	{0x8cf9,0xa3, 0, 0}, {0x8cfa,0xe0, 0, 0}, {0x8cfb,0xf5, 0, 0},
	{0x8cfc,0x71, 0, 0}, {0x8cfd,0x90, 0, 0}, {0x8cfe,0x60, 0, 0},
	{0x8cff,0x36, 0, 0}, {0x8d00,0x12, 0, 0}, {0x8d01,0x00, 0, 0},
	{0x8d02,0x16, 0, 0}, {0x8d03,0x12, 0, 0}, {0x8d04,0x01, 0, 0},
	{0x8d05,0xc9, 0, 0}, {0x8d06,0x40, 0, 0}, {0x8d07,0x06, 0, 0},
	{0x8d08,0x75, 0, 0}, {0x8d09,0x2a, 0, 0}, {0x8d0a,0xff, 0, 0},
	{0x8d0b,0x75, 0, 0}, {0x8d0c,0x2b, 0, 0}, {0x8d0d,0xff, 0, 0},
	{0x8d0e,0x85, 0, 0}, {0x8d0f,0x2a, 0, 0}, {0x8d10,0x74, 0, 0},
	{0x8d11,0x85, 0, 0}, {0x8d12,0x2b, 0, 0}, {0x8d13,0x75, 0, 0},
	{0x8d14,0x90, 0, 0}, {0x8d15,0x60, 0, 0}, {0x8d16,0x1a, 0, 0},
	{0x8d17,0xe0, 0, 0}, {0x8d18,0xf5, 0, 0}, {0x8d19,0x6a, 0, 0},
	{0x8d1a,0xa3, 0, 0}, {0x8d1b,0xe0, 0, 0}, {0x8d1c,0xf5, 0, 0},
	{0x8d1d,0x6b, 0, 0}, {0x8d1e,0x90, 0, 0}, {0x8d1f,0x60, 0, 0},
	{0x8d20,0x22, 0, 0}, {0x8d21,0xe0, 0, 0}, {0x8d22,0xf5, 0, 0},
	{0x8d23,0x6c, 0, 0}, {0x8d24,0xa3, 0, 0}, {0x8d25,0xe0, 0, 0},
	{0x8d26,0xf5, 0, 0}, {0x8d27,0x6d, 0, 0}, {0x8d28,0x90, 0, 0},
	{0x8d29,0x60, 0, 0}, {0x8d2a,0x2a, 0, 0}, {0x8d2b,0xe0, 0, 0},
	{0x8d2c,0xf5, 0, 0}, {0x8d2d,0x6e, 0, 0}, {0x8d2e,0xa3, 0, 0},
	{0x8d2f,0xe0, 0, 0}, {0x8d30,0xf5, 0, 0}, {0x8d31,0x6f, 0, 0},
	{0x8d32,0x90, 0, 0}, {0x8d33,0x60, 0, 0}, {0x8d34,0x32, 0, 0},
	{0x8d35,0xe0, 0, 0}, {0x8d36,0xf5, 0, 0}, {0x8d37,0x70, 0, 0},
	{0x8d38,0xa3, 0, 0}, {0x8d39,0xe0, 0, 0}, {0x8d3a,0xf5, 0, 0},
	{0x8d3b,0x71, 0, 0}, {0x8d3c,0x90, 0, 0}, {0x8d3d,0x60, 0, 0},
	{0x8d3e,0x3a, 0, 0}, {0x8d3f,0x12, 0, 0}, {0x8d40,0x00, 0, 0},
	{0x8d41,0x16, 0, 0}, {0x8d42,0x12, 0, 0}, {0x8d43,0x01, 0, 0},
	{0x8d44,0xc9, 0, 0}, {0x8d45,0x40, 0, 0}, {0x8d46,0x06, 0, 0},
	{0x8d47,0x75, 0, 0}, {0x8d48,0x2a, 0, 0}, {0x8d49,0xff, 0, 0},
	{0x8d4a,0x75, 0, 0}, {0x8d4b,0x2b, 0, 0}, {0x8d4c,0xff, 0, 0},
	{0x8d4d,0x85, 0, 0}, {0x8d4e,0x2a, 0, 0}, {0x8d4f,0x76, 0, 0},
	{0x8d50,0x85, 0, 0}, {0x8d51,0x2b, 0, 0}, {0x8d52,0x77, 0, 0},
	{0x8d53,0xd2, 0, 0}, {0x8d54,0x3d, 0, 0}, {0x8d55,0xe5, 0, 0},
	{0x8d56,0x08, 0, 0}, {0x8d57,0x30, 0, 0}, {0x8d58,0xe5, 0, 0},
	{0x8d59,0x44, 0, 0}, {0x8d5a,0x90, 0, 0}, {0x8d5b,0x56, 0, 0},
	{0x8d5c,0x90, 0, 0}, {0x8d5d,0xe0, 0, 0}, {0x8d5e,0xf5, 0, 0},
	{0x8d5f,0x55, 0, 0}, {0x8d60,0xe5, 0, 0}, {0x8d61,0x7b, 0, 0},
	{0x8d62,0x24, 0, 0}, {0x8d63,0x02, 0, 0}, {0x8d64,0xff, 0, 0},
	{0x8d65,0xe4, 0, 0}, {0x8d66,0x33, 0, 0}, {0x8d67,0xfe, 0, 0},
	{0x8d68,0xad, 0, 0}, {0x8d69,0x55, 0, 0}, {0x8d6a,0xc3, 0, 0},
	{0x8d6b,0xef, 0, 0}, {0x8d6c,0x9d, 0, 0}, {0x8d6d,0x74, 0, 0},
	{0x8d6e,0x80, 0, 0}, {0x8d6f,0xf8, 0, 0}, {0x8d70,0x6e, 0, 0},
	{0x8d71,0x98, 0, 0}, {0x8d72,0x50, 0, 0}, {0x8d73,0x02, 0, 0},
	{0x8d74,0x80, 0, 0}, {0x8d75,0x01, 0, 0}, {0x8d76,0xc3, 0, 0},
	{0x8d77,0x92, 0, 0}, {0x8d78,0x27, 0, 0}, {0x8d79,0xaf, 0, 0},
	{0x8d7a,0x55, 0, 0}, {0x8d7b,0xef, 0, 0}, {0x8d7c,0x24, 0, 0},
	{0x8d7d,0x02, 0, 0}, {0x8d7e,0xff, 0, 0}, {0x8d7f,0xe4, 0, 0},
	{0x8d80,0x33, 0, 0}, {0x8d81,0xfe, 0, 0}, {0x8d82,0xc3, 0, 0},
	{0x8d83,0xef, 0, 0}, {0x8d84,0x95, 0, 0}, {0x8d85,0x7b, 0, 0},
	{0x8d86,0x74, 0, 0}, {0x8d87,0x80, 0, 0}, {0x8d88,0xf8, 0, 0},
	{0x8d89,0x6e, 0, 0}, {0x8d8a,0x98, 0, 0}, {0x8d8b,0x50, 0, 0},
	{0x8d8c,0x02, 0, 0}, {0x8d8d,0x80, 0, 0}, {0x8d8e,0x02, 0, 0},
	{0x8d8f,0xa2, 0, 0}, {0x8d90,0x27, 0, 0}, {0x8d91,0x92, 0, 0},
	{0x8d92,0x27, 0, 0}, {0x8d93,0x30, 0, 0}, {0x8d94,0x27, 0, 0},
	{0x8d95,0x04, 0, 0}, {0x8d96,0xaf, 0, 0}, {0x8d97,0x55, 0, 0},
	{0x8d98,0x80, 0, 0}, {0x8d99,0x02, 0, 0}, {0x8d9a,0xaf, 0, 0},
	{0x8d9b,0x7b, 0, 0}, {0x8d9c,0x8f, 0, 0}, {0x8d9d,0x7b, 0, 0},
	{0x8d9e,0xe5, 0, 0}, {0x8d9f,0x08, 0, 0}, {0x8da0,0x30, 0, 0},
	{0x8da1,0xe1, 0, 0}, {0x8da2,0x08, 0, 0}, {0x8da3,0x90, 0, 0},
	{0x8da4,0x30, 0, 0}, {0x8da5,0x24, 0, 0}, {0x8da6,0xe0, 0, 0},
	{0x8da7,0xf5, 0, 0}, {0x8da8,0x33, 0, 0}, {0x8da9,0xe4, 0, 0},
	{0x8daa,0xf0, 0, 0}, {0x8dab,0x90, 0, 0}, {0x8dac,0x3f, 0, 0},
	{0x8dad,0x0c, 0, 0}, {0x8dae,0xe5, 0, 0}, {0x8daf,0x08, 0, 0},
	{0x8db0,0xf0, 0, 0}, {0x8db1,0xd0, 0, 0}, {0x8db2,0x07, 0, 0},
	{0x8db3,0xd0, 0, 0}, {0x8db4,0x06, 0, 0}, {0x8db5,0xd0, 0, 0},
	{0x8db6,0x05, 0, 0}, {0x8db7,0xd0, 0, 0}, {0x8db8,0x04, 0, 0},
	{0x8db9,0xd0, 0, 0}, {0x8dba,0x03, 0, 0}, {0x8dbb,0xd0, 0, 0},
	{0x8dbc,0x02, 0, 0}, {0x8dbd,0xd0, 0, 0}, {0x8dbe,0x01, 0, 0},
	{0x8dbf,0xd0, 0, 0}, {0x8dc0,0x00, 0, 0}, {0x8dc1,0xd0, 0, 0},
	{0x8dc2,0xd0, 0, 0}, {0x8dc3,0xd0, 0, 0}, {0x8dc4,0x82, 0, 0},
	{0x8dc5,0xd0, 0, 0}, {0x8dc6,0x83, 0, 0}, {0x8dc7,0xd0, 0, 0},
	{0x8dc8,0xf0, 0, 0}, {0x8dc9,0xd0, 0, 0}, {0x8dca,0xe0, 0, 0},
	{0x8dcb,0x32, 0, 0}, {0x8dcc,0x30, 0, 0}, {0x8dcd,0x04, 0, 0},
	{0x8dce,0x03, 0, 0}, {0x8dcf,0x02, 0, 0}, {0x8dd0,0x0e, 0, 0},
	{0x8dd1,0xac, 0, 0}, {0x8dd2,0xd2, 0, 0}, {0x8dd3,0x04, 0, 0},
	{0x8dd4,0xe5, 0, 0}, {0x8dd5,0x7e, 0, 0}, {0x8dd6,0xb4, 0, 0},
	{0x8dd7,0x01, 0, 0}, {0x8dd8,0x06, 0, 0}, {0x8dd9,0x12, 0, 0},
	{0x8dda,0x16, 0, 0}, {0x8ddb,0xa4, 0, 0}, {0x8ddc,0x02, 0, 0},
	{0x8ddd,0x0e, 0, 0}, {0x8dde,0xa5, 0, 0}, {0x8ddf,0xe5, 0, 0},
	{0x8de0,0x7e, 0, 0}, {0x8de1,0xb4, 0, 0}, {0x8de2,0x02, 0, 0},
	{0x8de3,0x06, 0, 0}, {0x8de4,0x12, 0, 0}, {0x8de5,0x16, 0, 0},
	{0x8de6,0xb5, 0, 0}, {0x8de7,0x02, 0, 0}, {0x8de8,0x0e, 0, 0},
	{0x8de9,0xa5, 0, 0}, {0x8dea,0xe5, 0, 0}, {0x8deb,0x7e, 0, 0},
	{0x8dec,0xb4, 0, 0}, {0x8ded,0x03, 0, 0}, {0x8dee,0x05, 0, 0},
	{0x8def,0xe4, 0, 0}, {0x8df0,0xf5, 0, 0}, {0x8df1,0x0c, 0, 0},
	{0x8df2,0x80, 0, 0}, {0x8df3,0x08, 0, 0}, {0x8df4,0xe5, 0, 0},
	{0x8df5,0x7e, 0, 0}, {0x8df6,0xb4, 0, 0}, {0x8df7,0x04, 0, 0},
	{0x8df8,0x09, 0, 0}, {0x8df9,0x85, 0, 0}, {0x8dfa,0x7c, 0, 0},
	{0x8dfb,0x0c, 0, 0}, {0x8dfc,0x12, 0, 0}, {0x8dfd,0x13, 0, 0},
	{0x8dfe,0x64, 0, 0}, {0x8dff,0x02, 0, 0}, {0x8e00,0x0e, 0, 0},
	{0x8e01,0xa5, 0, 0}, {0x8e02,0xe5, 0, 0}, {0x8e03,0x7e, 0, 0},
	{0x8e04,0x64, 0, 0}, {0x8e05,0x0f, 0, 0}, {0x8e06,0x70, 0, 0},
	{0x8e07,0x15, 0, 0}, {0x8e08,0x12, 0, 0}, {0x8e09,0x02, 0, 0},
	{0x8e0a,0xb3, 0, 0}, {0x8e0b,0x40, 0, 0}, {0x8e0c,0x06, 0, 0},
	{0x8e0d,0x7e, 0, 0}, {0x8e0e,0x03, 0, 0}, {0x8e0f,0x7f, 0, 0},
	{0x8e10,0xff, 0, 0}, {0x8e11,0x80, 0, 0}, {0x8e12,0x04, 0, 0},
	{0x8e13,0xae, 0, 0}, {0x8e14,0x68, 0, 0}, {0x8e15,0xaf, 0, 0},
	{0x8e16,0x69, 0, 0}, {0x8e17,0x12, 0, 0}, {0x8e18,0x0e, 0, 0},
	{0x8e19,0xad, 0, 0}, {0x8e1a,0x02, 0, 0}, {0x8e1b,0x0e, 0, 0},
	{0x8e1c,0xa5, 0, 0}, {0x8e1d,0xe5, 0, 0}, {0x8e1e,0x7e, 0, 0},
	{0x8e1f,0x64, 0, 0}, {0x8e20,0x10, 0, 0}, {0x8e21,0x60, 0, 0},
	{0x8e22,0x03, 0, 0}, {0x8e23,0x02, 0, 0}, {0x8e24,0x0e, 0, 0},
	{0x8e25,0xa5, 0, 0}, {0x8e26,0xf5, 0, 0}, {0x8e27,0x66, 0, 0},
	{0x8e28,0xf5, 0, 0}, {0x8e29,0x67, 0, 0}, {0x8e2a,0xf5, 0, 0},
	{0x8e2b,0x68, 0, 0}, {0x8e2c,0xab, 0, 0}, {0x8e2d,0x69, 0, 0},
	{0x8e2e,0xaa, 0, 0}, {0x8e2f,0x68, 0, 0}, {0x8e30,0xa9, 0, 0},
	{0x8e31,0x67, 0, 0}, {0x8e32,0xa8, 0, 0}, {0x8e33,0x66, 0, 0},
	{0x8e34,0x12, 0, 0}, {0x8e35,0x01, 0, 0}, {0x8e36,0x73, 0, 0},
	{0x8e37,0xfe, 0, 0}, {0x8e38,0xe4, 0, 0}, {0x8e39,0xfc, 0, 0},
	{0x8e3a,0xfd, 0, 0}, {0x8e3b,0x12, 0, 0}, {0x8e3c,0x01, 0, 0},
	{0x8e3d,0xf1, 0, 0}, {0x8e3e,0xe4, 0, 0}, {0x8e3f,0x7b, 0, 0},
	{0x8e40,0xff, 0, 0}, {0x8e41,0xfa, 0, 0}, {0x8e42,0xf9, 0, 0},
	{0x8e43,0xf8, 0, 0}, {0x8e44,0x12, 0, 0}, {0x8e45,0x02, 0, 0},
	{0x8e46,0x67, 0, 0}, {0x8e47,0x12, 0, 0}, {0x8e48,0x02, 0, 0},
	{0x8e49,0x39, 0, 0}, {0x8e4a,0xe4, 0, 0}, {0x8e4b,0x93, 0, 0},
	{0x8e4c,0xfe, 0, 0}, {0x8e4d,0x74, 0, 0}, {0x8e4e,0x01, 0, 0},
	{0x8e4f,0x93, 0, 0}, {0x8e50,0xff, 0, 0}, {0x8e51,0xe4, 0, 0},
	{0x8e52,0xfc, 0, 0}, {0x8e53,0xfd, 0, 0}, {0x8e54,0xe5, 0, 0},
	{0x8e55,0x69, 0, 0}, {0x8e56,0x2f, 0, 0}, {0x8e57,0xf5, 0, 0},
	{0x8e58,0x69, 0, 0}, {0x8e59,0xe5, 0, 0}, {0x8e5a,0x68, 0, 0},
	{0x8e5b,0x3e, 0, 0}, {0x8e5c,0xf5, 0, 0}, {0x8e5d,0x68, 0, 0},
	{0x8e5e,0xed, 0, 0}, {0x8e5f,0x35, 0, 0}, {0x8e60,0x67, 0, 0},
	{0x8e61,0xf5, 0, 0}, {0x8e62,0x67, 0, 0}, {0x8e63,0xec, 0, 0},
	{0x8e64,0x35, 0, 0}, {0x8e65,0x66, 0, 0}, {0x8e66,0xf5, 0, 0},
	{0x8e67,0x66, 0, 0}, {0x8e68,0x12, 0, 0}, {0x8e69,0x02, 0, 0},
	{0x8e6a,0xb3, 0, 0}, {0x8e6b,0x40, 0, 0}, {0x8e6c,0x06, 0, 0},
	{0x8e6d,0x7e, 0, 0}, {0x8e6e,0x03, 0, 0}, {0x8e6f,0x7f, 0, 0},
	{0x8e70,0xff, 0, 0}, {0x8e71,0x80, 0, 0}, {0x8e72,0x04, 0, 0},
	{0x8e73,0xae, 0, 0}, {0x8e74,0x68, 0, 0}, {0x8e75,0xaf, 0, 0},
	{0x8e76,0x69, 0, 0}, {0x8e77,0x12, 0, 0}, {0x8e78,0x0e, 0, 0},
	{0x8e79,0xad, 0, 0}, {0x8e7a,0xe4, 0, 0}, {0x8e7b,0xf5, 0, 0},
	{0x8e7c,0x67, 0, 0}, {0x8e7d,0xf5, 0, 0}, {0x8e7e,0x67, 0, 0},
	{0x8e7f,0xe5, 0, 0}, {0x8e80,0x67, 0, 0}, {0x8e81,0xd3, 0, 0},
	{0x8e82,0x95, 0, 0}, {0x8e83,0x7c, 0, 0}, {0x8e84,0x50, 0, 0},
	{0x8e85,0x1c, 0, 0}, {0x8e86,0x12, 0, 0}, {0x8e87,0x02, 0, 0},
	{0x8e88,0x39, 0, 0}, {0x8e89,0xaf, 0, 0}, {0x8e8a,0x67, 0, 0},
	{0x8e8b,0x75, 0, 0}, {0x8e8c,0xf0, 0, 0}, {0x8e8d,0x02, 0, 0},
	{0x8e8e,0xef, 0, 0}, {0x8e8f,0x12, 0, 0}, {0x8e90,0x09, 0, 0},
	{0x8e91,0xeb, 0, 0}, {0x8e92,0xc3, 0, 0}, {0x8e93,0x74, 0, 0},
	{0x8e94,0x01, 0, 0}, {0x8e95,0x93, 0, 0}, {0x8e96,0x95, 0, 0},
	{0x8e97,0x69, 0, 0}, {0x8e98,0xe4, 0, 0}, {0x8e99,0x93, 0, 0},
	{0x8e9a,0x95, 0, 0}, {0x8e9b,0x68, 0, 0}, {0x8e9c,0x50, 0, 0},
	{0x8e9d,0x04, 0, 0}, {0x8e9e,0x05, 0, 0}, {0x8e9f,0x67, 0, 0},
	{0x8ea0,0x80, 0, 0}, {0x8ea1,0xdd, 0, 0}, {0x8ea2,0x85, 0, 0},
	{0x8ea3,0x67, 0, 0}, {0x8ea4,0x7d, 0, 0}, {0x8ea5,0x90, 0, 0},
	{0x8ea6,0x30, 0, 0}, {0x8ea7,0x25, 0, 0}, {0x8ea8,0xe4, 0, 0},
	{0x8ea9,0xf0, 0, 0}, {0x8eaa,0xd2, 0, 0}, {0x8eab,0x34, 0, 0},
	{0x8eac,0x22, 0, 0}, {0x8ead,0x8e, 0, 0}, {0x8eae,0x68, 0, 0},
	{0x8eaf,0x8f, 0, 0}, {0x8eb0,0x69, 0, 0}, {0x8eb1,0x85, 0, 0},
	{0x8eb2,0x68, 0, 0}, {0x8eb3,0x64, 0, 0}, {0x8eb4,0x85, 0, 0},
	{0x8eb5,0x69, 0, 0}, {0x8eb6,0x65, 0, 0}, {0x8eb7,0xe5, 0, 0},
	{0x8eb8,0x69, 0, 0}, {0x8eb9,0xc4, 0, 0}, {0x8eba,0xf8, 0, 0},
	{0x8ebb,0x54, 0, 0}, {0x8ebc,0x0f, 0, 0}, {0x8ebd,0xc8, 0, 0},
	{0x8ebe,0x68, 0, 0}, {0x8ebf,0xf5, 0, 0}, {0x8ec0,0x69, 0, 0},
	{0x8ec1,0xe5, 0, 0}, {0x8ec2,0x68, 0, 0}, {0x8ec3,0xc4, 0, 0},
	{0x8ec4,0x54, 0, 0}, {0x8ec5,0xf0, 0, 0}, {0x8ec6,0x48, 0, 0},
	{0x8ec7,0xf5, 0, 0}, {0x8ec8,0x68, 0, 0}, {0x8ec9,0x85, 0, 0},
	{0x8eca,0x68, 0, 0}, {0x8ecb,0x0d, 0, 0}, {0x8ecc,0x85, 0, 0},
	{0x8ecd,0x69, 0, 0}, {0x8ece,0x0e, 0, 0}, {0x8ecf,0x12, 0, 0},
	{0x8ed0,0x0a, 0, 0}, {0x8ed1,0x1d, 0, 0}, {0x8ed2,0x22, 0, 0},
	{0x8ed3,0x12, 0, 0}, {0x8ed4,0x02, 0, 0}, {0x8ed5,0x7f, 0, 0},
	{0x8ed6,0xb5, 0, 0}, {0x8ed7,0x07, 0, 0}, {0x8ed8,0x03, 0, 0},
	{0x8ed9,0xd3, 0, 0}, {0x8eda,0x80, 0, 0}, {0x8edb,0x01, 0, 0},
	{0x8edc,0xc3, 0, 0}, {0x8edd,0x40, 0, 0}, {0x8ede,0x03, 0, 0},
	{0x8edf,0x02, 0, 0}, {0x8ee0,0x0f, 0, 0}, {0x8ee1,0x72, 0, 0},
	{0x8ee2,0x90, 0, 0}, {0x8ee3,0x30, 0, 0}, {0x8ee4,0x04, 0, 0},
	{0x8ee5,0xe0, 0, 0}, {0x8ee6,0x44, 0, 0}, {0x8ee7,0x20, 0, 0},
	{0x8ee8,0xf0, 0, 0}, {0x8ee9,0xa3, 0, 0}, {0x8eea,0xe0, 0, 0},
	{0x8eeb,0x44, 0, 0}, {0x8eec,0x40, 0, 0}, {0x8eed,0xf0, 0, 0},
	{0x8eee,0x90, 0, 0}, {0x8eef,0x50, 0, 0}, {0x8ef0,0x25, 0, 0},
	{0x8ef1,0xe0, 0, 0}, {0x8ef2,0x44, 0, 0}, {0x8ef3,0x04, 0, 0},
	{0x8ef4,0xf0, 0, 0}, {0x8ef5,0x90, 0, 0}, {0x8ef6,0x50, 0, 0},
	{0x8ef7,0x03, 0, 0}, {0x8ef8,0xe0, 0, 0}, {0x8ef9,0x54, 0, 0},
	{0x8efa,0xfd, 0, 0}, {0x8efb,0xf0, 0, 0}, {0x8efc,0x90, 0, 0},
	{0x8efd,0x50, 0, 0}, {0x8efe,0x27, 0, 0}, {0x8eff,0xe0, 0, 0},
	{0x8f00,0x44, 0, 0}, {0x8f01,0x01, 0, 0}, {0x8f02,0xf0, 0, 0},
	{0x8f03,0x90, 0, 0}, {0x8f04,0x50, 0, 0}, {0x8f05,0x31, 0, 0},
	{0x8f06,0xe4, 0, 0}, {0x8f07,0xf0, 0, 0}, {0x8f08,0x90, 0, 0},
	{0x8f09,0x50, 0, 0}, {0x8f0a,0x33, 0, 0}, {0x8f0b,0xf0, 0, 0},
	{0x8f0c,0x90, 0, 0}, {0x8f0d,0x30, 0, 0}, {0x8f0e,0x1e, 0, 0},
	{0x8f0f,0x12, 0, 0}, {0x8f10,0x02, 0, 0}, {0x8f11,0x05, 0, 0},
	{0x8f12,0x90, 0, 0}, {0x8f13,0x30, 0, 0}, {0x8f14,0x18, 0, 0},
	{0x8f15,0x12, 0, 0}, {0x8f16,0x02, 0, 0}, {0x8f17,0x05, 0, 0},
	{0x8f18,0x90, 0, 0}, {0x8f19,0x30, 0, 0}, {0x8f1a,0x1b, 0, 0},
	{0x8f1b,0x12, 0, 0}, {0x8f1c,0x02, 0, 0}, {0x8f1d,0x05, 0, 0},
	{0x8f1e,0xe0, 0, 0}, {0x8f1f,0xf5, 0, 0}, {0x8f20,0x25, 0, 0},
	{0x8f21,0x90, 0, 0}, {0x8f22,0x30, 0, 0}, {0x8f23,0x18, 0, 0},
	{0x8f24,0xe0, 0, 0}, {0x8f25,0xf5, 0, 0}, {0x8f26,0x21, 0, 0},
	{0x8f27,0x90, 0, 0}, {0x8f28,0x60, 0, 0}, {0x8f29,0x00, 0, 0},
	{0x8f2a,0x74, 0, 0}, {0x8f2b,0xf5, 0, 0}, {0x8f2c,0xf0, 0, 0},
	{0x8f2d,0x90, 0, 0}, {0x8f2e,0x3f, 0, 0}, {0x8f2f,0x01, 0, 0},
	{0x8f30,0xe4, 0, 0}, {0x8f31,0xf0, 0, 0}, {0x8f32,0xa3, 0, 0},
	{0x8f33,0xf0, 0, 0}, {0x8f34,0x90, 0, 0}, {0x8f35,0x3f, 0, 0},
	{0x8f36,0x01, 0, 0}, {0x8f37,0xe0, 0, 0}, {0x8f38,0x44, 0, 0},
	{0x8f39,0x08, 0, 0}, {0x8f3a,0xf0, 0, 0}, {0x8f3b,0xe0, 0, 0},
	{0x8f3c,0x44, 0, 0}, {0x8f3d,0x20, 0, 0}, {0x8f3e,0xf0, 0, 0},
	{0x8f3f,0x90, 0, 0}, {0x8f40,0x3f, 0, 0}, {0x8f41,0x05, 0, 0},
	{0x8f42,0x74, 0, 0}, {0x8f43,0x30, 0, 0}, {0x8f44,0xf0, 0, 0},
	{0x8f45,0xa3, 0, 0}, {0x8f46,0x74, 0, 0}, {0x8f47,0x24, 0, 0},
	{0x8f48,0xf0, 0, 0}, {0x8f49,0x90, 0, 0}, {0x8f4a,0x3f, 0, 0},
	{0x8f4b,0x0b, 0, 0}, {0x8f4c,0xe0, 0, 0}, {0x8f4d,0x44, 0, 0},
	{0x8f4e,0x0f, 0, 0}, {0x8f4f,0xf0, 0, 0}, {0x8f50,0x90, 0, 0},
	{0x8f51,0x3f, 0, 0}, {0x8f52,0x01, 0, 0}, {0x8f53,0xe0, 0, 0},
	{0x8f54,0x44, 0, 0}, {0x8f55,0x02, 0, 0}, {0x8f56,0xf0, 0, 0},
	{0x8f57,0xc2, 0, 0}, {0x8f58,0x8c, 0, 0}, {0x8f59,0x75, 0, 0},
	{0x8f5a,0x89, 0, 0}, {0x8f5b,0x03, 0, 0}, {0x8f5c,0x75, 0, 0},
	{0x8f5d,0xa8, 0, 0}, {0x8f5e,0x07, 0, 0}, {0x8f5f,0x75, 0, 0},
	{0x8f60,0xb8, 0, 0}, {0x8f61,0x04, 0, 0}, {0x8f62,0xe4, 0, 0},
	{0x8f63,0xf5, 0, 0}, {0x8f64,0xd8, 0, 0}, {0x8f65,0xf5, 0, 0},
	{0x8f66,0xe8, 0, 0}, {0x8f67,0x90, 0, 0}, {0x8f68,0x30, 0, 0},
	{0x8f69,0x01, 0, 0}, {0x8f6a,0xe0, 0, 0}, {0x8f6b,0x44, 0, 0},
	{0x8f6c,0x40, 0, 0}, {0x8f6d,0xf0, 0, 0}, {0x8f6e,0xe0, 0, 0},
	{0x8f6f,0x54, 0, 0}, {0x8f70,0xbf, 0, 0}, {0x8f71,0xf0, 0, 0},
	{0x8f72,0x22, 0, 0}, {0x8f73,0x30, 0, 0}, {0x8f74,0x3e, 0, 0},
	{0x8f75,0x09, 0, 0}, {0x8f76,0x30, 0, 0}, {0x8f77,0x20, 0, 0},
	{0x8f78,0x06, 0, 0}, {0x8f79,0xae, 0, 0}, {0x8f7a,0x56, 0, 0},
	{0x8f7b,0xaf, 0, 0}, {0x8f7c,0x57, 0, 0}, {0x8f7d,0x80, 0, 0},
	{0x8f7e,0x04, 0, 0}, {0x8f7f,0xae, 0, 0}, {0x8f80,0x6a, 0, 0},
	{0x8f81,0xaf, 0, 0}, {0x8f82,0x6b, 0, 0}, {0x8f83,0x8e, 0, 0},
	{0x8f84,0x56, 0, 0}, {0x8f85,0x8f, 0, 0}, {0x8f86,0x57, 0, 0},
	{0x8f87,0x30, 0, 0}, {0x8f88,0x3e, 0, 0}, {0x8f89,0x09, 0, 0},
	{0x8f8a,0x30, 0, 0}, {0x8f8b,0x21, 0, 0}, {0x8f8c,0x06, 0, 0},
	{0x8f8d,0xae, 0, 0}, {0x8f8e,0x58, 0, 0}, {0x8f8f,0xaf, 0, 0},
	{0x8f90,0x59, 0, 0}, {0x8f91,0x80, 0, 0}, {0x8f92,0x04, 0, 0},
	{0x8f93,0xae, 0, 0}, {0x8f94,0x6c, 0, 0}, {0x8f95,0xaf, 0, 0},
	{0x8f96,0x6d, 0, 0}, {0x8f97,0x8e, 0, 0}, {0x8f98,0x58, 0, 0},
	{0x8f99,0x8f, 0, 0}, {0x8f9a,0x59, 0, 0}, {0x8f9b,0x30, 0, 0},
	{0x8f9c,0x3e, 0, 0}, {0x8f9d,0x09, 0, 0}, {0x8f9e,0x30, 0, 0},
	{0x8f9f,0x22, 0, 0}, {0x8fa0,0x06, 0, 0}, {0x8fa1,0xae, 0, 0},
	{0x8fa2,0x5a, 0, 0}, {0x8fa3,0xaf, 0, 0}, {0x8fa4,0x5b, 0, 0},
	{0x8fa5,0x80, 0, 0}, {0x8fa6,0x04, 0, 0}, {0x8fa7,0xae, 0, 0},
	{0x8fa8,0x6e, 0, 0}, {0x8fa9,0xaf, 0, 0}, {0x8faa,0x6f, 0, 0},
	{0x8fab,0x8e, 0, 0}, {0x8fac,0x5a, 0, 0}, {0x8fad,0x8f, 0, 0},
	{0x8fae,0x5b, 0, 0}, {0x8faf,0x30, 0, 0}, {0x8fb0,0x3e, 0, 0},
	{0x8fb1,0x09, 0, 0}, {0x8fb2,0x30, 0, 0}, {0x8fb3,0x23, 0, 0},
	{0x8fb4,0x06, 0, 0}, {0x8fb5,0xae, 0, 0}, {0x8fb6,0x5c, 0, 0},
	{0x8fb7,0xaf, 0, 0}, {0x8fb8,0x5d, 0, 0}, {0x8fb9,0x80, 0, 0},
	{0x8fba,0x04, 0, 0}, {0x8fbb,0xae, 0, 0}, {0x8fbc,0x70, 0, 0},
	{0x8fbd,0xaf, 0, 0}, {0x8fbe,0x71, 0, 0}, {0x8fbf,0x8e, 0, 0},
	{0x8fc0,0x5c, 0, 0}, {0x8fc1,0x8f, 0, 0}, {0x8fc2,0x5d, 0, 0},
	{0x8fc3,0x30, 0, 0}, {0x8fc4,0x3e, 0, 0}, {0x8fc5,0x09, 0, 0},
	{0x8fc6,0x30, 0, 0}, {0x8fc7,0x24, 0, 0}, {0x8fc8,0x06, 0, 0},
	{0x8fc9,0xae, 0, 0}, {0x8fca,0x5e, 0, 0}, {0x8fcb,0xaf, 0, 0},
	{0x8fcc,0x5f, 0, 0}, {0x8fcd,0x80, 0, 0}, {0x8fce,0x04, 0, 0},
	{0x8fcf,0xae, 0, 0}, {0x8fd0,0x72, 0, 0}, {0x8fd1,0xaf, 0, 0},
	{0x8fd2,0x73, 0, 0}, {0x8fd3,0x8e, 0, 0}, {0x8fd4,0x5e, 0, 0},
	{0x8fd5,0x8f, 0, 0}, {0x8fd6,0x5f, 0, 0}, {0x8fd7,0x30, 0, 0},
	{0x8fd8,0x3e, 0, 0}, {0x8fd9,0x09, 0, 0}, {0x8fda,0x30, 0, 0},
	{0x8fdb,0x25, 0, 0}, {0x8fdc,0x06, 0, 0}, {0x8fdd,0xae, 0, 0},
	{0x8fde,0x60, 0, 0}, {0x8fdf,0xaf, 0, 0}, {0x8fe0,0x61, 0, 0},
	{0x8fe1,0x80, 0, 0}, {0x8fe2,0x04, 0, 0}, {0x8fe3,0xae, 0, 0},
	{0x8fe4,0x74, 0, 0}, {0x8fe5,0xaf, 0, 0}, {0x8fe6,0x75, 0, 0},
	{0x8fe7,0x8e, 0, 0}, {0x8fe8,0x60, 0, 0}, {0x8fe9,0x8f, 0, 0},
	{0x8fea,0x61, 0, 0}, {0x8feb,0x30, 0, 0}, {0x8fec,0x3e, 0, 0},
	{0x8fed,0x09, 0, 0}, {0x8fee,0x30, 0, 0}, {0x8fef,0x26, 0, 0},
	{0x8ff0,0x06, 0, 0}, {0x8ff1,0xae, 0, 0}, {0x8ff2,0x62, 0, 0},
	{0x8ff3,0xaf, 0, 0}, {0x8ff4,0x63, 0, 0}, {0x8ff5,0x80, 0, 0},
	{0x8ff6,0x04, 0, 0}, {0x8ff7,0xae, 0, 0}, {0x8ff8,0x76, 0, 0},
	{0x8ff9,0xaf, 0, 0}, {0x8ffa,0x77, 0, 0}, {0x8ffb,0x8e, 0, 0},
	{0x8ffc,0x62, 0, 0}, {0x8ffd,0x8f, 0, 0}, {0x8ffe,0x63, 0, 0},
	{0x8fff,0x22, 0, 0}, {0x9000,0x30, 0, 0}, {0x9001,0x36, 0, 0},
	{0x9002,0x4e, 0, 0}, {0x9003,0x12, 0, 0}, {0x9004,0x10, 0, 0},
	{0x9005,0x78, 0, 0}, {0x9006,0xe5, 0, 0}, {0x9007,0x32, 0, 0},
	{0x9008,0xd3, 0, 0}, {0x9009,0x95, 0, 0}, {0x900a,0x7d, 0, 0},
	{0x900b,0x40, 0, 0}, {0x900c,0x0c, 0, 0}, {0x900d,0xe5, 0, 0},
	{0x900e,0x32, 0, 0}, {0x900f,0x04, 0, 0}, {0x9010,0xf5, 0, 0},
	{0x9011,0x0c, 0, 0}, {0x9012,0x12, 0, 0}, {0x9013,0x13, 0, 0},
	{0x9014,0x64, 0, 0}, {0x9015,0xd2, 0, 0}, {0x9016,0x37, 0, 0},
	{0x9017,0x80, 0, 0}, {0x9018,0x44, 0, 0}, {0x9019,0xe5, 0, 0},
	{0x901a,0x7c, 0, 0}, {0x901b,0xb5, 0, 0}, {0x901c,0x7d, 0, 0},
	{0x901d,0x04, 0, 0}, {0x901e,0x7f, 0, 0}, {0x901f,0x01, 0, 0},
	{0x9020,0x80, 0, 0}, {0x9021,0x02, 0, 0}, {0x9022,0x7f, 0, 0},
	{0x9023,0x00, 0, 0}, {0x9024,0xef, 0, 0}, {0x9025,0x24, 0, 0},
	{0x9026,0xfb, 0, 0}, {0x9027,0xd3, 0, 0}, {0x9028,0x64, 0, 0},
	{0x9029,0x80, 0, 0}, {0x902a,0xf8, 0, 0}, {0x902b,0xe5, 0, 0},
	{0x902c,0x7d, 0, 0}, {0x902d,0x64, 0, 0}, {0x902e,0x80, 0, 0},
	{0x902f,0x98, 0, 0}, {0x9030,0x40, 0, 0}, {0x9031,0x12, 0, 0},
	{0x9032,0xc2, 0, 0}, {0x9033,0x37, 0, 0}, {0x9034,0xe5, 0, 0},
	{0x9035,0x7c, 0, 0}, {0x9036,0xb5, 0, 0}, {0x9037,0x7d, 0, 0},
	{0x9038,0x04, 0, 0}, {0x9039,0x7f, 0, 0}, {0x903a,0x01, 0, 0},
	{0x903b,0x80, 0, 0}, {0x903c,0x02, 0, 0}, {0x903d,0x7f, 0, 0},
	{0x903e,0x00, 0, 0}, {0x903f,0xef, 0, 0}, {0x9040,0x24, 0, 0},
	{0x9041,0xfa, 0, 0}, {0x9042,0x80, 0, 0}, {0x9043,0x06, 0, 0},
	{0x9044,0xd2, 0, 0}, {0x9045,0x37, 0, 0}, {0x9046,0xe5, 0, 0},
	{0x9047,0x7d, 0, 0}, {0x9048,0x24, 0, 0}, {0x9049,0x02, 0, 0},
	{0x904a,0xf5, 0, 0}, {0x904b,0x0c, 0, 0}, {0x904c,0x12, 0, 0},
	{0x904d,0x13, 0, 0}, {0x904e,0x64, 0, 0}, {0x904f,0x80, 0, 0},
	{0x9050,0x0c, 0, 0}, {0x9051,0xe5, 0, 0}, {0x9052,0x7d, 0, 0},
	{0x9053,0x70, 0, 0}, {0x9054,0x0f, 0, 0}, {0x9055,0x12, 0, 0},
	{0x9056,0x10, 0, 0}, {0x9057,0x78, 0, 0}, {0x9058,0xc2, 0, 0},
	{0x9059,0x03, 0, 0}, {0x905a,0x12, 0, 0}, {0x905b,0x16, 0, 0},
	{0x905c,0xa4, 0, 0}, {0x905d,0xd2, 0, 0}, {0x905e,0x02, 0, 0},
	{0x905f,0xd2, 0, 0}, {0x9060,0x01, 0, 0}, {0x9061,0xd2, 0, 0},
	{0x9062,0x00, 0, 0}, {0x9063,0x22, 0, 0}, {0x9064,0x30, 0, 0},
	{0x9065,0x03, 0, 0}, {0x9066,0x08, 0, 0}, {0x9067,0xc2, 0, 0},
	{0x9068,0x03, 0, 0}, {0x9069,0xc2, 0, 0}, {0x906a,0x04, 0, 0},
	{0x906b,0x12, 0, 0}, {0x906c,0x02, 0, 0}, {0x906d,0x9b, 0, 0},
	{0x906e,0x22, 0, 0}, {0x906f,0xe4, 0, 0}, {0x9070,0xf5, 0, 0},
	{0x9071,0x0c, 0, 0}, {0x9072,0x12, 0, 0}, {0x9073,0x13, 0, 0},
	{0x9074,0x64, 0, 0}, {0x9075,0xd2, 0, 0}, {0x9076,0x03, 0, 0},
	{0x9077,0x22, 0, 0}, {0x9078,0x12, 0, 0}, {0x9079,0x16, 0, 0},
	{0x907a,0x90, 0, 0}, {0x907b,0xc2, 0, 0}, {0x907c,0x3e, 0, 0},
	{0x907d,0x12, 0, 0}, {0x907e,0x0f, 0, 0}, {0x907f,0x73, 0, 0},
	{0x9080,0xc2, 0, 0}, {0x9081,0x3e, 0, 0}, {0x9082,0x12, 0, 0},
	{0x9083,0x13, 0, 0}, {0x9084,0x01, 0, 0}, {0x9085,0x22, 0, 0},
	{0x9086,0xd3, 0, 0}, {0x9087,0xe5, 0, 0}, {0x9088,0x57, 0, 0},
	{0x9089,0x95, 0, 0}, {0x908a,0x6b, 0, 0}, {0x908b,0xe5, 0, 0},
	{0x908c,0x56, 0, 0}, {0x908d,0x95, 0, 0}, {0x908e,0x6a, 0, 0},
	{0x908f,0x40, 0, 0}, {0x9090,0x03, 0, 0}, {0x9091,0xd3, 0, 0},
	{0x9092,0x80, 0, 0}, {0x9093,0x01, 0, 0}, {0x9094,0xc3, 0, 0},
	{0x9095,0x92, 0, 0}, {0x9096,0x20, 0, 0}, {0x9097,0xd3, 0, 0},
	{0x9098,0xe5, 0, 0}, {0x9099,0x59, 0, 0}, {0x909a,0x95, 0, 0},
	{0x909b,0x6d, 0, 0}, {0x909c,0xe5, 0, 0}, {0x909d,0x58, 0, 0},
	{0x909e,0x95, 0, 0}, {0x909f,0x6c, 0, 0}, {0x90a0,0x40, 0, 0},
	{0x90a1,0x03, 0, 0}, {0x90a2,0xd3, 0, 0}, {0x90a3,0x80, 0, 0},
	{0x90a4,0x01, 0, 0}, {0x90a5,0xc3, 0, 0}, {0x90a6,0x92, 0, 0},
	{0x90a7,0x21, 0, 0}, {0x90a8,0xd3, 0, 0}, {0x90a9,0xe5, 0, 0},
	{0x90aa,0x5b, 0, 0}, {0x90ab,0x95, 0, 0}, {0x90ac,0x6f, 0, 0},
	{0x90ad,0xe5, 0, 0}, {0x90ae,0x5a, 0, 0}, {0x90af,0x95, 0, 0},
	{0x90b0,0x6e, 0, 0}, {0x90b1,0x40, 0, 0}, {0x90b2,0x03, 0, 0},
	{0x90b3,0xd3, 0, 0}, {0x90b4,0x80, 0, 0}, {0x90b5,0x01, 0, 0},
	{0x90b6,0xc3, 0, 0}, {0x90b7,0x92, 0, 0}, {0x90b8,0x22, 0, 0},
	{0x90b9,0xd3, 0, 0}, {0x90ba,0xe5, 0, 0}, {0x90bb,0x5d, 0, 0},
	{0x90bc,0x95, 0, 0}, {0x90bd,0x71, 0, 0}, {0x90be,0xe5, 0, 0},
	{0x90bf,0x5c, 0, 0}, {0x90c0,0x95, 0, 0}, {0x90c1,0x70, 0, 0},
	{0x90c2,0x40, 0, 0}, {0x90c3,0x03, 0, 0}, {0x90c4,0xd3, 0, 0},
	{0x90c5,0x80, 0, 0}, {0x90c6,0x01, 0, 0}, {0x90c7,0xc3, 0, 0},
	{0x90c8,0x92, 0, 0}, {0x90c9,0x23, 0, 0}, {0x90ca,0xd3, 0, 0},
	{0x90cb,0xe5, 0, 0}, {0x90cc,0x5f, 0, 0}, {0x90cd,0x95, 0, 0},
	{0x90ce,0x73, 0, 0}, {0x90cf,0xe5, 0, 0}, {0x90d0,0x5e, 0, 0},
	{0x90d1,0x95, 0, 0}, {0x90d2,0x72, 0, 0}, {0x90d3,0x40, 0, 0},
	{0x90d4,0x03, 0, 0}, {0x90d5,0xd3, 0, 0}, {0x90d6,0x80, 0, 0},
	{0x90d7,0x01, 0, 0}, {0x90d8,0xc3, 0, 0}, {0x90d9,0x92, 0, 0},
	{0x90da,0x24, 0, 0}, {0x90db,0xd3, 0, 0}, {0x90dc,0xe5, 0, 0},
	{0x90dd,0x61, 0, 0}, {0x90de,0x95, 0, 0}, {0x90df,0x75, 0, 0},
	{0x90e0,0xe5, 0, 0}, {0x90e1,0x60, 0, 0}, {0x90e2,0x95, 0, 0},
	{0x90e3,0x74, 0, 0}, {0x90e4,0x40, 0, 0}, {0x90e5,0x03, 0, 0},
	{0x90e6,0xd3, 0, 0}, {0x90e7,0x80, 0, 0}, {0x90e8,0x01, 0, 0},
	{0x90e9,0xc3, 0, 0}, {0x90ea,0x92, 0, 0}, {0x90eb,0x25, 0, 0},
	{0x90ec,0xd3, 0, 0}, {0x90ed,0xe5, 0, 0}, {0x90ee,0x63, 0, 0},
	{0x90ef,0x95, 0, 0}, {0x90f0,0x77, 0, 0}, {0x90f1,0xe5, 0, 0},
	{0x90f2,0x62, 0, 0}, {0x90f3,0x95, 0, 0}, {0x90f4,0x76, 0, 0},
	{0x90f5,0x40, 0, 0}, {0x90f6,0x03, 0, 0}, {0x90f7,0xd3, 0, 0},
	{0x90f8,0x80, 0, 0}, {0x90f9,0x01, 0, 0}, {0x90fa,0xc3, 0, 0},
	{0x90fb,0x92, 0, 0}, {0x90fc,0x26, 0, 0}, {0x90fd,0x22, 0, 0},
	{0x90fe,0xe5, 0, 0}, {0x90ff,0x7e, 0, 0}, {0x9100,0x64, 0, 0},
	{0x9101,0x01, 0, 0}, {0x9102,0x70, 0, 0}, {0x9103,0x50, 0, 0},
	{0x9104,0x12, 0, 0}, {0x9105,0x02, 0, 0}, {0x9106,0x39, 0, 0},
	{0x9107,0xe5, 0, 0}, {0x9108,0x7d, 0, 0}, {0x9109,0x12, 0, 0},
	{0x910a,0x01, 0, 0}, {0x910b,0x7b, 0, 0}, {0x910c,0xfe, 0, 0},
	{0x910d,0xe4, 0, 0}, {0x910e,0x8f, 0, 0}, {0x910f,0x69, 0, 0},
	{0x9110,0x8e, 0, 0}, {0x9111,0x68, 0, 0}, {0x9112,0xf5, 0, 0},
	{0x9113,0x67, 0, 0}, {0x9114,0xf5, 0, 0}, {0x9115,0x66, 0, 0},
	{0x9116,0x12, 0, 0}, {0x9117,0x01, 0, 0}, {0x9118,0xfc, 0, 0},
	{0x9119,0x7b, 0, 0}, {0x911a,0xff, 0, 0}, {0x911b,0xfa, 0, 0},
	{0x911c,0xf9, 0, 0}, {0x911d,0xf8, 0, 0}, {0x911e,0x12, 0, 0},
	{0x911f,0x01, 0, 0}, {0x9120,0xf1, 0, 0}, {0x9121,0xc0, 0, 0},
	{0x9122,0x04, 0, 0}, {0x9123,0xc0, 0, 0}, {0x9124,0x05, 0, 0},
	{0x9125,0xc0, 0, 0}, {0x9126,0x06, 0, 0}, {0x9127,0xc0, 0, 0},
	{0x9128,0x07, 0, 0}, {0x9129,0x12, 0, 0}, {0x912a,0x01, 0, 0},
	{0x912b,0x73, 0, 0}, {0x912c,0xab, 0, 0}, {0x912d,0x07, 0, 0},
	{0x912e,0xfa, 0, 0}, {0x912f,0xe4, 0, 0}, {0x9130,0xf9, 0, 0},
	{0x9131,0xf8, 0, 0}, {0x9132,0xd0, 0, 0}, {0x9133,0x07, 0, 0},
	{0x9134,0xd0, 0, 0}, {0x9135,0x06, 0, 0}, {0x9136,0xd0, 0, 0},
	{0x9137,0x05, 0, 0}, {0x9138,0xd0, 0, 0}, {0x9139,0x04, 0, 0},
	{0x913a,0x12, 0, 0}, {0x913b,0x02, 0, 0}, {0x913c,0x67, 0, 0},
	{0x913d,0x85, 0, 0}, {0x913e,0x69, 0, 0}, {0x913f,0x66, 0, 0},
	{0x9140,0x85, 0, 0}, {0x9141,0x7d, 0, 0}, {0x9142,0x67, 0, 0},
	{0x9143,0x12, 0, 0}, {0x9144,0x02, 0, 0}, {0x9145,0x39, 0, 0},
	{0x9146,0xe5, 0, 0}, {0x9147,0x7d, 0, 0}, {0x9148,0x12, 0, 0},
	{0x9149,0x02, 0, 0}, {0x914a,0x95, 0, 0}, {0x914b,0xe4, 0, 0},
	{0x914c,0x93, 0, 0}, {0x914d,0xf5, 0, 0}, {0x914e,0x68, 0, 0},
	{0x914f,0x74, 0, 0}, {0x9150,0x01, 0, 0}, {0x9151,0x93, 0, 0},
	{0x9152,0xf5, 0, 0}, {0x9153,0x69, 0, 0}, {0x9154,0x90, 0, 0},
	{0x9155,0x50, 0, 0}, {0x9156,0x82, 0, 0}, {0x9157,0xe5, 0, 0},
	{0x9158,0x66, 0, 0}, {0x9159,0xf0, 0, 0}, {0x915a,0xa3, 0, 0},
	{0x915b,0xe5, 0, 0}, {0x915c,0x67, 0, 0}, {0x915d,0xf0, 0, 0},
	{0x915e,0xa3, 0, 0}, {0x915f,0xe5, 0, 0}, {0x9160,0x68, 0, 0},
	{0x9161,0xf0, 0, 0}, {0x9162,0xa3, 0, 0}, {0x9163,0xe5, 0, 0},
	{0x9164,0x69, 0, 0}, {0x9165,0xf0, 0, 0}, {0x9166,0x22, 0, 0},
	{0x9167,0x56, 0, 0}, {0x9168,0x0c, 0, 0}, {0x9169,0x04, 0, 0},
	{0x916a,0x00, 0, 0}, {0x916b,0x00, 0, 0}, {0x916c,0x00, 0, 0},
	{0x916d,0xc8, 0, 0}, {0x916e,0x01, 0, 0}, {0x916f,0x2c, 0, 0},
	{0x9170,0x01, 0, 0}, {0x9171,0x5e, 0, 0}, {0x9172,0x01, 0, 0},
	{0x9173,0x8b, 0, 0}, {0x9174,0x01, 0, 0}, {0x9175,0xb8, 0, 0},
	{0x9176,0x01, 0, 0}, {0x9177,0xe5, 0, 0}, {0x9178,0x02, 0, 0},
	{0x9179,0x12, 0, 0}, {0x917a,0x02, 0, 0}, {0x917b,0x3f, 0, 0},
	{0x917c,0x02, 0, 0}, {0x917d,0x6c, 0, 0}, {0x917e,0x02, 0, 0},
	{0x917f,0x99, 0, 0}, {0x9180,0x02, 0, 0}, {0x9181,0xc6, 0, 0},
	{0x9182,0x02, 0, 0}, {0x9183,0xf3, 0, 0}, {0x9184,0x07, 0, 0},
	{0x9185,0x20, 0, 0}, {0x9186,0x12, 0, 0}, {0x9187,0x28, 0, 0},
	{0x9188,0x1e, 0, 0}, {0x9189,0x18, 0, 0}, {0x918a,0x18, 0, 0},
	{0x918b,0x28, 0, 0}, {0x918c,0x1e, 0, 0}, {0x918d,0x18, 0, 0},
	{0x918e,0x12, 0, 0}, {0x918f,0x28, 0, 0}, {0x9190,0x1e, 0, 0},
	{0x9191,0x18, 0, 0}, {0x9192,0x12, 0, 0}, {0x9193,0x28, 0, 0},
	{0x9194,0x18, 0, 0}, {0x9195,0x18, 0, 0}, {0x9196,0x12, 0, 0},
	{0x9197,0x20, 0, 0}, {0x9198,0x18, 0, 0}, {0x9199,0x28, 0, 0},
	{0x919a,0x1c, 0, 0}, {0x919b,0x30, 0, 0}, {0x919c,0x24, 0, 0},
	{0x919d,0x10, 0, 0}, {0x919e,0x1c, 0, 0}, {0x919f,0x18, 0, 0},
	{0x91a0,0x24, 0, 0}, {0x91a1,0x1c, 0, 0}, {0x91a2,0x14, 0, 0},
	{0x91a3,0x24, 0, 0}, {0x91a4,0x1c, 0, 0}, {0x91a5,0x28, 0, 0},
	{0x91a6,0x0c, 0, 0}, {0x91a7,0x30, 0, 0}, {0x91a8,0x14, 0, 0},
	{0x91a9,0x10, 0, 0}, {0x91aa,0x0c, 0, 0}, {0x91ab,0x18, 0, 0},
	{0x91ac,0x14, 0, 0}, {0x91ad,0x1c, 0, 0}, {0x91ae,0x20, 0, 0},
	{0x91af,0x24, 0, 0}, {0x91b0,0x28, 0, 0}, {0x91b1,0x0c, 0, 0},
	{0x91b2,0x14, 0, 0}, {0x91b3,0x14, 0, 0}, {0x91b4,0x1c, 0, 0},
	{0x91b5,0x1c, 0, 0}, {0x91b6,0x14, 0, 0}, {0x91b7,0x24, 0, 0},
	{0x91b8,0x1c, 0, 0}, {0x91b9,0x2c, 0, 0}, {0x91ba,0x14, 0, 0},
	{0x91bb,0x34, 0, 0}, {0x91bc,0x1c, 0, 0}, {0x91bd,0x1c, 0, 0},
	{0x91be,0x08, 0, 0}, {0x91bf,0x24, 0, 0}, {0x91c0,0x10, 0, 0},
	{0x91c1,0x19, 0, 0}, {0x91c2,0x19, 0, 0}, {0x91c3,0x1c, 0, 0},
	{0x91c4,0x19, 0, 0}, {0x91c5,0x19, 0, 0}, {0x91c6,0x10, 0, 0},
	{0x91c7,0x10, 0, 0}, {0x91c8,0x10, 0, 0}, {0x91c9,0x10, 0, 0},
	{0x91ca,0x10, 0, 0}, {0x91cb,0x00, 0, 0}, {0x91cc,0x00, 0, 0},
	{0x91cd,0x00, 0, 0}, {0x91ce,0x00, 0, 0}, {0x91cf,0x00, 0, 0},
	{0x91d0,0x12, 0, 0}, {0x91d1,0x10, 0, 0}, {0x91d2,0x86, 0, 0},
	{0x91d3,0x30, 0, 0}, {0x91d4,0x36, 0, 0}, {0x91d5,0x40, 0, 0},
	{0x91d6,0xe5, 0, 0}, {0x91d7,0x24, 0, 0}, {0x91d8,0x54, 0, 0},
	{0x91d9,0x1f, 0, 0}, {0x91da,0xff, 0, 0}, {0x91db,0x60, 0, 0},
	{0x91dc,0x0e, 0, 0}, {0x91dd,0x64, 0, 0}, {0x91de,0x1f, 0, 0},
	{0x91df,0x60, 0, 0}, {0x91e0,0x0a, 0, 0}, {0x91e1,0xe5, 0, 0},
	{0x91e2,0x7c, 0, 0}, {0x91e3,0x65, 0, 0}, {0x91e4,0x7d, 0, 0},
	{0x91e5,0x60, 0, 0}, {0x91e6,0x04, 0, 0}, {0x91e7,0xe5, 0, 0},
	{0x91e8,0x7d, 0, 0}, {0x91e9,0x70, 0, 0}, {0x91ea,0x18, 0, 0},
	{0x91eb,0xbf, 0, 0}, {0x91ec,0x1f, 0, 0}, {0x91ed,0x02, 0, 0},
	{0x91ee,0xb2, 0, 0}, {0x91ef,0x37, 0, 0}, {0x91f0,0xe5, 0, 0},
	{0x91f1,0x7c, 0, 0}, {0x91f2,0xb5, 0, 0}, {0x91f3,0x7d, 0, 0},
	{0x91f4,0x02, 0, 0}, {0x91f5,0xc2, 0, 0}, {0x91f6,0x37, 0, 0},
	{0x91f7,0xe5, 0, 0}, {0x91f8,0x7d, 0, 0}, {0x91f9,0x70, 0, 0},
	{0x91fa,0x02, 0, 0}, {0x91fb,0xd2, 0, 0}, {0x91fc,0x37, 0, 0},
	{0x91fd,0xc2, 0, 0}, {0x91fe,0x02, 0, 0}, {0x91ff,0xd2, 0, 0},
	{0x9200,0x01, 0, 0}, {0x9201,0xd2, 0, 0}, {0x9202,0x00, 0, 0},
	{0x9203,0xc2, 0, 0}, {0x9204,0x3e, 0, 0}, {0x9205,0x12, 0, 0},
	{0x9206,0x0f, 0, 0}, {0x9207,0x73, 0, 0}, {0x9208,0xc2, 0, 0},
	{0x9209,0x3e, 0, 0}, {0x920a,0x12, 0, 0}, {0x920b,0x13, 0, 0},
	{0x920c,0x01, 0, 0}, {0x920d,0x30, 0, 0}, {0x920e,0x37, 0, 0},
	{0x920f,0x03, 0, 0}, {0x9210,0x02, 0, 0}, {0x9211,0x16, 0, 0},
	{0x9212,0xa4, 0, 0}, {0x9213,0x02, 0, 0}, {0x9214,0x16, 0, 0},
	{0x9215,0xb5, 0, 0}, {0x9216,0xd2, 0, 0}, {0x9217,0x3e, 0, 0},
	{0x9218,0x12, 0, 0}, {0x9219,0x0f, 0, 0}, {0x921a,0x73, 0, 0},
	{0x921b,0xd2, 0, 0}, {0x921c,0x3e, 0, 0}, {0x921d,0x12, 0, 0},
	{0x921e,0x13, 0, 0}, {0x921f,0x01, 0, 0}, {0x9220,0x12, 0, 0},
	{0x9221,0x16, 0, 0}, {0x9222,0xa4, 0, 0}, {0x9223,0xe5, 0, 0},
	{0x9224,0x32, 0, 0}, {0x9225,0xd3, 0, 0}, {0x9226,0x95, 0, 0},
	{0x9227,0x7d, 0, 0}, {0x9228,0x40, 0, 0}, {0x9229,0x05, 0, 0},
	{0x922a,0xe4, 0, 0}, {0x922b,0x95, 0, 0}, {0x922c,0x7d, 0, 0},
	{0x922d,0x40, 0, 0}, {0x922e,0x06, 0, 0}, {0x922f,0xc2, 0, 0},
	{0x9230,0x02, 0, 0}, {0x9231,0xd2, 0, 0}, {0x9232,0x01, 0, 0},
	{0x9233,0xd2, 0, 0}, {0x9234,0x00, 0, 0}, {0x9235,0x22, 0, 0},
	{0x9236,0x12, 0, 0}, {0x9237,0x10, 0, 0}, {0x9238,0x86, 0, 0},
	{0x9239,0xc3, 0, 0}, {0x923a,0x30, 0, 0}, {0x923b,0x25, 0, 0},
	{0x923c,0x0c, 0, 0}, {0x923d,0xe5, 0, 0}, {0x923e,0x61, 0, 0},
	{0x923f,0x95, 0, 0}, {0x9240,0x75, 0, 0}, {0x9241,0xff, 0, 0},
	{0x9242,0xe5, 0, 0}, {0x9243,0x60, 0, 0}, {0x9244,0x95, 0, 0},
	{0x9245,0x74, 0, 0}, {0x9246,0xfe, 0, 0}, {0x9247,0x80, 0, 0},
	{0x9248,0x0a, 0, 0}, {0x9249,0xe5, 0, 0}, {0x924a,0x75, 0, 0},
	{0x924b,0x95, 0, 0}, {0x924c,0x61, 0, 0}, {0x924d,0xff, 0, 0},
	{0x924e,0xe5, 0, 0}, {0x924f,0x74, 0, 0}, {0x9250,0x95, 0, 0},
	{0x9251,0x60, 0, 0}, {0x9252,0xfe, 0, 0}, {0x9253,0x8e, 0, 0},
	{0x9254,0x0a, 0, 0}, {0x9255,0x8f, 0, 0}, {0x9256,0x0b, 0, 0},
	{0x9257,0x30, 0, 0}, {0x9258,0x03, 0, 0}, {0x9259,0x26, 0, 0},
	{0x925a,0x12, 0, 0}, {0x925b,0x02, 0, 0}, {0x925c,0xa9, 0, 0},
	{0x925d,0x50, 0, 0}, {0x925e,0x03, 0, 0}, {0x925f,0x30, 0, 0},
	{0x9260,0x27, 0, 0}, {0x9261,0x07, 0, 0}, {0x9262,0xc2, 0, 0},
	{0x9263,0x3e, 0, 0}, {0x9264,0x12, 0, 0}, {0x9265,0x0f, 0, 0},
	{0x9266,0x73, 0, 0}, {0x9267,0x80, 0, 0}, {0x9268,0x2f, 0, 0},
	{0x9269,0x05, 0, 0}, {0x926a,0x31, 0, 0}, {0x926b,0xe5, 0, 0},
	{0x926c,0x31, 0, 0}, {0x926d,0xd3, 0, 0}, {0x926e,0x94, 0, 0},
	{0x926f,0x02, 0, 0}, {0x9270,0x40, 0, 0}, {0x9271,0x29, 0, 0},
	{0x9272,0xe4, 0, 0}, {0x9273,0xf5, 0, 0}, {0x9274,0x31, 0, 0},
	{0x9275,0xc2, 0, 0}, {0x9276,0x03, 0, 0}, {0x9277,0xc2, 0, 0},
	{0x9278,0x02, 0, 0}, {0x9279,0xc2, 0, 0}, {0x927a,0x01, 0, 0},
	{0x927b,0xd2, 0, 0}, {0x927c,0x00, 0, 0}, {0x927d,0xd2, 0, 0},
	{0x927e,0x34, 0, 0}, {0x927f,0x22, 0, 0}, {0x9280,0x12, 0, 0},
	{0x9281,0x02, 0, 0}, {0x9282,0xa9, 0, 0}, {0x9283,0x50, 0, 0},
	{0x9284,0x03, 0, 0}, {0x9285,0x30, 0, 0}, {0x9286,0x27, 0, 0},
	{0x9287,0x04, 0, 0}, {0x9288,0x05, 0, 0}, {0x9289,0x31, 0, 0},
	{0x928a,0x80, 0, 0}, {0x928b,0x03, 0, 0}, {0x928c,0xe4, 0, 0},
	{0x928d,0xf5, 0, 0}, {0x928e,0x31, 0, 0}, {0x928f,0xe5, 0, 0},
	{0x9290,0x31, 0, 0}, {0x9291,0xd3, 0, 0}, {0x9292,0x94, 0, 0},
	{0x9293,0x02, 0, 0}, {0x9294,0x40, 0, 0}, {0x9295,0x05, 0, 0},
	{0x9296,0xd2, 0, 0}, {0x9297,0x03, 0, 0}, {0x9298,0xe4, 0, 0},
	{0x9299,0xf5, 0, 0}, {0x929a,0x31, 0, 0}, {0x929b,0x22, 0, 0},
	{0x929c,0xe5, 0, 0}, {0x929d,0x0a, 0, 0}, {0x929e,0x70, 0, 0},
	{0x929f,0x04, 0, 0}, {0x92a0,0x7a, 0, 0}, {0x92a1,0x11, 0, 0},
	{0x92a2,0x7b, 0, 0}, {0x92a3,0x85, 0, 0}, {0x92a4,0xe5, 0, 0},
	{0x92a5,0x0a, 0, 0}, {0x92a6,0xb4, 0, 0}, {0x92a7,0x01, 0, 0},
	{0x92a8,0x04, 0, 0}, {0x92a9,0x7a, 0, 0}, {0x92aa,0x11, 0, 0},
	{0x92ab,0x7b, 0, 0}, {0x92ac,0x99, 0, 0}, {0x92ad,0xe5, 0, 0},
	{0x92ae,0x0a, 0, 0}, {0x92af,0xb4, 0, 0}, {0x92b0,0x02, 0, 0},
	{0x92b1,0x04, 0, 0}, {0x92b2,0x7a, 0, 0}, {0x92b3,0x11, 0, 0},
	{0x92b4,0x7b, 0, 0}, {0x92b5,0xad, 0, 0}, {0x92b6,0x8b, 0, 0},
	{0x92b7,0x82, 0, 0}, {0x92b8,0x8a, 0, 0}, {0x92b9,0x83, 0, 0},
	{0x92ba,0x12, 0, 0}, {0x92bb,0x09, 0, 0}, {0x92bc,0xdb, 0, 0},
	{0x92bd,0x8f, 0, 0}, {0x92be,0x37, 0, 0}, {0x92bf,0x8e, 0, 0},
	{0x92c0,0x36, 0, 0}, {0x92c1,0x8d, 0, 0}, {0x92c2,0x35, 0, 0},
	{0x92c3,0x8c, 0, 0}, {0x92c4,0x34, 0, 0}, {0x92c5,0xe5, 0, 0},
	{0x92c6,0x82, 0, 0}, {0x92c7,0x24, 0, 0}, {0x92c8,0x04, 0, 0},
	{0x92c9,0xf5, 0, 0}, {0x92ca,0x82, 0, 0}, {0x92cb,0xe4, 0, 0},
	{0x92cc,0x35, 0, 0}, {0x92cd,0x83, 0, 0}, {0x92ce,0xf5, 0, 0},
	{0x92cf,0x83, 0, 0}, {0x92d0,0x12, 0, 0}, {0x92d1,0x09, 0, 0},
	{0x92d2,0xdb, 0, 0}, {0x92d3,0x8f, 0, 0}, {0x92d4,0x3b, 0, 0},
	{0x92d5,0x8e, 0, 0}, {0x92d6,0x3a, 0, 0}, {0x92d7,0x8d, 0, 0},
	{0x92d8,0x39, 0, 0}, {0x92d9,0x8c, 0, 0}, {0x92da,0x38, 0, 0},
	{0x92db,0xeb, 0, 0}, {0x92dc,0x24, 0, 0}, {0x92dd,0x08, 0, 0},
	{0x92de,0x12, 0, 0}, {0x92df,0x02, 0, 0}, {0x92e0,0x40, 0, 0},
	{0x92e1,0x12, 0, 0}, {0x92e2,0x01, 0, 0}, {0x92e3,0xc0, 0, 0},
	{0x92e4,0xeb, 0, 0}, {0x92e5,0x24, 0, 0}, {0x92e6,0x0c, 0, 0},
	{0x92e7,0x12, 0, 0}, {0x92e8,0x02, 0, 0}, {0x92e9,0x40, 0, 0},
	{0x92ea,0x8f, 0, 0}, {0x92eb,0x43, 0, 0}, {0x92ec,0x8e, 0, 0},
	{0x92ed,0x42, 0, 0}, {0x92ee,0x8d, 0, 0}, {0x92ef,0x41, 0, 0},
	{0x92f0,0x8c, 0, 0}, {0x92f1,0x40, 0, 0}, {0x92f2,0xeb, 0, 0},
	{0x92f3,0x24, 0, 0}, {0x92f4,0x10, 0, 0}, {0x92f5,0x12, 0, 0},
	{0x92f6,0x02, 0, 0}, {0x92f7,0x40, 0, 0}, {0x92f8,0x8f, 0, 0},
	{0x92f9,0x47, 0, 0}, {0x92fa,0x8e, 0, 0}, {0x92fb,0x46, 0, 0},
	{0x92fc,0x8d, 0, 0}, {0x92fd,0x45, 0, 0}, {0x92fe,0x8c, 0, 0},
	{0x92ff,0x44, 0, 0}, {0x9300,0x22, 0, 0}, {0x9301,0x30, 0, 0},
	{0x9302,0x3e, 0, 0}, {0x9303,0x07, 0, 0}, {0x9304,0x30, 0, 0},
	{0x9305,0x20, 0, 0}, {0x9306,0x04, 0, 0}, {0x9307,0xaf, 0, 0},
	{0x9308,0x4a, 0, 0}, {0x9309,0x80, 0, 0}, {0x930a,0x02, 0, 0},
	{0x930b,0xaf, 0, 0}, {0x930c,0x7d, 0, 0}, {0x930d,0x8f, 0, 0},
	{0x930e,0x4a, 0, 0}, {0x930f,0x30, 0, 0}, {0x9310,0x3e, 0, 0},
	{0x9311,0x07, 0, 0}, {0x9312,0x30, 0, 0}, {0x9313,0x21, 0, 0},
	{0x9314,0x04, 0, 0}, {0x9315,0xaf, 0, 0}, {0x9316,0x4b, 0, 0},
	{0x9317,0x80, 0, 0}, {0x9318,0x02, 0, 0}, {0x9319,0xaf, 0, 0},
	{0x931a,0x7d, 0, 0}, {0x931b,0x8f, 0, 0}, {0x931c,0x4b, 0, 0},
	{0x931d,0x30, 0, 0}, {0x931e,0x3e, 0, 0}, {0x931f,0x07, 0, 0},
	{0x9320,0x30, 0, 0}, {0x9321,0x22, 0, 0}, {0x9322,0x04, 0, 0},
	{0x9323,0xaf, 0, 0}, {0x9324,0x4c, 0, 0}, {0x9325,0x80, 0, 0},
	{0x9326,0x02, 0, 0}, {0x9327,0xaf, 0, 0}, {0x9328,0x7d, 0, 0},
	{0x9329,0x8f, 0, 0}, {0x932a,0x4c, 0, 0}, {0x932b,0x30, 0, 0},
	{0x932c,0x3e, 0, 0}, {0x932d,0x07, 0, 0}, {0x932e,0x30, 0, 0},
	{0x932f,0x23, 0, 0}, {0x9330,0x04, 0, 0}, {0x9331,0xaf, 0, 0},
	{0x9332,0x4d, 0, 0}, {0x9333,0x80, 0, 0}, {0x9334,0x02, 0, 0},
	{0x9335,0xaf, 0, 0}, {0x9336,0x7d, 0, 0}, {0x9337,0x8f, 0, 0},
	{0x9338,0x4d, 0, 0}, {0x9339,0x30, 0, 0}, {0x933a,0x3e, 0, 0},
	{0x933b,0x07, 0, 0}, {0x933c,0x30, 0, 0}, {0x933d,0x24, 0, 0},
	{0x933e,0x04, 0, 0}, {0x933f,0xaf, 0, 0}, {0x9340,0x4e, 0, 0},
	{0x9341,0x80, 0, 0}, {0x9342,0x02, 0, 0}, {0x9343,0xaf, 0, 0},
	{0x9344,0x7d, 0, 0}, {0x9345,0x8f, 0, 0}, {0x9346,0x4e, 0, 0},
	{0x9347,0x30, 0, 0}, {0x9348,0x3e, 0, 0}, {0x9349,0x07, 0, 0},
	{0x934a,0x30, 0, 0}, {0x934b,0x25, 0, 0}, {0x934c,0x04, 0, 0},
	{0x934d,0xaf, 0, 0}, {0x934e,0x4f, 0, 0}, {0x934f,0x80, 0, 0},
	{0x9350,0x02, 0, 0}, {0x9351,0xaf, 0, 0}, {0x9352,0x7d, 0, 0},
	{0x9353,0x8f, 0, 0}, {0x9354,0x4f, 0, 0}, {0x9355,0x30, 0, 0},
	{0x9356,0x3e, 0, 0}, {0x9357,0x07, 0, 0}, {0x9358,0x30, 0, 0},
	{0x9359,0x26, 0, 0}, {0x935a,0x04, 0, 0}, {0x935b,0xaf, 0, 0},
	{0x935c,0x50, 0, 0}, {0x935d,0x80, 0, 0}, {0x935e,0x02, 0, 0},
	{0x935f,0xaf, 0, 0}, {0x9360,0x7d, 0, 0}, {0x9361,0x8f, 0, 0},
	{0x9362,0x50, 0, 0}, {0x9363,0x22, 0, 0}, {0x9364,0xe5, 0, 0},
	{0x9365,0x0c, 0, 0}, {0x9366,0xd3, 0, 0}, {0x9367,0x95, 0, 0},
	{0x9368,0x7c, 0, 0}, {0x9369,0x40, 0, 0}, {0x936a,0x01, 0, 0},
	{0x936b,0x22, 0, 0}, {0x936c,0x12, 0, 0}, {0x936d,0x02, 0, 0},
	{0x936e,0x39, 0, 0}, {0x936f,0xe5, 0, 0}, {0x9370,0x0c, 0, 0},
	{0x9371,0x12, 0, 0}, {0x9372,0x02, 0, 0}, {0x9373,0x95, 0, 0},
	{0x9374,0xe4, 0, 0}, {0x9375,0x93, 0, 0}, {0x9376,0xfe, 0, 0},
	{0x9377,0x74, 0, 0}, {0x9378,0x01, 0, 0}, {0x9379,0x93, 0, 0},
	{0x937a,0xff, 0, 0}, {0x937b,0x4e, 0, 0}, {0x937c,0x60, 0, 0},
	{0x937d,0x21, 0, 0}, {0x937e,0x8e, 0, 0}, {0x937f,0x64, 0, 0},
	{0x9380,0x8f, 0, 0}, {0x9381,0x65, 0, 0}, {0x9382,0xef, 0, 0},
	{0x9383,0xc4, 0, 0}, {0x9384,0xf8, 0, 0}, {0x9385,0x54, 0, 0},
	{0x9386,0x0f, 0, 0}, {0x9387,0xc8, 0, 0}, {0x9388,0x68, 0, 0},
	{0x9389,0xff, 0, 0}, {0x938a,0xee, 0, 0}, {0x938b,0xc4, 0, 0},
	{0x938c,0x54, 0, 0}, {0x938d,0xf0, 0, 0}, {0x938e,0x48, 0, 0},
	{0x938f,0xfe, 0, 0}, {0x9390,0x43, 0, 0}, {0x9391,0x07, 0, 0},
	{0x9392,0x0d, 0, 0}, {0x9393,0x8e, 0, 0}, {0x9394,0x0d, 0, 0},
	{0x9395,0x8f, 0, 0}, {0x9396,0x0e, 0, 0}, {0x9397,0x12, 0, 0},
	{0x9398,0x0a, 0, 0}, {0x9399,0x1d, 0, 0}, {0x939a,0x30, 0, 0},
	{0x939b,0x33, 0, 0}, {0x939c,0x22, 0, 0}, {0x939d,0xc3, 0, 0},
	{0x939e,0x22, 0, 0}, {0x939f,0x75, 0, 0}, {0x93a0,0x0d, 0, 0},
	{0x93a1,0x00, 0, 0}, {0x93a2,0x75, 0, 0}, {0x93a3,0x0e, 0, 0},
	{0x93a4,0x0d, 0, 0}, {0x93a5,0x12, 0, 0}, {0x93a6,0x0a, 0, 0},
	{0x93a7,0x1d, 0, 0}, {0x93a8,0x30, 0, 0}, {0x93a9,0x33, 0, 0},
	{0x93aa,0x02, 0, 0}, {0x93ab,0xc3, 0, 0}, {0x93ac,0x22, 0, 0},
	{0x93ad,0x75, 0, 0}, {0x93ae,0x0d, 0, 0}, {0x93af,0x00, 0, 0},
	{0x93b0,0x75, 0, 0}, {0x93b1,0x0e, 0, 0}, {0x93b2,0x64, 0, 0},
	{0x93b3,0x12, 0, 0}, {0x93b4,0x16, 0, 0}, {0x93b5,0x3d, 0, 0},
	{0x93b6,0x75, 0, 0}, {0x93b7,0x0d, 0, 0}, {0x93b8,0x80, 0, 0},
	{0x93b9,0x75, 0, 0}, {0x93ba,0x0e, 0, 0}, {0x93bb,0x00, 0, 0},
	{0x93bc,0x12, 0, 0}, {0x93bd,0x0a, 0, 0}, {0x93be,0x1d, 0, 0},
	{0x93bf,0x85, 0, 0}, {0x93c0,0x0c, 0, 0}, {0x93c1,0x7d, 0, 0},
	{0x93c2,0xd3, 0, 0}, {0x93c3,0x22, 0, 0}, {0x93c4,0x12, 0, 0},
	{0x93c5,0x10, 0, 0}, {0x93c6,0x86, 0, 0}, {0x93c7,0xd2, 0, 0},
	{0x93c8,0x3e, 0, 0}, {0x93c9,0x12, 0, 0}, {0x93ca,0x0f, 0, 0},
	{0x93cb,0x73, 0, 0}, {0x93cc,0xd2, 0, 0}, {0x93cd,0x3e, 0, 0},
	{0x93ce,0x12, 0, 0}, {0x93cf,0x13, 0, 0}, {0x93d0,0x01, 0, 0},
	{0x93d1,0x30, 0, 0}, {0x93d2,0x36, 0, 0}, {0x93d3,0x26, 0, 0},
	{0x93d4,0x30, 0, 0}, {0x93d5,0x37, 0, 0}, {0x93d6,0x06, 0, 0},
	{0x93d7,0xe5, 0, 0}, {0x93d8,0x7c, 0, 0}, {0x93d9,0x65, 0, 0},
	{0x93da,0x7d, 0, 0}, {0x93db,0x60, 0, 0}, {0x93dc,0x3e, 0, 0},
	{0x93dd,0x20, 0, 0}, {0x93de,0x37, 0, 0}, {0x93df,0x04, 0, 0},
	{0x93e0,0xe5, 0, 0}, {0x93e1,0x7d, 0, 0}, {0x93e2,0x60, 0, 0},
	{0x93e3,0x37, 0, 0}, {0x93e4,0xe5, 0, 0}, {0x93e5,0x24, 0, 0},
	{0x93e6,0x54, 0, 0}, {0x93e7,0x1f, 0, 0}, {0x93e8,0xff, 0, 0},
	{0x93e9,0xbf, 0, 0}, {0x93ea,0x1f, 0, 0}, {0x93eb,0x06, 0, 0},
	{0x93ec,0x30, 0, 0}, {0x93ed,0x25, 0, 0}, {0x93ee,0x03, 0, 0},
	{0x93ef,0x20, 0, 0}, {0x93f0,0x26, 0, 0}, {0x93f1,0x29, 0, 0},
	{0x93f2,0x30, 0, 0}, {0x93f3,0x37, 0, 0}, {0x93f4,0x02, 0, 0},
	{0x93f5,0x80, 0, 0}, {0x93f6,0x21, 0, 0}, {0x93f7,0x02, 0, 0},
	{0x93f8,0x16, 0, 0}, {0x93f9,0xb5, 0, 0}, {0x93fa,0xe5, 0, 0},
	{0x93fb,0x7c, 0, 0}, {0x93fc,0xd3, 0, 0}, {0x93fd,0x95, 0, 0},
	{0x93fe,0x7d, 0, 0}, {0x93ff,0x40, 0, 0}, {0x9400,0x03, 0, 0},
	{0x9401,0xd3, 0, 0}, {0x9402,0x80, 0, 0}, {0x9403,0x01, 0, 0},
	{0x9404,0xc3, 0, 0}, {0x9405,0x50, 0, 0}, {0x9406,0x14, 0, 0},
	{0x9407,0x20, 0, 0}, {0x9408,0x39, 0, 0}, {0x9409,0x0e, 0, 0},
	{0x940a,0xe5, 0, 0}, {0x940b,0x24, 0, 0}, {0x940c,0x54, 0, 0},
	{0x940d,0x1f, 0, 0}, {0x940e,0xff, 0, 0}, {0x940f,0xbf, 0, 0},
	{0x9410,0x1f, 0, 0}, {0x9411,0x06, 0, 0}, {0x9412,0x30, 0, 0},
	{0x9413,0x25, 0, 0}, {0x9414,0x03, 0, 0}, {0x9415,0x20, 0, 0},
	{0x9416,0x26, 0, 0}, {0x9417,0x03, 0, 0}, {0x9418,0x02, 0, 0},
	{0x9419,0x16, 0, 0}, {0x941a,0xa4, 0, 0}, {0x941b,0x12, 0, 0},
	{0x941c,0x02, 0, 0}, {0x941d,0xbd, 0, 0}, {0x941e,0x22, 0, 0},
	{0x941f,0xc2, 0, 0}, {0x9420,0x34, 0, 0}, {0x9421,0x20, 0, 0},
	{0x9422,0x05, 0, 0}, {0x9423,0x05, 0, 0}, {0x9424,0x75, 0, 0},
	{0x9425,0x0a, 0, 0}, {0x9426,0xee, 0, 0}, {0x9427,0x80, 0, 0},
	{0x9428,0x36, 0, 0}, {0x9429,0x20, 0, 0}, {0x942a,0x07, 0, 0},
	{0x942b,0x08, 0, 0}, {0x942c,0x20, 0, 0}, {0x942d,0x06, 0, 0},
	{0x942e,0x05, 0, 0}, {0x942f,0xe4, 0, 0}, {0x9430,0xf5, 0, 0},
	{0x9431,0x0a, 0, 0}, {0x9432,0x80, 0, 0}, {0x9433,0x2b, 0, 0},
	{0x9434,0x20, 0, 0}, {0x9435,0x07, 0, 0}, {0x9436,0x08, 0, 0},
	{0x9437,0x30, 0, 0}, {0x9438,0x06, 0, 0}, {0x9439,0x05, 0, 0},
	{0x943a,0x75, 0, 0}, {0x943b,0x0a, 0, 0}, {0x943c,0x20, 0, 0},
	{0x943d,0x80, 0, 0}, {0x943e,0x20, 0, 0}, {0x943f,0x30, 0, 0},
	{0x9440,0x00, 0, 0}, {0x9441,0x05, 0, 0}, {0x9442,0x75, 0, 0},
	{0x9443,0x0a, 0, 0}, {0x9444,0x01, 0, 0}, {0x9445,0x80, 0, 0},
	{0x9446,0x18, 0, 0}, {0x9447,0xe5, 0, 0}, {0x9448,0x20, 0, 0},
	{0x9449,0x54, 0, 0}, {0x944a,0x07, 0, 0}, {0x944b,0xff, 0, 0},
	{0x944c,0xbf, 0, 0}, {0x944d,0x06, 0, 0}, {0x944e,0x0d, 0, 0},
	{0x944f,0x30, 0, 0}, {0x9450,0x31, 0, 0}, {0x9451,0x04, 0, 0},
	{0x9452,0x7f, 0, 0}, {0x9453,0x12, 0, 0}, {0x9454,0x80, 0, 0},
	{0x9455,0x02, 0, 0}, {0x9456,0x7f, 0, 0}, {0x9457,0x02, 0, 0},
	{0x9458,0x8f, 0, 0}, {0x9459,0x0a, 0, 0}, {0x945a,0x80, 0, 0},
	{0x945b,0x03, 0, 0}, {0x945c,0x75, 0, 0}, {0x945d,0x0a, 0, 0},
	{0x945e,0xfe, 0, 0}, {0x945f,0x90, 0, 0}, {0x9460,0x30, 0, 0},
	{0x9461,0x27, 0, 0}, {0x9462,0xe5, 0, 0}, {0x9463,0x0a, 0, 0},
	{0x9464,0xf0, 0, 0}, {0x9465,0xe5, 0, 0}, {0x9466,0x23, 0, 0},
	{0x9467,0x54, 0, 0}, {0x9468,0xf8, 0, 0}, {0x9469,0xf5, 0, 0},
	{0x946a,0x0a, 0, 0}, {0x946b,0xe5, 0, 0}, {0x946c,0x79, 0, 0},
	{0x946d,0x25, 0, 0}, {0x946e,0x0a, 0, 0}, {0x946f,0xf5, 0, 0},
	{0x9470,0x0a, 0, 0}, {0x9471,0x90, 0, 0}, {0x9472,0x30, 0, 0},
	{0x9473,0x26, 0, 0}, {0x9474,0xe5, 0, 0}, {0x9475,0x0a, 0, 0},
	{0x9476,0xf0, 0, 0}, {0x9477,0x22, 0, 0}, {0x9478,0xe5, 0, 0},
	{0x9479,0x0a, 0, 0}, {0x947a,0x70, 0, 0}, {0x947b,0x04, 0, 0},
	{0x947c,0x7e, 0, 0}, {0x947d,0x11, 0, 0}, {0x947e,0x7f, 0, 0},
	{0x947f,0xc1, 0, 0}, {0x9480,0xe5, 0, 0}, {0x9481,0x0a, 0, 0},
	{0x9482,0xb4, 0, 0}, {0x9483,0x01, 0, 0}, {0x9484,0x04, 0, 0},
	{0x9485,0x7e, 0, 0}, {0x9486,0x11, 0, 0}, {0x9487,0x7f, 0, 0},
	{0x9488,0xc6, 0, 0}, {0x9489,0xe5, 0, 0}, {0x948a,0x0a, 0, 0},
	{0x948b,0xb4, 0, 0}, {0x948c,0x02, 0, 0}, {0x948d,0x04, 0, 0},
	{0x948e,0x7e, 0, 0}, {0x948f,0x11, 0, 0}, {0x9490,0x7f, 0, 0},
	{0x9491,0xcb, 0, 0}, {0x9492,0x8f, 0, 0}, {0x9493,0x82, 0, 0},
	{0x9494,0x8e, 0, 0}, {0x9495,0x83, 0, 0}, {0x9496,0xe4, 0, 0},
	{0x9497,0x93, 0, 0}, {0x9498,0xf5, 0, 0}, {0x9499,0x2c, 0, 0},
	{0x949a,0x74, 0, 0}, {0x949b,0x01, 0, 0}, {0x949c,0x93, 0, 0},
	{0x949d,0xf5, 0, 0}, {0x949e,0x2d, 0, 0}, {0x949f,0x74, 0, 0},
	{0x94a0,0x02, 0, 0}, {0x94a1,0x93, 0, 0}, {0x94a2,0xf5, 0, 0},
	{0x94a3,0x2e, 0, 0}, {0x94a4,0x74, 0, 0}, {0x94a5,0x03, 0, 0},
	{0x94a6,0x93, 0, 0}, {0x94a7,0xf5, 0, 0}, {0x94a8,0x2f, 0, 0},
	{0x94a9,0x74, 0, 0}, {0x94aa,0x04, 0, 0}, {0x94ab,0x93, 0, 0},
	{0x94ac,0xf5, 0, 0}, {0x94ad,0x30, 0, 0}, {0x94ae,0xe5, 0, 0},
	{0x94af,0x0a, 0, 0}, {0x94b0,0xb4, 0, 0}, {0x94b1,0x01, 0, 0},
	{0x94b2,0x07, 0, 0}, {0x94b3,0x74, 0, 0}, {0x94b4,0x2c, 0, 0},
	{0x94b5,0x25, 0, 0}, {0x94b6,0x79, 0, 0}, {0x94b7,0xf8, 0, 0},
	{0x94b8,0x76, 0, 0}, {0x94b9,0x40, 0, 0}, {0x94ba,0xe5, 0, 0},
	{0x94bb,0x0a, 0, 0}, {0x94bc,0xb4, 0, 0}, {0x94bd,0x02, 0, 0},
	{0x94be,0x07, 0, 0}, {0x94bf,0x74, 0, 0}, {0x94c0,0x2c, 0, 0},
	{0x94c1,0x25, 0, 0}, {0x94c2,0x79, 0, 0}, {0x94c3,0xf8, 0, 0},
	{0x94c4,0x76, 0, 0}, {0x94c5,0x80, 0, 0}, {0x94c6,0x22, 0, 0},
	{0x94c7,0xc2, 0, 0}, {0x94c8,0xaf, 0, 0}, {0x94c9,0x90, 0, 0},
	{0x94ca,0x30, 0, 0}, {0x94cb,0x27, 0, 0}, {0x94cc,0x74, 0, 0},
	{0x94cd,0xfa, 0, 0}, {0x94ce,0xf0, 0, 0}, {0x94cf,0x12, 0, 0},
	{0x94d0,0x0e, 0, 0}, {0x94d1,0xd3, 0, 0}, {0x94d2,0x12, 0, 0},
	{0x94d3,0x16, 0, 0}, {0x94d4,0x15, 0, 0}, {0x94d5,0xe4, 0, 0},
	{0x94d6,0xf5, 0, 0}, {0x94d7,0x33, 0, 0}, {0x94d8,0xd2, 0, 0},
	{0x94d9,0xaf, 0, 0}, {0x94da,0x12, 0, 0}, {0x94db,0x0b, 0, 0},
	{0x94dc,0x69, 0, 0}, {0x94dd,0x30, 0, 0}, {0x94de,0x30, 0, 0},
	{0x94df,0x03, 0, 0}, {0x94e0,0x12, 0, 0}, {0x94e1,0x06, 0, 0},
	{0x94e2,0xf8, 0, 0}, {0x94e3,0x30, 0, 0}, {0x94e4,0x34, 0, 0},
	{0x94e5,0x03, 0, 0}, {0x94e6,0x12, 0, 0}, {0x94e7,0x14, 0, 0},
	{0x94e8,0x1f, 0, 0}, {0x94e9,0x30, 0, 0}, {0x94ea,0x3d, 0, 0},
	{0x94eb,0xee, 0, 0}, {0x94ec,0xc2, 0, 0}, {0x94ed,0x3d, 0, 0},
	{0x94ee,0xd2, 0, 0}, {0x94ef,0x35, 0, 0}, {0x94f0,0x30, 0, 0},
	{0x94f1,0x00, 0, 0}, {0x94f2,0x05, 0, 0}, {0x94f3,0x12, 0, 0},
	{0x94f4,0x15, 0, 0}, {0x94f5,0xe8, 0, 0}, {0x94f6,0x80, 0, 0},
	{0x94f7,0x17, 0, 0}, {0x94f8,0x30, 0, 0}, {0x94f9,0x07, 0, 0},
	{0x94fa,0x0b, 0, 0}, {0x94fb,0x30, 0, 0}, {0x94fc,0x06, 0, 0},
	{0x94fd,0x08, 0, 0}, {0x94fe,0x20, 0, 0}, {0x94ff,0x31, 0, 0},
	{0x9500,0x0e, 0, 0}, {0x9501,0x12, 0, 0}, {0x9502,0x12, 0, 0},
	{0x9503,0x36, 0, 0}, {0x9504,0x80, 0, 0}, {0x9505,0x09, 0, 0},
	{0x9506,0x20, 0, 0}, {0x9507,0x07, 0, 0}, {0x9508,0x06, 0, 0},
	{0x9509,0x30, 0, 0}, {0x950a,0x06, 0, 0}, {0x950b,0x03, 0, 0},
	{0x950c,0x12, 0, 0}, {0x950d,0x0d, 0, 0}, {0x950e,0xcc, 0, 0},
	{0x950f,0xc2, 0, 0}, {0x9510,0x35, 0, 0}, {0x9511,0x80, 0, 0},
	{0x9512,0xc7, 0, 0}, {0x9513,0xc0, 0, 0}, {0x9514,0xe0, 0, 0},
	{0x9515,0xc0, 0, 0}, {0x9516,0x83, 0, 0}, {0x9517,0xc0, 0, 0},
	{0x9518,0x82, 0, 0}, {0x9519,0xc0, 0, 0}, {0x951a,0xd0, 0, 0},
	{0x951b,0x90, 0, 0}, {0x951c,0x3f, 0, 0}, {0x951d,0x0d, 0, 0},
	{0x951e,0xe0, 0, 0}, {0x951f,0xf5, 0, 0}, {0x9520,0x09, 0, 0},
	{0x9521,0xe5, 0, 0}, {0x9522,0x09, 0, 0}, {0x9523,0x30, 0, 0},
	{0x9524,0xe0, 0, 0}, {0x9525,0x2e, 0, 0}, {0x9526,0xe5, 0, 0},
	{0x9527,0x7a, 0, 0}, {0x9528,0xb4, 0, 0}, {0x9529,0x01, 0, 0},
	{0x952a,0x09, 0, 0}, {0x952b,0x90, 0, 0}, {0x952c,0x3a, 0, 0},
	{0x952d,0x00, 0, 0}, {0x952e,0xe0, 0, 0}, {0x952f,0xf5, 0, 0},
	{0x9530,0x78, 0, 0}, {0x9531,0x44, 0, 0}, {0x9532,0x01, 0, 0},
	{0x9533,0xf0, 0, 0}, {0x9534,0xe5, 0, 0}, {0x9535,0x7a, 0, 0},
	{0x9536,0xb4, 0, 0}, {0x9537,0x03, 0, 0}, {0x9538,0x09, 0, 0},
	{0x9539,0x90, 0, 0}, {0x953a,0x3a, 0, 0}, {0x953b,0x00, 0, 0},
	{0x953c,0xe0, 0, 0}, {0x953d,0xf5, 0, 0}, {0x953e,0x78, 0, 0},
	{0x953f,0x54, 0, 0}, {0x9540,0xfe, 0, 0}, {0x9541,0xf0, 0, 0},
	{0x9542,0xe5, 0, 0}, {0x9543,0x7a, 0, 0}, {0x9544,0xb4, 0, 0},
	{0x9545,0x03, 0, 0}, {0x9546,0x05, 0, 0}, {0x9547,0x75, 0, 0},
	{0x9548,0x7a, 0, 0}, {0x9549,0x00, 0, 0}, {0x954a,0x80, 0, 0},
	{0x954b,0x02, 0, 0}, {0x954c,0x05, 0, 0}, {0x954d,0x7a, 0, 0},
	{0x954e,0x90, 0, 0}, {0x954f,0x3f, 0, 0}, {0x9550,0x0d, 0, 0},
	{0x9551,0x74, 0, 0}, {0x9552,0x01, 0, 0}, {0x9553,0xf0, 0, 0},
	{0x9554,0xd0, 0, 0}, {0x9555,0xd0, 0, 0}, {0x9556,0xd0, 0, 0},
	{0x9557,0x82, 0, 0}, {0x9558,0xd0, 0, 0}, {0x9559,0x83, 0, 0},
	{0x955a,0xd0, 0, 0}, {0x955b,0xe0, 0, 0}, {0x955c,0x32, 0, 0},
	{0x955d,0x90, 0, 0}, {0x955e,0x50, 0, 0}, {0x955f,0x27, 0, 0},
	{0x9560,0xe0, 0, 0}, {0x9561,0x44, 0, 0}, {0x9562,0x01, 0, 0},
	{0x9563,0xf0, 0, 0}, {0x9564,0x90, 0, 0}, {0x9565,0x50, 0, 0},
	{0x9566,0x34, 0, 0}, {0x9567,0x74, 0, 0}, {0x9568,0x80, 0, 0},
	{0x9569,0xf0, 0, 0}, {0x956a,0xa3, 0, 0}, {0x956b,0x74, 0, 0},
	{0x956c,0x2a, 0, 0}, {0x956d,0xf0, 0, 0}, {0x956e,0xa3, 0, 0},
	{0x956f,0x74, 0, 0}, {0x9570,0x14, 0, 0}, {0x9571,0xf0, 0, 0},
	{0x9572,0x90, 0, 0}, {0x9573,0x50, 0, 0}, {0x9574,0x30, 0, 0},
	{0x9575,0xe4, 0, 0}, {0x9576,0xf0, 0, 0}, {0x9577,0xa3, 0, 0},
	{0x9578,0x74, 0, 0}, {0x9579,0x02, 0, 0}, {0x957a,0xf0, 0, 0},
	{0x957b,0xa3, 0, 0}, {0x957c,0xe4, 0, 0}, {0x957d,0xf0, 0, 0},
	{0x957e,0xa3, 0, 0}, {0x957f,0x74, 0, 0}, {0x9580,0x80, 0, 0},
	{0x9581,0xf0, 0, 0}, {0x9582,0xe4, 0, 0}, {0x9583,0xf5, 0, 0},
	{0x9584,0x0a, 0, 0}, {0x9585,0x12, 0, 0}, {0x9586,0x12, 0, 0},
	{0x9587,0x9c, 0, 0}, {0x9588,0x75, 0, 0}, {0x9589,0x79, 0, 0},
	{0x958a,0x02, 0, 0}, {0x958b,0x75, 0, 0}, {0x958c,0x0a, 0, 0},
	{0x958d,0x01, 0, 0}, {0x958e,0x12, 0, 0}, {0x958f,0x14, 0, 0},
	{0x9590,0x78, 0, 0}, {0x9591,0xd2, 0, 0}, {0x9592,0x18, 0, 0},
	{0x9593,0xd2, 0, 0}, {0x9594,0x19, 0, 0}, {0x9595,0xc2, 0, 0},
	{0x9596,0x3c, 0, 0}, {0x9597,0xc2, 0, 0}, {0x9598,0x3b, 0, 0},
	{0x9599,0xd2, 0, 0}, {0x959a,0x1a, 0, 0}, {0x959b,0xd2, 0, 0},
	{0x959c,0x38, 0, 0}, {0x959d,0xd2, 0, 0}, {0x959e,0x30, 0, 0},
	{0x959f,0xc2, 0, 0}, {0x95a0,0x35, 0, 0}, {0x95a1,0xc2, 0, 0},
	{0x95a2,0x3d, 0, 0}, {0x95a3,0x22, 0, 0}, {0x95a4,0x85, 0, 0},
	{0x95a5,0x10, 0, 0}, {0x95a6,0x11, 0, 0}, {0x95a7,0x7f, 0, 0},
	{0x95a8,0x08, 0, 0}, {0x95a9,0xe5, 0, 0}, {0x95aa,0x11, 0, 0},
	{0x95ab,0x30, 0, 0}, {0x95ac,0xe7, 0, 0}, {0x95ad,0x04, 0, 0},
	{0x95ae,0xd2, 0, 0}, {0x95af,0x29, 0, 0}, {0x95b0,0x80, 0, 0},
	{0x95b1,0x02, 0, 0}, {0x95b2,0xc2, 0, 0}, {0x95b3,0x29, 0, 0},
	{0x95b4,0x12, 0, 0}, {0x95b5,0x01, 0, 0}, {0x95b6,0x6c, 0, 0},
	{0x95b7,0x75, 0, 0}, {0x95b8,0x51, 0, 0}, {0x95b9,0x0a, 0, 0},
	{0x95ba,0xae, 0, 0}, {0x95bb,0x51, 0, 0}, {0x95bc,0x15, 0, 0},
	{0x95bd,0x51, 0, 0}, {0x95be,0xee, 0, 0}, {0x95bf,0x70, 0, 0},
	{0x95c0,0xf9, 0, 0}, {0x95c1,0xe5, 0, 0}, {0x95c2,0x11, 0, 0},
	{0x95c3,0x25, 0, 0}, {0x95c4,0xe0, 0, 0}, {0x95c5,0xf5, 0, 0},
	{0x95c6,0x11, 0, 0}, {0x95c7,0xd2, 0, 0}, {0x95c8,0x28, 0, 0},
	{0x95c9,0x12, 0, 0}, {0x95ca,0x01, 0, 0}, {0x95cb,0x6c, 0, 0},
	{0x95cc,0x75, 0, 0}, {0x95cd,0x51, 0, 0}, {0x95ce,0x0a, 0, 0},
	{0x95cf,0xae, 0, 0}, {0x95d0,0x51, 0, 0}, {0x95d1,0x15, 0, 0},
	{0x95d2,0x51, 0, 0}, {0x95d3,0xee, 0, 0}, {0x95d4,0x70, 0, 0},
	{0x95d5,0xf9, 0, 0}, {0x95d6,0xc2, 0, 0}, {0x95d7,0x28, 0, 0},
	{0x95d8,0x12, 0, 0}, {0x95d9,0x01, 0, 0}, {0x95da,0x6c, 0, 0},
	{0x95db,0x75, 0, 0}, {0x95dc,0x51, 0, 0}, {0x95dd,0x05, 0, 0},
	{0x95de,0xae, 0, 0}, {0x95df,0x51, 0, 0}, {0x95e0,0x15, 0, 0},
	{0x95e1,0x51, 0, 0}, {0x95e2,0xee, 0, 0}, {0x95e3,0x70, 0, 0},
	{0x95e4,0xf9, 0, 0}, {0x95e5,0xdf, 0, 0}, {0x95e6,0xc2, 0, 0},
	{0x95e7,0x22, 0, 0}, {0x95e8,0xe5, 0, 0}, {0x95e9,0x20, 0, 0},
	{0x95ea,0x54, 0, 0}, {0x95eb,0x07, 0, 0}, {0x95ec,0xff, 0, 0},
	{0x95ed,0xbf, 0, 0}, {0x95ee,0x01, 0, 0}, {0x95ef,0x03, 0, 0},
	{0x95f0,0x02, 0, 0}, {0x95f1,0x10, 0, 0}, {0x95f2,0x00, 0, 0},
	{0x95f3,0xe5, 0, 0}, {0x95f4,0x20, 0, 0}, {0x95f5,0x54, 0, 0},
	{0x95f6,0x07, 0, 0}, {0x95f7,0xff, 0, 0}, {0x95f8,0xbf, 0, 0},
	{0x95f9,0x07, 0, 0}, {0x95fa,0x03, 0, 0}, {0x95fb,0x02, 0, 0},
	{0x95fc,0x11, 0, 0}, {0x95fd,0xd0, 0, 0}, {0x95fe,0xe5, 0, 0},
	{0x95ff,0x20, 0, 0}, {0x9600,0x54, 0, 0}, {0x9601,0x07, 0, 0},
	{0x9602,0xff, 0, 0}, {0x9603,0xbf, 0, 0}, {0x9604,0x03, 0, 0},
	{0x9605,0x03, 0, 0}, {0x9606,0x02, 0, 0}, {0x9607,0x13, 0, 0},
	{0x9608,0xc4, 0, 0}, {0x9609,0xe5, 0, 0}, {0x960a,0x20, 0, 0},
	{0x960b,0x54, 0, 0}, {0x960c,0x07, 0, 0}, {0x960d,0xff, 0, 0},
	{0x960e,0xbf, 0, 0}, {0x960f,0x05, 0, 0}, {0x9610,0x03, 0, 0},
	{0x9611,0x12, 0, 0}, {0x9612,0x16, 0, 0}, {0x9613,0xe2, 0, 0},
	{0x9614,0x22, 0, 0}, {0x9615,0x12, 0, 0}, {0x9616,0x15, 0, 0},
	{0x9617,0x5d, 0, 0}, {0x9618,0x12, 0, 0}, {0x9619,0x16, 0, 0},
	{0x961a,0xf0, 0, 0}, {0x961b,0x50, 0, 0}, {0x961c,0x04, 0, 0},
	{0x961d,0xd2, 0, 0}, {0x961e,0x05, 0, 0}, {0x961f,0x80, 0, 0},
	{0x9620,0x02, 0, 0}, {0x9621,0xc2, 0, 0}, {0x9622,0x05, 0, 0},
	{0x9623,0x12, 0, 0}, {0x9624,0x02, 0, 0}, {0x9625,0x51, 0, 0},
	{0x9626,0xc2, 0, 0}, {0x9627,0x39, 0, 0}, {0x9628,0xc2, 0, 0},
	{0x9629,0x36, 0, 0}, {0x962a,0xc2, 0, 0}, {0x962b,0x31, 0, 0},
	{0x962c,0xd2, 0, 0}, {0x962d,0x34, 0, 0}, {0x962e,0x12, 0, 0},
	{0x962f,0x02, 0, 0}, {0x9630,0x7f, 0, 0}, {0x9631,0xb5, 0, 0},
	{0x9632,0x07, 0, 0}, {0x9633,0x03, 0, 0}, {0x9634,0xd3, 0, 0},
	{0x9635,0x80, 0, 0}, {0x9636,0x01, 0, 0}, {0x9637,0xc3, 0, 0},
	{0x9638,0x40, 0, 0}, {0x9639,0x02, 0, 0}, {0x963a,0xc2, 0, 0},
	{0x963b,0x05, 0, 0}, {0x963c,0x22, 0, 0}, {0x963d,0xe4, 0, 0},
	{0x963e,0xff, 0, 0}, {0x963f,0xfe, 0, 0}, {0x9640,0xc3, 0, 0},
	{0x9641,0xef, 0, 0}, {0x9642,0x95, 0, 0}, {0x9643,0x0e, 0, 0},
	{0x9644,0xee, 0, 0}, {0x9645,0x95, 0, 0}, {0x9646,0x0d, 0, 0},
	{0x9647,0x50, 0, 0}, {0x9648,0x15, 0, 0}, {0x9649,0x7d, 0, 0},
	{0x964a,0x8a, 0, 0}, {0x964b,0x7c, 0, 0}, {0x964c,0x02, 0, 0},
	{0x964d,0xed, 0, 0}, {0x964e,0x1d, 0, 0}, {0x964f,0xaa, 0, 0},
	{0x9650,0x04, 0, 0}, {0x9651,0x70, 0, 0}, {0x9652,0x01, 0, 0},
	{0x9653,0x1c, 0, 0}, {0x9654,0x4a, 0, 0}, {0x9655,0x70, 0, 0},
	{0x9656,0xf6, 0, 0}, {0x9657,0x0f, 0, 0}, {0x9658,0xbf, 0, 0},
	{0x9659,0x00, 0, 0}, {0x965a,0x01, 0, 0}, {0x965b,0x0e, 0, 0},
	{0x965c,0x80, 0, 0}, {0x965d,0xe2, 0, 0}, {0x965e,0x22, 0, 0},
	{0x965f,0xc2, 0, 0}, {0x9660,0x03, 0, 0}, {0x9661,0xd2, 0, 0},
	{0x9662,0x04, 0, 0}, {0x9663,0x12, 0, 0}, {0x9664,0x02, 0, 0},
	{0x9665,0x9b, 0, 0}, {0x9666,0x30, 0, 0}, {0x9667,0x07, 0, 0},
	{0x9668,0x05, 0, 0}, {0x9669,0x30, 0, 0}, {0x966a,0x06, 0, 0},
	{0x966b,0x02, 0, 0}, {0x966c,0xd2, 0, 0}, {0x966d,0x36, 0, 0},
	{0x966e,0xc2, 0, 0}, {0x966f,0x3e, 0, 0}, {0x9670,0x12, 0, 0},
	{0x9671,0x0f, 0, 0}, {0x9672,0x73, 0, 0}, {0x9673,0xc2, 0, 0},
	{0x9674,0x3e, 0, 0}, {0x9675,0x12, 0, 0}, {0x9676,0x13, 0, 0},
	{0x9677,0x01, 0, 0}, {0x9678,0xd2, 0, 0}, {0x9679,0x34, 0, 0},
	{0x967a,0x22, 0, 0}, {0x967b,0x75, 0, 0}, {0x967c,0x48, 0, 0},
	{0x967d,0x11, 0, 0}, {0x967e,0x75, 0, 0}, {0x967f,0x49, 0, 0},
	{0x9680,0x6a, 0, 0}, {0x9681,0x90, 0, 0}, {0x9682,0x11, 0, 0},
	{0x9683,0x68, 0, 0}, {0x9684,0xe4, 0, 0}, {0x9685,0x93, 0, 0},
	{0x9686,0xf5, 0, 0}, {0x9687,0x7c, 0, 0}, {0x9688,0xa3, 0, 0},
	{0x9689,0xe4, 0, 0}, {0x968a,0x93, 0, 0}, {0x968b,0xf5, 0, 0},
	{0x968c,0x32, 0, 0}, {0x968d,0xc2, 0, 0}, {0x968e,0x3a, 0, 0},
	{0x968f,0x22, 0, 0}, {0x9690,0xe4, 0, 0}, {0x9691,0xff, 0, 0},
	{0x9692,0xef, 0, 0}, {0x9693,0x25, 0, 0}, {0x9694,0xe0, 0, 0},
	{0x9695,0x24, 0, 0}, {0x9696,0x56, 0, 0}, {0x9697,0xf8, 0, 0},
	{0x9698,0xe4, 0, 0}, {0x9699,0xf6, 0, 0}, {0x969a,0x08, 0, 0},
	{0x969b,0xf6, 0, 0}, {0x969c,0x0f, 0, 0}, {0x969d,0xbf, 0, 0},
	{0x969e,0x07, 0, 0}, {0x969f,0xf2, 0, 0}, {0x96a0,0x53, 0, 0},
	{0x96a1,0x24, 0, 0}, {0x96a2,0x80, 0, 0}, {0x96a3,0x22, 0, 0},
	{0x96a4,0xe5, 0, 0}, {0x96a5,0x7d, 0, 0}, {0x96a6,0xc3, 0, 0},
	{0x96a7,0x95, 0, 0}, {0x96a8,0x7c, 0, 0}, {0x96a9,0x40, 0, 0},
	{0x96aa,0x01, 0, 0}, {0x96ab,0x22, 0, 0}, {0x96ac,0xe5, 0, 0},
	{0x96ad,0x7d, 0, 0}, {0x96ae,0x04, 0, 0}, {0x96af,0xf5, 0, 0},
	{0x96b0,0x0c, 0, 0}, {0x96b1,0x12, 0, 0}, {0x96b2,0x13, 0, 0},
	{0x96b3,0x64, 0, 0}, {0x96b4,0x22, 0, 0}, {0x96b5,0xe5, 0, 0},
	{0x96b6,0x7d, 0, 0}, {0x96b7,0x70, 0, 0}, {0x96b8,0x02, 0, 0},
	{0x96b9,0xc3, 0, 0}, {0x96ba,0x22, 0, 0}, {0x96bb,0xe5, 0, 0},
	{0x96bc,0x7d, 0, 0}, {0x96bd,0x14, 0, 0}, {0x96be,0xf5, 0, 0},
	{0x96bf,0x0c, 0, 0}, {0x96c0,0x12, 0, 0}, {0x96c1,0x13, 0, 0},
	{0x96c2,0x64, 0, 0}, {0x96c3,0x22, 0, 0}, {0x96c4,0xe5, 0, 0},
	{0x96c5,0x7e, 0, 0}, {0x96c6,0xb4, 0, 0}, {0x96c7,0x01, 0, 0},
	{0x96c8,0x09, 0, 0}, {0x96c9,0x12, 0, 0}, {0x96ca,0x16, 0, 0},
	{0x96cb,0x7b, 0, 0}, {0x96cc,0xe4, 0, 0}, {0x96cd,0xf5, 0, 0},
	{0x96ce,0x0c, 0, 0}, {0x96cf,0x12, 0, 0}, {0x96d0,0x13, 0, 0},
	{0x96d1,0x64, 0, 0}, {0x96d2,0x22, 0, 0}, {0x96d3,0xe5, 0, 0},
	{0x96d4,0x7e, 0, 0}, {0x96d5,0x24, 0, 0}, {0x96d6,0xfe, 0, 0},
	{0x96d7,0x60, 0, 0}, {0x96d8,0x06, 0, 0}, {0x96d9,0x04, 0, 0},
	{0x96da,0x70, 0, 0}, {0x96db,0x05, 0, 0}, {0x96dc,0xd2, 0, 0},
	{0x96dd,0x39, 0, 0}, {0x96de,0x22, 0, 0}, {0x96df,0xc2, 0, 0},
	{0x96e0,0x39, 0, 0}, {0x96e1,0x22, 0, 0}, {0x96e2,0xe5, 0, 0},
	{0x96e3,0x31, 0, 0}, {0x96e4,0xd3, 0, 0}, {0x96e5,0x94, 0, 0},
	{0x96e6,0x00, 0, 0}, {0x96e7,0x40, 0, 0}, {0x96e8,0x03, 0, 0},
	{0x96e9,0x15, 0, 0}, {0x96ea,0x31, 0, 0}, {0x96eb,0x22, 0, 0},
	{0x96ec,0x12, 0, 0}, {0x96ed,0x16, 0, 0}, {0x96ee,0x5f, 0, 0},
	{0x96ef,0x22, 0, 0}, {0x96f0,0x12, 0, 0}, {0x96f1,0x16, 0, 0},
	{0x96f2,0x7b, 0, 0}, {0x96f3,0xe4, 0, 0}, {0x96f4,0xf5, 0, 0},
	{0x96f5,0x0c, 0, 0}, {0x96f6,0x12, 0, 0}, {0x96f7,0x13, 0, 0},
	{0x96f8,0x64, 0, 0},
	{0x96f9,0x22, 0, 0},
};

static struct reg_value ov5642_setting_15fps_QCIF_176_144[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3501, 0x1e, 0, 0},
	{0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x3818, 0xc1, 0, 0},
	{0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0}, {0x3801, 0x80, 0, 0},
	{0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0},
	{0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a1a, 0x05, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0}, {0x3502, 0x00, 0, 0},
	{0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0}, {0x3503, 0x00, 0, 0},
	{0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0}, {0x528f, 0x10, 0, 0},
	{0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x02, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x02, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x02, 0, 0},
	{0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3c, 0, 0},
	{0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x3a02, 0x00, 0, 0}, {0x3a03, 0x7d, 0, 0},
	{0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0}, {0x3a15, 0x7d, 0, 0},
	{0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a08, 0x09, 0, 0},
	{0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0}, {0x3a0b, 0xd0, 0, 0},
	{0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x06, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0}, {0x401e, 0x20, 0, 0},
	{0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0}, {0x528a, 0x01, 0, 0},
	{0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0}, {0x528d, 0x10, 0, 0},
	{0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0}, {0x5290, 0x30, 0, 0},
	{0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0}, {0x5294, 0x00, 0, 0},
	{0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0}, {0x5297, 0x08, 0, 0},
	{0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0}, {0x529a, 0x00, 0, 0},
	{0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0}, {0x529d, 0x28, 0, 0},
	{0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0}, {0x5302, 0x00, 0, 0},
	{0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0}, {0x530d, 0x0c, 0, 0},
	{0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0}, {0x5310, 0x20, 0, 0},
	{0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0}, {0x5309, 0x40, 0, 0},
	{0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0}, {0x5306, 0x00, 0, 0},
	{0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0}, {0x5315, 0x20, 0, 0},
	{0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0}, {0x5317, 0x00, 0, 0},
	{0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0}, {0x5381, 0x00, 0, 0},
	{0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0}, {0x5384, 0x00, 0, 0},
	{0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0}, {0x5387, 0x00, 0, 0},
	{0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0}, {0x538a, 0x00, 0, 0},
	{0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0}, {0x538d, 0x00, 0, 0},
	{0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0}, {0x5390, 0x00, 0, 0},
	{0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0}, {0x5393, 0xa2, 0, 0},
	{0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0}, {0x5481, 0x21, 0, 0},
	{0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0}, {0x5484, 0x65, 0, 0},
	{0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0}, {0x5487, 0x87, 0, 0},
	{0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0}, {0x548a, 0xaa, 0, 0},
	{0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0}, {0x548d, 0xdd, 0, 0},
	{0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0}, {0x5490, 0x05, 0, 0},
	{0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0}, {0x5493, 0x20, 0, 0},
	{0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0}, {0x5496, 0x02, 0, 0},
	{0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0}, {0x5499, 0x86, 0, 0},
	{0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0}, {0x549c, 0x02, 0, 0},
	{0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0}, {0x549f, 0x1c, 0, 0},
	{0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0}, {0x54a2, 0x01, 0, 0},
	{0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0}, {0x54a5, 0xc5, 0, 0},
	{0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0}, {0x54a8, 0x01, 0, 0},
	{0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0}, {0x54ab, 0x41, 0, 0},
	{0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0}, {0x54ae, 0x00, 0, 0},
	{0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0}, {0x54b1, 0x20, 0, 0},
	{0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0}, {0x54b4, 0x00, 0, 0},
	{0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0}, {0x54b7, 0xdf, 0, 0},
	{0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0}, {0x3406, 0x00, 0, 0},
	{0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0}, {0x5182, 0x11, 0, 0},
	{0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0}, {0x5185, 0x24, 0, 0},
	{0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0}, {0x5188, 0x08, 0, 0},
	{0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0}, {0x518b, 0xb2, 0, 0},
	{0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0}, {0x518e, 0x3d, 0, 0},
	{0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0}, {0x5191, 0xff, 0, 0},
	{0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0}, {0x5194, 0xf0, 0, 0},
	{0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0}, {0x519a, 0x04, 0, 0},
	{0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0}, {0x519d, 0x82, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0}, {0x3a0f, 0x38, 0, 0},
	{0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0}, {0x3a1e, 0x2e, 0, 0},
	{0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0}, {0x5688, 0xa6, 0, 0},
	{0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0}, {0x568b, 0xae, 0, 0},
	{0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0}, {0x568e, 0x62, 0, 0},
	{0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0}, {0x5584, 0x60, 0, 0},
	{0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0}, {0x5800, 0x27, 0, 0},
	{0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0}, {0x5803, 0x0f, 0, 0},
	{0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0}, {0x5806, 0x1e, 0, 0},
	{0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0}, {0x5809, 0x0d, 0, 0},
	{0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0}, {0x580c, 0x0a, 0, 0},
	{0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0}, {0x580f, 0x19, 0, 0},
	{0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0}, {0x5812, 0x04, 0, 0},
	{0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0}, {0x5815, 0x06, 0, 0},
	{0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0}, {0x5818, 0x0a, 0, 0},
	{0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0}, {0x581b, 0x00, 0, 0},
	{0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0}, {0x581e, 0x08, 0, 0},
	{0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0}, {0x5821, 0x05, 0, 0},
	{0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0}, {0x5824, 0x00, 0, 0},
	{0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0}, {0x5827, 0x0c, 0, 0},
	{0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0}, {0x582a, 0x06, 0, 0},
	{0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0}, {0x582d, 0x07, 0, 0},
	{0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0}, {0x5830, 0x18, 0, 0},
	{0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0}, {0x5833, 0x0a, 0, 0},
	{0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0}, {0x5836, 0x15, 0, 0},
	{0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0}, {0x5839, 0x1f, 0, 0},
	{0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0}, {0x583c, 0x17, 0, 0},
	{0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0}, {0x583f, 0x53, 0, 0},
	{0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0}, {0x5842, 0x0d, 0, 0},
	{0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0}, {0x5845, 0x09, 0, 0},
	{0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0}, {0x5848, 0x10, 0, 0},
	{0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0}, {0x584b, 0x0e, 0, 0},
	{0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0}, {0x584e, 0x11, 0, 0},
	{0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0}, {0x5851, 0x0c, 0, 0},
	{0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0}, {0x5854, 0x10, 0, 0},
	{0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0}, {0x5857, 0x0b, 0, 0},
	{0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0}, {0x585a, 0x0d, 0, 0},
	{0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0}, {0x585d, 0x0c, 0, 0},
	{0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0}, {0x5860, 0x0c, 0, 0},
	{0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0}, {0x5863, 0x08, 0, 0},
	{0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0}, {0x5866, 0x18, 0, 0},
	{0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0}, {0x5869, 0x19, 0, 0},
	{0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0}, {0x586c, 0x13, 0, 0},
	{0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0}, {0x586f, 0x16, 0, 0},
	{0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0}, {0x5872, 0x10, 0, 0},
	{0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0}, {0x5875, 0x16, 0, 0},
	{0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0}, {0x5878, 0x10, 0, 0},
	{0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0}, {0x587b, 0x14, 0, 0},
	{0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0}, {0x587e, 0x11, 0, 0},
	{0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0}, {0x5881, 0x15, 0, 0},
	{0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0}, {0x5884, 0x15, 0, 0},
	{0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0}, {0x5887, 0x17, 0, 0},
	{0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0}, {0x3702, 0x10, 0, 0},
	{0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0}, {0x370b, 0x40, 0, 0},
	{0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0}, {0x3632, 0x52, 0, 0},
	{0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0}, {0x5785, 0x07, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0}, {0x370f, 0xc0, 0, 0},
	{0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0}, {0x5007, 0x00, 0, 0},
	{0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0}, {0x5013, 0x00, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0}, {0x5087, 0x00, 0, 0},
	{0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0}, {0x302b, 0x00, 0, 0},
	{0x3808, 0x00, 0, 0}, {0x3809, 0xb0, 0, 0}, {0x380a, 0x00, 0, 0},
	{0x380b, 0x90, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0},	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_30fps_QCIF_176_144[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x10, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3501, 0x1e, 0, 0},
	{0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x3818, 0xc1, 0, 0},
	{0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0}, {0x3801, 0x80, 0, 0},
	{0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0},
	{0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a1a, 0x05, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0}, {0x3502, 0x00, 0, 0},
	{0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0}, {0x3503, 0x00, 0, 0},
	{0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0}, {0x528f, 0x10, 0, 0},
	{0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x02, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x02, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x02, 0, 0},
	{0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3c, 0, 0},
	{0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x3a02, 0x00, 0, 0}, {0x3a03, 0x7d, 0, 0},
	{0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0}, {0x3a15, 0x7d, 0, 0},
	{0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a08, 0x09, 0, 0},
	{0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0}, {0x3a0b, 0xd0, 0, 0},
	{0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x06, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0}, {0x401e, 0x20, 0, 0},
	{0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0}, {0x528a, 0x01, 0, 0},
	{0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0}, {0x528d, 0x10, 0, 0},
	{0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0}, {0x5290, 0x30, 0, 0},
	{0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0}, {0x5294, 0x00, 0, 0},
	{0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0}, {0x5297, 0x08, 0, 0},
	{0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0}, {0x529a, 0x00, 0, 0},
	{0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0}, {0x529d, 0x28, 0, 0},
	{0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0}, {0x5302, 0x00, 0, 0},
	{0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0}, {0x530d, 0x0c, 0, 0},
	{0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0}, {0x5310, 0x20, 0, 0},
	{0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0}, {0x5309, 0x40, 0, 0},
	{0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0}, {0x5306, 0x00, 0, 0},
	{0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0}, {0x5315, 0x20, 0, 0},
	{0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0}, {0x5317, 0x00, 0, 0},
	{0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0}, {0x5381, 0x00, 0, 0},
	{0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0}, {0x5384, 0x00, 0, 0},
	{0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0}, {0x5387, 0x00, 0, 0},
	{0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0}, {0x538a, 0x00, 0, 0},
	{0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0}, {0x538d, 0x00, 0, 0},
	{0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0}, {0x5390, 0x00, 0, 0},
	{0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0}, {0x5393, 0xa2, 0, 0},
	{0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0}, {0x5481, 0x21, 0, 0},
	{0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0}, {0x5484, 0x65, 0, 0},
	{0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0}, {0x5487, 0x87, 0, 0},
	{0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0}, {0x548a, 0xaa, 0, 0},
	{0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0}, {0x548d, 0xdd, 0, 0},
	{0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0}, {0x5490, 0x05, 0, 0},
	{0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0}, {0x5493, 0x20, 0, 0},
	{0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0}, {0x5496, 0x02, 0, 0},
	{0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0}, {0x5499, 0x86, 0, 0},
	{0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0}, {0x549c, 0x02, 0, 0},
	{0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0}, {0x549f, 0x1c, 0, 0},
	{0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0}, {0x54a2, 0x01, 0, 0},
	{0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0}, {0x54a5, 0xc5, 0, 0},
	{0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0}, {0x54a8, 0x01, 0, 0},
	{0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0}, {0x54ab, 0x41, 0, 0},
	{0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0}, {0x54ae, 0x00, 0, 0},
	{0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0}, {0x54b1, 0x20, 0, 0},
	{0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0}, {0x54b4, 0x00, 0, 0},
	{0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0}, {0x54b7, 0xdf, 0, 0},
	{0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0}, {0x3406, 0x00, 0, 0},
	{0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0}, {0x5182, 0x11, 0, 0},
	{0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0}, {0x5185, 0x24, 0, 0},
	{0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0}, {0x5188, 0x08, 0, 0},
	{0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0}, {0x518b, 0xb2, 0, 0},
	{0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0}, {0x518e, 0x3d, 0, 0},
	{0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0}, {0x5191, 0xff, 0, 0},
	{0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0}, {0x5194, 0xf0, 0, 0},
	{0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0}, {0x519a, 0x04, 0, 0},
	{0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0}, {0x519d, 0x82, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0}, {0x3a0f, 0x38, 0, 0},
	{0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0}, {0x3a1e, 0x2e, 0, 0},
	{0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0}, {0x5688, 0xa6, 0, 0},
	{0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0}, {0x568b, 0xae, 0, 0},
	{0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0}, {0x568e, 0x62, 0, 0},
	{0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0}, {0x5584, 0x60, 0, 0},
	{0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0}, {0x5800, 0x27, 0, 0},
	{0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0}, {0x5803, 0x0f, 0, 0},
	{0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0}, {0x5806, 0x1e, 0, 0},
	{0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0}, {0x5809, 0x0d, 0, 0},
	{0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0}, {0x580c, 0x0a, 0, 0},
	{0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0}, {0x580f, 0x19, 0, 0},
	{0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0}, {0x5812, 0x04, 0, 0},
	{0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0}, {0x5815, 0x06, 0, 0},
	{0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0}, {0x5818, 0x0a, 0, 0},
	{0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0}, {0x581b, 0x00, 0, 0},
	{0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0}, {0x581e, 0x08, 0, 0},
	{0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0}, {0x5821, 0x05, 0, 0},
	{0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0}, {0x5824, 0x00, 0, 0},
	{0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0}, {0x5827, 0x0c, 0, 0},
	{0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0}, {0x582a, 0x06, 0, 0},
	{0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0}, {0x582d, 0x07, 0, 0},
	{0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0}, {0x5830, 0x18, 0, 0},
	{0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0}, {0x5833, 0x0a, 0, 0},
	{0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0}, {0x5836, 0x15, 0, 0},
	{0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0}, {0x5839, 0x1f, 0, 0},
	{0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0}, {0x583c, 0x17, 0, 0},
	{0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0}, {0x583f, 0x53, 0, 0},
	{0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0}, {0x5842, 0x0d, 0, 0},
	{0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0}, {0x5845, 0x09, 0, 0},
	{0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0}, {0x5848, 0x10, 0, 0},
	{0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0}, {0x584b, 0x0e, 0, 0},
	{0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0}, {0x584e, 0x11, 0, 0},
	{0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0}, {0x5851, 0x0c, 0, 0},
	{0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0}, {0x5854, 0x10, 0, 0},
	{0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0}, {0x5857, 0x0b, 0, 0},
	{0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0}, {0x585a, 0x0d, 0, 0},
	{0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0}, {0x585d, 0x0c, 0, 0},
	{0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0}, {0x5860, 0x0c, 0, 0},
	{0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0}, {0x5863, 0x08, 0, 0},
	{0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0}, {0x5866, 0x18, 0, 0},
	{0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0}, {0x5869, 0x19, 0, 0},
	{0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0}, {0x586c, 0x13, 0, 0},
	{0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0}, {0x586f, 0x16, 0, 0},
	{0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0}, {0x5872, 0x10, 0, 0},
	{0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0}, {0x5875, 0x16, 0, 0},
	{0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0}, {0x5878, 0x10, 0, 0},
	{0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0}, {0x587b, 0x14, 0, 0},
	{0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0}, {0x587e, 0x11, 0, 0},
	{0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0}, {0x5881, 0x15, 0, 0},
	{0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0}, {0x5884, 0x15, 0, 0},
	{0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0}, {0x5887, 0x17, 0, 0},
	{0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0}, {0x3702, 0x10, 0, 0},
	{0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0}, {0x370b, 0x40, 0, 0},
	{0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0}, {0x3632, 0x52, 0, 0},
	{0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0}, {0x5785, 0x07, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0}, {0x370f, 0xc0, 0, 0},
	{0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0}, {0x5007, 0x00, 0, 0},
	{0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0}, {0x5013, 0x00, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0}, {0x5087, 0x00, 0, 0},
	{0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0}, {0x302b, 0x00, 0, 0},
	{0x3808, 0x00, 0, 0}, {0x3809, 0xb0, 0, 0}, {0x380a, 0x00, 0, 0},
	{0x380b, 0x90, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_QSXGA_2592_1944[] = {
	{0x3503, 0x07, 0, 0}, {0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0},
	{0x3002, 0x00, 0, 0}, {0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0},
	{0x3005, 0xff, 0, 0}, {0x3006, 0xff, 0, 0}, {0x3007, 0x3f, 0, 0},
	{0x3011, 0x08, 0, 0}, {0x3010, 0x10, 0, 0}, {0x3818, 0xc0, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3602, 0xe4, 0, 0}, {0x3612, 0xac, 0, 0}, {0x3613, 0x44, 0, 0},
	{0x3622, 0x60, 0, 0}, {0x3623, 0x22, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3705, 0xda, 0, 0}, {0x370a, 0x80, 0, 0}, {0x3801, 0x95, 0, 0},
	{0x3803, 0x0e, 0, 0}, {0x3804, 0x0a, 0, 0}, {0x3805, 0x20, 0, 0},
	{0x3806, 0x07, 0, 0}, {0x3807, 0x98, 0, 0}, {0x3808, 0x0a, 0, 0},
	{0x3809, 0x20, 0, 0}, {0x380a, 0x07, 0, 0}, {0x380b, 0x98, 0, 0},
	{0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3815, 0x44, 0, 0},
	{0x3824, 0x11, 0, 0}, {0x3825, 0xac, 0, 0}, {0x3827, 0x0c, 0, 0},
	{0x3a00, 0x78, 0, 0}, {0x3a0d, 0x10, 0, 0}, {0x3a0e, 0x0d, 0, 0},
	{0x5682, 0x0a, 0, 0}, {0x5683, 0x20, 0, 0}, {0x5686, 0x07, 0, 0},
	{0x5687, 0x98, 0, 0}, {0x5001, 0xff, 0, 0}, {0x589b, 0x00, 0, 0},
	{0x589a, 0xc0, 0, 0}, {0x4407, 0x04, 0, 0}, {0x3008, 0x02, 0, 0},
	{0x460b, 0x37, 0, 0}, {0x460c, 0x22, 0, 0}, {0x471d, 0x05, 0, 0},
	{0x4713, 0x03, 0, 0}, {0x471c, 0xd0, 0, 0}, {0x3815, 0x01, 0, 0},
	{0x501f, 0x00, 0, 0}, {0x3002, 0x1c, 0, 0}, {0x3819, 0x80, 0, 0},
	{0x5002, 0xe0, 0, 0}, {0x530a, 0x01, 0, 0}, {0x530d, 0x10, 0, 0},
	{0x530c, 0x04, 0, 0}, {0x5312, 0x20, 0, 0}, {0x5282, 0x01, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x3012, 0x00, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};


static struct reg_value ov5642_setting_VGA_2_QVGA[] = {
	{0x3808, 0x01, 0, 0}, {0x3809, 0x40, 0, 0}, {0x380a, 0x00, 0, 0},
	{0x380b, 0xf0, 0, 0}, {0x3815, 0x04, 0, 0},
};

static struct reg_value ov5642_setting_QSXGA_2_VGA[] = {
	{0x3503, 0x00, 0, 0}, {0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0},
	{0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0},
	{0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x3818, 0xc1, 0, 0}, {0x3621, 0x87, 0, 0},
	{0x350c, 0x03, 0, 0}, {0x350d, 0xe8, 0, 0}, {0x3602, 0xfc, 0, 0},
	{0x3612, 0xff, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3622, 0x60, 0, 0},
	{0x3623, 0x01, 0, 0}, {0x3604, 0x48, 0, 0}, {0x3705, 0xdb, 0, 0},
	{0x370a, 0x81, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x3806, 0x03, 0, 0},
	{0x3807, 0xc0, 0, 0}, {0x3808, 0x02, 0, 0}, {0x3809, 0x80, 0, 0},
	{0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3810, 0x40, 0, 0}, {0x3815, 0x04, 0, 0}, {0x3824, 0x11, 0, 0},
	{0x3825, 0xb4, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0},
	{0x5001, 0xff, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x4407, 0x0c, 0, 0}, {0x3008, 0x02, 0, 0}, {0x460b, 0x37, 0, 0},
	{0x460c, 0x22, 0, 0}, {0x471d, 0x05, 0, 0}, {0x4713, 0x02, 0, 0},
	{0x471c, 0xd0, 0, 0}, {0x3815, 0x04, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x3002, 0x5c, 0, 0}, {0x3819, 0x80, 0, 0}, {0x5002, 0xe0, 0, 0},
	{0x530a, 0x01, 0, 0}, {0x530d, 0x0c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x5312, 0x40, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x3012, 0x02, 0, 0}, {0x3010, 0x00, 0, 0},
};

static struct reg_value ov5642_setting_30fps_VGA_640_480[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0},
	{0x302c, 0x40, 0x60, 0},
	{0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_VGA_640_480[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0},
	{0x302c, 0x40, 0x60, 0},
	{0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};


static struct reg_value ov5642_setting_30fps_XGA_1024_768[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3808, 0x04, 0, 0}, {0x3809, 0x00, 0, 0}, {0x380a, 0x03, 0, 0},
	{0x380b, 0x00, 0, 0}, {0x3815, 0x02, 0, 0}, {0x302c, 0x60, 0x60, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_XGA_1024_768[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3808, 0x04, 0, 0}, {0x3809, 0x00, 0, 0}, {0x380a, 0x03, 0, 0},
	{0x380b, 0x00, 0, 0}, {0x3815, 0x02, 0, 0}, {0x302c, 0x60, 0x60, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_30fps_QVGA_320_240[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x08, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0},
	{0x302c, 0x40, 0x60, 0},
	{0x3621, 0x87, 0, 0}, {0x3808, 0x01, 0, 0},
	{0x3809, 0x40, 0, 0}, {0x380a, 0x00, 0, 0}, {0x380b, 0xf0, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_30fps_NTSC_720_480[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0xd0, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xb0, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x3c, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0},
	{0x5683, 0x00, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0x58, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0x58, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x302c, 0x60, 0x60, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_30fps_PAL_720_576[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3615, 0xf0, 0, 0}, {0x3000, 0x00, 0, 0},
	{0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0}, {0x3003, 0x00, 0, 0},
	{0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0}, {0x3006, 0x43, 0, 0},
	{0x3007, 0x37, 0, 0}, {0x3011, 0x09, 0, 0}, {0x3012, 0x02, 0, 0},
	{0x3010, 0x00, 0, 0}, {0x460c, 0x20, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0xd0, 0, 0}, {0x380a, 0x02, 0, 0}, {0x380b, 0x40, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3825, 0xd8, 0, 0},
	{0x3501, 0x1e, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0},
	{0x380c, 0x07, 0, 0}, {0x380d, 0x2a, 0, 0}, {0x380e, 0x03, 0, 0},
	{0x380f, 0xe8, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x3818, 0xc1, 0, 0}, {0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0},
	{0x3801, 0x80, 0, 0}, {0x3621, 0xc7, 0, 0}, {0x3801, 0x50, 0, 0},
	{0x3803, 0x08, 0, 0}, {0x3827, 0x3c, 0, 0}, {0x3810, 0x80, 0, 0},
	{0x3804, 0x04, 0, 0}, {0x3805, 0xb0, 0, 0}, {0x5682, 0x04, 0, 0},
	{0x5683, 0xb0, 0, 0}, {0x3806, 0x03, 0, 0}, {0x3807, 0x58, 0, 0},
	{0x5686, 0x03, 0, 0}, {0x5687, 0x58, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a1a, 0x05, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0},
	{0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0},
	{0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0},
	{0x350d, 0xd0, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0},
	{0x528c, 0x08, 0, 0}, {0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0},
	{0x528f, 0x10, 0, 0}, {0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0},
	{0x5293, 0x02, 0, 0}, {0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0},
	{0x5296, 0x00, 0, 0}, {0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0},
	{0x5299, 0x02, 0, 0}, {0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0},
	{0x529c, 0x00, 0, 0}, {0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0},
	{0x529f, 0x02, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0},
	{0x3a1f, 0x10, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0},
	{0x5193, 0x70, 0, 0}, {0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0},
	{0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3621, 0x87, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x302c, 0x60, 0x60, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_720P_1280_720[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x00, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, 
	{0x3011, 0x08, 0, 0}, {0x3010, 0x10, 0, 0}, {0x3604, 0x60, 0, 0},
	{0x3622, 0x60, 0, 0}, {0x3621, 0x09, 0, 0}, {0x3709, 0x00, 0, 0},
	{0x4000, 0x21, 0, 0}, {0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0},
	{0x3605, 0x04, 0, 0}, {0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0},
	{0x300d, 0x22, 0, 0}, {0x3623, 0x22, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5500, 0x0a, 0, 0},
	{0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0}, {0x5080, 0x08, 0, 0},
	{0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0}, {0x471d, 0x05, 0, 0},
	{0x4708, 0x06, 0, 0}, {0x370c, 0xa0, 0, 0}, {0x3808, 0x0a, 0, 0},
	{0x3809, 0x20, 0, 0}, {0x380a, 0x07, 0, 0}, {0x380b, 0x98, 0, 0},
	{0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x5687, 0x94, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x5001, 0xcf, 0, 0}, {0x4300, 0x30, 0, 0},
	{0x4300, 0x30, 0, 0}, {0x460b, 0x35, 0, 0}, {0x471d, 0x00, 0, 0},
	{0x3002, 0x0c, 0, 0}, {0x3002, 0x00, 0, 0}, {0x4713, 0x03, 0, 0},
	{0x471c, 0x50, 0, 0}, {0x4721, 0x02, 0, 0}, {0x4402, 0x90, 0, 0},
	{0x460c, 0x22, 0, 0}, {0x3815, 0x44, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3818, 0xc8, 0, 0}, {0x3801, 0x88, 0, 0}, {0x3824, 0x11, 0, 0},
	{0x3a00, 0x78, 0, 0}, {0x3a1a, 0x04, 0, 0}, {0x3a13, 0x30, 0, 0},
	{0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0},
	{0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0},
	{0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0}, {0x3a0d, 0x08, 0, 0},
	{0x3a0e, 0x06, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x32, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x32, 0, 0}, {0x3a11, 0x80, 0, 0},
	{0x3a1f, 0x20, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x09, 0, 0}, {0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0},
	{0x3a0b, 0xd0, 0, 0}, {0x3a0d, 0x10, 0, 0}, {0x3a0e, 0x0d, 0, 0},
	{0x4407, 0x04, 0, 0}, {0x5193, 0x70, 0, 0}, {0x589b, 0x00, 0, 0},
	{0x589a, 0xc0, 0, 0}, {0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0},
	{0x401c, 0x06, 0, 0}, {0x3825, 0xac, 0, 0}, {0x3827, 0x0c, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3503, 0x07, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x350c, 0x02, 0, 0}, {0x350d, 0xe4, 0, 0}, {0x3621, 0xc9, 0, 0},
	{0x370a, 0x81, 0, 0}, {0x3803, 0x08, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x3806, 0x02, 0, 0}, {0x3807, 0xd0, 0, 0},
	{0x3808, 0x05, 0, 0}, {0x3809, 0x00, 0, 0}, {0x380a, 0x02, 0, 0},
	{0x380b, 0xd0, 0, 0}, {0x380c, 0x08, 0, 0}, {0x380d, 0x72, 0, 0},
	{0x380e, 0x02, 0, 0}, {0x380f, 0xe4, 0, 0}, {0x3810, 0xc0, 0, 0},
	{0x3818, 0xc9, 0, 0}, {0x381c, 0x10, 0, 0}, {0x381d, 0xa0, 0, 0},
	{0x381e, 0x05, 0, 0}, {0x381f, 0xb0, 0, 0}, {0x3820, 0x00, 0, 0},
	{0x3821, 0x00, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3a08, 0x1b, 0, 0},
	{0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x17, 0, 0}, {0x3a0b, 0x20, 0, 0},
	{0x3a0d, 0x02, 0, 0}, {0x3a0e, 0x01, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0}, {0x5686, 0x02, 0, 0},
	{0x5687, 0xcc, 0, 0}, {0x5001, 0x7f, 0, 0}, {0x589b, 0x06, 0, 0},
	{0x589a, 0xc5, 0, 0}, {0x3503, 0x00, 0, 0}, {0x3010, 0x10, 0, 0},
	{0x460c, 0x20, 0, 0}, {0x460b, 0x37, 0, 0}, {0x471c, 0xd0, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x3815, 0x01, 0, 0}, {0x3818, 0x00, 0x08, 0},
	{0x501f, 0x00, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3002, 0x1c, 0, 0},
	{0x3819, 0x80, 0, 0}, {0x5002, 0xe0, 0, 0}, {0x3010, 0x30, 0, 0},
	{0x3a08, 0x06, 0, 0}, {0x3a09, 0x60, 0, 0}, {0x3a0a, 0x05, 0, 0},
	{0x3a0b, 0x50, 0, 0}, {0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x07, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_30fps_720P_1280_720[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x00, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, 
	{0x3011, 0x08, 0, 0}, {0x3010, 0x10, 0, 0}, {0x3604, 0x60, 0, 0},
	{0x3622, 0x60, 0, 0}, {0x3621, 0x09, 0, 0}, {0x3709, 0x00, 0, 0},
	{0x4000, 0x21, 0, 0}, {0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0},
	{0x3605, 0x04, 0, 0}, {0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0},
	{0x300d, 0x22, 0, 0}, {0x3623, 0x22, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5500, 0x0a, 0, 0},
	{0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0}, {0x5080, 0x08, 0, 0},
	{0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0}, {0x471d, 0x05, 0, 0},
	{0x4708, 0x06, 0, 0}, {0x370c, 0xa0, 0, 0}, {0x3808, 0x0a, 0, 0},
	{0x3809, 0x20, 0, 0}, {0x380a, 0x07, 0, 0}, {0x380b, 0x98, 0, 0},
	{0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x5687, 0x94, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x5001, 0xcf, 0, 0}, {0x4300, 0x30, 0, 0},
	{0x4300, 0x30, 0, 0}, {0x460b, 0x35, 0, 0}, {0x471d, 0x00, 0, 0},
	{0x3002, 0x0c, 0, 0}, {0x3002, 0x00, 0, 0}, {0x4713, 0x03, 0, 0},
	{0x471c, 0x50, 0, 0}, {0x4721, 0x02, 0, 0}, {0x4402, 0x90, 0, 0},
	{0x460c, 0x22, 0, 0}, {0x3815, 0x44, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3818, 0xc8, 0, 0}, {0x3801, 0x88, 0, 0}, {0x3824, 0x11, 0, 0},
	{0x3a00, 0x78, 0, 0}, {0x3a1a, 0x04, 0, 0}, {0x3a13, 0x30, 0, 0},
	{0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0},
	{0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0},
	{0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0}, {0x3a0d, 0x08, 0, 0},
	{0x3a0e, 0x06, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x32, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x32, 0, 0}, {0x3a11, 0x80, 0, 0},
	{0x3a1f, 0x20, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x09, 0, 0}, {0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0},
	{0x3a0b, 0xd0, 0, 0}, {0x3a0d, 0x10, 0, 0}, {0x3a0e, 0x0d, 0, 0},
	{0x4407, 0x04, 0, 0}, {0x5193, 0x70, 0, 0}, {0x589b, 0x00, 0, 0},
	{0x589a, 0xc0, 0, 0}, {0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0},
	{0x401c, 0x06, 0, 0}, {0x3825, 0xac, 0, 0}, {0x3827, 0x0c, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3503, 0x07, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x350c, 0x02, 0, 0}, {0x350d, 0xe4, 0, 0}, {0x3621, 0xc9, 0, 0},
	{0x370a, 0x81, 0, 0}, {0x3803, 0x08, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x3806, 0x02, 0, 0}, {0x3807, 0xd0, 0, 0},
	{0x3808, 0x05, 0, 0}, {0x3809, 0x00, 0, 0}, {0x380a, 0x02, 0, 0},
	{0x380b, 0xd0, 0, 0}, {0x380c, 0x08, 0, 0}, {0x380d, 0x72, 0, 0},
	{0x380e, 0x02, 0, 0}, {0x380f, 0xe4, 0, 0}, {0x3810, 0xc0, 0, 0},
	{0x3818, 0xc9, 0, 0}, {0x381c, 0x10, 0, 0}, {0x381d, 0xa0, 0, 0},
	{0x381e, 0x05, 0, 0}, {0x381f, 0xb0, 0, 0}, {0x3820, 0x00, 0, 0},
	{0x3821, 0x00, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3a08, 0x1b, 0, 0},
	{0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x17, 0, 0}, {0x3a0b, 0x20, 0, 0},
	{0x3a0d, 0x02, 0, 0}, {0x3a0e, 0x01, 0, 0}, {0x401c, 0x04, 0, 0},
	{0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0}, {0x5686, 0x02, 0, 0},
	{0x5687, 0xcc, 0, 0}, {0x5001, 0x7f, 0, 0}, {0x589b, 0x06, 0, 0},
	{0x589a, 0xc5, 0, 0}, {0x3503, 0x00, 0, 0}, {0x3010, 0x10, 0, 0},
	{0x460c, 0x20, 0, 0}, {0x460b, 0x37, 0, 0}, {0x471c, 0xd0, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x3815, 0x01, 0, 0}, {0x3818, 0x00, 0x08, 0},
	{0x501f, 0x00, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3002, 0x1c, 0, 0},
	{0x3819, 0x80, 0, 0}, {0x5002, 0xe0, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_1080P_1920_1080[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x00, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, 
	{0x3011, 0x08, 0, 0}, {0x3010, 0x10, 0, 0}, {0x3604, 0x60, 0, 0},
	{0x3622, 0x60, 0, 0}, {0x3621, 0x09, 0, 0}, {0x3709, 0x00, 0, 0},
	{0x4000, 0x21, 0, 0}, {0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0},
	{0x3605, 0x04, 0, 0}, {0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0},
	{0x300d, 0x22, 0, 0}, {0x3623, 0x22, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5500, 0x0a, 0, 0},
	{0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0}, {0x5080, 0x08, 0, 0},
	{0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0}, {0x471d, 0x05, 0, 0},
	{0x4708, 0x06, 0, 0}, {0x370c, 0xa0, 0, 0}, {0x3808, 0x0a, 0, 0},
	{0x3809, 0x20, 0, 0}, {0x380a, 0x07, 0, 0}, {0x380b, 0x98, 0, 0},
	{0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0}, {0x380e, 0x07, 0, 0},
	{0x380f, 0xd0, 0, 0}, {0x5687, 0x94, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x5001, 0xcf, 0, 0}, {0x4300, 0x30, 0, 0},
	{0x4300, 0x30, 0, 0}, {0x460b, 0x35, 0, 0}, {0x471d, 0x00, 0, 0},
	{0x3002, 0x0c, 0, 0}, {0x3002, 0x00, 0, 0}, {0x4713, 0x03, 0, 0},
	{0x471c, 0x50, 0, 0}, {0x4721, 0x02, 0, 0}, {0x4402, 0x90, 0, 0},
	{0x460c, 0x22, 0, 0}, {0x3815, 0x44, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3818, 0xc8, 0, 0}, {0x3801, 0x88, 0, 0}, {0x3824, 0x11, 0, 0},
	{0x3a00, 0x78, 0, 0}, {0x3a1a, 0x04, 0, 0}, {0x3a13, 0x30, 0, 0},
	{0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0}, {0x3a08, 0x12, 0, 0},
	{0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0}, {0x3a0b, 0xa0, 0, 0},
	{0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0}, {0x3a0d, 0x08, 0, 0},
	{0x3a0e, 0x06, 0, 0}, {0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0},
	{0x3502, 0x00, 0, 0}, {0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x32, 0, 0},
	{0x3a1b, 0x3c, 0, 0}, {0x3a1e, 0x32, 0, 0}, {0x3a11, 0x80, 0, 0},
	{0x3a1f, 0x20, 0, 0}, {0x3a02, 0x00, 0, 0},
	{0x3a03, 0x7d, 0, 0}, {0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0},
	{0x3a15, 0x7d, 0, 0}, {0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x3a08, 0x09, 0, 0}, {0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0},
	{0x3a0b, 0xd0, 0, 0}, {0x3a0d, 0x10, 0, 0}, {0x3a0e, 0x0d, 0, 0},
	{0x4407, 0x04, 0, 0}, {0x5193, 0x70, 0, 0}, {0x589b, 0x00, 0, 0},
	{0x589a, 0xc0, 0, 0}, {0x401e, 0x20, 0, 0}, {0x4001, 0x42, 0, 0},
	{0x401c, 0x06, 0, 0}, {0x3825, 0xac, 0, 0}, {0x3827, 0x0c, 0, 0},
	{0x528a, 0x01, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x10, 0, 0}, {0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0},
	{0x5290, 0x30, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x08, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x28, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0},
	{0x5282, 0x00, 0, 0}, {0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0},
	{0x5302, 0x00, 0, 0}, {0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0},
	{0x530d, 0x0c, 0, 0}, {0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0},
	{0x5310, 0x20, 0, 0}, {0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0},
	{0x5309, 0x40, 0, 0}, {0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0},
	{0x5306, 0x00, 0, 0}, {0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0},
	{0x5315, 0x20, 0, 0}, {0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0},
	{0x5317, 0x00, 0, 0}, {0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0},
	{0x5381, 0x00, 0, 0}, {0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0},
	{0x5384, 0x00, 0, 0}, {0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0},
	{0x5387, 0x00, 0, 0}, {0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0},
	{0x538a, 0x00, 0, 0}, {0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0},
	{0x538d, 0x00, 0, 0}, {0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0},
	{0x5390, 0x00, 0, 0}, {0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0},
	{0x5393, 0xa2, 0, 0}, {0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0},
	{0x5481, 0x21, 0, 0}, {0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0},
	{0x5484, 0x65, 0, 0}, {0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0},
	{0x5487, 0x87, 0, 0}, {0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0},
	{0x548a, 0xaa, 0, 0}, {0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0},
	{0x548d, 0xdd, 0, 0}, {0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0},
	{0x5490, 0x05, 0, 0}, {0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0},
	{0x5493, 0x20, 0, 0}, {0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0},
	{0x5496, 0x02, 0, 0}, {0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0},
	{0x5499, 0x86, 0, 0}, {0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0},
	{0x549c, 0x02, 0, 0}, {0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0},
	{0x549f, 0x1c, 0, 0}, {0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0},
	{0x54a2, 0x01, 0, 0}, {0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0},
	{0x54a5, 0xc5, 0, 0}, {0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0},
	{0x54a8, 0x01, 0, 0}, {0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0},
	{0x54ab, 0x41, 0, 0}, {0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0},
	{0x54ae, 0x00, 0, 0}, {0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0},
	{0x54b1, 0x20, 0, 0}, {0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0},
	{0x54b4, 0x00, 0, 0}, {0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0},
	{0x54b7, 0xdf, 0, 0}, {0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0},
	{0x3406, 0x00, 0, 0}, {0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0},
	{0x5182, 0x11, 0, 0}, {0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0},
	{0x5185, 0x24, 0, 0}, {0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0},
	{0x5188, 0x08, 0, 0}, {0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0},
	{0x518b, 0xb2, 0, 0}, {0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0},
	{0x518e, 0x3d, 0, 0}, {0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0},
	{0x5191, 0xff, 0, 0}, {0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x5194, 0xf0, 0, 0}, {0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0},
	{0x5197, 0x01, 0, 0}, {0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0},
	{0x519a, 0x04, 0, 0}, {0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0},
	{0x519d, 0x82, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0},
	{0x3a0f, 0x38, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0},
	{0x3a1e, 0x2e, 0, 0}, {0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x5688, 0xa6, 0, 0}, {0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0},
	{0x568b, 0xae, 0, 0}, {0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0},
	{0x568e, 0x62, 0, 0}, {0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0},
	{0x5584, 0x60, 0, 0}, {0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0},
	{0x5800, 0x27, 0, 0}, {0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0},
	{0x5803, 0x0f, 0, 0}, {0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0},
	{0x5806, 0x1e, 0, 0}, {0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0},
	{0x5809, 0x0d, 0, 0}, {0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0},
	{0x580c, 0x0a, 0, 0}, {0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0},
	{0x580f, 0x19, 0, 0}, {0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0},
	{0x5812, 0x04, 0, 0}, {0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0},
	{0x5815, 0x06, 0, 0}, {0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0},
	{0x5818, 0x0a, 0, 0}, {0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0},
	{0x581b, 0x00, 0, 0}, {0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0},
	{0x581e, 0x08, 0, 0}, {0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0},
	{0x5821, 0x05, 0, 0}, {0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0},
	{0x5824, 0x00, 0, 0}, {0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0},
	{0x5827, 0x0c, 0, 0}, {0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0},
	{0x582a, 0x06, 0, 0}, {0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0},
	{0x582d, 0x07, 0, 0}, {0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0},
	{0x5830, 0x18, 0, 0}, {0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0},
	{0x5833, 0x0a, 0, 0}, {0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0},
	{0x5836, 0x15, 0, 0}, {0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0},
	{0x5839, 0x1f, 0, 0}, {0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0},
	{0x583c, 0x17, 0, 0}, {0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0},
	{0x583f, 0x53, 0, 0}, {0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0},
	{0x5842, 0x0d, 0, 0}, {0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0},
	{0x5845, 0x09, 0, 0}, {0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0},
	{0x5848, 0x10, 0, 0}, {0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0},
	{0x584b, 0x0e, 0, 0}, {0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0},
	{0x584e, 0x11, 0, 0}, {0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0},
	{0x5851, 0x0c, 0, 0}, {0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0},
	{0x5854, 0x10, 0, 0}, {0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0},
	{0x5857, 0x0b, 0, 0}, {0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0},
	{0x585a, 0x0d, 0, 0}, {0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0},
	{0x585d, 0x0c, 0, 0}, {0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0},
	{0x5860, 0x0c, 0, 0}, {0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0},
	{0x5863, 0x08, 0, 0}, {0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0},
	{0x5866, 0x18, 0, 0}, {0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0},
	{0x5869, 0x19, 0, 0}, {0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0},
	{0x586c, 0x13, 0, 0}, {0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0},
	{0x586f, 0x16, 0, 0}, {0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0},
	{0x5872, 0x10, 0, 0}, {0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0},
	{0x5875, 0x16, 0, 0}, {0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0},
	{0x5878, 0x10, 0, 0}, {0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0},
	{0x587b, 0x14, 0, 0}, {0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0},
	{0x587e, 0x11, 0, 0}, {0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0},
	{0x5881, 0x15, 0, 0}, {0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0},
	{0x5884, 0x15, 0, 0}, {0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0},
	{0x5887, 0x17, 0, 0}, {0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0},
	{0x3702, 0x10, 0, 0}, {0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0},
	{0x370b, 0x40, 0, 0}, {0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0},
	{0x3632, 0x52, 0, 0}, {0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0},
	{0x5785, 0x07, 0, 0}, {0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0},
	{0x3604, 0x48, 0, 0}, {0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0},
	{0x370f, 0xc0, 0, 0}, {0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0},
	{0x5007, 0x00, 0, 0}, {0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0},
	{0x5013, 0x00, 0, 0}, {0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0},
	{0x5087, 0x00, 0, 0}, {0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0},
	{0x302b, 0x00, 0, 0}, {0x3503, 0x07, 0, 0}, {0x3011, 0x07, 0, 0},
	{0x350c, 0x04, 0, 0}, {0x350d, 0x58, 0, 0}, {0x3801, 0x8a, 0, 0},
	{0x3803, 0x0a, 0, 0}, {0x3804, 0x07, 0, 0}, {0x3805, 0x80, 0, 0},
	{0x3806, 0x04, 0, 0}, {0x3807, 0x39, 0, 0}, {0x3808, 0x07, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x04, 0, 0}, {0x380b, 0x38, 0, 0},
	{0x380c, 0x09, 0, 0}, {0x380d, 0xd6, 0, 0}, {0x380e, 0x04, 0, 0},
	{0x380f, 0x58, 0, 0}, {0x381c, 0x11, 0, 0}, {0x381d, 0xba, 0, 0},
	{0x381e, 0x04, 0, 0}, {0x381f, 0x48, 0, 0}, {0x3820, 0x04, 0, 0},
	{0x3821, 0x18, 0, 0}, {0x3a08, 0x14, 0, 0}, {0x3a09, 0xe0, 0, 0},
	{0x3a0a, 0x11, 0, 0}, {0x3a0b, 0x60, 0, 0}, {0x3a0d, 0x04, 0, 0},
	{0x3a0e, 0x03, 0, 0}, {0x5682, 0x07, 0, 0}, {0x5683, 0x60, 0, 0},
	{0x5686, 0x04, 0, 0}, {0x5687, 0x1c, 0, 0}, {0x5001, 0x7f, 0, 0},
	{0x3503, 0x00, 0, 0}, {0x3010, 0x10, 0, 0}, {0x460c, 0x20, 0, 0},
	{0x460b, 0x37, 0, 0}, {0x471c, 0xd0, 0, 0}, {0x471d, 0x05, 0, 0},
	{0x3815, 0x01, 0, 0}, {0x3818, 0x00, 0x08, 0}, {0x501f, 0x00, 0, 0},
	{0x4300, 0x30, 0, 0}, {0x3002, 0x1c, 0, 0}, {0x3819, 0x80, 0, 0},
	{0x5002, 0xe0, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_QVGA_320_240[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3501, 0x1e, 0, 0},
	{0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x3818, 0xc1, 0, 0},
	{0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0}, {0x3801, 0x80, 0, 0},
	{0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0},
	{0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a1a, 0x05, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0}, {0x3502, 0x00, 0, 0},
	{0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0}, {0x3503, 0x00, 0, 0},
	{0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0}, {0x528f, 0x10, 0, 0},
	{0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x02, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x02, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x02, 0, 0},
	{0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3c, 0, 0},
	{0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x3a02, 0x00, 0, 0}, {0x3a03, 0x7d, 0, 0},
	{0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0}, {0x3a15, 0x7d, 0, 0},
	{0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a08, 0x09, 0, 0},
	{0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0}, {0x3a0b, 0xd0, 0, 0},
	{0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x06, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0}, {0x401e, 0x20, 0, 0},
	{0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0}, {0x528a, 0x01, 0, 0},
	{0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0}, {0x528d, 0x10, 0, 0},
	{0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0}, {0x5290, 0x30, 0, 0},
	{0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0}, {0x5294, 0x00, 0, 0},
	{0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0}, {0x5297, 0x08, 0, 0},
	{0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0}, {0x529a, 0x00, 0, 0},
	{0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0}, {0x529d, 0x28, 0, 0},
	{0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0}, {0x5302, 0x00, 0, 0},
	{0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0}, {0x530d, 0x0c, 0, 0},
	{0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0}, {0x5310, 0x20, 0, 0},
	{0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0}, {0x5309, 0x40, 0, 0},
	{0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0}, {0x5306, 0x00, 0, 0},
	{0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0}, {0x5315, 0x20, 0, 0},
	{0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0}, {0x5317, 0x00, 0, 0},
	{0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0}, {0x5381, 0x00, 0, 0},
	{0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0}, {0x5384, 0x00, 0, 0},
	{0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0}, {0x5387, 0x00, 0, 0},
	{0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0}, {0x538a, 0x00, 0, 0},
	{0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0}, {0x538d, 0x00, 0, 0},
	{0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0}, {0x5390, 0x00, 0, 0},
	{0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0}, {0x5393, 0xa2, 0, 0},
	{0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0}, {0x5481, 0x21, 0, 0},
	{0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0}, {0x5484, 0x65, 0, 0},
	{0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0}, {0x5487, 0x87, 0, 0},
	{0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0}, {0x548a, 0xaa, 0, 0},
	{0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0}, {0x548d, 0xdd, 0, 0},
	{0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0}, {0x5490, 0x05, 0, 0},
	{0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0}, {0x5493, 0x20, 0, 0},
	{0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0}, {0x5496, 0x02, 0, 0},
	{0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0}, {0x5499, 0x86, 0, 0},
	{0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0}, {0x549c, 0x02, 0, 0},
	{0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0}, {0x549f, 0x1c, 0, 0},
	{0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0}, {0x54a2, 0x01, 0, 0},
	{0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0}, {0x54a5, 0xc5, 0, 0},
	{0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0}, {0x54a8, 0x01, 0, 0},
	{0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0}, {0x54ab, 0x41, 0, 0},
	{0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0}, {0x54ae, 0x00, 0, 0},
	{0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0}, {0x54b1, 0x20, 0, 0},
	{0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0}, {0x54b4, 0x00, 0, 0},
	{0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0}, {0x54b7, 0xdf, 0, 0},
	{0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0}, {0x3406, 0x00, 0, 0},
	{0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0}, {0x5182, 0x11, 0, 0},
	{0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0}, {0x5185, 0x24, 0, 0},
	{0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0}, {0x5188, 0x08, 0, 0},
	{0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0}, {0x518b, 0xb2, 0, 0},
	{0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0}, {0x518e, 0x3d, 0, 0},
	{0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0}, {0x5191, 0xff, 0, 0},
	{0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0}, {0x5194, 0xf0, 0, 0},
	{0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0}, {0x519a, 0x04, 0, 0},
	{0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0}, {0x519d, 0x82, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0}, {0x3a0f, 0x38, 0, 0},
	{0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0}, {0x3a1e, 0x2e, 0, 0},
	{0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0}, {0x5688, 0xa6, 0, 0},
	{0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0}, {0x568b, 0xae, 0, 0},
	{0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0}, {0x568e, 0x62, 0, 0},
	{0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0}, {0x5584, 0x60, 0, 0},
	{0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0}, {0x5800, 0x27, 0, 0},
	{0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0}, {0x5803, 0x0f, 0, 0},
	{0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0}, {0x5806, 0x1e, 0, 0},
	{0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0}, {0x5809, 0x0d, 0, 0},
	{0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0}, {0x580c, 0x0a, 0, 0},
	{0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0}, {0x580f, 0x19, 0, 0},
	{0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0}, {0x5812, 0x04, 0, 0},
	{0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0}, {0x5815, 0x06, 0, 0},
	{0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0}, {0x5818, 0x0a, 0, 0},
	{0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0}, {0x581b, 0x00, 0, 0},
	{0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0}, {0x581e, 0x08, 0, 0},
	{0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0}, {0x5821, 0x05, 0, 0},
	{0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0}, {0x5824, 0x00, 0, 0},
	{0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0}, {0x5827, 0x0c, 0, 0},
	{0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0}, {0x582a, 0x06, 0, 0},
	{0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0}, {0x582d, 0x07, 0, 0},
	{0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0}, {0x5830, 0x18, 0, 0},
	{0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0}, {0x5833, 0x0a, 0, 0},
	{0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0}, {0x5836, 0x15, 0, 0},
	{0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0}, {0x5839, 0x1f, 0, 0},
	{0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0}, {0x583c, 0x17, 0, 0},
	{0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0}, {0x583f, 0x53, 0, 0},
	{0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0}, {0x5842, 0x0d, 0, 0},
	{0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0}, {0x5845, 0x09, 0, 0},
	{0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0}, {0x5848, 0x10, 0, 0},
	{0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0}, {0x584b, 0x0e, 0, 0},
	{0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0}, {0x584e, 0x11, 0, 0},
	{0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0}, {0x5851, 0x0c, 0, 0},
	{0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0}, {0x5854, 0x10, 0, 0},
	{0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0}, {0x5857, 0x0b, 0, 0},
	{0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0}, {0x585a, 0x0d, 0, 0},
	{0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0}, {0x585d, 0x0c, 0, 0},
	{0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0}, {0x5860, 0x0c, 0, 0},
	{0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0}, {0x5863, 0x08, 0, 0},
	{0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0}, {0x5866, 0x18, 0, 0},
	{0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0}, {0x5869, 0x19, 0, 0},
	{0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0}, {0x586c, 0x13, 0, 0},
	{0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0}, {0x586f, 0x16, 0, 0},
	{0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0}, {0x5872, 0x10, 0, 0},
	{0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0}, {0x5875, 0x16, 0, 0},
	{0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0}, {0x5878, 0x10, 0, 0},
	{0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0}, {0x587b, 0x14, 0, 0},
	{0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0}, {0x587e, 0x11, 0, 0},
	{0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0}, {0x5881, 0x15, 0, 0},
	{0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0}, {0x5884, 0x15, 0, 0},
	{0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0}, {0x5887, 0x17, 0, 0},
	{0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0}, {0x3702, 0x10, 0, 0},
	{0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0}, {0x370b, 0x40, 0, 0},
	{0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0}, {0x3632, 0x52, 0, 0},
	{0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0}, {0x5785, 0x07, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0}, {0x370f, 0xc0, 0, 0},
	{0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0}, {0x5007, 0x00, 0, 0},
	{0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0}, {0x5013, 0x00, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0}, {0x5087, 0x00, 0, 0},
	{0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0}, {0x302b, 0x00, 0, 0},
	{0x302c, 0x40, 0x60, 0},
	{0x3808, 0x01, 0, 0}, {0x3809, 0x40, 0, 0}, {0x380a, 0x00, 0, 0},
	{0x380b, 0xf0, 0, 0}, {0x3a00, 0x78, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_NTSC_720_480[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3501, 0x1e, 0, 0},
	{0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x3818, 0xc1, 0, 0},
	{0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0}, {0x3801, 0x80, 0, 0},
	{0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0},
	{0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a1a, 0x05, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0}, {0x3502, 0x00, 0, 0},
	{0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0}, {0x3503, 0x00, 0, 0},
	{0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0}, {0x528f, 0x10, 0, 0},
	{0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x02, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x02, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x02, 0, 0},
	{0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3c, 0, 0},
	{0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x3a02, 0x00, 0, 0}, {0x3a03, 0x7d, 0, 0},
	{0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0}, {0x3a15, 0x7d, 0, 0},
	{0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a08, 0x09, 0, 0},
	{0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0}, {0x3a0b, 0xd0, 0, 0},
	{0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x06, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0}, {0x401e, 0x20, 0, 0},
	{0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0}, {0x528a, 0x01, 0, 0},
	{0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0}, {0x528d, 0x10, 0, 0},
	{0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0}, {0x5290, 0x30, 0, 0},
	{0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0}, {0x5294, 0x00, 0, 0},
	{0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0}, {0x5297, 0x08, 0, 0},
	{0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0}, {0x529a, 0x00, 0, 0},
	{0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0}, {0x529d, 0x28, 0, 0},
	{0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0}, {0x5302, 0x00, 0, 0},
	{0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0}, {0x530d, 0x0c, 0, 0},
	{0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0}, {0x5310, 0x20, 0, 0},
	{0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0}, {0x5309, 0x40, 0, 0},
	{0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0}, {0x5306, 0x00, 0, 0},
	{0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0}, {0x5315, 0x20, 0, 0},
	{0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0}, {0x5317, 0x00, 0, 0},
	{0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0}, {0x5381, 0x00, 0, 0},
	{0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0}, {0x5384, 0x00, 0, 0},
	{0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0}, {0x5387, 0x00, 0, 0},
	{0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0}, {0x538a, 0x00, 0, 0},
	{0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0}, {0x538d, 0x00, 0, 0},
	{0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0}, {0x5390, 0x00, 0, 0},
	{0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0}, {0x5393, 0xa2, 0, 0},
	{0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0}, {0x5481, 0x21, 0, 0},
	{0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0}, {0x5484, 0x65, 0, 0},
	{0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0}, {0x5487, 0x87, 0, 0},
	{0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0}, {0x548a, 0xaa, 0, 0},
	{0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0}, {0x548d, 0xdd, 0, 0},
	{0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0}, {0x5490, 0x05, 0, 0},
	{0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0}, {0x5493, 0x20, 0, 0},
	{0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0}, {0x5496, 0x02, 0, 0},
	{0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0}, {0x5499, 0x86, 0, 0},
	{0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0}, {0x549c, 0x02, 0, 0},
	{0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0}, {0x549f, 0x1c, 0, 0},
	{0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0}, {0x54a2, 0x01, 0, 0},
	{0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0}, {0x54a5, 0xc5, 0, 0},
	{0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0}, {0x54a8, 0x01, 0, 0},
	{0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0}, {0x54ab, 0x41, 0, 0},
	{0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0}, {0x54ae, 0x00, 0, 0},
	{0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0}, {0x54b1, 0x20, 0, 0},
	{0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0}, {0x54b4, 0x00, 0, 0},
	{0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0}, {0x54b7, 0xdf, 0, 0},
	{0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0}, {0x3406, 0x00, 0, 0},
	{0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0}, {0x5182, 0x11, 0, 0},
	{0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0}, {0x5185, 0x24, 0, 0},
	{0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0}, {0x5188, 0x08, 0, 0},
	{0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0}, {0x518b, 0xb2, 0, 0},
	{0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0}, {0x518e, 0x3d, 0, 0},
	{0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0}, {0x5191, 0xff, 0, 0},
	{0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0}, {0x5194, 0xf0, 0, 0},
	{0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0}, {0x519a, 0x04, 0, 0},
	{0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0}, {0x519d, 0x82, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0}, {0x3a0f, 0x38, 0, 0},
	{0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0}, {0x3a1e, 0x2e, 0, 0},
	{0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0}, {0x5688, 0xa6, 0, 0},
	{0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0}, {0x568b, 0xae, 0, 0},
	{0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0}, {0x568e, 0x62, 0, 0},
	{0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0}, {0x5584, 0x60, 0, 0},
	{0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0}, {0x5800, 0x27, 0, 0},
	{0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0}, {0x5803, 0x0f, 0, 0},
	{0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0}, {0x5806, 0x1e, 0, 0},
	{0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0}, {0x5809, 0x0d, 0, 0},
	{0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0}, {0x580c, 0x0a, 0, 0},
	{0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0}, {0x580f, 0x19, 0, 0},
	{0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0}, {0x5812, 0x04, 0, 0},
	{0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0}, {0x5815, 0x06, 0, 0},
	{0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0}, {0x5818, 0x0a, 0, 0},
	{0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0}, {0x581b, 0x00, 0, 0},
	{0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0}, {0x581e, 0x08, 0, 0},
	{0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0}, {0x5821, 0x05, 0, 0},
	{0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0}, {0x5824, 0x00, 0, 0},
	{0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0}, {0x5827, 0x0c, 0, 0},
	{0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0}, {0x582a, 0x06, 0, 0},
	{0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0}, {0x582d, 0x07, 0, 0},
	{0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0}, {0x5830, 0x18, 0, 0},
	{0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0}, {0x5833, 0x0a, 0, 0},
	{0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0}, {0x5836, 0x15, 0, 0},
	{0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0}, {0x5839, 0x1f, 0, 0},
	{0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0}, {0x583c, 0x17, 0, 0},
	{0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0}, {0x583f, 0x53, 0, 0},
	{0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0}, {0x5842, 0x0d, 0, 0},
	{0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0}, {0x5845, 0x09, 0, 0},
	{0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0}, {0x5848, 0x10, 0, 0},
	{0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0}, {0x584b, 0x0e, 0, 0},
	{0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0}, {0x584e, 0x11, 0, 0},
	{0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0}, {0x5851, 0x0c, 0, 0},
	{0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0}, {0x5854, 0x10, 0, 0},
	{0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0}, {0x5857, 0x0b, 0, 0},
	{0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0}, {0x585a, 0x0d, 0, 0},
	{0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0}, {0x585d, 0x0c, 0, 0},
	{0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0}, {0x5860, 0x0c, 0, 0},
	{0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0}, {0x5863, 0x08, 0, 0},
	{0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0}, {0x5866, 0x18, 0, 0},
	{0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0}, {0x5869, 0x19, 0, 0},
	{0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0}, {0x586c, 0x13, 0, 0},
	{0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0}, {0x586f, 0x16, 0, 0},
	{0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0}, {0x5872, 0x10, 0, 0},
	{0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0}, {0x5875, 0x16, 0, 0},
	{0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0}, {0x5878, 0x10, 0, 0},
	{0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0}, {0x587b, 0x14, 0, 0},
	{0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0}, {0x587e, 0x11, 0, 0},
	{0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0}, {0x5881, 0x15, 0, 0},
	{0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0}, {0x5884, 0x15, 0, 0},
	{0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0}, {0x5887, 0x17, 0, 0},
	{0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0}, {0x3702, 0x10, 0, 0},
	{0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0}, {0x370b, 0x40, 0, 0},
	{0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0}, {0x3632, 0x52, 0, 0},
	{0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0}, {0x5785, 0x07, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0}, {0x370f, 0xc0, 0, 0},
	{0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0}, {0x5007, 0x00, 0, 0},
	{0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0}, {0x5013, 0x00, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0}, {0x5087, 0x00, 0, 0},
	{0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0}, {0x302b, 0x00, 0, 0},
	{0x3824, 0x11, 0, 0}, {0x3825, 0xb4, 0, 0}, {0x3826, 0x00, 0, 0},
	{0x3827, 0x3d, 0, 0}, {0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0},
	{0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0xd0, 0, 0}, {0x380A, 0x01, 0, 0}, {0x380B, 0xe0, 0, 0},
	{0x3804, 0x05, 0, 0}, {0x3805, 0x00, 0, 0}, {0x3806, 0x03, 0, 0},
	{0x3807, 0x55, 0, 0}, {0x5686, 0x03, 0, 0}, {0x5687, 0x55, 0, 0},
	{0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct reg_value ov5642_setting_15fps_PAL_720_576[] = {
	{0x3103, 0x93, 0, 0}, {0x3008, 0x82, 0, 0}, {0x3017, 0x7f, 0, 0},
	{0x3018, 0xfc, 0, 0}, {0x3810, 0xc2, 0, 0}, {0x3615, 0xf0, 0, 0},
	{0x3000, 0x00, 0, 0}, {0x3001, 0x00, 0, 0}, {0x3002, 0x5c, 0, 0},
	{0x3003, 0x00, 0, 0}, {0x3004, 0xff, 0, 0}, {0x3005, 0xff, 0, 0},
	{0x3006, 0x43, 0, 0}, {0x3007, 0x37, 0, 0}, {0x3011, 0x08, 0, 0},
	{0x3010, 0x10, 0, 0}, {0x460c, 0x22, 0, 0}, {0x3815, 0x04, 0, 0},
	{0x370c, 0xa0, 0, 0}, {0x3602, 0xfc, 0, 0}, {0x3612, 0xff, 0, 0},
	{0x3634, 0xc0, 0, 0}, {0x3613, 0x00, 0, 0}, {0x3605, 0x7c, 0, 0},
	{0x3621, 0x09, 0, 0}, {0x3622, 0x60, 0, 0}, {0x3604, 0x40, 0, 0},
	{0x3603, 0xa7, 0, 0}, {0x3603, 0x27, 0, 0}, {0x4000, 0x21, 0, 0},
	{0x401d, 0x22, 0, 0}, {0x3600, 0x54, 0, 0}, {0x3605, 0x04, 0, 0},
	{0x3606, 0x3f, 0, 0}, {0x3c01, 0x80, 0, 0}, {0x5000, 0x4f, 0, 0},
	{0x5020, 0x04, 0, 0}, {0x5181, 0x79, 0, 0}, {0x5182, 0x00, 0, 0},
	{0x5185, 0x22, 0, 0}, {0x5197, 0x01, 0, 0}, {0x5001, 0xff, 0, 0},
	{0x5500, 0x0a, 0, 0}, {0x5504, 0x00, 0, 0}, {0x5505, 0x7f, 0, 0},
	{0x5080, 0x08, 0, 0}, {0x300e, 0x18, 0, 0}, {0x4610, 0x00, 0, 0},
	{0x471d, 0x05, 0, 0}, {0x4708, 0x06, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0x80, 0, 0}, {0x380a, 0x01, 0, 0}, {0x380b, 0xe0, 0, 0},
	{0x380e, 0x07, 0, 0}, {0x380f, 0xd0, 0, 0}, {0x501f, 0x00, 0, 0},
	{0x5000, 0x4f, 0, 0}, {0x4300, 0x30, 0, 0}, {0x3503, 0x07, 0, 0},
	{0x3501, 0x73, 0, 0}, {0x3502, 0x80, 0, 0}, {0x350b, 0x00, 0, 0},
	{0x3503, 0x07, 0, 0}, {0x3824, 0x11, 0, 0}, {0x3501, 0x1e, 0, 0},
	{0x3502, 0x80, 0, 0}, {0x350b, 0x7f, 0, 0}, {0x380c, 0x0c, 0, 0},
	{0x380d, 0x80, 0, 0}, {0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0},
	{0x3a0d, 0x04, 0, 0}, {0x3a0e, 0x03, 0, 0}, {0x3818, 0xc1, 0, 0},
	{0x3705, 0xdb, 0, 0}, {0x370a, 0x81, 0, 0}, {0x3801, 0x80, 0, 0},
	{0x3621, 0x87, 0, 0}, {0x3801, 0x50, 0, 0}, {0x3803, 0x08, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x3810, 0x40, 0, 0}, {0x3804, 0x05, 0, 0},
	{0x3805, 0x00, 0, 0}, {0x5682, 0x05, 0, 0}, {0x5683, 0x00, 0, 0},
	{0x3806, 0x03, 0, 0}, {0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0},
	{0x5687, 0xbc, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a1a, 0x05, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3a18, 0x00, 0, 0}, {0x3a19, 0x7c, 0, 0},
	{0x3a08, 0x12, 0, 0}, {0x3a09, 0xc0, 0, 0}, {0x3a0a, 0x0f, 0, 0},
	{0x3a0b, 0xa0, 0, 0}, {0x350c, 0x07, 0, 0}, {0x350d, 0xd0, 0, 0},
	{0x3500, 0x00, 0, 0}, {0x3501, 0x00, 0, 0}, {0x3502, 0x00, 0, 0},
	{0x350a, 0x00, 0, 0}, {0x350b, 0x00, 0, 0}, {0x3503, 0x00, 0, 0},
	{0x528a, 0x02, 0, 0}, {0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0},
	{0x528d, 0x08, 0, 0}, {0x528e, 0x08, 0, 0}, {0x528f, 0x10, 0, 0},
	{0x5290, 0x10, 0, 0}, {0x5292, 0x00, 0, 0}, {0x5293, 0x02, 0, 0},
	{0x5294, 0x00, 0, 0}, {0x5295, 0x02, 0, 0}, {0x5296, 0x00, 0, 0},
	{0x5297, 0x02, 0, 0}, {0x5298, 0x00, 0, 0}, {0x5299, 0x02, 0, 0},
	{0x529a, 0x00, 0, 0}, {0x529b, 0x02, 0, 0}, {0x529c, 0x00, 0, 0},
	{0x529d, 0x02, 0, 0}, {0x529e, 0x00, 0, 0}, {0x529f, 0x02, 0, 0},
	{0x3a0f, 0x3c, 0, 0}, {0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3c, 0, 0},
	{0x3a1e, 0x30, 0, 0}, {0x3a11, 0x70, 0, 0}, {0x3a1f, 0x10, 0, 0},
	{0x3a02, 0x00, 0, 0}, {0x3a03, 0x7d, 0, 0},
	{0x3a04, 0x00, 0, 0}, {0x3a14, 0x00, 0, 0}, {0x3a15, 0x7d, 0, 0},
	{0x3a16, 0x00, 0, 0}, {0x3a00, 0x78, 0, 0}, {0x3a08, 0x09, 0, 0},
	{0x3a09, 0x60, 0, 0}, {0x3a0a, 0x07, 0, 0}, {0x3a0b, 0xd0, 0, 0},
	{0x3a0d, 0x08, 0, 0}, {0x3a0e, 0x06, 0, 0}, {0x5193, 0x70, 0, 0},
	{0x589b, 0x04, 0, 0}, {0x589a, 0xc5, 0, 0}, {0x401e, 0x20, 0, 0},
	{0x4001, 0x42, 0, 0}, {0x401c, 0x04, 0, 0}, {0x528a, 0x01, 0, 0},
	{0x528b, 0x04, 0, 0}, {0x528c, 0x08, 0, 0}, {0x528d, 0x10, 0, 0},
	{0x528e, 0x20, 0, 0}, {0x528f, 0x28, 0, 0}, {0x5290, 0x30, 0, 0},
	{0x5292, 0x00, 0, 0}, {0x5293, 0x01, 0, 0}, {0x5294, 0x00, 0, 0},
	{0x5295, 0x04, 0, 0}, {0x5296, 0x00, 0, 0}, {0x5297, 0x08, 0, 0},
	{0x5298, 0x00, 0, 0}, {0x5299, 0x10, 0, 0}, {0x529a, 0x00, 0, 0},
	{0x529b, 0x20, 0, 0}, {0x529c, 0x00, 0, 0}, {0x529d, 0x28, 0, 0},
	{0x529e, 0x00, 0, 0}, {0x529f, 0x30, 0, 0}, {0x5282, 0x00, 0, 0},
	{0x5300, 0x00, 0, 0}, {0x5301, 0x20, 0, 0}, {0x5302, 0x00, 0, 0},
	{0x5303, 0x7c, 0, 0}, {0x530c, 0x00, 0, 0}, {0x530d, 0x0c, 0, 0},
	{0x530e, 0x20, 0, 0}, {0x530f, 0x80, 0, 0}, {0x5310, 0x20, 0, 0},
	{0x5311, 0x80, 0, 0}, {0x5308, 0x20, 0, 0}, {0x5309, 0x40, 0, 0},
	{0x5304, 0x00, 0, 0}, {0x5305, 0x30, 0, 0}, {0x5306, 0x00, 0, 0},
	{0x5307, 0x80, 0, 0}, {0x5314, 0x08, 0, 0}, {0x5315, 0x20, 0, 0},
	{0x5319, 0x30, 0, 0}, {0x5316, 0x10, 0, 0}, {0x5317, 0x00, 0, 0},
	{0x5318, 0x02, 0, 0}, {0x5380, 0x01, 0, 0}, {0x5381, 0x00, 0, 0},
	{0x5382, 0x00, 0, 0}, {0x5383, 0x4e, 0, 0}, {0x5384, 0x00, 0, 0},
	{0x5385, 0x0f, 0, 0}, {0x5386, 0x00, 0, 0}, {0x5387, 0x00, 0, 0},
	{0x5388, 0x01, 0, 0}, {0x5389, 0x15, 0, 0}, {0x538a, 0x00, 0, 0},
	{0x538b, 0x31, 0, 0}, {0x538c, 0x00, 0, 0}, {0x538d, 0x00, 0, 0},
	{0x538e, 0x00, 0, 0}, {0x538f, 0x0f, 0, 0}, {0x5390, 0x00, 0, 0},
	{0x5391, 0xab, 0, 0}, {0x5392, 0x00, 0, 0}, {0x5393, 0xa2, 0, 0},
	{0x5394, 0x08, 0, 0}, {0x5480, 0x14, 0, 0}, {0x5481, 0x21, 0, 0},
	{0x5482, 0x36, 0, 0}, {0x5483, 0x57, 0, 0}, {0x5484, 0x65, 0, 0},
	{0x5485, 0x71, 0, 0}, {0x5486, 0x7d, 0, 0}, {0x5487, 0x87, 0, 0},
	{0x5488, 0x91, 0, 0}, {0x5489, 0x9a, 0, 0}, {0x548a, 0xaa, 0, 0},
	{0x548b, 0xb8, 0, 0}, {0x548c, 0xcd, 0, 0}, {0x548d, 0xdd, 0, 0},
	{0x548e, 0xea, 0, 0}, {0x548f, 0x1d, 0, 0}, {0x5490, 0x05, 0, 0},
	{0x5491, 0x00, 0, 0}, {0x5492, 0x04, 0, 0}, {0x5493, 0x20, 0, 0},
	{0x5494, 0x03, 0, 0}, {0x5495, 0x60, 0, 0}, {0x5496, 0x02, 0, 0},
	{0x5497, 0xb8, 0, 0}, {0x5498, 0x02, 0, 0}, {0x5499, 0x86, 0, 0},
	{0x549a, 0x02, 0, 0}, {0x549b, 0x5b, 0, 0}, {0x549c, 0x02, 0, 0},
	{0x549d, 0x3b, 0, 0}, {0x549e, 0x02, 0, 0}, {0x549f, 0x1c, 0, 0},
	{0x54a0, 0x02, 0, 0}, {0x54a1, 0x04, 0, 0}, {0x54a2, 0x01, 0, 0},
	{0x54a3, 0xed, 0, 0}, {0x54a4, 0x01, 0, 0}, {0x54a5, 0xc5, 0, 0},
	{0x54a6, 0x01, 0, 0}, {0x54a7, 0xa5, 0, 0}, {0x54a8, 0x01, 0, 0},
	{0x54a9, 0x6c, 0, 0}, {0x54aa, 0x01, 0, 0}, {0x54ab, 0x41, 0, 0},
	{0x54ac, 0x01, 0, 0}, {0x54ad, 0x20, 0, 0}, {0x54ae, 0x00, 0, 0},
	{0x54af, 0x16, 0, 0}, {0x54b0, 0x01, 0, 0}, {0x54b1, 0x20, 0, 0},
	{0x54b2, 0x00, 0, 0}, {0x54b3, 0x10, 0, 0}, {0x54b4, 0x00, 0, 0},
	{0x54b5, 0xf0, 0, 0}, {0x54b6, 0x00, 0, 0}, {0x54b7, 0xdf, 0, 0},
	{0x5402, 0x3f, 0, 0}, {0x5403, 0x00, 0, 0}, {0x3406, 0x00, 0, 0},
	{0x5180, 0xff, 0, 0}, {0x5181, 0x52, 0, 0}, {0x5182, 0x11, 0, 0},
	{0x5183, 0x94, 0, 0}, {0x5184, 0x25, 0, 0}, {0x5185, 0x24, 0, 0},
	{0x5186, 0x06, 0, 0}, {0x5187, 0x08, 0, 0}, {0x5188, 0x08, 0, 0},
	{0x5189, 0x7c, 0, 0}, {0x518a, 0x60, 0, 0}, {0x518b, 0xb2, 0, 0},
	{0x518c, 0xb2, 0, 0}, {0x518d, 0x44, 0, 0}, {0x518e, 0x3d, 0, 0},
	{0x518f, 0x58, 0, 0}, {0x5190, 0x46, 0, 0}, {0x5191, 0xff, 0, 0},
	{0x5192, 0x00, 0, 0}, {0x5193, 0x70, 0, 0}, {0x5194, 0xf0, 0, 0},
	{0x5195, 0xf0, 0, 0}, {0x5196, 0x03, 0, 0}, {0x5197, 0x01, 0, 0},
	{0x5198, 0x04, 0, 0}, {0x5199, 0x12, 0, 0}, {0x519a, 0x04, 0, 0},
	{0x519b, 0x00, 0, 0}, {0x519c, 0x06, 0, 0}, {0x519d, 0x82, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5025, 0x80, 0, 0}, {0x3a0f, 0x38, 0, 0},
	{0x3a10, 0x30, 0, 0}, {0x3a1b, 0x3a, 0, 0}, {0x3a1e, 0x2e, 0, 0},
	{0x3a11, 0x60, 0, 0}, {0x3a1f, 0x10, 0, 0}, {0x5688, 0xa6, 0, 0},
	{0x5689, 0x6a, 0, 0}, {0x568a, 0xea, 0, 0}, {0x568b, 0xae, 0, 0},
	{0x568c, 0xa6, 0, 0}, {0x568d, 0x6a, 0, 0}, {0x568e, 0x62, 0, 0},
	{0x568f, 0x26, 0, 0}, {0x5583, 0x60, 0, 0}, {0x5584, 0x60, 0, 0},
	{0x5580, 0x02, 0, 0}, {0x5000, 0xcf, 0, 0}, {0x5800, 0x27, 0, 0},
	{0x5801, 0x19, 0, 0}, {0x5802, 0x12, 0, 0}, {0x5803, 0x0f, 0, 0},
	{0x5804, 0x10, 0, 0}, {0x5805, 0x15, 0, 0}, {0x5806, 0x1e, 0, 0},
	{0x5807, 0x2f, 0, 0}, {0x5808, 0x15, 0, 0}, {0x5809, 0x0d, 0, 0},
	{0x580a, 0x0a, 0, 0}, {0x580b, 0x09, 0, 0}, {0x580c, 0x0a, 0, 0},
	{0x580d, 0x0c, 0, 0}, {0x580e, 0x12, 0, 0}, {0x580f, 0x19, 0, 0},
	{0x5810, 0x0b, 0, 0}, {0x5811, 0x07, 0, 0}, {0x5812, 0x04, 0, 0},
	{0x5813, 0x03, 0, 0}, {0x5814, 0x03, 0, 0}, {0x5815, 0x06, 0, 0},
	{0x5816, 0x0a, 0, 0}, {0x5817, 0x0f, 0, 0}, {0x5818, 0x0a, 0, 0},
	{0x5819, 0x05, 0, 0}, {0x581a, 0x01, 0, 0}, {0x581b, 0x00, 0, 0},
	{0x581c, 0x00, 0, 0}, {0x581d, 0x03, 0, 0}, {0x581e, 0x08, 0, 0},
	{0x581f, 0x0c, 0, 0}, {0x5820, 0x0a, 0, 0}, {0x5821, 0x05, 0, 0},
	{0x5822, 0x01, 0, 0}, {0x5823, 0x00, 0, 0}, {0x5824, 0x00, 0, 0},
	{0x5825, 0x03, 0, 0}, {0x5826, 0x08, 0, 0}, {0x5827, 0x0c, 0, 0},
	{0x5828, 0x0e, 0, 0}, {0x5829, 0x08, 0, 0}, {0x582a, 0x06, 0, 0},
	{0x582b, 0x04, 0, 0}, {0x582c, 0x05, 0, 0}, {0x582d, 0x07, 0, 0},
	{0x582e, 0x0b, 0, 0}, {0x582f, 0x12, 0, 0}, {0x5830, 0x18, 0, 0},
	{0x5831, 0x10, 0, 0}, {0x5832, 0x0c, 0, 0}, {0x5833, 0x0a, 0, 0},
	{0x5834, 0x0b, 0, 0}, {0x5835, 0x0e, 0, 0}, {0x5836, 0x15, 0, 0},
	{0x5837, 0x19, 0, 0}, {0x5838, 0x32, 0, 0}, {0x5839, 0x1f, 0, 0},
	{0x583a, 0x18, 0, 0}, {0x583b, 0x16, 0, 0}, {0x583c, 0x17, 0, 0},
	{0x583d, 0x1e, 0, 0}, {0x583e, 0x26, 0, 0}, {0x583f, 0x53, 0, 0},
	{0x5840, 0x10, 0, 0}, {0x5841, 0x0f, 0, 0}, {0x5842, 0x0d, 0, 0},
	{0x5843, 0x0c, 0, 0}, {0x5844, 0x0e, 0, 0}, {0x5845, 0x09, 0, 0},
	{0x5846, 0x11, 0, 0}, {0x5847, 0x10, 0, 0}, {0x5848, 0x10, 0, 0},
	{0x5849, 0x10, 0, 0}, {0x584a, 0x10, 0, 0}, {0x584b, 0x0e, 0, 0},
	{0x584c, 0x10, 0, 0}, {0x584d, 0x10, 0, 0}, {0x584e, 0x11, 0, 0},
	{0x584f, 0x10, 0, 0}, {0x5850, 0x0f, 0, 0}, {0x5851, 0x0c, 0, 0},
	{0x5852, 0x0f, 0, 0}, {0x5853, 0x10, 0, 0}, {0x5854, 0x10, 0, 0},
	{0x5855, 0x0f, 0, 0}, {0x5856, 0x0e, 0, 0}, {0x5857, 0x0b, 0, 0},
	{0x5858, 0x10, 0, 0}, {0x5859, 0x0d, 0, 0}, {0x585a, 0x0d, 0, 0},
	{0x585b, 0x0c, 0, 0}, {0x585c, 0x0c, 0, 0}, {0x585d, 0x0c, 0, 0},
	{0x585e, 0x0b, 0, 0}, {0x585f, 0x0c, 0, 0}, {0x5860, 0x0c, 0, 0},
	{0x5861, 0x0c, 0, 0}, {0x5862, 0x0d, 0, 0}, {0x5863, 0x08, 0, 0},
	{0x5864, 0x11, 0, 0}, {0x5865, 0x18, 0, 0}, {0x5866, 0x18, 0, 0},
	{0x5867, 0x19, 0, 0}, {0x5868, 0x17, 0, 0}, {0x5869, 0x19, 0, 0},
	{0x586a, 0x16, 0, 0}, {0x586b, 0x13, 0, 0}, {0x586c, 0x13, 0, 0},
	{0x586d, 0x12, 0, 0}, {0x586e, 0x13, 0, 0}, {0x586f, 0x16, 0, 0},
	{0x5870, 0x14, 0, 0}, {0x5871, 0x12, 0, 0}, {0x5872, 0x10, 0, 0},
	{0x5873, 0x11, 0, 0}, {0x5874, 0x11, 0, 0}, {0x5875, 0x16, 0, 0},
	{0x5876, 0x14, 0, 0}, {0x5877, 0x11, 0, 0}, {0x5878, 0x10, 0, 0},
	{0x5879, 0x0f, 0, 0}, {0x587a, 0x10, 0, 0}, {0x587b, 0x14, 0, 0},
	{0x587c, 0x13, 0, 0}, {0x587d, 0x12, 0, 0}, {0x587e, 0x11, 0, 0},
	{0x587f, 0x11, 0, 0}, {0x5880, 0x12, 0, 0}, {0x5881, 0x15, 0, 0},
	{0x5882, 0x14, 0, 0}, {0x5883, 0x15, 0, 0}, {0x5884, 0x15, 0, 0},
	{0x5885, 0x15, 0, 0}, {0x5886, 0x13, 0, 0}, {0x5887, 0x17, 0, 0},
	{0x3710, 0x10, 0, 0}, {0x3632, 0x51, 0, 0}, {0x3702, 0x10, 0, 0},
	{0x3703, 0xb2, 0, 0}, {0x3704, 0x18, 0, 0}, {0x370b, 0x40, 0, 0},
	{0x370d, 0x03, 0, 0}, {0x3631, 0x01, 0, 0}, {0x3632, 0x52, 0, 0},
	{0x3606, 0x24, 0, 0}, {0x3620, 0x96, 0, 0}, {0x5785, 0x07, 0, 0},
	{0x3a13, 0x30, 0, 0}, {0x3600, 0x52, 0, 0}, {0x3604, 0x48, 0, 0},
	{0x3606, 0x1b, 0, 0}, {0x370d, 0x0b, 0, 0}, {0x370f, 0xc0, 0, 0},
	{0x3709, 0x01, 0, 0}, {0x3823, 0x00, 0, 0}, {0x5007, 0x00, 0, 0},
	{0x5009, 0x00, 0, 0}, {0x5011, 0x00, 0, 0}, {0x5013, 0x00, 0, 0},
	{0x519e, 0x00, 0, 0}, {0x5086, 0x00, 0, 0}, {0x5087, 0x00, 0, 0},
	{0x5088, 0x00, 0, 0}, {0x5089, 0x00, 0, 0}, {0x302b, 0x00, 0, 0},
	{0x3824, 0x11, 0, 0}, {0x3825, 0xdc, 0, 0}, {0x3826, 0x00, 0, 0},
	{0x3827, 0x08, 0, 0}, {0x380c, 0x0c, 0, 0}, {0x380d, 0x80, 0, 0},
	{0x380e, 0x03, 0, 0}, {0x380f, 0xe8, 0, 0}, {0x3808, 0x02, 0, 0},
	{0x3809, 0xd0, 0, 0}, {0x380A, 0x02, 0, 0}, {0x380B, 0x40, 0, 0},
	{0x3804, 0x04, 0, 0}, {0x3805, 0xb0, 0, 0}, {0x3806, 0x03, 0, 0},
	{0x3807, 0xc0, 0, 0}, {0x5686, 0x03, 0, 0}, {0x5687, 0xc0, 0, 0},
	{0x5682, 0x04, 0, 0}, {0x5683, 0xb0, 0, 0},
	{0x4302, 0x03, 0, 0}, {0x4303, 0xac, 0, 0}, 	/* Clip Y range to [16..235] (*4) */
	{0x4304, 0x00, 0, 0}, {0x4305, 0x40, 0, 0},
};

static struct ov5642_mode_info ov5642_mode_info_data[2][ov5642_mode_MAX + 1] = {
	{
		{ov5642_mode_VGA_640_480,    640,  480,
		ov5642_setting_15fps_VGA_640_480,
		ARRAY_SIZE(ov5642_setting_15fps_VGA_640_480)},
		{ov5642_mode_QVGA_320_240,   320,  240,
		ov5642_setting_15fps_QVGA_320_240,
		ARRAY_SIZE(ov5642_setting_15fps_QVGA_320_240)},
		{ov5642_mode_NTSC_720_480,   720,  480,
		ov5642_setting_15fps_NTSC_720_480,
		ARRAY_SIZE(ov5642_setting_15fps_NTSC_720_480)},
		{ov5642_mode_PAL_720_576,   720,  576,
		ov5642_setting_15fps_PAL_720_576,
		ARRAY_SIZE(ov5642_setting_15fps_PAL_720_576)},
		{ov5642_mode_720P_1280_720,  1280, 720,
		ov5642_setting_15fps_720P_1280_720,
		ARRAY_SIZE(ov5642_setting_15fps_720P_1280_720)},
		{ov5642_mode_1080P_1920_1080, 1920, 1080,
		ov5642_setting_15fps_1080P_1920_1080,
		ARRAY_SIZE(ov5642_setting_15fps_1080P_1920_1080)},
		{ov5642_mode_QSXGA_2592_1944, 2592, 1944,
		ov5642_setting_15fps_QSXGA_2592_1944,
		ARRAY_SIZE(ov5642_setting_15fps_QSXGA_2592_1944)},
		{ov5642_mode_QCIF_176_144, 176, 144,
		ov5642_setting_15fps_QCIF_176_144,
		ARRAY_SIZE(ov5642_setting_15fps_QCIF_176_144)},
		{ov5642_mode_XGA_1024_768, 1024, 768,
		ov5642_setting_15fps_XGA_1024_768,
		ARRAY_SIZE(ov5642_setting_15fps_XGA_1024_768)},
	},
	{
		{ov5642_mode_VGA_640_480,    640,  480,
		ov5642_setting_30fps_VGA_640_480,
		ARRAY_SIZE(ov5642_setting_30fps_VGA_640_480)},
		{ov5642_mode_QVGA_320_240,   320,  240,
		ov5642_setting_30fps_QVGA_320_240,
		ARRAY_SIZE(ov5642_setting_30fps_QVGA_320_240)},
		{ov5642_mode_NTSC_720_480,   720, 480,
		ov5642_setting_30fps_NTSC_720_480,
		ARRAY_SIZE(ov5642_setting_30fps_NTSC_720_480)},
		{ov5642_mode_PAL_720_576,    720, 576,
		ov5642_setting_30fps_PAL_720_576,
		ARRAY_SIZE(ov5642_setting_30fps_PAL_720_576)},
		{ov5642_mode_720P_1280_720,  1280, 720,
		ov5642_setting_30fps_720P_1280_720,
		ARRAY_SIZE(ov5642_setting_30fps_720P_1280_720)},
		{ov5642_mode_1080P_1920_1080, 0, 0, NULL, 0},
		{ov5642_mode_QSXGA_2592_1944, 0, 0, NULL, 0},
		{ov5642_mode_QCIF_176_144, 176, 144,
		ov5642_setting_30fps_QCIF_176_144,
		ARRAY_SIZE(ov5642_setting_30fps_QCIF_176_144)},
		{ov5642_mode_XGA_1024_768, 1024, 768,
		ov5642_setting_30fps_XGA_1024_768,
		ARRAY_SIZE(ov5642_setting_30fps_XGA_1024_768)},
	},
};

static struct regulator *io_regulator;
static struct regulator *core_regulator;
static struct regulator *analog_regulator;
static struct regulator *gpo_regulator;

static int ov5642_probe(struct i2c_client *adapter,
				const struct i2c_device_id *device_id);
static int ov5642_remove(struct i2c_client *client);

static s32 ov5642_read_reg(u16 reg, u8 *val);
static s32 ov5642_write_reg(u16 reg, u8 val);

#ifdef CONFIG_OF
static const struct of_device_id ov5642_of_match[] = {
	{ .compatible = "ovti," "ov5642",
	},
	{ /* sentinel */ }
};

MODULE_DEVICE_TABLE(of, ov5642_of_match);
#endif

static const struct i2c_device_id ov5642_id[] = {
	{"ov5642", 0},
	{"ov564x", 0},
	{},
};

MODULE_DEVICE_TABLE(i2c, ov5642_id);

static struct i2c_driver ov5642_i2c_driver = {
	.driver = {
		  .owner = THIS_MODULE,
		  .name  = "ov5642",
#ifdef CONFIG_OF
		  .of_match_table = of_match_ptr(ov5642_of_match),
#endif
		  },
	.probe  = ov5642_probe,
	.remove = ov5642_remove,
	.id_table = ov5642_id,
};

static void ov5642_power_down_sensor(int enable)
{
	if (!gpiod_pwdn)
		return;
	gpiod_set_value_cansleep(gpiod_pwdn, enable ? 1 : 0);

	msleep(2);
	pr_debug("ov5640_mipi_camera_powerdown: powerdown=%x\n", enable);
}

static s32 update_device_addr(struct sensor_data *sensor)
{
	int ret;
	u8 buf[4];
	unsigned reg = 0x3100;
	unsigned default_addr = 0x3c;
	struct i2c_msg msg;

	if (sensor->i2c_client->addr == default_addr)
		return 0;

	buf[0] = reg >> 8;
	buf[1] = reg & 0xff;
	buf[2] = sensor->i2c_client->addr << 1;
	msg.addr = default_addr;
	msg.flags = 0;
	msg.len = 3;
	msg.buf = buf;


	ret = i2c_transfer(sensor->i2c_client->adapter, &msg, 1);
	if (ret < 0)
		pr_err("%s: ov5642 ret=%d\n", __func__, ret);
	return ret;
}

static void ov5642_reset(void)
{
	if (!gpiod_rst && !gpiod_pwdn)
		return;

	if (gpiod_rst2)
		gpiod_set_value_cansleep(gpiod_rst2, 1); /* camera reset */
	gpiod_set_value_cansleep(gpiod_rst, 1);		/* camera reset */
	gpiod_set_value_cansleep(gpiod_pwdn, 1);	/* camera power down */

	/* >= 5 ms, Let power supply stabilize */
	msleep(5);
	mxc_camera_common_lock();

	gpiod_set_value_cansleep(gpiod_pwdn, 0);

	/* >= 1ms from powerup, to reset release*/
	msleep(1);
	if (gpiod_rst2)
		gpiod_set_value_cansleep(gpiod_rst2, 0);
	gpiod_set_value_cansleep(gpiod_rst, 0);

	/* >= 20 ms from reset high to SCCB initialized */
	msleep(20);
	pr_debug("%s(mipi): reset released\n", __func__);

	update_device_addr(&ov5642_data);
	mxc_camera_common_unlock();

	gpiod_set_value_cansleep(gpiod_pwdn, 1);
}

static int ov5642_power_on(struct device *dev)
{
	int ret = 0;

	io_regulator = devm_regulator_get(dev, "DOVDD");
	if (!IS_ERR(io_regulator)) {
		regulator_set_voltage(io_regulator,
				      OV5642_VOLTAGE_DIGITAL_IO,
				      OV5642_VOLTAGE_DIGITAL_IO);
		ret = regulator_enable(io_regulator);
		if (ret) {
			pr_err("%s:io set voltage error\n", __func__);
			return ret;
		} else {
			dev_dbg(dev,
				"%s:io set voltage ok\n", __func__);
		}
	} else {
		pr_err("%s: cannot get io voltage error\n", __func__);
		io_regulator = NULL;
	}

	core_regulator = devm_regulator_get(dev, "DVDD");
	if (!IS_ERR(core_regulator)) {
		regulator_set_voltage(core_regulator,
				      OV5642_VOLTAGE_DIGITAL_CORE,
				      OV5642_VOLTAGE_DIGITAL_CORE);
		ret = regulator_enable(core_regulator);
		if (ret) {
			pr_err("%s:core set voltage error\n", __func__);
			return ret;
		} else {
			dev_dbg(dev,
				"%s:core set voltage ok\n", __func__);
		}
	} else {
		core_regulator = NULL;
		pr_err("%s: cannot get core voltage error\n", __func__);
	}

	analog_regulator = devm_regulator_get(dev, "AVDD");
	if (!IS_ERR(analog_regulator)) {
		regulator_set_voltage(analog_regulator,
				      OV5642_VOLTAGE_ANALOG,
				      OV5642_VOLTAGE_ANALOG);
		ret = regulator_enable(analog_regulator);
		if (ret) {
			pr_err("%s:analog set voltage error\n",
				__func__);
			return ret;
		} else {
			dev_dbg(dev,
				"%s:analog set voltage ok\n", __func__);
		}
	} else {
		analog_regulator = NULL;
		pr_err("%s: cannot get analog voltage error\n", __func__);
	}

	return ret;
}

static void ov5642_power_off(struct sensor_data *sensor)
{
	ov5642_power_down_sensor(1);
	if (gpo_regulator)
		regulator_disable(gpo_regulator);
	if (core_regulator)
		regulator_disable(core_regulator);
	if (analog_regulator)
		regulator_disable(analog_regulator);
	if (io_regulator)
		regulator_disable(io_regulator);
	sensor->on = 0;
}

static s32 ov5642_write_reg(u16 reg, u8 val)
{
	int ret;
	u8 au8Buf[3] = {0};

	au8Buf[0] = reg >> 8;
	au8Buf[1] = reg & 0xff;
	au8Buf[2] = val;

	if ((reg == 0x3008) && (val & 0x80)) {
		mxc_camera_common_lock();

		ret = i2c_master_send(ov5642_data.i2c_client, au8Buf, 3);
		update_device_addr(&ov5642_data);

		mxc_camera_common_unlock();
	} else {
		ret = i2c_master_send(ov5642_data.i2c_client, au8Buf, 3);
	}

	if (ret < 0) {
		pr_err("%s:write reg error:reg=%x,val=%x ret=%d\n",
			__func__, reg, val, ret);
		return ret;
	}
	return 0;
}

static s32 ov5642_read_reg(u16 reg, u8 *val)
{
	u8 au8RegBuf[2] = {0};
	u8 u8RdVal = 0;

	au8RegBuf[0] = reg >> 8;
	au8RegBuf[1] = reg & 0xff;

	if (2 != i2c_master_send(ov5642_data.i2c_client, au8RegBuf, 2)) {
		pr_err("%s:write reg error:reg=%x\n",
				__func__, reg);
		return -1;
	}

	if (1 != i2c_master_recv(ov5642_data.i2c_client, &u8RdVal, 1)) {
		pr_err("%s:read reg error:reg=%x,val=%x\n",
				__func__, reg, u8RdVal);
		return -1;
	}

	*val = u8RdVal;

	return u8RdVal;
}

static int ov5642_set_idle_mode(void) {
	register u16 RegAddr = 0;
	u8 ReadVal = 0;
	u8 WriteVal = 0;
	int retval = 0;
	int lc = 0;

	ReadVal = -1;
	RegAddr = REG_STA_FOCUS;
	retval = ov5642_read_reg(RegAddr, &ReadVal);
	if (retval < 0) {
		pr_err("%s, read reg 0x%x failed\n", __FUNCTION__, RegAddr);
	}

	for (lc = 0; (lc < 100) && (ReadVal != S_IDLE); ++lc) {
		WriteVal = CMD_IDLE_MODE;
		RegAddr = REG_CMD_MAIN;
		retval = ov5642_write_reg(RegAddr, WriteVal);
		if (retval < 0) {
			pr_err("%s, write reg 0x%x failed\n", __FUNCTION__, RegAddr);
		}

		mdelay(1);

		ReadVal = -1;
		RegAddr = REG_STA_FOCUS;
		retval = ov5642_read_reg(RegAddr, &ReadVal);
		if (retval < 0) {
			pr_err("%s, read reg 0x%x failed\n", __FUNCTION__, RegAddr);
		}
	}

	if (ReadVal != S_IDLE)
		retval = -1;
	else
		retval = 0;
	return retval;
}

static int ov5642_config_auto_focus(void){
	ov5642_write_reg(REG_CMD_TAG, 0x01);
	ov5642_write_reg(REG_CMD_MAIN, 0x10);
	return 0;
}

static int ov5642_apply_settings(const struct reg_value *regv, int cnt)
{
	int i = 0;
	int retval = 0;

	for (i = 0; i < cnt; ++i, ++regv) {
		u32 delay_ms = regv->u32Delay_ms;
		u16 reg_addr = regv->u16RegAddr;
		u8 val = regv->u8Val;
		u8 mask = regv->u8Mask;
		u8 orig_val = 0;

		if (mask) {
			retval = ov5642_read_reg(reg_addr, &orig_val);
			if (retval < 0) {
				pr_err("%s, read reg 0x%x failed\n",
						__func__, reg_addr);
				return retval;
			}

			val = (val & mask) | (orig_val & ~mask);
		}

		retval = ov5642_write_reg(reg_addr, val);
		if (retval < 0) {
			pr_err("%s, write reg 0x%x failed\n", __func__,
					reg_addr);
			return retval;
		}

		if (delay_ms)
			mdelay(delay_ms);
	}
	return retval;
}

static int ov5642_set_rot_mode(struct reg_value *rot_mode)
{
	return ov5642_apply_settings(rot_mode, 2);
}

static int ov5642_af_init(void)
{
	int ret;

	/* Set firmware's interface registers to initial values */
	ret = ov5642_apply_settings(ov5642_af_firmware,
				    ARRAY_SIZE(ov5642_af_firmware));
	if (ret < 0) {
		pr_err("%s: AF firmware init failed\n", __func__);
		goto out;
	}

	/* Wait for firmware to become ready */
	ret = ov5642_set_idle_mode();
	if (ret < 0) {
		pr_err("%s: firmware is not responding\n", __func__);
		ret = -EBUSY;
		goto out;
	}

	pr_info("ov5642: initialized AF successufully\n");
	ret = 0;
out:
	return ret;
}

static int ov5642_af_get_status(int *status)
{
	int err;
	u8 val = 0;

	err = ov5642_read_reg(REG_STA_FOCUS, &val);
	if (err < 0)
		return err;

	switch (val) {
	case S_DRVICERR:
	case S_ERROR:
	case S_FIRWARE:
		/* firmware is downloaded and not to be initialized */
		*status = V4L2_AUTO_FOCUS_STATUS_FAILED;
		break;
	case S_CAPTURE:
	case S_FOCUSING:
	case S_STEP:
		/* focus is running */
		*status = V4L2_AUTO_FOCUS_STATUS_BUSY;
		break;
	case S_FOCUSED:
		/* focus is finished */
		*status = V4L2_AUTO_FOCUS_STATUS_REACHED;
		break;
	case S_STARTUP:
		/* firmware is initializing */
	case S_IDLE:
		/* firmware is idle */
	default:
		/*
		 * when focus is idle, the status value is variable,
		 * but it is different with status value above, and
		 * the value is 0x20 in most of cases.
		 */
		*status = V4L2_AUTO_FOCUS_STATUS_IDLE;
	}

	return 0;
}

static int ov5642_af_set_mode(int mode)
{
	int err = 0;

	switch (mode) {
	case V4L2_CID_AUTO_FOCUS_START:
		/* start single focus */
		err = ov5642_write_reg(REG_CMD_MAIN, CMD_SINGLE_FOCUS_MODE);
		focus_range = V4L2_AUTO_FOCUS_RANGE_NORMAL;
		break;
	case V4L2_CID_FOCUS_AUTO:
		/* start continuous focus */
		err = ov5642_write_reg(REG_CMD_MAIN, CMD_CONST_FOCUS_MODE);
		focus_range = V4L2_AUTO_FOCUS_RANGE_NORMAL;
		break;
	case V4L2_CID_3A_LOCK:
		/* pause focus */
		err = ov5642_write_reg(REG_CMD_MAIN, CMD_PAUSE);
		break;
	case V4L2_CID_AUTO_FOCUS_STOP:
		/* release focus to infinity */
		err = ov5642_write_reg(REG_CMD_MAIN, CMD_IDLE_MODE);
		break;
	default:
		pr_err("Invalid mode: %d\n", mode);
		return -EINVAL;
	}

	if (err) {
		pr_err("Setting focus mode failed!\n");
		return err;
	}

	focus_mode = mode;

	return 0;
}

static int ov5642_af_start_single(void)
{
	return ov5642_af_set_mode(V4L2_CID_AUTO_FOCUS_START);
}

static int ov5642_af_stop_single(void)
{
	return ov5642_af_set_mode(V4L2_CID_AUTO_FOCUS_STOP);
}

static int ov5642_af_set_auto(int enable)
{
	int ret = 0;

	if (enable)
		ret = ov5642_af_set_mode(V4L2_CID_FOCUS_AUTO);
	else
		ret = ov5642_af_set_mode(V4L2_CID_AUTO_FOCUS_STOP);

	return ret;
}

static int ov5642_af_set_focus_lock(int value)
{
	if (value != V4L2_LOCK_FOCUS) {
		pr_err("%s: Invalid focus cmd (%#x)\n", __func__, value);
		return -EINVAL;
	}

	return ov5642_af_set_mode(V4L2_CID_3A_LOCK);
}

static int ov5642_af_set_focus_range(int value)
{
	int ret = 0;

	switch (value) {
	case V4L2_AUTO_FOCUS_RANGE_INFINITY:
		ret = ov5642_af_set_mode(V4L2_CID_AUTO_FOCUS_STOP);
		focus_range = V4L2_AUTO_FOCUS_RANGE_INFINITY;
		break;
	case V4L2_AUTO_FOCUS_RANGE_NORMAL:
	default:
		focus_range = V4L2_AUTO_FOCUS_RANGE_NORMAL;
	}
	return ret;
}

static int ov5642_af_set_region(uint16_t x, uint16_t y)
{
	int err = 0;

	/* Sanity check */
	if ((x > OV5642_AF_ZONE_ARRAY_WIDTH) ||
	    (y > OV5642_AF_ZONE_ARRAY_HEIGHT)) {
		pr_err("Wrong region values: x %d y %d\n", x, y);
		return -EINVAL;
	}

	/* Setting the region */
	err = ov5642_write_reg(REG_CMD_PARA0, x);
	if (err) {
		pr_err("Setting param 0 failed!\n");
		return err;
	}

	err = ov5642_write_reg(REG_CMD_PARA1, y);
	if (err) {
		pr_err("Setting param 1 failed!\n");
		return err;
	}

	err = ov5642_write_reg(REG_CMD_MAIN, CMD_SET_ZONE_MODE);
	if (err) {
		pr_err("Resetting focus region failed!\n");
		return err;
	}

	roi_x = x;
	roi_y = y;

	return 0;
}

static int ov5642_init_mode(enum ov5642_frame_rate frame_rate,
		enum ov5642_mode mode);
static int ov5642_write_snapshot_para(enum ov5642_frame_rate frame_rate,
		enum ov5642_mode mode);
static int ov5642_change_mode(enum ov5642_frame_rate new_frame_rate,
		enum ov5642_frame_rate old_frame_rate,
		enum ov5642_mode new_mode,
		enum ov5642_mode orig_mode)
{
	struct reg_value *pModeSetting = NULL;
	s32 iModeSettingArySize = 0;
	int retval = 0;

	if (new_mode > ov5642_mode_MAX || new_mode < ov5642_mode_MIN) {
		pr_err("Wrong ov5642 mode detected!\n");
		return -1;
	}

	if ((new_frame_rate == old_frame_rate) &&
	    (new_mode == ov5642_mode_VGA_640_480) &&
		(orig_mode == ov5642_mode_QSXGA_2592_1944)) {
		pModeSetting = ov5642_setting_QSXGA_2_VGA;
		iModeSettingArySize = ARRAY_SIZE(ov5642_setting_QSXGA_2_VGA);
		ov5642_data.pix.width = 640;
		ov5642_data.pix.height = 480;
	} else if ((new_frame_rate == old_frame_rate) &&
	    (new_mode == ov5642_mode_QVGA_320_240) &&
		(orig_mode == ov5642_mode_VGA_640_480)) {
		pModeSetting = ov5642_setting_VGA_2_QVGA;
		iModeSettingArySize = ARRAY_SIZE(ov5642_setting_VGA_2_QVGA);
		ov5642_data.pix.width = 320;
		ov5642_data.pix.height = 240;
	} else {
		retval = ov5642_write_snapshot_para(new_frame_rate, new_mode);
		return retval;
	}

	if (ov5642_data.pix.width == 0 || ov5642_data.pix.height == 0 ||
			pModeSetting == NULL || iModeSettingArySize == 0)
		return -EINVAL;

	return ov5642_apply_settings(pModeSetting, iModeSettingArySize);
}

static int ov5642_init_mode(enum ov5642_frame_rate frame_rate,
			    enum ov5642_mode mode)
{
	struct reg_value *pModeSetting = NULL;
	s32 iModeSettingArySize = 0;

	if (mode > ov5642_mode_MAX || mode < ov5642_mode_MIN) {
		pr_err("Wrong ov5642 mode detected!\n");
		return -1;
	}

	pModeSetting = ov5642_mode_info_data[frame_rate][mode].init_data_ptr;
	iModeSettingArySize =
		ov5642_mode_info_data[frame_rate][mode].init_data_size;

	ov5642_data.pix.width = ov5642_mode_info_data[frame_rate][mode].width;
	ov5642_data.pix.height = ov5642_mode_info_data[frame_rate][mode].height;

	if (ov5642_data.pix.width == 0 || ov5642_data.pix.height == 0 ||
	    pModeSetting == NULL || iModeSettingArySize == 0)
		return -EINVAL;

	return ov5642_apply_settings(pModeSetting, iModeSettingArySize);
}

static int ov5642_write_snapshot_para(enum ov5642_frame_rate frame_rate,
       enum ov5642_mode mode)
{
	int ret = 0;
	bool m_60Hz = false;
	u16 cap_frame_rate = 50;
	u16 g_prev_frame_rate = 225;

	u8 ev_low, ev_mid, ev_high;
	u8 ret_l, ret_m, ret_h, gain, lines_10ms;
	u16 ulcap_ev, icap_gain, prev_maxlines;
	u32 ulcap_ev_gain, cap_maxlines, g_prev_ev;

	ov5642_write_reg(0x3503, 0x07);

	ret_h = ret_m = ret_l = 0;
	g_prev_ev = 0;
	ov5642_read_reg(0x3500, &ret_h);
	ov5642_read_reg(0x3501, &ret_m);
	ov5642_read_reg(0x3502, &ret_l);
	g_prev_ev = (ret_h << 12) + (ret_m << 4) + (ret_l >> 4);

	ret_h = ret_m = ret_l = 0;
	prev_maxlines = 0;
	ov5642_read_reg(0x380e, &ret_h);
	ov5642_read_reg(0x380f, &ret_l);
	prev_maxlines = (ret_h << 8) + ret_l;
	/*Read back AGC Gain for preview*/
	gain = 0;
	ov5642_read_reg(0x350b, &gain);

	ret = ov5642_init_mode(frame_rate, mode);
	if (ret < 0)
		return ret;

	ret_h = ret_m = ret_l = 0;
	ov5642_read_reg(0x380e, &ret_h);
	ov5642_read_reg(0x380f, &ret_l);
	cap_maxlines = (ret_h << 8) + ret_l;
	if (m_60Hz == true)
		lines_10ms = cap_frame_rate * cap_maxlines/12000;
	else
		lines_10ms = cap_frame_rate * cap_maxlines/10000;

	if (prev_maxlines == 0)
		prev_maxlines = 1;

	ulcap_ev = (g_prev_ev*(cap_frame_rate)*(cap_maxlines))/
		(((prev_maxlines)*(g_prev_frame_rate)));
	icap_gain = (gain & 0x0f) + 16;
	if (gain & 0x10)
		icap_gain = icap_gain << 1;

	if (gain & 0x20)
		icap_gain = icap_gain << 1;

	if (gain & 0x40)
		icap_gain = icap_gain << 1;

	if (gain & 0x80)
		icap_gain = icap_gain << 1;

	ulcap_ev_gain = 2 * ulcap_ev * icap_gain;

	if (ulcap_ev_gain < cap_maxlines*16) {
		ulcap_ev = ulcap_ev_gain/16;
		if (ulcap_ev > lines_10ms) {
			ulcap_ev /= lines_10ms;
			ulcap_ev *= lines_10ms;
		}
	} else
		ulcap_ev = cap_maxlines;

	if (ulcap_ev == 0)
		ulcap_ev = 1;

	icap_gain = (ulcap_ev_gain*2/ulcap_ev + 1)/2;
	ev_low = ((unsigned char)ulcap_ev)<<4;
	ev_mid = (unsigned char)(ulcap_ev >> 4) & 0xff;
	ev_high = (unsigned char)(ulcap_ev >> 12);

	gain = 0;
	if (icap_gain > 31) {
		gain |= 0x10;
		icap_gain = icap_gain >> 1;
	}
	if (icap_gain > 31) {
		gain |= 0x20;
		icap_gain = icap_gain >> 1;
	}
	if (icap_gain > 31) {
		gain |= 0x40;
		icap_gain = icap_gain >> 1;
	}
	if (icap_gain > 31) {
		gain |= 0x80;
		icap_gain = icap_gain >> 1;
	}
	if (icap_gain > 16)
		gain |= ((icap_gain - 16) & 0x0f);

	if (gain == 0x10)
		gain = 0x11;

	ov5642_write_reg(0x350b, gain);
	ov5642_write_reg(0x3502, ev_low);
	ov5642_write_reg(0x3501, ev_mid);
	ov5642_write_reg(0x3500, ev_high);
	msleep(500);

	return ret ;
}


/* --------------- IOCTL functions from v4l2_int_ioctl_desc --------------- */

static int ioctl_g_ifparm(struct v4l2_int_device *s, struct v4l2_ifparm *p)
{
	if (s == NULL) {
		pr_err("   ERROR!! no slave device set!\n");
		return -1;
	}

	memset(p, 0, sizeof(*p));
	p->u.bt656.clock_curr = ov5642_data.mclk;
	pr_debug("   clock_curr=mclk=%d\n", ov5642_data.mclk);
	p->if_type = V4L2_IF_TYPE_BT656;
	p->u.bt656.mode = V4L2_IF_TYPE_BT656_MODE_NOBT_8BIT;
	p->u.bt656.clock_min = OV5642_XCLK_MIN;
	p->u.bt656.clock_max = OV5642_XCLK_MAX;
	p->u.bt656.bt_sync_correct = 1;  /* Indicate external vsync */

	return 0;
}

/*!
 * ioctl_s_power - V4L2 sensor interface handler for VIDIOC_S_POWER ioctl
 * @s: pointer to standard V4L2 device structure
 * @on: indicates power mode (on or off)
 *
 * Turns the power on or off, depending on the value of on and returns the
 * appropriate error code.
 */
static int ioctl_s_power(struct v4l2_int_device *s, int on)
{
	struct sensor_data *sensor = s->priv;

	if (on && !sensor->on) {
		if (io_regulator)
			if (regulator_enable(io_regulator) != 0)
				return -EIO;
		if (core_regulator)
			if (regulator_enable(core_regulator) != 0)
				return -EIO;
		if (gpo_regulator)
			if (regulator_enable(gpo_regulator) != 0)
				return -EIO;
		if (analog_regulator)
			if (regulator_enable(analog_regulator) != 0)
				return -EIO;
		/* Make sure power on */
		ov5642_power_down_sensor(0);
	} else if (!on && sensor->on) {
		ov5642_power_down_sensor(1);
		if (analog_regulator)
			regulator_disable(analog_regulator);
		if (core_regulator)
			regulator_disable(core_regulator);
		if (io_regulator)
			regulator_disable(io_regulator);
		if (gpo_regulator)
			regulator_disable(gpo_regulator);
	}

	sensor->on = on;

	return 0;
}

/*!
 * ioctl_g_parm - V4L2 sensor interface handler for VIDIOC_G_PARM ioctl
 * @s: pointer to standard V4L2 device structure
 * @a: pointer to standard V4L2 VIDIOC_G_PARM ioctl structure
 *
 * Returns the sensor's video CAPTURE parameters.
 */
static int ioctl_g_parm(struct v4l2_int_device *s, struct v4l2_streamparm *a)
{
	struct sensor_data *sensor = s->priv;
	struct v4l2_captureparm *cparm = &a->parm.capture;
	int ret = 0;

	switch (a->type) {
	/* This is the only case currently handled. */
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		memset(a, 0, sizeof(*a));
		a->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		cparm->capability = sensor->streamcap.capability;
		cparm->timeperframe = sensor->streamcap.timeperframe;
		cparm->capturemode = sensor->streamcap.capturemode;
		ret = 0;
		break;

	/* These are all the possible cases. */
	case V4L2_BUF_TYPE_VIDEO_OUTPUT:
	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
	case V4L2_BUF_TYPE_VBI_CAPTURE:
	case V4L2_BUF_TYPE_VBI_OUTPUT:
	case V4L2_BUF_TYPE_SLICED_VBI_CAPTURE:
	case V4L2_BUF_TYPE_SLICED_VBI_OUTPUT:
		ret = -EINVAL;
		break;

	default:
		pr_debug("   type is unknown - %d\n", a->type);
		ret = -EINVAL;
		break;
	}

	return ret;
}

/*!
 * ioctl_s_parm - V4L2 sensor interface handler for VIDIOC_S_PARM ioctl
 * @s: pointer to standard V4L2 device structure
 * @a: pointer to standard V4L2 VIDIOC_S_PARM ioctl structure
 *
 * Configures the sensor to use the input parameters, if possible.  If
 * not possible, reverts to the old parameters and returns the
 * appropriate error code.
 */
static int ioctl_s_parm(struct v4l2_int_device *s, struct v4l2_streamparm *a)
{
	struct sensor_data *sensor = s->priv;
	struct v4l2_fract *timeperframe = &a->parm.capture.timeperframe;
	u32 tgt_fps, old_fps;	/* target frames per secound */
	enum ov5642_frame_rate new_frame_rate, old_frame_rate;
	int ret = 0;

	/* Make sure power on */
	ov5642_power_down_sensor(0);

	switch (a->type) {
	/* This is the only case currently handled. */
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		/* Check that the new frame rate is allowed. */
		if ((timeperframe->numerator == 0) ||
		    (timeperframe->denominator == 0)) {
			timeperframe->denominator = DEFAULT_FPS;
			timeperframe->numerator = 1;
		}

		tgt_fps = timeperframe->denominator /
			  timeperframe->numerator;

		if (tgt_fps > MAX_FPS) {
			timeperframe->denominator = MAX_FPS;
			timeperframe->numerator = 1;
		} else if (tgt_fps < MIN_FPS) {
			timeperframe->denominator = MIN_FPS;
			timeperframe->numerator = 1;
		}

		/* Actual frame rate we use */
		tgt_fps = timeperframe->denominator /
			  timeperframe->numerator;

		if (tgt_fps == 15)
			new_frame_rate = ov5642_15_fps;
		else if (tgt_fps == 30)
			new_frame_rate = ov5642_30_fps;
		else {
			pr_err(" The camera frame rate is not supported!\n");
			return -EINVAL;
		}

		if (sensor->streamcap.timeperframe.numerator != 0)
			old_fps = sensor->streamcap.timeperframe.denominator /
				sensor->streamcap.timeperframe.numerator;
		else
			old_fps = 30;

		if (old_fps == 15)
			old_frame_rate = ov5642_15_fps;
		else if (old_fps == 30)
			old_frame_rate = ov5642_30_fps;
		else {
			pr_warning(" No valid frame rate set!\n");
			old_frame_rate = ov5642_30_fps;
		}

		ret = ov5642_change_mode(new_frame_rate, old_frame_rate,
				a->parm.capture.capturemode,
				sensor->streamcap.capturemode);
		if (ret < 0)
			return ret;

		sensor->streamcap.timeperframe = *timeperframe;
		sensor->streamcap.capturemode =
				(u32)a->parm.capture.capturemode;
		break;

	/* These are all the possible cases. */
	case V4L2_BUF_TYPE_VIDEO_OUTPUT:
	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
	case V4L2_BUF_TYPE_VBI_CAPTURE:
	case V4L2_BUF_TYPE_VBI_OUTPUT:
	case V4L2_BUF_TYPE_SLICED_VBI_CAPTURE:
	case V4L2_BUF_TYPE_SLICED_VBI_OUTPUT:
		pr_debug("   type is not " \
			"V4L2_BUF_TYPE_VIDEO_CAPTURE but %d\n",
			a->type);
		ret = -EINVAL;
		break;

	default:
		pr_debug("   type is unknown - %d\n", a->type);
		ret = -EINVAL;
		break;
	}

	return ret;
}

/*!
 * ioctl_g_fmt_cap - V4L2 sensor interface handler for ioctl_g_fmt_cap
 * @s: pointer to standard V4L2 device structure
 * @f: pointer to standard V4L2 v4l2_format structure
 *
 * Returns the sensor's current pixel format in the v4l2_format
 * parameter.
 */
static int ioctl_g_fmt_cap(struct v4l2_int_device *s, struct v4l2_format *f)
{
	struct sensor_data *sensor = s->priv;

	switch (f->type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		f->fmt.pix = sensor->pix;
		pr_debug("%s: %dx%d\n", __func__, sensor->pix.width, sensor->pix.height);
		break;

	case V4L2_BUF_TYPE_SENSOR:
		pr_debug("%s: left=%d, top=%d, %dx%d\n", __func__,
			sensor->spix.left, sensor->spix.top,
			sensor->spix.swidth, sensor->spix.sheight);
		f->fmt.spix = sensor->spix;
		break;

	case V4L2_BUF_TYPE_PRIVATE:
		break;

	default:
		f->fmt.pix = sensor->pix;
		break;
	}
	return 0;
}

/*!
 * ioctl_g_ctrl - V4L2 sensor interface handler for VIDIOC_G_CTRL ioctl
 * @s: pointer to standard V4L2 device structure
 * @vc: standard V4L2 VIDIOC_G_CTRL ioctl structure
 *
 * If the requested control is supported, returns the control's current
 * value from the video_control[] array.  Otherwise, returns -EINVAL
 * if the control is not supported.
 */
static int ioctl_g_ctrl(struct v4l2_int_device *s, struct v4l2_control *vc)
{
	int ret = 0;

	switch (vc->id) {
	case V4L2_CID_BRIGHTNESS:
		vc->value = ov5642_data.brightness;
		break;
	case V4L2_CID_HUE:
		vc->value = ov5642_data.hue;
		break;
	case V4L2_CID_CONTRAST:
		vc->value = ov5642_data.contrast;
		break;
	case V4L2_CID_SATURATION:
		vc->value = ov5642_data.saturation;
		break;
	case V4L2_CID_RED_BALANCE:
		vc->value = ov5642_data.red;
		break;
	case V4L2_CID_BLUE_BALANCE:
		vc->value = ov5642_data.blue;
		break;
	case V4L2_CID_EXPOSURE:
		vc->value = ov5642_data.ae_mode;
		break;
	case V4L2_CID_3A_LOCK:
		vc->value = (focus_mode == V4L2_CID_3A_LOCK);
		break;
	case V4L2_CID_AUTO_FOCUS_RANGE:
		vc->value = focus_range;
		break;
	case V4L2_CID_AUTO_FOCUS_STATUS:
		ret = ov5642_af_get_status(&vc->value);
		break;
	case V4L2_CID_FOCUS_ABSOLUTE:
		/* Custom implementation that returns the ROI */
		vc->value = (roi_x << 16) + roi_y;
		break;
	case V4L2_CID_FOCUS_AUTO:
		vc->value = (focus_mode == V4L2_CID_FOCUS_AUTO);
		break;
	default:
		ret = -EINVAL;
	}

	return ret;
}

/*!
 * ioctl_s_ctrl - V4L2 sensor interface handler for VIDIOC_S_CTRL ioctl
 * @s: pointer to standard V4L2 device structure
 * @vc: standard V4L2 VIDIOC_S_CTRL ioctl structure
 *
 * If the requested control is supported, sets the control's current
 * value in HW (and updates the video_control[] array).  Otherwise,
 * returns -EINVAL if the control is not supported.
 */
static int ioctl_s_ctrl(struct v4l2_int_device *s, struct v4l2_control *vc)
{
	int retval = 0;
	struct sensor_data *sensor = s->priv;
	__u32 captureMode = sensor->streamcap.capturemode;
	struct reg_value *rot_mode = NULL;

	pr_debug("In ov5642:ioctl_s_ctrl %d\n",
		 vc->id);

	switch (vc->id) {
	case V4L2_CID_BRIGHTNESS:
		break;
	case V4L2_CID_3A_LOCK:
		retval = ov5642_af_set_focus_lock(vc->value);
		break;
	case V4L2_CID_AUTO_FOCUS_RANGE:
		retval = ov5642_af_set_focus_range(vc->value);
		break;
	case V4L2_CID_AUTO_FOCUS_START:
		retval = ov5642_af_start_single();
		break;
	case V4L2_CID_AUTO_FOCUS_STOP:
		retval = ov5642_af_stop_single();
		break;
	case V4L2_CID_FOCUS_ABSOLUTE:
		/* Custom implementation that sets the ROI */
		retval = ov5642_af_set_region((vc->value >> 16),
					      (vc->value & 0xffff));
		break;
	case V4L2_CID_FOCUS_AUTO:
		retval = ov5642_af_set_auto(!!vc->value);
		break;
	case V4L2_CID_CONTRAST:
		break;
	case V4L2_CID_SATURATION:
		break;
	case V4L2_CID_HUE:
		break;
	case V4L2_CID_AUTO_WHITE_BALANCE:
		break;
	case V4L2_CID_DO_WHITE_BALANCE:
		break;
	case V4L2_CID_RED_BALANCE:
		break;
	case V4L2_CID_BLUE_BALANCE:
		break;
	case V4L2_CID_GAMMA:
		break;
	case V4L2_CID_EXPOSURE:
		break;
	case V4L2_CID_AUTOGAIN:
		break;
	case V4L2_CID_GAIN:
		break;
	case V4L2_CID_HFLIP:
		break;
	case V4L2_CID_VFLIP:
		break;
	case V4L2_CID_MXC_ROT:
	case V4L2_CID_MXC_VF_ROT:
		switch (vc->value) {
		case V4L2_MXC_ROTATE_NONE:
			if (captureMode == ov5642_mode_QSXGA_2592_1944)
				rot_mode = ov5642_rot_none_FULL;
			else
				rot_mode = ov5642_rot_none_VGA;

			if (ov5642_set_rot_mode(rot_mode))
				retval = -EPERM;
			break;
		case V4L2_MXC_ROTATE_VERT_FLIP:
			if (captureMode == ov5642_mode_QSXGA_2592_1944)
				rot_mode = ov5642_rot_vert_flip_FULL;
			else
				rot_mode = ov5642_rot_vert_flip_VGA ;

			if (ov5642_set_rot_mode(rot_mode))
				retval = -EPERM;
			break;
		case V4L2_MXC_ROTATE_HORIZ_FLIP:
			if (captureMode == ov5642_mode_QSXGA_2592_1944)
				rot_mode = ov5642_rot_horiz_flip_FULL;
			else
				rot_mode = ov5642_rot_horiz_flip_VGA;

			if (ov5642_set_rot_mode(rot_mode))
				retval = -EPERM;
			break;
		case V4L2_MXC_ROTATE_180:
			if (captureMode == ov5642_mode_QSXGA_2592_1944)
				rot_mode = ov5642_rot_180_FULL;
			else
				rot_mode = ov5642_rot_180_VGA;

			if (ov5642_set_rot_mode(rot_mode))
				retval = -EPERM;
			break;
		default:
			retval = -EPERM;
			break;
		}
		break;
	default:
		retval = -EPERM;
		break;
	}

	return retval;
}

static int ioctl_send_command(struct v4l2_int_device *s, struct v4l2_send_command_control *vc) {
	int ret = -1;
	int retval1,retval2;
	u8 loca_val=0;

	ret = ov5642_set_idle_mode();
	if (0 != ret)
		pr_err("error %d setting idle mode\n", ret);
	ov5642_config_auto_focus();
	switch (vc->id) {
		case 101: //step to near
			pr_debug("Stepping to near object\n");
			retval1=ov5642_write_reg(REG_CMD_TAG, 0x01);
			retval2=ov5642_write_reg(REG_CMD_MAIN, 0x05);
			if(retval1 == 0 && retval2 == 0)
				ret = 0;
			break;
		case 102: //step to far
			pr_debug("Stepping to far object\n");
			retval1=ov5642_write_reg(REG_CMD_TAG, 0x02);
			retval2=ov5642_write_reg(REG_CMD_MAIN, 0x05);
			if(retval1 == 0 && retval2 == 0)
				ret = 0;
			break;

		case 103: //step to furthest
			pr_debug("Stepping to furthest object\n");
			retval1=ov5642_write_reg(REG_CMD_TAG, 0x03);
			retval2=ov5642_write_reg(REG_CMD_MAIN, 0x05);
			if(retval1 == 0 && retval2 == 0)
				ret = 0;
			break;

		case 104: //step to nearest
			pr_debug("Stepping to nearest object\n");
			retval1=ov5642_write_reg(REG_CMD_TAG, 0x04);
			retval2=ov5642_write_reg(REG_CMD_MAIN, 0x05);
			if(retval1 == 0 && retval2 == 0)
				ret = 0;
			break;


		case 105: //step to specified position
			pr_debug("Stepping to position: %d\n", vc->value0);
			if(vc->value0 < 0 || vc->value0 > 255)
				return ret;
			loca_val = vc->value0;
			retval1=ov5642_write_reg(REG_CMD_TAG, 0x10);
			retval2=ov5642_write_reg(REG_CMD_PARA0, loca_val);
			ret=ov5642_write_reg(REG_CMD_MAIN, 0x05);
			if(retval1 != 0 && retval2 != 0 && ret != 0)
				ret = -1;
			break;
		default:
			break;
	}

	return ret;
}

/*!
 * ioctl_enum_framesizes - V4L2 sensor interface handler for
 *			   VIDIOC_ENUM_FRAMESIZES ioctl
 * @s: pointer to standard V4L2 device structure
 * @fsize: standard V4L2 VIDIOC_ENUM_FRAMESIZES ioctl structure
 *
 * Return 0 if successful, otherwise -EINVAL.
 */
static int ioctl_enum_framesizes(struct v4l2_int_device *s,
				 struct v4l2_frmsizeenum *fsize)
{
	if (fsize->index > ov5642_mode_MAX)
		return -EINVAL;

	fsize->pixel_format = ov5642_data.pix.pixelformat;
	fsize->discrete.width =
			max(ov5642_mode_info_data[0][fsize->index].width,
			    ov5642_mode_info_data[1][fsize->index].width);
	fsize->discrete.height =
			max(ov5642_mode_info_data[0][fsize->index].height,
			    ov5642_mode_info_data[1][fsize->index].height);
	return 0;
}

/*!
 * ioctl_enum_frameintervals - V4L2 sensor interface handler for
 *			       VIDIOC_ENUM_FRAMEINTERVALS ioctl
 * @s: pointer to standard V4L2 device structure
 * @fival: standard V4L2 VIDIOC_ENUM_FRAMEINTERVALS ioctl structure
 *
 * Return 0 if successful, otherwise -EINVAL.
 */
static int ioctl_enum_frameintervals(struct v4l2_int_device *s,
					 struct v4l2_frmivalenum *fival)
{
	int i, j, count;

	if (fival->index < 0 || fival->index > ov5642_mode_MAX)
		return -EINVAL;

	if (fival->pixel_format == 0 || fival->width == 0 ||
			fival->height == 0) {
		pr_warning("Please assign pixelformat, width and height.\n");
		return -EINVAL;
	}

	fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
	fival->discrete.numerator = 1;

	count = 0;
	for (i = 0; i < ARRAY_SIZE(ov5642_mode_info_data); i++) {
		for (j = 0; j < (ov5642_mode_MAX + 1); j++) {
			if (fival->pixel_format == ov5642_data.pix.pixelformat
			 && fival->width == ov5642_mode_info_data[i][j].width
			 && fival->height == ov5642_mode_info_data[i][j].height
			 && ov5642_mode_info_data[i][j].init_data_ptr != NULL) {
				count++;
			}
			if (fival->index == (count - 1)) {
				fival->discrete.denominator =
						ov5642_framerates[i];
				return 0;
			}
		}
	}

	return -EINVAL;
}

/*!
 * ioctl_g_chip_ident - V4L2 sensor interface handler for
 *			VIDIOC_DBG_G_CHIP_IDENT ioctl
 * @s: pointer to standard V4L2 device structure
 * @id: pointer to int
 *
 * Return 0.
 */
static int ioctl_g_chip_ident(struct v4l2_int_device *s, int *id)
{
	((struct v4l2_dbg_chip_ident *)id)->match.type =
					V4L2_CHIP_MATCH_I2C_DRIVER;
	strcpy(((struct v4l2_dbg_chip_ident *)id)->match.name, "ov5642_camera");

	return 0;
}

/*!
 * ioctl_init - V4L2 sensor interface handler for VIDIOC_INT_INIT
 * @s: pointer to standard V4L2 device structure
 */
static int ioctl_init(struct v4l2_int_device *s)
{

	return 0;
}

/*!
 * ioctl_enum_fmt_cap - V4L2 sensor interface handler for VIDIOC_ENUM_FMT
 * @s: pointer to standard V4L2 device structure
 * @fmt: pointer to standard V4L2 fmt description structure
 *
 * Return 0.
 */
static int ioctl_enum_fmt_cap(struct v4l2_int_device *s,
			      struct v4l2_fmtdesc *fmt)
{
	if (fmt->index > 0)	/* only 1 pixelformat support so far */
		return -EINVAL;

	fmt->pixelformat = ov5642_data.pix.pixelformat;

	return 0;
}

/*!
 * ioctl_dev_init - V4L2 sensor interface handler for vidioc_int_dev_init_num
 * @s: pointer to standard V4L2 device structure
 *
 * Initialise the device when slave attaches to the master.
 */
static int ioctl_dev_init(struct v4l2_int_device *s)
{
	struct reg_value *pModeSetting = NULL;
	s32 iModeSettingArySize = 0;

	struct sensor_data *sensor = s->priv;
	u32 tgt_xclk;	/* target xclk */
	u32 tgt_fps;	/* target frames per secound */
	enum ov5642_frame_rate frame_rate;

	ov5642_data.on = true;

	/* mclk */
	tgt_xclk = ov5642_data.mclk;
	tgt_xclk = min(tgt_xclk, (u32)OV5642_XCLK_MAX);
	tgt_xclk = max(tgt_xclk, (u32)OV5642_XCLK_MIN);
	ov5642_data.mclk = tgt_xclk;

	pr_debug("   Setting mclk to %d MHz\n", tgt_xclk / 1000000);

	/* Default camera frame rate is set in probe */
	tgt_fps = sensor->streamcap.timeperframe.denominator /
		  sensor->streamcap.timeperframe.numerator;

	if (tgt_fps == 15)
		frame_rate = ov5642_15_fps;
	else if (tgt_fps == 30)
		frame_rate = ov5642_30_fps;
	else
		return -EINVAL; /* Only support 15fps or 30fps now. */

	pModeSetting = ov5642_initial_setting;
	iModeSettingArySize = ARRAY_SIZE(ov5642_initial_setting);

	return ov5642_apply_settings(pModeSetting, iModeSettingArySize);
}

/*!
 * ioctl_dev_exit - V4L2 sensor interface handler for vidioc_int_dev_exit_num
 * @s: pointer to standard V4L2 device structure
 *
 * Delinitialise the device when slave detaches to the master.
 */
static int ioctl_dev_exit(struct v4l2_int_device *s)
{
	return 0;
}

/*!
 * This structure defines all the ioctls for this module and links them to the
 * enumeration.
 */
static struct v4l2_int_ioctl_desc ov5642_ioctl_desc[] = {
	{ vidioc_int_dev_init_num,
	  (v4l2_int_ioctl_func *)ioctl_dev_init },
	{ vidioc_int_dev_exit_num, ioctl_dev_exit},
	{ vidioc_int_s_power_num,
	  (v4l2_int_ioctl_func *)ioctl_s_power },
	{ vidioc_int_g_ifparm_num,
	  (v4l2_int_ioctl_func *)ioctl_g_ifparm },
/*	{ vidioc_int_g_needs_reset_num,
	  (v4l2_int_ioctl_func *)ioctl_g_needs_reset }, */
/*	{ vidioc_int_reset_num,
	  (v4l2_int_ioctl_func *)ioctl_reset }, */
	{ vidioc_int_init_num,
	  (v4l2_int_ioctl_func *)ioctl_init },
	{ vidioc_int_enum_fmt_cap_num,
	  (v4l2_int_ioctl_func *)ioctl_enum_fmt_cap },
/*	{ vidioc_int_try_fmt_cap_num,
	  (v4l2_int_ioctl_func *)ioctl_try_fmt_cap }, */
	{ vidioc_int_g_fmt_cap_num,
	  (v4l2_int_ioctl_func *)ioctl_g_fmt_cap },
/*	{ vidioc_int_s_fmt_cap_num,
	  (v4l2_int_ioctl_func *)ioctl_s_fmt_cap }, */
	{ vidioc_int_g_parm_num,
	  (v4l2_int_ioctl_func *)ioctl_g_parm },
	{ vidioc_int_s_parm_num,
	  (v4l2_int_ioctl_func *)ioctl_s_parm },
/*	{ vidioc_int_queryctrl_num,
	  (v4l2_int_ioctl_func *)ioctl_queryctrl }, */
	{ vidioc_int_g_ctrl_num,
	  (v4l2_int_ioctl_func *)ioctl_g_ctrl },
	{ vidioc_int_s_ctrl_num,
	  (v4l2_int_ioctl_func *)ioctl_s_ctrl },
	{ vidioc_int_enum_framesizes_num,
	  (v4l2_int_ioctl_func *)ioctl_enum_framesizes },
	{ vidioc_int_enum_frameintervals_num,
	  (v4l2_int_ioctl_func *)ioctl_enum_frameintervals },
	{ vidioc_int_g_chip_ident_num,
	  (v4l2_int_ioctl_func *)ioctl_g_chip_ident },
	{vidioc_int_send_command_num,
	  (v4l2_int_ioctl_func *) ioctl_send_command},
};

static struct v4l2_int_slave ov5642_slave = {
	.ioctls = ov5642_ioctl_desc,
	.num_ioctls = ARRAY_SIZE(ov5642_ioctl_desc),
};

static struct v4l2_int_device ov5642_int_device = {
	.module = THIS_MODULE,
	.name = "ov5642",
	.type = v4l2_int_type_slave,
	.u = {
		.slave = &ov5642_slave,
	},
};

#define MINREG 0x3000
#define MAXREG 0x603C

static int last_reg = -1 ;

static ssize_t show_reg(struct device *dev,
			struct device_attribute *attr, char *buf)
{
	if ((MINREG <= last_reg) && (MAXREG >= last_reg)) {
		u8 val;
		s32 rval = ov5642_read_reg(last_reg, &val);
		if (0 <= rval)
			return sprintf(buf, "0x%02x", val);
		else
			return rval;
	} else
		return 0;
}

static ssize_t store_reg
	(struct device *dev,
	 struct device_attribute *attr,
	 const char *buf, size_t count)
{
	unsigned reg, value;
	int numscanned = sscanf(buf,"%x %x\n", &reg, &value);
	if (2 == numscanned) {
		int rval = ov5642_write_reg(reg,value);
		if (0 != rval)
			dev_err(dev,"%s: error %d setting reg 0x%04x to 0x%02x\n",
				__func__, rval, reg, value);
	} else if (1 == numscanned) {
		last_reg = reg;
	} else
		dev_err(dev,"%s: invalid register: use form 0xREG [0xVAL]\n",
			__func__);
	return count;
}
static DEVICE_ATTR(ov5642_reg, S_IRUGO|S_IWUSR|S_IWGRP, show_reg, store_reg);

/*!
 * ov5642 I2C probe function
 *
 * @param adapter            struct i2c_adapter *
 * @return  Error code indicating success or failure
 */
static int ov5642_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct pinctrl *pinctrl;
	struct device *dev = &client->dev;
	int retval;
	u8 chip_id_high, chip_id_low;
	struct sensor_data *sensor = &ov5642_data;
	struct gpio_desc *gd;

	/* ov5642 pinctrl */
	pinctrl = devm_pinctrl_get_select_default(dev);
	if (IS_ERR(pinctrl)) {
		dev_err(dev, "ov5642 setup pinctrl failed!");
		return PTR_ERR(pinctrl);
	}

	/* request power down pin */
	gd = devm_gpiod_get_index_optional(dev, "pwn", 0, GPIOD_OUT_HIGH);
	if (IS_ERR(gd))
		return PTR_ERR(gd);
	if (!gd)
		dev_warn(dev, "no sensor pwdn pin available");
	gpiod_pwdn = gd;

	/* request reset pin */
	gd = devm_gpiod_get_index_optional(dev, "rst", 0, GPIOD_OUT_HIGH);
	if (IS_ERR(gd))
		return PTR_ERR(gd);
	if (!gd)
		dev_warn(dev, "no sensor reset pin available");
	gpiod_rst = gd;
	gd = devm_gpiod_get_index_optional(dev, "rst", 1, GPIOD_OUT_HIGH);
	if (IS_ERR(gd))
		return PTR_ERR(gd);
	gpiod_rst2 = gd;

	/* Set initial values for the sensor struct. */
	memset(&ov5642_data, 0, sizeof(ov5642_data));
	ov5642_data.sensor_clk = devm_clk_get(dev, "csi_mclk");
	if (IS_ERR(ov5642_data.sensor_clk)) {
		/* assuming clock enabled by default */
		ov5642_data.sensor_clk = NULL;
		dev_err(dev, "clock-frequency missing or invalid\n");
		return PTR_ERR(ov5642_data.sensor_clk);
	}

	retval = of_property_read_u32(dev->of_node, "mclk",
					(u32 *) &(ov5642_data.mclk));
	if (retval) {
		dev_err(dev, "mclk missing or invalid\n");
		return retval;
	}

	retval = of_property_read_u32(dev->of_node, "mclk_source",
					(u32 *) &(ov5642_data.mclk_source));
	if (retval) {
		dev_err(dev, "mclk_source missing or invalid\n");
		return retval;
	}

	retval = of_property_read_u32(dev->of_node, "ipu_id",
					&sensor->ipu_id);
	if (retval) {
		dev_err(dev, "ipu_id missing or invalid\n");
		return retval;
	}

	retval = of_property_read_u32(dev->of_node, "csi_id",
					&(ov5642_data.csi));
	if (retval) {
		dev_err(dev, "csi_id missing or invalid\n");
		return retval;
	}

	clk_prepare_enable(ov5642_data.sensor_clk);

	ov5642_data.io_init = ov5642_reset;
	ov5642_data.i2c_client = client;
	ov5642_data.pix.pixelformat = V4L2_PIX_FMT_YUYV;
	ov5642_data.pix.width = 640;
	ov5642_data.pix.height = 480;
	ov5642_data.streamcap.capability = V4L2_MODE_HIGHQUALITY |
					   V4L2_CAP_TIMEPERFRAME;
	ov5642_data.streamcap.capturemode = 0;
	ov5642_data.streamcap.timeperframe.denominator = DEFAULT_FPS;
	ov5642_data.streamcap.timeperframe.numerator = 1;

	ov5642_power_on(&client->dev);

	ov5642_reset();

	ov5642_power_down_sensor(0);

	retval = ov5642_read_reg(OV5642_CHIP_ID_HIGH_BYTE, &chip_id_high);
	if (retval < 0 || chip_id_high != 0x56) {
		pr_warning("camera ov5642 is not found(%d) %x\n", retval, chip_id_high);
		goto err;
	}
	retval = ov5642_read_reg(OV5642_CHIP_ID_LOW_BYTE, &chip_id_low);
	if (retval < 0 || chip_id_low != 0x42) {
		pr_warning("camera ov5642 is not found(%d) low %x\n", retval, chip_id_low);
		goto err;
	}

	retval = ov5642_af_init();
	if (retval < 0)
		pr_err("%s: error downloading autofocus firmware\n", __func__);

	ov5642_power_down_sensor(1);

	ov5642_int_device.priv = &ov5642_data;

	retval = v4l2_int_device_register(&ov5642_int_device);

	clk_disable_unprepare(ov5642_data.sensor_clk);

	if (device_create_file(&client->dev,
			&dev_attr_ov5642_reg))
		dev_err(&client->dev, "Error on creating sysfs file for ov5642_reg\n");
	else
		dev_err(&client->dev, "created sysfs entry for reading regs\n");

	pr_info("camera ov5642 is found\n");
	return retval;
err:
	ov5642_power_off(sensor);
	clk_disable_unprepare(ov5642_data.sensor_clk);
	return -ENODEV;
}

/*!
 * ov5642 I2C detach function
 *
 * @param client            struct i2c_client *
 * @return  Error code indicating success or failure
 */
static int ov5642_remove(struct i2c_client *client)
{
	v4l2_int_device_unregister(&ov5642_int_device);

	if (gpo_regulator)
		regulator_disable(gpo_regulator);

	if (analog_regulator)
		regulator_disable(analog_regulator);

	if (core_regulator)
		regulator_disable(core_regulator);

	if (io_regulator)
		regulator_disable(io_regulator);

	device_remove_file(&client->dev,
		&dev_attr_ov5642_reg);

	return 0;
}

/*!
 * ov5642 init function
 * Called by insmod ov5642_camera.ko.
 *
 * @return  Error code indicating success or failure
 */
static __init int ov5642_init(void)
{
	u8 err;

	err = i2c_add_driver(&ov5642_i2c_driver);
	if (err != 0)
		pr_err("%s:driver registration failed, error=%d\n",
			__func__, err);

	return err;
}

/*!
 * OV5642 cleanup function
 * Called on rmmod ov5642_camera.ko
 *
 * @return  Error code indicating success or failure
 */
static void __exit ov5642_clean(void)
{
	i2c_del_driver(&ov5642_i2c_driver);
}

module_init(ov5642_init);
module_exit(ov5642_clean);

MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_DESCRIPTION("OV5642 Camera Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
MODULE_ALIAS("CSI");
